#! /usr/bin/perl

use strict;
use warnings;

use FindBin;
use lib "$FindBin::Bin/../../corelib";
use lib "$FindBin::Bin/../../lib/";
use lib "$FindBin::Bin/../../lib/BioFileConverter";

# Probably useful with Ensemble (cf --helpensembl)
# TODO - Arranger ca et verifier que ca fonctionne toujours avec ensembl !
# To be sure that result of -I is BEFORE the standard Bio location !
#@INC = reverse @INC;

use Error qw(:try);
use LipmError;
use General;
use ParamParser;
use AnnotatedSeq;
use Assembly;
use EnzymeValidation;
use AccWarn;
use FileCompressedAppend;
use File::Basename;
use File::Spec;
use File::Temp;
use Submit;

# hou what a vilaine verrue - Ony useful for ProcessHeaderEmblFile 
$main::RH_HEADERS={};

use constant EMBL_FILE_VALIDATOR => 'http://www.ebi.ac.uk/~reddyk/embl-client.jar';

MAIN:
{
	#&TryMain(1);	# WHEN DEBUGGING CODE 
	#&TryMain(0);	# WHEN DEBUGGING CODE
	&TryMain(-1);	# FOR PRODUCTION CODE
};

sub Main
{
	DataReader::DisableInlineC();
	ParamParser::SetDefaultBehaviour('exit_on_getopt_error');
	my $o_param;
	if ( @ARGV==2 && $ARGV[0] !~ /^--/ )
	{
		$o_param = New ParamParser();
		$o_param -> Set ( 'input-ann-file', $ARGV[0] );
		$o_param -> Set ( 'output-ann-file', $ARGV[1] );
	}
	else
	{
		$o_param =  New ParamParser("GETOPTLONG",\&Usage,"species=s","kingdom=s",
									"assembly-file=s","group-label=s","chromosomes=s",
									"input-ann-file=s","input-ann-format=s","input-seq-file=s",
									"output-ann-file=s","output-ann-format=s","output-seq-file=s","with-sequence-embedded","with-partial",
									"stop-after=i","contigs=s","all-contigs","output-utr","output-intron",
									"tmpdir=s","debug=s","codon-usage-table=i",
									"enzyme-validation","build-assembly", "disassemble-assembly","reassemble-assembly=s",
									"load-desc-file=s","load-header-file=s",
									"translational-frameshift=i","create-cds-if-necessary",
									"formats","help","helpensembl","submit=s","submit-pre=s","submit-post=s","tbl2asn",
									"protein-namespace=s",
									"renumber-locus_tags=s","new-rep-origin=i","create-utr-if-necessary",
									"filter-small-utr","filter-bad-proteins=i","generate-id","compute-gaps","keep-gap-size","check-gaps",
									"validate", "no-clean-fasta-header");
	
		if ( $o_param -> IsDefined ( 'helpensembl' ) )
		{
			&UsageEnsembl();
			exit;
		}
	}

	# Translational frameshift (yet undocumentated switch)
	if ( $o_param -> IsDefined ( 'translational-frameshift' ) )
	{
		&Locus::InitAllowedOverlap( $o_param -> Get ( 'translational-frameshift' ) );
	}
	
	# generate-id may be useful to get validated gff3 files
	if ($o_param->IsDefined("generate-id"))
	{
		&Locus::InitGenerateId(&TRUE);
	}
	
	my $o_fmt = New FileFormats;
	if ( $o_param -> IsDefined ( 'formats' ) )
	{
		print $o_fmt -> GetFormattedInfo();
		print "\n";
		print AnnotatedSeq::GetAllowedFormats();
		print "\n";
		exit;
	}
	
	$o_param -> SetUnlessDefined("kingdom","eukaryote");
	$o_param -> AssertAllowedPattern("kingdom",'^(eukaryote|bacteria)$');
	$o_param -> SetUnlessDefined("species", &AnnotatedSeq::UNKNOWN );
	if ( $o_param -> IsDefined ( 'check-gaps' ) && $o_param -> IsDefined ( 'compute-gaps' ) )
	{
		throw Error::Simple("Only one of check-gaps and compute-gaps should be specified");
	}
	if ( $o_param -> IsDefined ( 'assembly-file' ) )
	{
		#$o_param -> AssertFileExists("assembly-file");
		if ( $o_param -> IsDefined ( 'renumber-locus_tags' ) )
		{
			throw Error::Simple("renumber-locus_tags cannot be specified with assembly-file");
		}
		if ( $o_param -> IsDefined ( 'new-rep-origin' ) )
		{
			throw Error::Simple("new-rep-origin cannot be specified with assembly-file");
		}
	}
	else
	{
		for my $prm ( 'group-label', 'random_chromosome', 'chromosomes' )
		{
			if ( $o_param -> IsDefined ( $prm ) )
			{
				throw Error::Simple("$prm cannot be defined without assembly-file");
			}
		}
	}
	
	
	if ( $o_param -> IsDefined ( "new-rep-origin" ) )
	{
		$o_param -> AssertInteger( "new-rep-origin" );
		my $new_rep = $o_param -> Get ( "new-rep-origin" );
		if ( $new_rep == 0 )
		{
			throw Error::Simple("The new replication origin should be a POSITIVE or NEGATIVE (meaning strand=-1) number, NOT 0");
		}
		if ( $o_param -> Get ( 'kingdom' ) ne 'bacteria' )
		{
			throw Error::Simple("Cannot change the replication origin if not a bacteria");
		}
	}
	
	#$o_param -> AssertDefined("input-ann-file");
	if (  !$o_param->IsDefined("input-ann-format") && $o_param -> IsDefined ( "input-ann-file" ) ) 
	{
		my $f   = $o_param->Get("input-ann-file");
		my $fmt = $o_fmt->GuessFormat($f);
		if ( $fmt eq '' )
		{
			throw Error::Simple ( "ERROR - Cannot the guess the format of $f" );
		}
		$o_param -> SetUnlessDefined("input-ann-format",$fmt);
	}

	$o_param -> Set ( "input-ann-format",  uc($o_param -> Get ("input-ann-format") ) ) if ( $o_param->IsDefined("input-ann-format") );
	$o_param -> Set ( "output-ann-format", uc($o_param -> Get ("output-ann-format") ) ) if ( $o_param->IsDefined("output-ann-format") );
	$o_param -> Set ( "input-seq-format", "FASTA" );
	$o_param -> SetUnlessDefined("tmpdir","/tmp");
	$o_param -> SetUnlessDefined("debug","0");
	if ( $o_param -> IsDefined ( 'contigs' ) && $o_param -> IsDefined ( 'all-contigs' ) )
	{
		throw Error::Simple("You CANNOT USE the --contigs SWITCH TOGETHER WITH the --all-contigs SWITCH !");
	}
	if ( $o_param -> IsDefined ( 'stop-after' ) && !$o_param -> IsDefined ( 'all-contigs' ) && !$o_param -> IsDefined ( 'build-assembly' ) )
	{
		throw Error::Simple("You CANNOT USE the --stop-after SWITCH WITHOUT the --all-contigs SWITCH !");
	}
	if ( $o_param -> IsDefined ( 'all-contigs' ) )
	{
		$o_param -> SetUnlessDefined("stop-after",0);
	}
	if ( $o_param -> IsDefined( 'build-assembly' ) && $o_param -> IsDefined ( 'disassemble-assembly' ) )
	{
		throw Error::Simple("You CANNOT USE the switches --build-assembly and --disassemble-assembly together");
	}
	if ( $o_param -> IsDefined( 'build-assembly' ) && $o_param -> IsDefined ( 'reassemble-assembly' ) )
	{
		throw Error::Simple("You CANNOT USE the switches --build-assembly and --reassemble-assembly together");
	}
	if ( $o_param -> IsDefined( 'reassemble-assembly' ) && $o_param -> IsDefined ( 'disassemble-assembly' ) )
	{
		throw Error::Simple("You CANNOT USE the switches --reassemble-assembly and --disassemble-assembly together");
	}
	
	my $codon_usage_table = ($o_param -> IsDefined("codon-usage-table")) ? $o_param -> Get ( "codon-usage-table" ) : undef;
	
	# $debug: 0 ==> no debug
	#         1 ==> tell the user when you start reading/writing etc
	#         2 ==> do not use the AccWarn object, this will generate many many warnings
	#         3 ==> Call the SetDebug() method of $o_annot, generating still more messages and probably stopping at first error
	my $debug = $o_param -> Get ( "debug" );

	my $o_warnings = undef;
	$o_warnings = New AccWarn( '^(WARNING-?[A-Z]? [^ ]+ )',3,'^(ERROR )') if ( $debug < 2 );

	# Create the submit object
	my $submit;
	if ( $o_param -> IsDefined ( 'submit' ) )
	{
		$o_param->Set('filter-small-utr',1);
		$o_param->SetUnlessDefined('filter-bad-proteins',50);
		my %h_params;
		if ( $o_param -> IsDefined ( 'submit-pre' ) )
		{
			$h_params{ 'pre' } = $o_param -> Get ( 'submit-pre' ) ;
		}
		if ( $o_param -> IsDefined ( 'submit-post' ) )
		{
			$h_params{ 'post' } = $o_param -> Get ( 'submit-post' ) ;
		}
		if ( $o_param -> IsDefined ( 'protein-namespace' ) )
		{
			$h_params{ 'protein-namespace' } = $o_param -> Get ( 'protein-namespace' );
		}
		if ( $o_param -> Get ( 'submit' ) =~ /ToNcbi/i ) 
		{
			eval 'use Submit::ToNcbi';
			die $@ if ( $@ ne '' );
			$submit = New Submit::ToNcbi(%h_params);
		}
		elsif ( $o_param -> Get ( 'submit' ) =~ /ToEbi/i ) 
		{
			eval 'use Submit::ToEbi';
			die $@ if ( $@ ne '' );
			$submit = New Submit::ToEbi(%h_params);
			$o_param->AssertDefined('load-header-file');
			$o_param->AssertNonEmptyFile( $o_param->Get('load-header-file'));
			$o_param->Set('compute-gaps', 1);
		}
		else
		{
			$submit = New Submit();
		}
	}
	else
	{
		$submit = New Submit();
	};
			
	# A hash to store and may be accumulate the statistics
	my %h_stats = ();

	# Create the tmp directory
	my $tmpdir = $o_param -> Get("tmpdir");
	$tmpdir    = File::Temp->newdir("BioFileConverterXXXXXX", DIR => "$tmpdir");

	# The contigs, the assembly, the enzyme validation object
	my ($o_ann_splitter,$o_seq_splitter,$o_enz_val) = (undef,undef,undef);
	if ( $o_param -> IsDefined('enzyme-validation') )
	{
		$o_enz_val = New EnzymeValidation( $debug );
		&Locus::InitEnzymeValidation( $o_enz_val );
	}

	if ( $o_param -> IsDefined("contigs") || $o_param -> IsDefined("all-contigs") || $o_param -> IsDefined ( "build-assembly" ) )
	{
		# split the annotation file and the sequence file
		if ( $o_param -> IsDefined( 'input-ann-file') )
		{
			$o_ann_splitter = &SplitterFactory( $o_param -> Get ("input-ann-file"), $o_fmt->GetClassName($o_param->Get("input-ann-format")));
		}
		if ( $o_param -> IsDefined( 'input-seq-file') )
		{
			$o_seq_splitter = &SplitterFactory( $o_param -> Get ("input-seq-file"),'Fasta');
		};
	}

	# The list of output files and formats
	my @a_out_ann_files = split ( ',', $o_param -> Get ( 'output-ann-file' ) );
	my @a_out_ann_fmt;
	my $out_seq_file = $o_param -> Get ( 'output-seq-file' );
	

	if ( $o_param -> IsDefined ( 'output-ann-format' ) )
	{
		@a_out_ann_fmt = split ( ',', $o_param -> Get ( 'output-ann-format' ) );
		if ( $#a_out_ann_files != $#a_out_ann_fmt )
		{
			throw Error::Simple("There are " . @a_out_ann_files . " output annotation files, but " . @a_out_ann_fmt . " output format specifications");
		}
	}
	else
	{
		foreach my $f (@a_out_ann_files) 
		{
			my $fmt = $o_fmt -> GuessFormat($f);
			if ( $fmt eq '' )
			{
				throw Error::Simple ( "ERROR - Cannot the guess the format of $f" );
			}
			else
			{
				push @a_out_ann_fmt, $fmt;
			}
		}
	}

	# If there is FASTA among the formats, arrange to treat FASTA first
	# Fragment proteins are detected while dumping fasta files, and this is dumped to other formats (ie gbk/embl)
	for ( my $i=0; $i < @a_out_ann_fmt; $i++ )
	{
		my $fmt = $a_out_ann_fmt[$i];
		if ($fmt eq 'FASTA' && $i > 0)
		{
			my $sav_fmt = $a_out_ann_fmt[0];
			my $sav_fil = $a_out_ann_files[0];
			$a_out_ann_fmt[0]   = $a_out_ann_fmt[$i];
			$a_out_ann_files[0] = $a_out_ann_files[$i];
			$a_out_ann_fmt[$i]   = $sav_fmt;
			$a_out_ann_files[$i] = $sav_fil;
		}
	}

	# Prepare the parameters for the procedures to call
	my %h_params = ( o_param => $o_param, tmpdir => $tmpdir, o_submit => $submit, o_fmt => $o_fmt,o_warnings => $o_warnings, 
						ra_out_ann_files => \@a_out_ann_files, ra_out_ann_fmt => \@a_out_ann_fmt, out_seq_file => $out_seq_file, 
						rh_stats => \%h_stats,o_ann_splitter => $o_ann_splitter, o_seq_splitter => $o_seq_splitter );

	# Several contigs, with assembly
	#   - create a new AnnotatedSeq
	#   - For each contig:
	#        - Load the AnnotatedSeq with the contig
	#   - At the end, dump the AnnotatedSeq to the output files
	#
	if ( $o_param->IsDefined("build-assembly") )
	{
		&ProcessSeveralContigsWithAssembly(%h_params);
	}

	# Several contigs, without assembly
	#   - For each contig:
	#       - create a new AnnotatedSeq
	#       - Load it with the contig
	#       - Dump it to the output files, in append mode (except for the first contig)
	#
	elsif ( $o_param->IsDefined("all-contigs") || $o_param->IsDefined("contigs") )
	{
		&ProcessSeveralContigs(%h_params);
	}

	# Load only first contig, hoping there is only one !
	else
	{
		&ProcessOneContig(%h_params);
	}

	# Printing the statistics about this file
	print "\n\n";
	&PrintStat(\%h_stats);   	# Printing some stats about the read file
	&PrintEcStat($o_enz_val);	# Adding some stats about the enzyme validation
	
	if ( $o_param->IsDefined('tab2asn') )
	{
		&ProcessTbl2Asn($o_param);
	}
	
	&ProcessHeaderEmblFile($o_param,\@a_out_ann_files,\@a_out_ann_fmt);
	&ProcessHeaderGbkFile($o_param,\@a_out_ann_files,\@a_out_ann_fmt);

	if ($o_param->IsDefined('validate'))
	{
		if ( ! $o_param -> IsDefined ( 'submit' ) )
		{
			print STDERR "WARNING - The validate parameter is useful ONLY with submit\n";
		}
		elsif ( $o_param->Get('submit') eq 'ToEbi' )
		{
			&ValidateEmbl(\@a_out_ann_files,\@a_out_ann_fmt);
		}
		else
		{
			print STDERR "WARNING - The validate parameter is USELESS with --submit ToEBI\n";
		}
	}
	
	# Print the warnings
	&PrintWarnings( $o_warnings, $o_param );

}

sub ValidateEmbl
{
	my ($ra_out_ann_files,$ra_out_ann_fmt) = @_;
	
	my $validator_prog = &EMBL_FILE_VALIDATOR;
	`wget $validator_prog`;
	for ( my $i=0; $i < @$ra_out_ann_fmt; $i++ )
	{
		my $fmt = $ra_out_ann_fmt -> [$i];
		if ($fmt eq 'EMBL')
		{
			my $file_to_validate = $ra_out_ann_files -> [$i];
			`java -classpath embl-client.jar uk.ac.ebi.client.EnaValidator -r $file_to_validate;`
		}
	}
}

=head2 procedure ProcessHeaderEmblFile
 Title     : ProcessHeaderEmblFile
 Usage     : &ProcessHeaderEmblFile($o_param,\@a_out_ann_files);
 Procedure : THIS IS A VERY PAS BELLE VERRUE !
             Read the header the global hash and correct the header
             il y a des trucs qu'on ne peut/sait pas faire avec bioperl !
 Args      : See the call 

=cut
sub ProcessHeaderEmblFile
{
	my ($o_param,$ra_out_ann_file,$ra_out_ann_fmt) = @_;
	my $rh_headers = $main::RH_HEADERS;
	
	return if ( !defined ($rh_headers->{'data_class'}) && 
	            !defined ($rh_headers->{'release_date'})   &&
	            !defined ($rh_headers->{'project'}) );
	
	warn "WARNING HEADERS Cleaning and re-writing some headers\nSee ftp://ftp.ebi.ac.uk/pub/databases/embl/doc/usrman.txt, specially part 3.3\n";

	for ( my $i=0; $i<=@$ra_out_ann_fmt;++$i)
	{
		next unless ( $ra_out_ann_fmt->[$i] eq 'EMBL' );
		my $infile = $ra_out_ann_file->[$i];
		my $outfile= $infile.'.new';
		my $fh_input = &GetStreamIn ( $infile );
		my $fh_output= &GetStreamOut( $outfile );
		my $wgs='.';
		my $rn_flg=0;
		my $rt_flg=0;
		my $rl_flg=0;
		while ( my $line = <$fh_input> )
		{
			if ( $line =~ /^ID/ )
			{
				chomp $line;
				my @a_line = split ';', $line;
				if ( $o_param->IsDefined('submit'))
				{
					$a_line[0] = 'ID   XXX';
				}
				if (defined ($rh_headers->{'data_class'}))
				{
					$a_line[4] = ' ' . $rh_headers->{'data_class'};
					if ($rh_headers->{'data_class'} eq 'WGS')
					{
						$wgs = 'WGS.';
					};
				}
				$line = join ';',@a_line;
				print $fh_output "$line\n";
				if ( defined $rh_headers->{'release_date'})
				{
					print $fh_output "XX\n";
					print $fh_output "ST * ".$rh_headers->{'release_date'}."\n";
					print $fh_output "XX\n";
				}
			}
			elsif ( $line =~ /^AC/ )
			{
				if ( $o_param->IsDefined('submit') && $o_param->Get('submit') eq "ToNcbi")
				{
					$line =~ s/AC   /AC * /;
				}
				print $fh_output $line;
				if ( defined $rh_headers->{'project'})
				{
					print $fh_output "XX\n";
					print $fh_output "PR   Project:" . $rh_headers->{'project'} . ";\n";
					print $fh_output "XX\n";
				}
			}
			elsif ( $line =~ /^DE/ )
			{
				print $fh_output $line;
				print $fh_output "KW   $wgs\n";
			}
			elsif ( $line =~ /^RN/ )
			{
				$rn_flg=1;
				print $fh_output $line;
			}
			elsif ( $line =~ /^RT/ )
			{
				$rt_flg=1;
				print $fh_output $line;
			}
			elsif ( $line =~ /^RL/ )
			{
				$rl_flg=1;
				print $fh_output $line;
			}
			elsif ($line =~ /^FT\s+gap\s+(\d+)$/)
			{
				my $position = $1;
				next if (! ($o_param->IsDefined('submit') && 
			                $o_param->Get('submit') =~ /ToEbi/i ));
				# 1-length gaps written with only one position are not allowed by the EBI validator
				# => Replace "FT   gap   pos"  by "FT   gap   pos..pos"
				$line =~ s/$position/$position..$position/;
				print $fh_output $line;
			}
			else
			{
				print $fh_output $line;
			}
		}
	
		$fh_input->close();
		$fh_output->close();
		unlink $infile;
		link $outfile, $infile;
		unlink $outfile;
		
		if ( $o_param->IsDefined('submit') )
		{
			if ($rn_flg==0)
			{
				warn "WARNING RN line (reference number), this is required for a submission to embl";
			}
			if ($rl_flg==0)
			{
				warn "WARNING RL line (reference location), this is required for a submission to embl";
			}
			if ($rt_flg==0)
			{
				warn "WARNING RT line (reference title), this is required for a submission to embl";
			}
		}
	}
}

=head2 procedure ProcessHeaderGbkFile
 Title     : ProcessHeaderGbkFile
 Usage     : &ProcessHeaderGbkFile($o_param,\@a_out_ann_files);
 Procedure : THIS IS A VERY PAS BELLE VERRUE !
             Read the header the global hash and correct the header
             il y a des trucs qu'on ne peut/sait pas faire avec bioperl !
 Args      : See the call 

=cut
sub ProcessHeaderGbkFile
{
	my ($o_param,$ra_out_ann_file,$ra_out_ann_fmt) = @_;
	
	# rien a faire pour l'instant
	return;

	my $rh_headers = $main::RH_HEADERS;

	for ( my $i=0; $i<=@$ra_out_ann_fmt;++$i)
	{
		next unless ( $ra_out_ann_fmt->[$i] eq 'GENBANK' );
		my $infile = $ra_out_ann_file->[$i];
		my $outfile= $infile.'.new';
		my $fh_input = &GetStreamIn ( $infile );
		my $fh_output= &GetStreamOut( $outfile );
		while ( my $line = <$fh_input> )
		{
			print $fh_output $line;
		}

		$fh_input->close();
		$fh_output->close();
		unlink $infile;
		link $outfile, $infile;
		unlink $outfile;
	}
}

=head2 procedure ProcessOneContig
 Title     : ProcessOneContig
 Prerequisite : A lot of objects should be already alive
 Usage     : &ProcessOneContig(%h_params)
 Procedure : Called when only the FIRST CONTIG of is processed (generally when there is only 1 contig to process)
 Args      : See the call 
 NOTE - Please pass to this procedure ALL THE ARGUMENTS, however several may be undef

=cut
sub ProcessOneContig
{
	my %h_params = @_;
	my ($o_param,$tmpdir,$o_submit,$o_warnings,$ra_out_ann_files,$ra_out_ann_fmt,$out_seq_file,$rh_stats) =
	@h_params{'o_param','tmpdir','o_submit','o_warnings','ra_out_ann_files','ra_out_ann_fmt','out_seq_file','rh_stats'};
	my $debug = $o_param -> Get ( "debug" );
	my $codon_usage_table = ($o_param -> IsDefined("codon-usage-table")) ? $o_param -> Get ( "codon-usage-table" ) : undef;

	my $input_ann_file = $o_param -> Get("input-ann-file");
	my $tmp_flg = &FALSE;
	if ( $input_ann_file =~ /,/ )
	{
		my @a_filenames = split ( ',',$input_ann_file );
		my $o_fmt = New FileFormats();
		my $fmt    = $o_fmt -> GuessFormat ($a_filenames[0]);
		my $suffix = $o_fmt -> GetExtension ($fmt);
		my $fh_out = new File::Temp( UNLINK => 0, SUFFIX => $suffix );
		foreach my $src_fn (@a_filenames)
		{
			my $fh_in  = &GetStreamIn ( $src_fn );
			while ( my $line = <$fh_in> )
			{
				print $fh_out $line;
			}
		}
		$input_ann_file = $fh_out -> filename();
		$tmp_flg = &TRUE;
	}
	
	my $o_assembly = &AssemblyFactory($o_param);
	my $o_reassemble_assembly = undef;
	if ( $o_param -> Get ( 'reassemble-assembly' ) )
	{
		$o_reassemble_assembly = &AssemblyFactory($o_param,undef,$o_param -> Get ( 'reassemble-assembly' ));
	}
	my $o_ann_seq = New AnnotatedSeq( 
						kingdom       => $o_param -> Get("kingdom"),
						species       => $o_param -> Get("species"),
						-chromosomes  => $o_param -> Get("chromosomes"),
						-o_assembly   => $o_assembly,
						-disassemble_assembly => $o_param -> IsDefined ( 'disassemble-assembly' ) ? 1 : 0,
						-o_reassemble_assembly=> $o_reassemble_assembly,
						-in_ann_file  => $input_ann_file,
						-in_ann_fmt   => $o_param -> Get("input-ann-format"),
						-in_seq_file  => $o_param -> Get("input-seq-file"),
						-in_seq_fmt   => $o_param -> Get("input-seq-format"),
						-with_sequence_embedded  => $o_param -> IsDefined ( 'with-sequence-embedded' ) ? 1 : 0,
						-with_partial  => $o_param -> IsDefined ( 'with-partial' ) ? 1 : 0,
						-codon_usage_table => $codon_usage_table,
						-tmpdir       => $tmpdir,
						-create_cds   => $o_param -> IsDefined ( 'create-cds-if-necessary' ),
						-desc_file    => $o_param -> Get ( 'load-desc-file' ),
						-header_file  => $o_param -> Get ( 'load-header-file' ),
						-submit       => $o_param -> IsDefined ( 'submit' ),
						-renumber_lt  => $o_param -> IsDefined ( 'renumber-locus_tags' ) ? $o_param -> Get ( 'renumber-locus_tags' ) : undef,
						-new_rep_origin    => $o_param -> IsDefined ( 'new-rep-origin' ) ? $o_param -> Get ( 'new-rep-origin' ) : undef,
						-filter_small_utr  => $o_param -> IsDefined ( 'filter-small-utr' ) ? 1 : 0,
						-filter_bad_proteins => $o_param -> Get ( 'filter-bad-proteins' ),
						-create_utr   => $o_param -> IsDefined ( 'create-utr-if-necessary' ) ? 1 : 0,
						-compute_gaps => $o_param -> IsDefined ( 'compute-gaps' ) ? 1 : 0,
						-keep_gap_size => $o_param -> IsDefined ( 'keep-gap-size' ) ? 1 : 0,
						-check_gaps   => $o_param -> IsDefined ( 'check-gaps' ) ? 1 : 0,
						-output_utr    => $o_param -> IsDefined ( 'output-utr' ) ? 1 : 0,
						-output_intron => $o_param -> IsDefined ( 'output-intron' ) ? 1 : 0,
						-noclean_fasta_header => $o_param -> IsDefined ( 'no-clean-fasta-header' ) ? 1 : 0,
						);

	# Reading input file(s) and accumulating the stats
	$o_ann_seq -> SetDebug()                  if ( $debug > 2 );
	print STDERR "Reading the input file(s)\n" if ( $debug > 1 );
	&LoadAll ( $o_ann_seq, $o_submit, $o_warnings, &TRUE );
	
	# Dumping output files
	print STDERR "Writing the output file\n" if ( $debug > 1);
	&DumpAll( $o_ann_seq, $ra_out_ann_files, $ra_out_ann_fmt, $out_seq_file );
	
	# Accumulating the stats (after dump because some things may change when dumping, see Locus::Mrna::__CreateCDS
	&AccStat ( $o_ann_seq,$rh_stats);
	
	# cleaning temporary file if necessary
	unlink ($input_ann_file) if ( $tmp_flg == &TRUE );
}

=head2 procedure ProcessSeveralContigs
 Title     : ProcessSeveralContigs
 Prerequisite : A lot of objects should be already alive
 Usage     : &ProcessSeveralContigs(%h_params)
 Procedure : Called when SEVERAL CONTIGS of the input file are processed WITHOUT any assembly
 Args      : See the call 
 NOTE - Please pass to this procedure ALL THE ARGUMENTS, however several may be undef

=cut
sub ProcessSeveralContigs
{
	my %h_params = @_;
	my ($o_param,$tmpdir,$o_submit,$o_warnings,$o_fmt,$ra_out_ann_files,$ra_out_ann_fmt,$out_seq_file,$rh_stats,$o_ann_splitter,$o_seq_splitter) =
	@h_params{'o_param','tmpdir','o_submit','o_warnings','o_fmt','ra_out_ann_files','ra_out_ann_fmt','out_seq_file','rh_stats','o_ann_splitter','o_seq_splitter'};
	my $debug = $o_param -> Get ( "debug" );
	my $codon_usage_table = ($o_param -> IsDefined("codon-usage-table")) ? $o_param -> Get ( "codon-usage-table" ) : undef;

	my @a_id_contigs;
	if ( $o_param -> IsDefined("contigs") )
	{
		my $contigs = $o_param -> Get ( "contigs" );
		@a_id_contigs = split ( ',',$contigs );
	}
	
	elsif ( $o_param -> IsDefined("all-contigs") )
	{
		# the contigs are those discovered by the annotation splitter
		my $ra_id_contigs;
		if ( defined $o_ann_splitter )
		{
			$ra_id_contigs = $o_ann_splitter -> GetContigNames();
		}
		elsif ( defined $o_seq_splitter )
		{
			$ra_id_contigs = $o_seq_splitter -> GetContigNames();
		}
		else
		{
			throw Error::Simple("ERROR - Cannot find any contig !");
		}
		@a_id_contigs = @$ra_id_contigs;
	}

	my $o_assembly = &AssemblyFactory($o_param);
	my $o_reassemble_assembly = undef;
	if ( $o_param -> Get ( 'reassemble-assembly' ) )
	{
		$o_reassemble_assembly = &AssemblyFactory($o_param,undef,$o_param -> Get ( 'reassemble-assembly' ));
	}

	# Process the file for every contig found in the annotation
	my $nb_contigs = @a_id_contigs;
	my $stop_after = $o_param -> Get ( "stop-after" );
	my $i = 0;
	$|=1;
	my @a_compressed_files;
	my $o_ann_seq = undef;
	foreach my $id_contig ( @a_id_contigs )
	{
		my ($input_ann_file, $input_seq_file) = ( '', '' );
		if ( defined $o_ann_splitter )
		{
			$input_ann_file = File::Spec->catfile($tmpdir,$id_contig.'.'.$o_fmt->GetExtension($o_param -> Get("input-ann-format")));
			$o_ann_splitter -> CopyContig( $input_ann_file, $id_contig );
		}
		if ( defined $o_seq_splitter )
		{
			$input_seq_file = File::Spec->catfile($tmpdir,$id_contig.'.'.$o_fmt->GetExtension($o_param -> Get("input-seq-format")));
			$o_seq_splitter -> CopyContig( $input_seq_file, $id_contig );
		}

		# The assembly must be set to the good chromosome BEFORE the object creation !
		if ( defined $o_assembly )
		{
			$o_assembly -> SetCurrentChromosome( $id_contig );
		}

		# NOTE - The -contig switch is NOT used, it is easier and better using the DetectContig method of each 
		#        reader object
		my $o_ann_seq = New AnnotatedSeq( 
							kingdom       => $o_param -> Get("kingdom"),
							species       => $o_param -> Get("species"),
							-disassemble_assembly => $o_param -> IsDefined ( 'disassemble-assembly' ) ? 1 : 0,
							-o_reassemble_assembly=> $o_reassemble_assembly,
							-o_assembly   => $o_assembly,
							-in_ann_file  => $input_ann_file,
							-in_ann_fmt   => $o_param -> Get("input-ann-format"),
							-in_seq_file  => $input_seq_file,
							-in_seq_fmt   => $o_param -> Get("input-seq-format"),
							-with_sequence_embedded  => $o_param -> IsDefined ( 'with-sequence-embedded' ) ? 1 : 0,
							-with_partial  => $o_param -> IsDefined ( 'with-partial' ) ? 1 : 0,
							-codon_usage_table => $codon_usage_table,
							-tmpdir       => $tmpdir,
							-create_cds   => $o_param -> IsDefined ( 'create-cds-if-necessary' ),
							-desc_file    => $o_param -> Get ( 'load-desc-file' ),
							-header_file  => $o_param -> Get ( 'load-header-file' ),
							-submit       => $o_param -> IsDefined ( 'submit' ),
							-renumber_lt  => $o_param -> IsDefined ( 'renumber-locus_tags' ) ? $o_param -> Get ( 'renumber-locus_tags' ) : undef,
							-new_rep_origin    => $o_param -> IsDefined ( 'new-rep-origin' ) ? $o_param -> Get ( 'new-rep-origin' ) : undef,
							-filter_small_utr  => $o_param -> IsDefined ( 'filter-small-utr' ) ? 1 : 0,
							-filter_bad_proteins => $o_param -> Get ( 'filter-bad-proteins' ),
							-create_utr   => $o_param -> IsDefined ( 'create-utr-if-necessary' ) ? 1 : 0,
							-compute_gaps => $o_param -> IsDefined ( 'compute-gaps' ) ? 1 : 0,
							-keep_gap_size => $o_param -> IsDefined ( 'keep-gap-size' ) ? 1 : 0,
							-check_gaps   => $o_param -> IsDefined ( 'check-gaps' ) ? 1 : 0,
							-output_utr    => $o_param -> IsDefined ( 'output-utr' ) ? 1 : 0,
							-output_intron => $o_param -> IsDefined ( 'output-intron' ) ? 1 : 0,
							-noclean_fasta_header => $o_param -> IsDefined ( 'no-clean-fasta-header' ) ? 1 : 0,
							);
		$i++;							
		print "\rContig $i" . "/" . "$nb_contigs ($id_contig)";

		# Reading input file(s) and accumulating the stats
		$o_ann_seq -> SetDebug()                   if ( $debug > 2 );
		print STDERR "Reading the input file(s)\n" if ( $debug > 1 );

		&LoadAll ( $o_ann_seq, $o_submit, $o_warnings, &TRUE );

		# Dumping output files, appending data unless at first iteration
		print STDERR "Writing the output file\n" if ( $debug );
		my $append = ($i==1)?&FALSE:&TRUE;
		&DumpAll( $o_ann_seq, $ra_out_ann_files, $ra_out_ann_fmt, $out_seq_file, $append );

		# Accumulating the stats (after dump because some things may change when dumping, see Locus::Mrna::__CreateCDS
		&AccStat ( $o_ann_seq,$rh_stats);

		# At first iteration: Keep a ref to the FileCompressedAppend objects, delaying the compression phase
		push @a_compressed_files, @{$o_ann_seq->GetDumped()} if ( $i==1);
		
		last if ($stop_after ne '' && $i == $stop_after);
	}
	# We are leaving the scope of @a_compressed_files, so that each file pushed here is now compressed if necessary
}

=head2 procedure ProcessSeveralContigsWithAssembly
 Title     : ProcessSeveralContigsWithAssembly
 Prerequisite : A lot of objects should be already alive
 Usage     : &ProcessSeveralContigsWithAssembly(%h_params)
 Procedure : Called when SEVERAL CONTIGS of the input file are processed WITH assembly (switch --build-assembly)
 Args      : See the call 
 NOTE - Please pass to this procedure ALL THE ARGUMENTS, however several may be undef

=cut
sub ProcessSeveralContigsWithAssembly 
{
	my %h_params = @_;
	my ($o_param,$tmpdir,$o_submit,$o_warnings,$o_fmt,$ra_out_ann_files,$ra_out_ann_fmt,$out_seq_file,$rh_stats,$o_ann_splitter,$o_seq_splitter) =
	@h_params{'o_param','tmpdir','o_submit','o_warnings','o_fmt','ra_out_ann_files','ra_out_ann_fmt','out_seq_file','rh_stats','o_ann_splitter','o_seq_splitter'};
	my $debug = $o_param -> Get ( "debug" );
	my $codon_usage_table = ($o_param -> IsDefined("codon-usage-table")) ? $o_param -> Get ( "codon-usage-table" ) : undef;

	my $o_assembly;
	
	my $ra_chromosomes;
	if ($o_param->IsDefined ( 'chromosomes') )
	{
		# This will work with Assembly::Contigmd AND Assembly::Agp objects
		# Assembly::Contigmd objects support ONLY ONE chromosome
		my @a_chromosomes = split ',',$o_param->Get( 'chromosomes' );
		$ra_chromosomes = \@a_chromosomes;
		$o_assembly = &AssemblyFactory ( $o_param, $ra_chromosomes -> [0] );
	}
	else
	{
		# This will work ONLY with Assembly::Agp objects
		$o_assembly = &AssemblyFactory ( $o_param, '' );
		$ra_chromosomes = $o_assembly->GetChromosomeLabels();
	}

	###&Locus::InitAssembly( $o_assembly );

	my $j=0;
	my $nb_chroms = @$ra_chromosomes;
	foreach my $chromosome (@$ra_chromosomes)
	{
		$j++;
		$o_assembly->SetCurrentChromosome($chromosome);

		# Convert the file for every contig found in the annotation
		# Allocate an AnnotatedSeq
		my $o_ann_seq = New AnnotatedSeq( 
							kingdom       => $o_param -> Get("kingdom"),
							species       => $o_param -> Get("species"),
							-with_sequence_embedded  => $o_param -> IsDefined ( 'with-sequence-embedded' ) ? 1 : 0,
							-with_partial  => $o_param -> IsDefined ( 'with-partial' ) ? 1 : 0,
							-o_assembly   => $o_assembly,
							-codon_usage_table => $codon_usage_table,
							-tmpdir       => $tmpdir,
							-create_cds   => $o_param -> IsDefined ( 'create-cds-if-necessary' ),
							-desc_file    => $o_param -> Get ( 'load-desc-file' ),
							-header_file  => $o_param -> Get ( 'load-header-file' ),
							-submit       => $o_param -> IsDefined ( 'submit' ),
							-filter_small_utr  => $o_param -> IsDefined ( 'filter-small-utr' ) ? 1 : 0,
							-filter_bad_proteins => $o_param -> Get ( 'filter-bad-proteins' ),
							-create_utr   => $o_param -> IsDefined ( 'create-utr-if-necessary' ) ? 1 : 0,
							-build_assembly => &TRUE,
							-output_utr    => $o_param -> IsDefined ( 'output-utr' ) ? 1 : 0,
							-output_intron => $o_param -> IsDefined ( 'output-intron' ) ? 1 : 0,
							-noclean_fasta_header => $o_param -> IsDefined ( 'no-clean-fasta-header' ) ? 1 : 0,
							
						);
	
		my $nb_contigs = @{$o_assembly -> GetContigNames()};
		
		#use Data::Dumper;
		#my $ra_1 = $o_assembly -> GetContigNames();
		#print Dumper($ra_1)."\n";
		#print "======\n";
		#my $ra_2 = $o_ann_splitter -> GetContigNames();
		
		#print Dumper($ra_2)."\n";
		#print "======\n";

		my $i = 0;
		$|=1;
		
		# Consider the first contig and load the genomicsequence, ie the first gap, if any
		$o_assembly -> SetFirstContig ();
		$o_ann_seq -> LoadGenomicSequence();
		while ( $o_assembly->NextContig() ne 'end' )
		{
			my $id_contig = $o_assembly -> GetCurrentContigName();
			$i++;
			print "\rChromosome $j" . "/" . $nb_chroms . " ($chromosome) Contig $i" . "/" . "$nb_contigs ($id_contig)";
	
			my ($input_ann_file, $input_seq_file) = ( '', '' );
			if ( defined $o_ann_splitter )
			{
				$input_ann_file = File::Spec->catfile($tmpdir,$id_contig.'.'.$o_fmt->GetExtension($o_param -> Get("input-ann-format")));
				try
				{
					$o_ann_splitter -> CopyContig( $input_ann_file, $id_contig );
				}
				catch Error with
				{
					$input_ann_file = "";
				};
			}
			if ( defined $o_seq_splitter )
			{
				$input_seq_file = File::Spec->catfile($tmpdir,$id_contig.'.'.$o_fmt->GetExtension($o_param -> Get("input-seq-format")));
				$o_seq_splitter -> CopyContig( $input_seq_file, $id_contig );
			}
			
			next if ( $input_ann_file eq '' && $input_seq_file eq '');
			
			# $set to FALSE if the contig is NOT in the annotation file (this is quite possible)
			my $annotated_ctg = &TRUE;
			try
			{
				$o_ann_splitter -> CopyContig( $input_ann_file, $id_contig );
			}
			catch Error with
			{
				print STDERR "$id_contig NOT FOUND in the annotation" if ( $debug > 1 );
				$annotated_ctg = &FALSE;
			};
			
			$o_ann_seq -> SetInput (
								-in_ann_file  => $input_ann_file,
								-in_ann_fmt   => $o_param -> Get("input-ann-format"),
								-in_seq_file  => $input_seq_file,
								-in_seq_fmt   => $o_param -> Get("input-seq-format"),
								-contig       => $id_contig
								);
								
			# If the contig is not annotated, we just have to load the genomic sequence
			if ($annotated_ctg == &FALSE)
			{
				$o_ann_seq -> LoadGenomicSequence();
				next;
			}
	
			# Reading input file(s) and accumulating the stats
			$o_ann_seq -> SetDebug()                   if ( $debug > 2 );
			print STDERR "Reading the input file(s)\n" if ( $debug > 1 );
			&LoadAll ( $o_ann_seq, $o_submit, $o_warnings, &TRUE, &FALSE );
		}
		
		# Calling post loading tasks after everything was loaded and assembled
		$o_ann_seq -> PostLoadingTasks();
	
		# Dumping annotation file(s)
		my $append = ($j==1)?&FALSE:&TRUE;
		print STDERR "Writing the output file\n" if ( $debug > 1);
		&DumpAll( $o_ann_seq, $ra_out_ann_files, $ra_out_ann_fmt, $out_seq_file, $append );
	
		# Accumulating the stats (after dump because some things may change when dumping, see Locus::Mrna::__CreateCDS
		&AccStat ( $o_ann_seq,$rh_stats);
	}
}

=head2 procedure LoadAll
 Title     : LoadAll
 Prerequisite : The AnnotatedSeq object should be loaded
 Usage     : &LoadAll( $o_annot )
 Function  : Use the Load function of $o_annot to load annotation and genomic data
             If an exception is generated during the Load: print the warnings and exit
             If an exception is generated during the LoadGenomicSequence: ignore it (this may be normal, 
             if the input files did not get genomic data)
 Args      : $o_annot               An AnnotatedSeq object
             $o_submit              A Submit objectoutput a circular genome genbank
             $o_warnings (optional) An AccWarn object
			 $genomic    (optional) If &TRUE, load also the genomic sequence
			 $postloading (optional, default &TRUE) If &TRUE, call PostLoadingTasks, then ApplyRules and GetHeaders

=cut
sub LoadAll
{
	my ($o_annot, $o_submit, $o_warnings, $genomic, $postloading) = @_;
	$genomic = &FALSE unless ( defined $genomic );
	$postloading = &TRUE unless (defined $postloading);
	if ( $genomic == &TRUE )
	{
		try
		{
			$o_annot -> LoadGenomicSequence();
		}
		catch Error with
		{};
	}

	try
	{
		$o_annot -> Load();
		if ($postloading == &TRUE)
		{
			$o_annot -> PostLoadingTasks();
			$o_submit->ApplyRules($o_annot);
			
			# hou what a very dirty trick - Only useful for ProcessHeaderEmblFile 
			$main::RH_HEADERS = $o_annot->GetHeaders();
		}
	}
	catch Error with
	{
		my ($o_error) = @_;
        if ($Error::Debug == 1)
        {
            print STDERR $o_error->stacktrace();
			print STDERR "ERROR - FILE  = " . $o_error->file . "\n";
			print STDERR "        LINE  = " . $o_error->line . "\n";
			print STDERR "        VALUE = " . $o_error->value . "\n" if ( defined $o_error->value );
			print STDERR "        TEXT  = " . $o_error->text . "\n";
        }
        else
        {
			print STDERR "\nERROR WHILE LOADING DATA - ";
			print STDERR $o_error -> text() . "\n";
		}
        &PrintWarnings ( $o_warnings );
        exit 1;
	};
}

=head2 procedure DumpAll
 Title     : DumpAll
 Prerequisite : The AnnotatedSeq object should be loaded
 Usage     : &DumpAll( $o_annot, \@a_out_ann_files, \@a_out_ann_fmts, $out_seq_file, $append)
 Function  : Use the Dump function of $o_annot to dump data to various annotation or sequence files
 Args      : $o_annot An AnnotatedSeq object
             $ra_out_ann_files An array of file names
             $ra_out_ann_fmt   An array of formats
             $out_seq_file     A sequence file name (only 1)
             $append (default 0) It &TRUE, call Dump in append mode

=cut
sub DumpAll
{
	my ( $o_annot, $ra_out_ann_files, $ra_out_ann_fmt, $out_seq_file, $append ) = @_;
	$append = &FALSE if ( ! defined ( $append ) );
	
	for (my $i=0; $i<@$ra_out_ann_files; $i++)
	{
		$o_annot -> SetOutput( 
			-out_ann_file => $ra_out_ann_files -> [$i],
			-out_ann_fmt  => $ra_out_ann_fmt   -> [$i]
		);
		$o_annot -> Dump( -append => $append );
	}
	$o_annot -> SetOutput (
		-out_seq_file => $out_seq_file
	);
	$o_annot -> Dump( -append => $append );
}

=head2 procedure PrintWarnings
 Title     : PrintWarnings
 Usage     : &PrintWarnings ( $o_warnings )
 Function  : If $o_warnings is defined, print the accumulated warnings
             Else, return without doing anything
 Args      : $o_warnings, $o_param

=cut

sub PrintWarnings
{
	my ( $o_warnings, $o_param ) = @_;
	if ( defined ($o_param) && $o_param->Get('debug')>0 )
	{
		my $output = "";
		$o_param->Dump('STRING', \$output, "");
		print STDERR '=' x 40 ."\n$0\n". '=' x 40 .$output. "\n";
	}
	if ( defined ( $o_warnings ) )
	{
		my $warnings = $o_warnings -> GetNbWarnings();
		if ( $warnings > 0 )
		{
			print STDERR "\n" . '=' x 40 . "\nTHERE ARE $warnings WARNINGS\n";
			$o_warnings -> DisplayWarnings();
		}
		else
		{
			print STDERR "No warnings.\n";
		}
	}
}

=head2 Functions AssemblyFactory,SplitterFactory
 Title     : AssemblyFactory,SplitterFactory
 Create a new Assembly or Splitter object
 Arg (AssemblyFactory): $o_param, $chromosome, $assembly_file. 
                        $chromosome is optional, if not specified read the chromosome from $o_param
                        $assembly_file is optional, if not specified read the file from $o_param
                        (yes this is a crappy thing, see ProcessSeveralContigsWithAssembly for the details)
 Return                 undef (if no assembly file is specified in $o_param), or an Assembly object
=cut

sub AssemblyFactory
{
	my ($o_param, $chromosome, $assembly_file) = @_;
	$chromosome = $o_param -> Get ( 'chromosomes' ) unless defined ($chromosome);
	$assembly_file = $o_param -> Get ( 'assembly-file' ) unless defined ($assembly_file);
	if ( $assembly_file eq '' )
	{
		return undef;
	}
	
	my $group_label   = $o_param -> Get ( 'group-label' );
	my $random         = $o_param -> IsDefined ( 'random-chromosome' );
	
	my $compression_path = &GetCompressPath( $assembly_file );
	
	my $ext;
	
	# NOTE - The switch assembly_file may be in the form: file1.agp,file2.agp, meaning that the object is built from the concatenation of several agp files
	#        We only check the type of the LAST file, of course if you mix md and agp files it will probably not work
	# file.agp
	if ( $compression_path eq "" )
	{
		($ext) = ($assembly_file =~ /\.(\w+)$/);
	}
	
	# file.agp.gz or seq_contig.md.gz
	else
	{
		($ext) = ($assembly_file =~ /\.(\w+)\.\w+$/);
	}
	
	my $o_assembly_class;
	if ( $ext eq 'md' )
	{	
		$o_assembly_class = "Assembly::Contigmd";
	}
	elsif ( $ext eq 'agp' )
	{
		$o_assembly_class = "Assembly::Agp";
	}
	else
	{
		throw Error::Simple("Extension $ext: Unknown format for the assembly file");
	}
	
	eval "use $o_assembly_class";
	throw Error::Simple( $@ ) if ( $@ ne "" );
	my $o_assembly = New $o_assembly_class ( $assembly_file, $group_label, $chromosome, $random );
	throw Error::Simple( $@ ) if ( $@ ne "" );

	return $o_assembly;
}
	
sub SplitterFactory
{
	my ($file, $format) = @_;
	my $o_splitter_class = "Splitter::" . $format;
	eval "use $o_splitter_class";
	throw Error::Simple( $@ ) if ( $@ ne "" );
	my $o_splitter       = New $o_splitter_class( "$file" );
	throw Error::Simple( $@ ) if ( $@ ne "" );
	return $o_splitter;
}

sub PrintStat
{
	my ($rh_stats) = @_;
	my ($o_ann_seq) = @_;
	my $species      = $rh_stats -> { species };
	my $chrom        = $rh_stats -> { chrom   };
	my $length       = $rh_stats -> { length  };
	my $kingdom      = $rh_stats -> { kingdom };
	my $table        = $rh_stats -> { table   };
	my $nb_genes     = $rh_stats -> { nb_genes };
	my $nb_o_genes   = $rh_stats -> { nb_o_genes };
	my $nb_p_genes   = $rh_stats -> { nb_p_genes };
	my $nb_r_genes   = $rh_stats -> { nb_r_genes };	
	my $nb_mrnas     = $rh_stats -> { nb_mrnas };
	my $nb_cds_regions  = $rh_stats -> { nb_cds_regions };
	my $nb_exon_regions = $rh_stats -> { nb_exon_regions };
	my $nb_ncrnas    = $rh_stats -> { nb_ncrnas};
	my $nb_repeats   = $rh_stats -> { nb_repeats };
	
	print "==========\nSTATISTICS\n==========\n";
	print "Species               = $species\n";
	print "Kingdom               = $kingdom (codon table = $table)\n";
	if ( $chrom ne "" )
	{
		print "Sequence name         = $chrom (length $length)\n";
	}
	print "Number of genes       = $nb_genes";
	print " ($nb_p_genes protein coding genes)\n";
	if ( $nb_p_genes != 0 )
	{
		print "Number of mrna        = $nb_mrnas ";
		printf ("(%6.2f mrna/coding_gene)\n",$nb_mrnas/$nb_p_genes);
	}
	if ( $nb_mrnas != 0 )
	{
		print "Number of cds regions = $nb_cds_regions ";
		printf ("(%6.2f cds regions/mrna)\n",$nb_cds_regions/$nb_mrnas);
		print "Number of exons       = $nb_exon_regions ";
		printf ("(%6.2f exons/mrna)\n",$nb_exon_regions/$nb_mrnas);
	}
	print "Number of repeats     = $nb_repeats\n";
	print "Number of ncrnas      = $nb_ncrnas\n";
	print "\n";
}

sub PrintEcStat
{
	my ($o_enz_val) = @_;
	return unless defined $o_enz_val;
	
	my $rh_validated  = $o_enz_val->GetValidated();
	my $rh_transferred= $o_enz_val->GetTransferred();
	my $rh_deleted    = $o_enz_val->GetDeleted();
	
	my (@a_ec,$cpt);

	print "====================\nVALIDATED EC NUMBERS\n====================\n";
	$cpt = 0;
	@a_ec = keys %$rh_validated;
	if ( @a_ec == 0 )
	{
		print "none\n\n";
	}
	else
	{
		while ( defined(my $ec = shift @a_ec) )
		{
			print "$ec ";
			if ( $cpt++==5 )
			{
				print "\n";
				$cpt = 0;
			}
		}
		print "\n\n";
	}
	
	print "====================\nTRANSFERRED EC NUMBERS\n====================\n";
	$cpt = 0;
	@a_ec = keys %$rh_transferred;
	if ( @a_ec == 0 )
	{
		print "none\n\n";
	}
	else
	{
		while ( defined(my $ec = shift @a_ec) )
		{
			print "$ec->$rh_transferred->{$ec} ";
			if ( $cpt++==5 )
			{
				print "\n";
				$cpt = 0;
			}
		}
	}
	print "\n\n";
	
	print "====================\nDELETED EC NUMBERS\n====================\n";
	$cpt = 0;
	@a_ec = keys %$rh_deleted;
	if ( @a_ec == 0 )
	{
		print "none\n\n";
	}
	else
	{
		while ( defined(my $ec = shift @a_ec) )
		{
			print "$ec ";
			if ( $cpt++==5 )
			{
				print "\n";
				$cpt = 0;
			}
		}
	}
	print "\n\n";
}

sub AccStat
{
	my ($o_ann_seq,$rh_stats) = @_;
	$rh_stats -> { species }  = $o_ann_seq -> GetSpecies();
	if ( !defined ($rh_stats -> { chrom }) )
	{
		$rh_stats -> { chrom } = $o_ann_seq -> GetChromosomeLabel();
	}
	else
	{
		$rh_stats -> { chrom } = "";
	}
	
	if ( $o_ann_seq -> GetLength() != 0 )
	{
		if ( ! defined ($rh_stats -> {length}))
		{
			$rh_stats -> {length} = $o_ann_seq -> GetLength();
		}
		else
		{
			if ( $rh_stats -> {length} ne "unknown" )
			{
				$rh_stats -> {length} += $o_ann_seq -> GetLength();
			}
		}
	}
	else
	{
		$rh_stats -> {length} = "unknown";
	}
	
	$rh_stats -> { kingdom   } = $o_ann_seq -> GetKingdom();
	$rh_stats -> { table     } = $o_ann_seq -> GetCodonTable();
	$rh_stats -> { nb_genes  } += $o_ann_seq -> GetNbGenes();
	my ($nb_o_genes,$nb_p_genes,$nb_r_genes) = $o_ann_seq -> GetNbGenes();
	$rh_stats -> { nb_o_genes  } += $nb_o_genes;
	$rh_stats -> { nb_p_genes  } += $nb_p_genes;
	$rh_stats -> { nb_r_genes  } += $nb_r_genes;
	$rh_stats -> { nb_mrnas  } += $o_ann_seq -> GetNbTranscripts();
	my ($nb_cds,$nb_exons) = $o_ann_seq -> GetNbExons();
	$rh_stats -> { nb_cds_regions }  += $nb_cds;
	$rh_stats -> { nb_exon_regions } += $nb_exons;
	$rh_stats -> { nb_ncrnas } += $o_ann_seq -> GetNbNcrnas();
	$rh_stats -> { nb_repeats} += $o_ann_seq -> GetNbRepeats();
}

sub ProcessTbl2Asn
{
	my ( $o_param) = @_;
	
	print "\n========================================\n";
	print "NOW EXECUTING tbl2asn\n";
	print "==========================================\n\n";
	
	# tbl2asn should be in the path
	my $cmd = 'tbl2asn';
	my @a_output = split ',',$o_param->Get('output-ann-file');
	my $prefix='';
	foreach my $f (@a_output)
	{
		if ( $f =~ /^(.+)\.tbl$/ )
		{
			$prefix = $1;
			last;
		}
	}
	if ( $prefix eq '' )
	{
		throw Error::Simple ( 'ERROR - There is NO .tbl file in the input files !' );
	}
	my $codon_usage_table;
	if ( $o_param -> IsDefined( 'codon-usage-table' ) )
	{
		$codon_usage_table = $o_param -> Get ( 'codon-usage-table' );
	}
	else
	{
		my $kingdom = $o_param -> Get ( 'codon-usage-table' );
		$codon_usage_table = $o_param->Get('kingdom') eq 'bacteria' ? 11 : '';
		$codon_usage_table = "-j '[gcode=$codon_usage_table]'" if ($codon_usage_table ne '');
	}
	
	my @a_time = localtime;
	$a_time[5] += 1900;
	$a_time[4] += 1;
	my $date = $a_time[4].'/'.$a_time[3].'/'.$a_time[5];

	$cmd .= " -i ${prefix}.fna -t ${prefix}.sbt -a s -f ${prefix}.tbl -V vb -k m -H '$date' $codon_usage_table";
	print "Using the command line = $cmd\n";
	my $status = system($cmd);
	if ($? == -1) 
	{
		print "failed to execute: $!\n";
	}
	elsif ($? & 127) 
	{
		printf "tbl2asn died with signal %d, %s coredump\n",
		($? & 127),  ($? & 128) ? 'with' : 'without';
	}
	else
	{
		my $status = $? >> 8;
		if ($status ne '0')
		{
			printf "tbl2asn exited with value %d\n", $status;
		}
		else
		{
			print "tbl2asn OK\n";
		}
	}
	print "========================================\n";
}

sub Usage
{
    my $prgm = basename($0);
    print STDERR <<END

Usage : $prgm 
        REQUIRED:
              --input-ann-file  <file>        Annotation file name
              --output-ann-file <file1,file2> Output annotation files

        THE SEQUENCE:
              --input-seq-file  <file>        Sequence file name (fasta), not always useful
              --output-seq-file <file>        Sequence file (fasta) generated from the input annotation or seq files
              --with-sequence-embedded        If specified and if possible, embed the sequence in the annotaiotn file
                                              Currently used ONLY with gff3 files
              --with-partial                  If specified, fragment proteins are tagged (gff3 only)
                                              this leads to a not valid gff3 file, because Partial is not yet an official property
              --all-contigs                   Consider ALL contigs in file (required if several contigs in the input file)

        RECOMMANDED:
              --species <species>             The species
              --kingdom <eukaryote|bacteria>  Kingdom, def eukaryote

        ASSEMBLY:
              The program may assemble the contigs in a single chromosome, or read some information from the assembly file
              --build-assembly                      Build the assembly from an assembly file and chromosome specifications
              --disassemble-assembly                Re-write the seq and annotation file to several contigs, using the assembly file (ONLY agp)
              --reassemble-assembly <file.agp1>     Disassemble using file1.agp, then use file specified with --assembly-file to build a new assembly
              --assembly-file       <file>          Assembly file: .md or .agp If .agp, you may have several file names (separator = ,)
                                                    If .agp, gap information is read from the agp file and written to the annotation output file (assembly_gap)
              --group-label       <label>           The assembly label to consider in the assembly file (.md only)
              --chromosomes       <chr1,chr2>       The chromosomes to consider in the assembly file. Consider all chromosomes if not specified

        ENZYME VALIDATION:
              --enzyme-validation                   download the enzyme.dat and enzclass.dat files from expasy and update the EC_Numbers
        
        OPTIONAL:
              --codon-usage-table <nb>              Codon usage table for translation, default is 1 for eukaryote, 11 for bacteria
              --contigs           <contig1,contig2> Consider only those contigs
              --stop-after        <nb>              Consider only the first nb contigs (useful for debugging data)
              NOTE - If none of --contigs or --all-contigs is specified, only the FIRST contig on the annotation 
                     file is considered. This is convenient if this is the only contig of the file
              --load-desc-file <file.desc>          read functional annotation from this "desc" file
                                                    WORKS ONLY WITH gff3 FILES !
              --translational-frameshift <1|2>      Accept an overlap of 1 or 2 between regions to take into account translational frameshift
             
              --debug             <0|1|2|3>         Few messages, more messages, a lot of messages

        FILE FORMATS: 
              The program tries to guess the file format from its name, but other formats can be forced
              If you want to read data from the Ensembl! databasae, please try the switch --helpensembl
              --formats                             Print the supported formats, and the corresponding file extensions
              --input-ann-format  <format>          Annotation file format
              --output-ann-format <format1,format2> Output annotation file formats
              
        SUBMITTING TO GENBANK OF OTHER DATABASE:
              Some rules, specific to the database, are executed
              --load-header-file file.cfg>           Load header information from this file
              --submit ToNcbi                        Apply rules for submission to Genbank
              --submit ToEbi                         Apply rules for submission to EMBL
              --submit_pre file.pl                   Execute this code through an eval BEFORE Submit::ApplyRules
              --submit_post file.pl                  Execute this code through an eval AFTER Submit::ApplyRules
              --tbl2asn                              Execute tbl2asn before leaving (useful for genbank submission)
              --protein-namespace xxx                Use this namespace instead of default BFC in the protein-id specifications
              --validate                             Call an external validator (works currently ONLY with ToEbi)
              
        RENUMBERING LOCUS_TAGS, CHANGING REP ORIGIN
              --new-rep-origin  100                  Change the replication origin. If the parameter is <0, change the strand too
              --renumber-locus_tags SMc....[,123[,10]]    Use this template to renumber the locus_tags, .... will be replaced with numbers
                                                     If ,123 is specified, start renumbering from 123
                                                     If ,10 is specified, renumber with step 10 (default = 1)
              
        FILTERING OR CORRECTING DATA
              --filter-small-utr                     Remove small (ie 1 bp) utr and correct exon, mRNA, gene boundaries
              --filter-bad-proteins 50               Declare as inconsistent proteins embedding 'X' in sequence
                                                     Completely remove mRNA if there are more than 50% 'X' in the sequence
              --create-utr-if-necessary              If Exon and CDS do not start/end at same position, create an utr
              --output-utr                           Works only with fasta ann output: Create the utr as above, but also output the utr to a fasta file
              --output-intron                        Works only with fasta ann output: Create the utr as above, but also output the introns to a fasta file
              --no-clean-fasta-header                Works only with fasta ann output: display long sequence identifiers (add the sequence type as prefix)
              --generate-id                          Useful when outputting gff3 file, sometimes necessary to get a gff3 valid file
              --create-cds-if-necessary              If CDS is NOT present in the input file, try to create it (from the exons/UTR)
              --check-gaps                           Check the coherency of the gaps annotations with the sequence
              --compute-gaps                         Clear the gap annotations, and re-create them from the sequence                                       
              --keep-gap-size                        Only useful with --compute-gaps: keep gap size in estimated_length instead of unknown
        
        ASKING FOR HELP:
              --help        Display this help screen
              --help
              --helpensembl Display more documentation about Ensembl
              
END
;
}

sub UsageEnsembl
{
	print STDERR <<END
	
	
---- Reading data from Ensembl ----
	
	Please note the Ensembl API releases AND the database releases must be synchronized. 
	So, it may be a good idea to install the API to your home directory
	
    1/ Installing the Ensembl API:
       see http://www.ensembl.org/info/docs/api/api_cvs.html
	
	1a/ Trying the current installation:
	    Try:   $0 --input-ann-file 'all.ensembl'
	    If this results in an error message (Can't locate Bio/EnsEMBL/Registry.pm in \@INC), the Ensembl API is NOT installed
	    
	1b/ Installing the API rel 66 inside your home directory:
	    cd
	    cvs -d :pserver:anonymous\@cvs.sanger.ac.uk:/cvsroot/ensembl checkout -r branch-ensembl-66 ensembl
	    mv ensembl ensembl66
	    
	1c/ Using Ensembl release 66:
	    perl -I ~/ensembl66/modules/ $0 --input-ann-file 'all.ensembl'
	    
    2/ Using an alternate configuration (other server, other species etc).
       see http://www.ensembl.org/info/docs/api/registry.html
	
	2a/ You may write a configuration file, let's call it ~/test.reg
	    Here is an example of configuration file:
	
                          use strict;
                          use Bio::EnsEMBL::DBSQL::DBAdaptor;

                          new Bio::EnsEMBL::DBSQL::DBAdaptor(
                              -host    => 'ensembl.example.org',
                              -user    => 'anonymous',
                              -species => 'homo_sapiens',
                              -port    => '9306',
                              -dbname  => 'homo_sapiens_core_39_36z'
                           );

                           my \\\@aliases = ( 'H_Sapiens', 'human', 'homo' );
                           Bio::EnsEMBL::Utils::ConfigRegistry->add_alias(
                                -species => "homo_sapiens",
                                -alias   => \\\@aliases
                           );
 
                           1;

	2b/ Set the environment variable called ENSEMBL_REGISTRY:
	    export ENSEMBL_REGISTRY="~/test.reg"
        
	2c/ Do not forget to use the version of the API corresponfing to your database schema (see 1)
    
    3/ Using the Ensembl database to retrieve data:
    
	3a/ Getting the list of the databases available on the server:
	    $0 --input-ann-file 'all.ensembl'
        
	3b/ Getting the list of available chromosomes for the species human:
	    $0 --input-ann-file 'human.all.ensembl'
        
	3c/ Getting data from Ensembl, for the chromosome 21, and storing data to a genbank file:
	    $0 --input-ann-file 'human.21.ensembl' --output-ann-file human.21.gbk
        
END
;
}

