#!/usr/bin/perl

=pod 

=head1 NAME

convert_trnascan_gff3.pl

=head1 SYNOPSIS

convert_trnascan_gff3.pl --infile trnascan-SE.out --outfile trnascan-SE.gff3 [--add_parent gene]

=head1 DESCRIPTION

Convert tRNAScan-SE output (v 1.23) in GFF3 format (http://www.sequenceontology.org/gff3.shtml)
Note: intron information is not integrated. Intron positions are just saved in the attributes field.

=cut

BEGIN
{
	my ($dir,$file) = $0 =~ /(.+)\/(.+)/;
	unshift (@INC,"$dir/../../corelib");
	unshift (@INC,"$dir/../../lib");
}

use strict; 
use warnings;
use Carp;

use ParamParser;
use Logger;
use General;

MAIN:
{
	my $o_param = New ParamParser('GETOPTLONG', "infile=s", "outfile=s", "add_parent=s");

	$o_param->SetUsage(my $usage=sub { &Usage(); } );
	$o_param->SetBehaviour('assert_empty_file_allowed', 1);
	$o_param->AssertFileExists('infile');
	$o_param->AssertDefined('outfile');

	my $infile  = $o_param->Get('infile');
	my $outfile = $o_param->Get('outfile');
	my $f_in    = &General::GetStreamIn($infile); 
	my $f_out   = &General::GetStreamOut($outfile); 

	# Skip the first three lines
	my $line = <$f_in>;
	$line    = <$f_in>;
	$line    = <$f_in>;
	# Print gff3 header
	print $f_out "##gff-version 3\n";

	while ( my $line = <$f_in> )
	{
		chomp($line);
        next if ( $line =~ /^#/ );
		# CELF22B7 	1	12619	12738	Leu	CAA	12657	12692	60.01
		$line  =~ s/\?/X/g;
        my @a_ = split(' ', $line);
		$a_[0] =~ s/ /_/;
		my ($beg, $end, $strand) = ( $a_[2] < $a_[3] ) ? ( $a_[2],$a_[3],'+' ) : ( $a_[3],$a_[2],'-' );
		my $id        = "$a_[0].tRNA.$a_[1]";
		my $intron    = '';
		my $type      = $a_[4];
		my $anticodon = $a_[5];

		# If there is intron information, add the position in the attributes field
		if (  $a_[6] != 0 ) 
		{
			$intron = ( $strand eq '+' ) ? "intron=$a_[6]-$a_[7]" : "intron=$a_[7]-$a_[6]";
		}
		my $more_att = "anti_codon=$anticodon;product=tRNA-$type;$intron";

		if ( $o_param->IsDefined('add_parent') )
		{
			my $feature_type = $o_param->Get('add_parent');
			print $f_out "$a_[0]\ttRNAscan-SE\t$feature_type\t$beg\t$end\t$a_[8]\t$strand\t.\tID=P_$id;Name=$a_[4]_$a_[5];Ontology_term=SO:0001272;$more_att\n";
			print $f_out "$a_[0]\ttRNAscan-SE\ttRNA\t$beg\t$end\t$a_[8]\t$strand\t.\tID=$id;Name=$a_[4]_$a_[5];Parent=P_$id;Ontology_term=SO:0001272;$more_att\n";
		}
		else
		{
			print $f_out "$a_[0]\ttRNAscan-SE\ttRNA\t$beg\t$end\t$a_[8]\t$strand\t.\tID=$id;Name=${id}_$a_[4]_$a_[5];Ontology_term=SO:0001272;$more_att\n";
		}
	}
	$f_in->close;
	$f_out->close;
}

sub Usage
{
print STDERR<<END
$0 
	--help
	--infile    filename   tRNAScan-SE output format (gzipped allowed)
	--outfile   filename   gff3 format 
[Optional]
	--add_parent string    If defined, add systematically a parent element from this type. (column 3 in GFF3) Examples: 'gene', 'ncRNA'

END
;
}
