#!/usr/bin/perl

# Copyright INRA

# Erika.Sallet@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=pod 

=head1 NAME

detect_start_transcription.pl

=head1 SYNOPSIS

$EGNP/bin/int/detect_start_transcription.pl --in GMI11495-Rm2011G.b.est30-31-32.position.gff3.xz 
--seq_len 1683348 --out out --seq_name GMI11495-Rm2011G.b --derivative_threshold 30 --log_threshold 1 --boostscore_factor 2

=head1 DESCRIPTION

Look for breaks of expression in a set of scores.
Read gff3 input and save the result in both gff3 and wig files.


=cut

BEGIN
{
	$ENV{LANG} = "C";
    use File::Basename;
    my $dirprg = dirname($0);
}

use FindBin;
use lib "$FindBin::Bin/../../corelib/"; 
use lib "$FindBin::Bin/../../lib/"; 
use strict;
use warnings;
use Carp;

use ParamParser;
use Logger;
use General;
use GeneralBioinfo;
use Heap::Simple;


my $DEFAULT_WINSIZE                 = 3; # default window size for smooth
my $DEFAULT_THRESHOLD_DERIVATIVE    = 0; # Default mininum value after obtaining the derivative
my $DEFAULT_THRESHOLD_NORMALIZATION = 0; # Default minimum value after the normalization 
my $DEFAULT_BOOST_SCORE             = 1; 
my $DEFAULT_BOXCONTEXT              = 20;
my $DEFAULT_TSTART_RATIO            = 1000;
my $BOX_BONUS = 2;
my $DEBUG     = 0;


MAIN:
{
    my $o_param = New ParamParser(
                                  'GETOPTLONG',  "in=s", "out=s", "format=s", "seq_len=i", "seq_name=s", "log_threshold=s", 
		"derivative_threshold=i", "boostscore_factor=i", "percentile=i", "boost_ig", "win_size=i", 
		"smooth_method=s","no_smooth", "box_in=s", "box_context=i", "strand=s", "tstart_ratio=i"
                                  );
	$o_param->SetUsage(my $usage=sub { &Usage(); } );

	$o_param->AssertFileExists('in');	
	$o_param->AssertDefined('out', 'seq_name');
	$o_param->SetUnlessDefined('format', 'gff3');
	$o_param->SetUnlessDefined('smooth_method','max');
	$o_param->SetUnlessDefined('win_size',             $DEFAULT_WINSIZE);
	$o_param->SetUnlessDefined('log_threshold',        $DEFAULT_THRESHOLD_NORMALIZATION);	
	$o_param->SetUnlessDefined('box_context',          $DEFAULT_BOXCONTEXT);	
	$o_param->SetUnlessDefined('boostscore_factor',    $DEFAULT_BOOST_SCORE);
	$o_param->SetUnlessDefined('derivative_threshold', $DEFAULT_THRESHOLD_DERIVATIVE);
	$o_param->SetUnlessDefined('tstart_ratio',         $DEFAULT_TSTART_RATIO);

	$o_param->AssertInteger('seq_len', 'win_size','derivative_threshold', 'box_context'); #'boost');
	$o_param->AssertFileExists('box_in')  if $o_param->IsDefined('box_in');
	$o_param->AssertInteger('percentile') if $o_param->IsDefined('percentile');
	$o_param->AssertAllowedValue('smooth_method','min','max','mean');
	$o_param->AssertAllowedValue('format','gff3','wig');

	my $format               = $o_param->Get('format');
	$o_param->AssertAllowedValue('strand', "+", "-") if ($format eq "wig");
		
	my $derivative_threshold = $o_param->Get('derivative_threshold');	
	my $log_threshold        = $o_param->Get('log_threshold');
	my $seq_len              = $o_param->Get('seq_len');
	my $seq_name             = $o_param->Get('seq_name');
	my $boost_score          = $o_param->Get('boostscore_factor');


	# Read the score in the input file on the two strands
	my @a_plus;
	my @a_minus;
	&ReadScore($o_param->Get('in'), $format, \@a_plus, \@a_minus, $seq_len, $o_param);

	# Smooth the scores
	my @a_smoothed_plus;	
	my @a_smoothed_minus;
	&Smooth(\@a_smoothed_plus, \@a_smoothed_minus, \@a_plus,\@a_minus, $seq_len, $o_param);

	# Compute derivative and of the result is upper to a threshold, normalize this value
	my @a_derivative_plus;
	my @a_derivative_minus;	
	my @a_deriv_norm_plus;
	my @a_deriv_norm_minus;
	&DerivativeNormalize(\@a_smoothed_plus,  \@a_derivative_plus,  
						 \@a_deriv_norm_plus, $seq_len, $derivative_threshold);
	&DerivativeNormalize(\@a_smoothed_minus, \@a_derivative_minus, 
						 \@a_deriv_norm_minus, $seq_len, $derivative_threshold);

	# Compute the threshold according to the percentile if specified
	$log_threshold  = &GetPercentile(\@a_deriv_norm_plus, \@a_deriv_norm_minus, $o_param->Get('percentile')) if ($o_param->IsDefined('percentile'));

	
	# number of maximum expected tstart
	my $expect_tstart_nb = int($seq_len/$o_param->Get('tstart_ratio'));
	# Detect the starts of transcription
	my ($ra_start_plus, $ra_start_minus) = &DetectStart(\@a_deriv_norm_plus, \@a_deriv_norm_minus, 
														\@a_derivative_plus, \@a_derivative_minus, 
														$seq_len, $log_threshold, $boost_score, 
														$expect_tstart_nb);


	&PrintLog(\@a_plus, \@a_minus, \@a_smoothed_plus, \@a_smoothed_minus, \@a_derivative_plus,\@a_derivative_minus, 
			  $ra_start_plus, $ra_start_minus, $seq_len) if ($DEBUG);

	# parse boxes if specified
	if ($o_param->IsDefined('box_in'))
	{
		&IntegrateBox($ra_start_plus, $ra_start_minus, $o_param);
		&PrintLog(\@a_plus, \@a_minus, \@a_smoothed_plus, \@a_smoothed_minus, \@a_derivative_plus,\@a_derivative_minus, 
				  $ra_start_plus, $ra_start_minus, $seq_len) if ($DEBUG);
	}

	# Write the tStarts in a wig and gff3 files
	&WriteStarts($ra_start_plus, $ra_start_minus, $seq_len, $seq_name, $o_param);
}




=head2 procedure ReadScore

 Title        : ReadScore
 Usage        : &ReadScore($in, $ra_plus, $ra_minus, $seq_len, $o_param)
 Prerequisite : $in is a gff3 format like stream
 Function     : Fill arrays with scores read in the gff3 infile
 Args         : $in: input file stream
                $ra_plus :ref to an array to fill with values of strand +
                $ra_minus :ref to an array to fill with values of strand -
                $seq_len : length of the seq
 Globals      : none

=cut
sub ReadScore
{
	my ($in, $format, $ra_plus, $ra_minus, $seq_len, $o_param) = @_;
	
    # Init arrays to 0
	for (my $i=0; $i < $seq_len; $i++)
	{
		$ra_plus->[$i]  = 0;
		$ra_minus->[$i] = 0;
	}

	if ($format eq "gff3")
	{
		my $score_max=0;
		# Read the score in the gff3 file and fill @a_plus and @a_minus array
		my @a_line;
		my $f_in  = &General::GetStreamIn($in);
		foreach my $gff3_line (<$f_in>)
		{
			next if ($gff3_line =~ /^#/);
			
			@a_line = split(/\t/, $gff3_line);
			$score_max = $a_line[5] if($a_line[5]> $score_max);
			if ($a_line[6] eq '+')
			{
				$ra_plus->[$a_line[3]-1]  = $a_line[5]; 
			}
			else 
			{
				$ra_minus->[$a_line[3]-1] = $a_line[5]; 
			}
		}
		print "Score max : $score_max\n" if ($DEBUG);
	}
	# wig format
	else
	{
		print "WIG format reading.not yet implemenet \n";
		exit;
		#my $f_in  = &General::GetStreamIn($in);
		#foreach my $line (<$f_in>)
		#{
		#	chomp($line);
		#	next if ($line !~ /^\d+/);
		#	@a_line = split(/\t/, $gff3_line);
			
		
	}

	return;
}

=head2 procedure Smooth
 Title        : Smooth
 Usage        : &Smooth($ra_smoothed_plus, $ra_smoothed_minus, $ra_plus, $ra_minus, $miwin_size, $seq_len, $smooth_method )
 Prerequisite : $o_param has to contain 'win_size' and 'smooth_method' keys
 Function     : smooth data with the 'smooth_method'. Fill arrays $ra_smoothed_plus and $ra_smoothed_minus
 Args         : $ra_smoothed_plus: ref to an array to fill with smooth values (strand +)
	            $ra_smoothed_minus: ref to an array to fill with smooth values (strand -)
                $ra_plus :ref to an array of scores (strand +)
                $ra_minus :ref to an array of scores (strand -)
                $seq_len : length of the seq
	            $o_param: ParamParser object
 Globals      : none

=cut
sub Smooth
{
	my ($ra_smoothed_plus, $ra_smoothed_minus, $ra_plus, $ra_minus, $seq_len, $o_param) = @_;

	my $win_size      = $o_param->Get('win_size');	
	my $miwin_size    = int($win_size/2);
	my $smooth_method = $o_param->Get('smooth_method');

	if (!$o_param->IsDefined('no_smooth'))
	{
		for (my $i=0; $i < $miwin_size; $i++)
		{	
			$ra_smoothed_plus->[$i]  = $ra_plus->[$i];
			$ra_smoothed_minus->[$i] = $ra_minus->[$i];	
		}
		for (my $i=$miwin_size; $i < ($seq_len-$miwin_size); $i++)
		{
			$ra_smoothed_plus->[$i]  = &GetSmoothVal($ra_plus,  $i, $miwin_size, $smooth_method);
			$ra_smoothed_minus->[$i] = &GetSmoothVal($ra_minus, $i, $miwin_size, $smooth_method);
		}
		for (my $i=1; $i <= $miwin_size; $i++)
		{		
			$ra_smoothed_plus->[$seq_len-$i]   = $ra_plus->[$seq_len-$i];
			$ra_smoothed_minus->[$seq_len-$i]  = $ra_minus->[$seq_len-$i];	
		}
	}
	else
	{
		print " --> No smooth step\n";
		for (my $i=0; $i < $seq_len; $i++)
		{	
			$ra_smoothed_plus->[$i]  = $ra_plus->[$i];
			$ra_smoothed_minus->[$i] = $ra_minus->[$i];	
		}
	}

}

=head2 procedure DerivativeNormalize
 Title        : DerivativeNormalize
 Usage        : &DerivativeNormalize(\@a_array, \@a_deriv_norm, $len)
 Prerequisite : -
 Function     : compute the derivative of each value of the array
                if the derivative is upper than a threshold, normalize
 Args         : $ra_array : reference to the initial array
	            $ra_derivative: ref to the array in which save the derivatives
                $seq_len : length of the seq
                $deriv_threshold
 Globals      : none

=cut
sub DerivativeNormalize
{
	my ($ra_, $ra_der, $ra_der_norm, $seq_len, $deriv_threshold) = @_; 

	my $x;
	my $xplus1;
	my $deriv;
	for (my $i=0; $i < $seq_len-1; $i++)
	{
		$x      = $ra_->[$i];
		$xplus1 = $ra_->[$i+1];
		$deriv  = &GetDerivative($x, $xplus1);
		$ra_der->[$i] = $deriv;

		if (abs($deriv) >= $deriv_threshold)
		{
			$ra_der_norm->[$i] = &GetNormalize($deriv, $x, $xplus1);
		}
		else
		{
			$ra_der_norm->[$i] = 0;
		}
	}
	$ra_der_norm->[$seq_len-1]  = 0; # Last value=0
	$ra_der->[$seq_len-1]       = 0; # Last value=0

	return;
}


=head2 function GetDerivative

 Title        : GetDerivative
 Usage        : &GetDerivative($x, $xplus1)
 Function     : Return the derivative of x
 Args         : $x : value of x
                $xplus1 : value (x+1) 
 Globals      : none

=cut
sub GetDerivative
{
	my ($x, $xplus1) = @_;
	return ($xplus1 - $x);
}

=head2 function GetNormalize

 Title        : GetNormalize
 Usage        : &GetNormalize($val, $x, $xplus1)
 Function     : Return the normalization = $val/(($x+xplus1)/2)
 Args         : $val
                $x : value of x
                $xplus1 : value (x+1) 
 Globals      : none

=cut
sub GetNormalize
{
	my ($val, $x, $xplus1) = @_;

	return 0 if ( ($x + $xplus1) == 0);
	return ( ($val * 2 / ($x + $xplus1) ) );
}

=head2 function Normalize

 Title        : GetPercentile
 Usage        : &GetPercentile(\@a_1, \@a_2, 95)
 Prerequisite : 
 Function     : Return the percentile $percentile of the two arrays
 Args         : $ra_plus: ref to a first array
                $ra_minus: ref to a second array
                $percentile : wanted percentile
 Globals      : none

=cut
sub GetPercentile
{
	my ($ra_plus, $ra_minus, $percentile) = @_;
	
	use Statistics::Descriptive;
	my $stat = Statistics::Descriptive::Full->new();
	my @a_tmp ;
	foreach my $a (@{$ra_plus})
	{
		push(@a_tmp, abs($a)) if ($a != 0);
	}	
	foreach my $a (@{$ra_minus})
	{
		push(@a_tmp, abs($a)) if ($a != 0);
	}
	$stat->add_data(@a_tmp);	
	my $threshold = $stat->percentile($percentile);	

	print "Threshold deduced with percentile ". $percentile." : ".	$threshold."\n";

	return $threshold;
}


=head2 procedure DetectStart
 Title        : DetectStart
 Usage        : -
 Prerequisite : -
 Function     : Detect the starts of transcription according to the values
 Return         $ra_start_plus and $ra_start_minus,  ref to the result array strand +/-
 Args         : $ra_deriv_norm_plus  ref to an array of deriv and normalized values strand +
	            $ra_deriv_norm_minus ref to an array of deriv and normalized values strand -
                $ra_deriv_plus       ref to an array of deriv  values strand +
                $ra_deriv_minus      ref to an array of deriv  values strand -
                $seq_len
                $threshold
                $bootscore_factor
                $expected_nb_tstart
 Globals      : none

=cut
sub DetectStart
{
	my ($ra_deriv_norm_plus, $ra_deriv_norm_minus, $ra_deriv_plus, $ra_deriv_minus, 
		$seq_len, $threshold, $boost_score, $expect_tstart_nb) = @_;

	my $heap       = Heap::Simple->new(elements => "Any", order=> "<");
	my $nb_tstart  = 0;

	# Results arrays
	my $ra_tmp_plus; 
	my $ra_tmp_minus;
	my $ra_start_plus; 
	my $ra_start_minus;
	
	$ra_tmp_plus->[0]   = 0;	
	$ra_tmp_plus->[1]   = 0;	
	$ra_start_plus->[0] = 0;	
	$ra_start_plus->[1] = 0;


	# Strand +
	for (my $i=2; $i < $seq_len; $i++)
	{
		# strand +: we seek positive derivative, higher than threshold
		if ( $ra_deriv_norm_plus->[$i-2]  >= $threshold )
		{
			$ra_tmp_plus->[$i] = abs($ra_deriv_norm_plus->[$i-2])*$boost_score;
			&InsertHeap($heap, abs($ra_deriv_plus->[$i-2]), $i, $expect_tstart_nb, \$nb_tstart);
		}
		else
		{
			$ra_tmp_plus->[$i] = 0;
		}
		$ra_start_plus->[$i] = 0; # init final result array
	}

	# Strand minus
	for (my $i=0; $i < $seq_len-1; $i++)
	{
		# strand -: we seek negative derivative, lower than -threshold
		if ($ra_deriv_norm_minus->[$i+1] <= -$threshold)
		{
			$ra_tmp_minus->[$i]  = abs($ra_deriv_norm_minus->[$i+1])*$boost_score;
			&InsertHeap($heap, abs($ra_deriv_minus->[$i+1]), -$i, $expect_tstart_nb, \$nb_tstart);
		}
		else
		{
			$ra_tmp_minus->[$i] = 0;
		}
		$ra_start_minus->[$i] = 0; # init final result array
	}
	$ra_tmp_minus->[$seq_len-1]   = 0;
	$ra_start_minus->[$seq_len-1] = 0;
	
	if ($DEBUG)
	{
		print "Number of tstart  : ".$heap->count." (Max expected: $expect_tstart_nb)\n";
		print "Minimum derivate value " . $heap->top_key. " (position " . $heap->top . ")\n";
	}

	while ($heap->count > 0)
	{
		my $pos = $heap->top;
		# strand+
		if ($pos > 0)
		{
			$ra_start_plus->[$pos] = $ra_tmp_plus->[$pos];
		}
		else
		{
			$ra_start_minus->[abs($pos)] = $ra_tmp_minus->[abs($pos)];
		}
		$heap->extract_top;
	}
	
	return ($ra_start_plus, $ra_start_minus);
}




=head2 procedure InsertHeap
 Title        : InsertHeap
 Usage        : -
 Prerequisite : -
 Function     : Insert $key (associated with $elt) in the $heap. The max nb is $max_elt_nb
 Return         -
 Args         : $heap
                $key
                $val
                $max_elt_nb Max number of elt to insert in the heap
                $r_elt_nb   Ref to the number of already inserted elts
 Globals      : none

=cut
sub InsertHeap
{
	my ($heap, $key, $elt, $max_elt_nb, $r_elt_nb) = @_;
	
	if ($$r_elt_nb < $max_elt_nb)
	{
		$heap->key_insert($key, $elt);
		$$r_elt_nb++;
	}
	else
	{
		if ($key > $heap->top_key)
		{
			$heap->extract_top;
			$heap->key_insert($key, $elt);
		}
	}

	return;
}


sub WriteStarts
{
	my ($ra_start_plus, $ra_start_minus, $seq_len, $seq_name, $o_param) = @_;

	my $out        = $o_param->Get('out');
	my $f_out      = &General::GetStreamOut($out );
	my $f_gff3_out = &General::GetStreamOut($out.".gff3" );

	my $id_root = $seq_name.$$;	
	my $cpt     = 1;

	print $f_gff3_out "##gff-version 3\n";
	
	# strand plus
	print $f_out "track color=255,0,0\nvariableStep chrom=$seq_name\n";
	for (my $i=0; $i < $seq_len; $i++)
	{
		next if ($ra_start_plus->[$i] == 0);
		
		print $f_out      "".($i+2)."\t".$ra_start_plus->[$i]."\n";
		print $f_gff3_out "$seq_name\tpatscan\ttranscript_region\t".($i+1)."\t".($i+1)."\t".$ra_start_plus->[$i]."\t+\t.\tID=$id_root:$cpt\n";		
		$cpt++;
		if ($o_param->IsDefined('boost_ig'))
		{
			print $f_gff3_out "$seq_name\tpatscan\tintergenic_region\t".($i)."\t".($i)."\t".$ra_start_plus->[$i]."\t+\t.\tID=$id_root:$cpt\n";
			$cpt++;
		}
	}
	
	# strand minus
	print $f_out "track color=100,100,100\nvariableStep chrom=$seq_name\n";
	for (my $i=0; $i < $seq_len; $i++)
	{
		next if ($ra_start_minus->[$i] == 0);
		print $f_out      "".($i)."\t-".$ra_start_minus->[$i]."\n";
		print $f_gff3_out "$seq_name\tpatscan\ttranscript_region\t".($i+1)."\t".($i+1)."\t".$ra_start_minus->[$i]."\t-\t.\tID=$id_root:$cpt\n";
		$cpt++;
		if ($o_param->IsDefined('boost_ig'))
		{
			print $f_gff3_out "$seq_name\tpatscan\tintergenic_region\t".($i+2)."\t".($i+2)."\t".$ra_start_minus->[$i]."\t-\t.\tID=$id_root:$cpt\n";
			$cpt++;
		}
	}
	$f_out->close();
	$f_gff3_out->close();
}


sub IntegrateBox
{
	my ($ra_start_plus, $ra_start_minus, $o_param) = @_;
	
	my $box_context = $o_param->Get('box_context');
	my $box_in      = $o_param->Get('box_in');
	my $f_box_in    = &General::GetStreamIn($box_in);
	foreach my $gff3_line (<$f_box_in>)
	{			
		next if ($gff3_line =~ /^#/);
		print $gff3_line if ($DEBUG);
		my @a_line = split(/\t/, $gff3_line);
		if ( $a_line[6] eq '+' )
		{
			&GetBonus($ra_start_plus, $a_line[4]+1, $a_line[4]+1+$box_context); 
		}
		else
		{
			&GetBonus($ra_start_minus, $a_line[3]-1-$box_context, $a_line[3]-1); 
		}
	}
}


sub GetBonus
{
	my ($ra_, $begin, $end) = @_;
	
	if ($begin > $end)
	{
		print "GetBonus: warning: $begin > $end\n";
		my $tmp = $begin;
		$begin  = $end;
		$end    = $tmp;
	}
	
	for (my $i = $begin-1; $i < $end-1; $i++)
	{
		if (defined($ra_->[$i]) && $ra_->[$i] != 0)
		{
			print "Bonus pos ".($i+1)." : ". $ra_->[$i]. " --> ".$ra_->[$i]*$BOX_BONUS."\n" if ($DEBUG);
			$ra_->[$i] = $ra_->[$i]*$BOX_BONUS;
		}
	}
	return;
	
}

sub max { @_ = sort { $a <=> $b } @_; pop }
sub min { @_ = sort { $b <=> $a } @_; pop }
sub mean
{
	my (@a_) = @_;
   
	my $mean  = 0;	
	my $sum   = 0;
	my $nb    = scalar(@a_);
	foreach my $element (@a_)
	{
		$sum += $element ;
	}
	$mean = $sum/$nb if ( $nb > 0 );

	return $mean;
}

=head2 function GetSmoothVal

 Title        : GetSmoothVal
 Usage        : &GetSmoothVal(\@a_, $pos, $size, $max)
 Prerequisite : 
 Function     : Return the smoothed value at the position i, with the smooth method, in a window of size ŝize
 Args         : $ra_: ref to the array
                $i: position in the array
                $size : window size in which do the smooth
               $method: smoothed method
 Globals      : none

=cut
sub GetSmoothVal
{
	my ($ra_, $i, $size, $method) = @_;

	if ($method eq "max")
	{
		return &max(@$ra_ [ ($i-$size)..($i+$size) ]);
	}
	elsif ($method eq "min")
	{
		return &min(@$ra_ [ ($i-$size)..($i+$size) ]);
	}
	elsif ($method eq "mean")
	{
		return &mean( @$ra_ [ ($i-$size)..($i+$size) ] );
	}
	else
	{
		print "Error: invalid smooth method: $method\n";
		return;
	}
}


sub PrintLog
{
	my ($ra_p, $ra_m, $ra_smoothed_p, $ra_smoothed_m, $ra_derivative_p, $ra_derivative_m, $ra_start_p, $ra_start_m, $seq_len) = @_;

	print "pos\tval+\tsmoothed val\tderivate\t start\tval-\tsmothedval-\tderivate-\tstart-\n";
	for (my $i=0; $i < $seq_len; $i++)
	{
		print "".($i+1)."\t".$ra_p->[$i]."\t".$ra_smoothed_p->[$i]."\t".$ra_derivative_p->[$i]."\t".$ra_start_p->[$i]."\t".$ra_m->[$i]."\t".$ra_smoothed_m->[$i]."\t".$ra_derivative_m->[$i]."\t".$ra_start_m->[$i]."\n";
	}
	print "END display\n";
}

sub Usage
{
    print STDERR<<END  
$0
    --in              filename|stdin   file containing expression data
    --out             filename|stdout  output files root name
    --seq_len         integer          Sequence length
    --seq_name        string           Sequence name

[Optional]
  --format               string  [Default gff3]   Format of input file: gff3 or wig (WIG not yet implemented!!!)
  --strand               +/-                      ONLY FOR WIG FORMAT
  --derivative_threshold integer [Default 0]      Derivative minimum threshold. 
  --log_threshold        float   [Default 0]      Threshold value to keep a break. 
                                                  Note that this threshold is applyed after log, smooth and derivate the values.
  --percentile           integer                  Threshold to keep a break is the percentile of this value instead of 'threshold' [1-99] 
  --boost_ig                                      To boost intergenic before the start
  --smooth_method        string  [Default max]    Smooth method. Allowed values :min max mean  
  --no_smooth                                     Dont smooth data
  --win_size             integer [Default 3]      Length of the window in which apply smoothing method
  --boostscore_factor    integer                  Multiply the score by this factor
  --box_in               filename                 GFF3 file of a promotor boxes 
  --box_context          integer [Default 20]     Length of the region after the box, in which get a bonus to all tstart signals   
	--tstart_ratio       integer [Default 1000]   Maximum expected tstart number is ses_len/tstart_ratio
  --help
END
}


=cut
