
# $Id: SGE.pm 239 2009-06-26 09:54:10Z manu $

package PARALOOP::SCHEDULER::SLURM;

=pod

=head1 Copyright Notice

 Copyright I.N.R.A. - C.N.R.S. 

 emmanuel.courcelle@toulouse.inra.fr
 jerome.gouzy@toulouse.inra.fr

 This software is a computer program whose purpose is to help you run 
 your scripts on parallel computers.

 This software is governed by the CeCILL  license under French law and
 abiding by the rules of distribution of free software.  You can  use, 
 modify and/ or redistribute the software under the terms of the CeCILL
 license as circulated by CEA, CNRS and INRIA at the following URL
 http://www.cecill.info" 

 As a counterpart to the access to the source code and  rights to copy,
 modify and redistribute granted by the license, users are provided only
 with a limited warranty  and the software's author,  the holder of the
 economic rights,  and the successive licensors  have only  limited
 liability. 

 In this respect, the user's attention is drawn to the risks associated
 with loading,  using,  modifying and/or developing or reproducing the
 software by the user in light of its specific status of free software,
 that may mean  that it is complicated to manipulate,  and  that  also
 therefore means  that it is reserved for developers  and  experienced
 professionals having in-depth computer knowledge. Users are therefore
 encouraged to load and test the software's suitability as regards their
 requirements in conditions enabling the security of their systems and/or 
 data to be ensured and,  more generally, to use and operate it in the 
 same conditions as regards security. 

 The fact that you are presently reading this means that you have had
 knowledge of the CeCILL license and that you accept its terms.

=head1 NAME

PARALOOP::SCHEDULER::SLURM - Scheduler object for use with SLURM

=head1 Derives from...

PARALOOP::SCHEDULER::Scheduler

=head1 DESCRIPTION

This scheduler is used for a cluster, equipped with the batch system SLURM. 

=head1 CONFIGURATION FILE PARAMETERS

 PARALOOP_Scheduler 
     SLURM
 PARALOOP_error_directory 
     The name of the error directory
 PARALOOP_qsub_params     
     String to pass to qsub
 PARALOOP_account         
     String to pass to the -A switch of SGE

=head1 PUBLIC METHODS

 WhatParaloopScheduler
    my $msg=$obj->WhatParaloopScheduler()

 Parameters
    my $msg=$obj->Parameters()

 New
    my $obj = PARALOOP::Dummy->New(-params => $params, -log => $log);

     
 IsSelfPossible
    If (&$obj->IsSelPossible()==1)...

 Submit
    my $job_id = $obj->Submit();

 RemTime
    my $rem_time = $obj->RemTime(); 

 JobId
    my $job_id = $obj->JobId();

=head1 METHODS AND SUBROUTINES - DETAILED DESCRIPTION

=cut

use strict;
use warnings;
use Cwd;
use File::Basename;
use Carp;
use Sys::Hostname;
# The NAMESPACE constant are also set in paraloop
use constant NAMESPACE => 'PARALOOP_';
    
use Logger;
use PARALOOP::SCHEDULER::Scheduler;

@PARALOOP::SCHEDULER::SLURM::ISA = qw( PARALOOP::SCHEDULER::Scheduler );

BEGIN
{
    our $VERSION = do {my @r = (q$Rev: 239 $ =~ /\d+/g); $r[0]};
}

=pod

=head2 Function WhatParaloopScheduler

 Usage:    my $msg = PARALOOP::SCHEDULER::SLURM::WhatParaloopScheduler()
        or my $msg = $obj->WhatParaloopScheduler()
 Function: Return some information (2 lines max) on the scheduler
           An object is NOT considered as a scheduler if this function is not implemented
 Arg:      none ($self is ignored)
 Return:   A message describing the plugin (1 or 2 lines)

=cut

sub WhatParaloopScheduler
{
    my $rvl = "";
    $rvl =
      "PARALOOP::SCHEDULER::SLURM v$PARALOOP::SCHEDULER::SLURM::VERSION - Deriving from Scheduler v$PARALOOP::SCHEDULER::Scheduler::VERSION\n";
    $rvl .= "This scheduler object connects paraloop to the SLURM queing system\n";
    return $rvl;
}

=pod

=head2 Function Parameters

 Usage:    my $msg = PARALOOP::SCHEDULER::SLURM::Parameters()
        or my $msg = $obj->Parameters()
 Function: Return the list of allowed parameters for this scheduler
           If your scheduler defines some parameters, it is good pratice adding this function,
           to display those parameters when the --parameters switch of paraloop.pl is used.
           However, this is not mandatory
 Arg:      none ($self is ignored)
 Return:   A message describing the scheduler parameters (1 line / parameter)

=cut

sub Parameters
{
    my $rvl = "";
    $rvl = "\nParameters for the SLURM scheduler\n=================\n";
    $rvl .= "PARALOOP_Scheduler\tMust be set to SLURM for using this scheduler\n";
    $rvl .= "PARALOOP_sbatch_params\tstring to pass to qsub\n";
    $rvl .= "                    \tExample:  --begin=12:00 for starting the job at 12:00\n";
    $rvl .= "PARALOOP_account\tString to pass to the -A switch of SLURM\n";
    return $rvl;
}

=pod

=head2 Function Submit

 Usage:     &$obj->Submit()
 Function:  Call a sbatch
 Arg :      none
 Return:    the job Id

=cut

sub Submit
{
    my $self    = shift;
    my $o_param = $self->{'__params'};
    my $o_log   = $self->{'__log'};

    my $sbatch = $o_param->Get('sbatch');

    my $subcount = $self->_IncSubCount();
    my $plp_id = $o_param->Get('plp_id');
    my $partition = '';
    if ($o_param->IsDefined('partition'))
    {
        $partition = '-p ' . $o_param->Get('partition');
    }

    my $account = '';
    if ($o_param->IsDefined('account'))
    {
        $account = '-A ' . $o_param->Get('account');
    }

    my $sbatch_params = $o_param->Get('sbatch_params');
    my $start       = $o_param->Get('start');
    my $error       = $o_param->Get('error_directory');

    # Avoiding PATH inflate at each submit in some conditions
    if (exists $ENV{'SLURM_O_PATH'})
    {
        $ENV{'PATH'}    = $ENV{'SLURM_O_PATH'};
    }
    $o_param->Dump('ENV');
    my $cwddir = cwd;
    #my $qsub_cmd    = "$qsub $qsub_params -cwd -b y -V $queue -o $error -e $error -N EUGENE$plp_id $account $0";
    
    # -cwd : repertoire de travail =>  -D
    # -b y : accept binary script. Je n'ai pas trouve d equivalent avec sbatch
    # -V Specifies that all environment variables active  within the qsub utility be exported to the context of the job. -->  --get-user-env
    # -N => -J
    my $sbatch_cmd = "$sbatch $sbatch_params -D $cwddir --get-user-env $partition  -o $error/EUGENE$plp_id.o -e $error/EUGENE$plp_id.e -J EUGENE$plp_id $account $0";

    print "calling $sbatch_cmd\n" if ($o_param->IsDefined('verbose'));

    my $job_id = `$sbatch_cmd`;
    chomp($job_id);
    if ($job_id =~ /Submitted batch job +(\d+)/)
    {
       $job_id = $1;
    }
    else
    {
       die "ERROR - , sbatch returned the following: $job_id. Looks strange, dying";
    }

    $self->_SyncLockFile($job_id);
    $o_log->Trace("$sbatch_cmd\n", '', '2');

    return $job_id;
}

=pod

=head2 Function JobId

 Usage:     my $job_id = $obj->JobId()
 Function:  Ask the system for the job id
            The environment variable JOB_ID is returned
 Arg :      none
 Return :   the job Id

=cut

sub JobId {
    my $host = hostname;
    
    # Remove domain name, if any
    $host =~ s/(.+)\..+$/$1/;

    return $host.'.'.$ENV{'SLURM_JOB_ID'};
}

=pod

=head2 Function RemTime

 Usage:    my $rem_time = &$obj->RemTime();
 Function: Asks the system for the number of seconds remainig before a kill
 Arg:      none
 Return:   $rem_time = The number of seconds remaining (in fact subtract 5% for security)

=cut

sub RemTime
{
    return -1;				# Not already implemented. Need to check if slurm support this function
}

=pod

=head2 Function IsLaunched

 Usage:    If (&$obj->IsLaunched($job_id)==1)...
 Function: Check the presence of the job in the system
           Call scontrol show job $job_id
 Arg:      job_id: the job Id to check
 Return:   0 ==> The job is not known by the system
           1 ==> The job is known (may be in different states: Run, Queue, etc)

=cut

sub IsLaunched
{
    my $self         = shift;
    my $job_id       = shift;

    # Keep only the job id, NOT the host name part
    $job_id =~ s /^.+\.(\d+)$/$1/;
    
    my $o_param      = $self->{'__params'};
 
    my $scontrol = $o_param->Get('scontrol');
 
    my $cmd = "$scontrol show job $job_id 2> /dev/null|wc -l";
    my $rvl = `$cmd`;
    chomp $rvl;
    return ($rvl eq '0') ? 0 : 1;
}

=pod

=head2 Procedure KillJob

 Usage:    $obj->KillJob($job_id);
 Function: Kill the job passed by parameter
           Call qdel $job_id
 Arg:      job_id: the job Id to kill
 Return:   none

=cut

sub KillJob
{
    my $self         = shift;
    my $o_param = $self->{'__params'};
    my $scancel = $o_param->Get('scancel');

    my $job_id       = shift;
    my $kill_cmd = "$scancel $job_id";
    print "calling $kill_cmd\n" if ($o_param->IsDefined('verbose'));
    `$kill_cmd`;
}

=pod

=head2 procedure IsSelfPossible

 Usage:  if ($obj->IsSelfPossible()) ... 
 Access: public (use New instead)
 function: return 1 if this scheduler accepts to submit himself, 0 if it doesn't
 Return:   0/1
 Args:     none
 globals:  none

=cut

sub IsSelfPossible {
    return 1;
};


