#!/usr/bin/perl

#
# $Id
#

# Copyright INRA

# Erika.Sallet@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Schiex@toulouse.inra.fr

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=pod 

=head1 NAME


=head1 SYNOPSIS

bin/egn-prok.pl --indir $PWD/Testus_erikatis/ --outdir $PWD/out/ --cfg cfg/egn-prok.cfg

=head1 DESCRIPTION


=cut


BEGIN
{
	$ENV{LANG} = "C";

	use Cwd 'abs_path';
	use File::Basename;
	my ($dirprg) =  dirname(abs_path($0));
	unshift(@INC,"$dirprg/../../lib");
	unshift(@INC,"$dirprg/../../corelib");

}

use strict;
use warnings;
use Carp;
use ParamParser;
use General;
use GeneralBioinfo;
use Logger;

my $LOGGER_LEVEL_EXEC = 'DE';

my $DEBUG;

my $DEBUG_DEFAULT  = 0;
my $TRNASCAN_EXT   = "trnascan";
my $RNAMMER_EXT    = "rnammer";
my $RFAMSCAN_EXT   = "rfamscan";
my $PRODIGAL_EXT   = "prodigal";
my $MAPPING_EXT    = "bed.specific.gz";
my $WIG_EXT        = "wig2expr";
my $EXPR_POS_EXT   = "position"; 
my $EXPR_REG_EXT   = "region";
my $TSTART_EXT     = "tstart";
my $TSTART_POS_EXT = "tstart_position";
my $ALLSEQFILE_NAME = "sequences";

MAIN:
{
    my $o_param = New ParamParser('GETOPTLONG',\&Usage,'indir=s', 'outdir=s', 'workingdir=s', 'cfg=s', 
								  'no_ncrna_detection', 'no_prodigal', 'no_rfamscan', 'no_trnascan', 
								  'no_rnammer', 'debug=i', 'uniq=s');
	$o_param->SetBehaviour('ignore_space');
	$o_param->Update('ENV', "A");

	$o_param->AssertDirExists('EGNPP');
	$o_param->AssertDirExists('EUGENEDIR');
	$o_param->AssertFileExists('cfg');
	$o_param->AssertFullPath('cfg');
	$o_param->Update($o_param->Get('cfg'),'A');
	$o_param->SetBehaviour('use_substitution_table');
	$o_param->SetSubstitution('%i',$o_param->Get('EGNPP'));	
	$o_param->SetSubstitution('%e',$o_param->Get('EUGENEDIR'));	
	$o_param->Update($o_param->Get('cfg'),'U');

	&AssertParameters($o_param);
	
	###################  Global parameters ##############################

	$DEBUG             = $o_param->Get('debug');
	my $outdir         = $o_param->Get('outdir');     # Output directory
	my $rootworkingdir = $o_param->Get('workingdir'); # Working directory
	my $allseqfile     = "$rootworkingdir/$ALLSEQFILE_NAME"; # multifasta of all seq

	# Uniq code to generate temporary files/directory
	$o_param->SetUnlessDefined('uniq',  time . "." . $$); # Uniq value
	my $uniq         =  $o_param->Get('uniq');
	my $o_logger     = New Logger ("$rootworkingdir/logger.$uniq.txt", 
								   $LOGGER_LEVEL_EXEC, 'NOCOMPRESSION');
	my $logger_stamp = $o_logger->CalcStamp();
	print STDERR "Log file : $rootworkingdir/logger.$uniq.txt\n" if ($DEBUG);

	
    ##################  Check and compute global variables ################

	# Check and get the evidence input files
	my %h_files;
	&AssertEvidencePatternKeys($o_param, $o_param->Get('datadir'), \%h_files);
	
	# Check and get the BlastX protein databases
	my %h_database;   # BlastX databases
	my %h_db_level;   # BlastX level (for EUGENE)
	&AssertBlastX($o_param, \%h_database, \%h_db_level);

	# Load the template configuration files for EGN
	my %h_cfg_template;
	&GetTemplate($o_param, \%h_cfg_template);

	&PrintModeInfo($o_param, $o_logger, $logger_stamp);


	################## Create the working tree #####################

	my @a_seq;        # Array of the genomic sequences
	&CreateWorkSeqTree($o_param, $o_param->Get('genomedir'), 
					   \@a_seq, $rootworkingdir, $ALLSEQFILE_NAME, 
					   $o_logger, $logger_stamp, $uniq);
 
	
	################### Global compute ##############################

	&GetGlobalEvidence($o_param, \%h_files, $rootworkingdir, $allseqfile, 
					   $ALLSEQFILE_NAME,  $o_logger, $logger_stamp, $uniq);
				

	################### Sequence specific compute ###################

	my @a_res_gff3;   # Array of EuGene gff3 result files
	my $annotastruct_cpt = -1;
	# for each fasta: create all evidences
	foreach my $seq (@a_seq)
	{
		printf STDERR "ANALYSE sequence: $seq\n" if ($DEBUG);

		my ($seqpath, $seqname, $seqid, $seqlen) = &GetSeqInfo($seq);

		my $workingdir = $seqpath."/work." .$uniq."/";
		$o_param->Set('workingdir', $workingdir);
		my $matrixfile = "$seq.mat";
		my $cfg_file   = "$seq.eugene.par";

		my $annotastruct_content = "";
		$annotastruct_cpt        = -1; # Reinit the counter for this sequence
   

		# 1 Launch All BlastX
		foreach my $nb (keys %h_database)
		{
			&GetBlastX($o_param, $seq, $h_database{$nb}, $nb, $workingdir, $o_logger, $logger_stamp);
		}
		
		# 2 Train EuGene
		&BuildEGNMatrix($o_param, $seq, $workingdir, $matrixfile, $o_logger, $logger_stamp);
		
		# 3 Manage all the kinds of RNAseq data
		foreach my $nb (sort keys(%h_files))
		{
			my $pat = "evidence_" . $nb;
			# Prepare the evidences for egn
			&GetEvidences($o_param, $seq, $seqid, $seqlen, $seqname, $workingdir, $pat, $h_files{$nb},
						  \$annotastruct_cpt, \%h_cfg_template, \$annotastruct_content, 
						  $o_logger, $logger_stamp, $uniq);
		}
		
		# 4 ncRNA computing
		&GettRNAscanSE($o_param, $seq, \$annotastruct_cpt, \%h_cfg_template, \$annotastruct_content, 
					   $o_logger, $logger_stamp);
		&GetRnammer($o_param, $seq, \$annotastruct_cpt, \%h_cfg_template, \$annotastruct_content, 
					$o_logger, $logger_stamp);
		&GetRfamScan($o_param, $seq, \$annotastruct_cpt, \%h_cfg_template, \$annotastruct_content, 
					 $o_logger, $logger_stamp);

		# 5 Launch Prodigal
		&GetProdigal($o_param, $seq, $seqlen, \$annotastruct_cpt, \%h_cfg_template, 
					 \$annotastruct_content, $o_logger, $logger_stamp);
		
		# 6 Build the cfg file for egn
		&BuildEGNProkParameterFile($o_param, $cfg_file, \%h_db_level, $annotastruct_content, $annotastruct_cpt);

		# 7 Launch EuGene
		&LaunchEGN( $o_param, $seq, $cfg_file, $matrixfile, $outdir, $o_logger, $logger_stamp);

		# Compress all the temporary files
		my $cmd = "gzip -qr $workingdir/";
		&ExeCmd($cmd, $o_logger, $logger_stamp);

		printf STDERR "ANALYSE sequence: $seq DONE.\n" if ($DEBUG);	

		push(@a_res_gff3, "$outdir/$seqname.gff3");
	}

	
	# Compute statistics about annotation. Extract protein sequences.
	&Consolidate($o_param, $outdir, \@a_res_gff3, $allseqfile,
				 $o_logger, $logger_stamp);
}


=head2 procedure PrintModeInfo
	
 Title        : PrintModeInfo
 Usage        : &PrintModeInfo($o_param, $o_logger, $stamp)
 Function     : Print to STDERR and in the log file information about the mode
               (No prodigal, No ncrna)
 Args         : $o_param 
 Globals      : none

=cut
sub PrintModeInfo
{
	my ($o_param, $o_logger, $logger_stamp) = @_;
	
	if ($o_param->IsDefined('no_ncrna_detection'))
	{
		printf STDERR "No ncRNA mode (tRNAscan-SE, rfam_scam and RNAmmer not launch)\n" if ($DEBUG);
		$o_logger->Trace("No ncRNA mode (tRNAscan-SE, rfam_scam and RNAmmer not launch)\n", 
						 $logger_stamp, $LOGGER_LEVEL_EXEC);
	}
	if ($o_param->IsDefined('no_prodigal'))
	{
		printf STDERR "No prodigal mode\n" if ($DEBUG);
		$o_logger->Trace("No prodigal\n", 
						 $logger_stamp, $LOGGER_LEVEL_EXEC);
	}
	if ($o_param->IsDefined('no_rfamscan'))
	{
		printf STDERR "No rfam_scan mode\n" if ($DEBUG);
		$o_logger->Trace("No rfam_scan\n", 
						 $logger_stamp, $LOGGER_LEVEL_EXEC);
	}
	if ($o_param->IsDefined('no_trnascan'))
	{
		printf STDERR "No trnascan-SE mode\n" if ($DEBUG);
		$o_logger->Trace("No trnascan-SE\n", 
						 $logger_stamp, $LOGGER_LEVEL_EXEC);
	}
	if ($o_param->IsDefined('no_rnammer'))
	{
		printf STDERR "No Rnammer mode\n" if ($DEBUG);
		$o_logger->Trace("No Rnammer\n", 
						 $logger_stamp, $LOGGER_LEVEL_EXEC);
	}
}



=head2 procedure AssertParameters

 Title        : AssertParameters
 Usage        : 
 Prerequisite : 
 Function     : assert tte required parameters
 Args         : $o_param 
 Globals      : none

=cut
sub AssertParameters
{
	my ($o_param) = @_;
	
	$o_param->SetUnlessDefined('debug', $DEBUG_DEFAULT);
	$o_param->AssertInteger('debug');

	$o_param->AssertDirExists('indir', 'outdir');
	$o_param->AssertFullPath('indir', 'outdir');
	$o_param->SetUnlessDefined('genomedir', $o_param->Get('indir')."/genome");
	$o_param->SetUnlessDefined('datadir'  , $o_param->Get('indir')."/data");
	$o_param->AssertDirExists('genomedir', 'datadir');
	
	$o_param->SetUnlessDefined('workingdir', $o_param->Get('outdir'));
	$o_param->AssertFullPath('workingdir', 'prg_egn-prok_train', 'prg_get_blastx');
	$o_param->AssertDirExists('workingdir');

	$o_param->AssertFileExists('prg_eugene_cfg_model', 
							   'prg_eugene_cfg_low_gc', 
							   'prg_eugene_cfg_high_gc', 
							   'prg_eugene_cfg_rnaseq_annotastruct', 
							   'prg_eugene_cfg_ncrna_annotastruct', 
							   'prg_eugene_cfg_blastx', 
							   'prg_eugene_cfg_region_annotastruct', 
							   'prg_eugene_cfg_tstart_annotastruct', 
							   'prg_eugene_cfg_start_annotastruct',
							   'prg_shuffleseq', 
							   'prg_prodigal','prg_fasta2tree',
                               'prg_biofileconverter',
							   'prg_genome_statistics',
							   'prg_split_bed', 
							   'prg_samtools', 'prg_bam2fastq',
							   'prg_glint', 'prg_bed_to_expr', 
							   'prg_wig_to_expr', 'prg_bed_filter', 
							   'prg_tstart_detection');

	$o_param->AssertFileExists('prg_rnammer') if (!$o_param->IsDefined('no_rnammer') && 
	                                              !$o_param->IsDefined('no_ncrna_detection'));

	$o_param->AssertFileExists('rfam_cm_file') if (!$o_param->IsDefined('no_rfamscan') && 
	                                              !$o_param->IsDefined('no_ncrna_detection'));
	
	$o_param->AssertDefined('prg_eugene', 'organism',
                            'prg_glint_param', 'prg_glint_small_param', 
							'prg_bed_to_expr_tstart_param', 
							'prg_bed_to_expr_tstart_scoremax',
							'prg_bed_to_expr_tstart_scalingfactor', 
							'prg_bed_to_expr_tstart_min_coverage',
							'prg_bed_to_expr_annot_param', 
							'prg_bed_to_expr_annot_scoremax', 
							'prg_bed_to_expr_annot_scalingfactor', 
							'prg_bed_to_expr_annot_min_coverage_percent', 
							'prg_bed_filter_param', 
                            'prg_bed_filter_pe_param',
                            'prg_bed_filter_small_pe_param',
							'prg_tstart_detection_param',
                            'prg_tstart_detection_log_threshold',
		                    'prg_tstart_detection_boostscore_factor', 
							'prg_wig_to_expr_param',
							'prg_egn-prok_train_hsp_training_evalue'
		);
	$o_param->AssertInteger('prg_glint_mate_dist', 
							'prg_glint_mm_se', 
							'prg_glint_mm_pe', 
							'prg_glint_mm_small_se', 
							'prg_glint_mm_small_pe',
							'prg_tstart_detection_derivative_threshold', 
							'prg_tstart_detection_tstart_ratio',
							'prg_egn-prok_train_hsp_training_min_nb',
							'prg_egn-prok_train_hsp_training_length', 
							'prg_egn-prok_train_hsp_training_pci',
							'prg_prodigal_minlen'
                           );

	$o_param->AssertAllowedValue("prg_tstart_detection_boost_ig", "TRUE", "FALSE");
	
	return;
}


=head2 function AssertBlastX

 Title        : AssertBlastX
 Usage        : 
 Prerequisite : blastx_db_list, training_db_file are defined in o_param
 Function     : assert the database exists and are formated. Fill $rh_db_list
 Args         : $o_param 
                $rh_db_list Ref to a hash: Key=nb, Value=path opf the database
 Globals      : none

=cut
sub AssertBlastX
{
	my ($o_param, $rh_db_list, $rh_db_level) = @_;
	
	$o_param->AssertDefined('blastx_db_list');

	my $list    = $o_param->Get('blastx_db_list'); # "2 5 6"
	my @a_db_nb = split(/\s+/, $list); 
	

	foreach my $nb (@a_db_nb)
	{
		my $db = $o_param->Get("blastx_db_$nb\_file");
		$db    =~ s/\s+$//;

		&AssertProteinDatabase($db);
		
		$o_param->SetUnlessDefined("blastx_db_$nb\_weight", 1);
		my $level = $o_param->Get("blastx_db_$nb\_weight");
		
		$$rh_db_list{$nb}  = $db;
		$$rh_db_level{$nb} = $level;
	}
	
	# assert protein database for training is defined
	&AssertProteinDatabase($o_param->Get('training_db_file'));

	
	return;
}




=head2 function AssertProteinDatabase

 Title        : AssertProteinDatabase
 Usage        : 
 Prerequisite : 
 Function     : Assert the protein database exists and is formated. 
                If the database is a .pal, check the subdatabases are formated
 Args         : $db
 Globals      : none

=cut
sub AssertProteinDatabase
{
	my ($db) = @_;

	if ( -e $db)
	{
		Carp::croak "The database \"$db\" isn't indexed. " . 
			"Use 'formatdb' program to format it.\n" if (! (-f $db.".pin" ));
	}
	elsif (-e $db.".pal")
	{
		my ($paldir, $file) =  $db =~ /(.+)\/(.+)$/;
		my $f_ = &General::GetStreamIn($db.".pal");
		while (my $line = <$f_>)
		{
			if ($line =~ /^DBLIST (.+)$/)
			{
				my @a_ff = split(/\s+/, $1);
				foreach my $elt (@a_ff)
				{
					# if it is not an absolute path
					$elt = "$paldir/$elt" if (dirname($elt) eq ".");
					Carp::croak "The database \"$elt\" isn't indexed. " . 
						"Use 'formatdb' program to format it.\n" if (! (-f "$elt.pin"));
				}
				last;
			}
		}
	}
	else
	{
		Carp::croak "ERROR : the protein database \"$db\" doesn't exist.\n";
	}
	return;
}

=head2 procedure AssertEvidencePatternKeys

 Title        : AssertEvidencePatternKeys
 Usage        : 
 Prerequisite : -
 Function     : Assert 'evidence_list' is defined. According to this value,  
                assert all the attending keys are defined in o_param. 
                Fill $rh_files with 
                the evidence numbers and the corresponding files
 Args         : $o_param     ParamParser object
                $indir       directory of input data
                $rh_files    Ref to a hask Key: evidence type number, 
                             value: ref to an array of files
 Globals      : none

=cut
sub AssertEvidencePatternKeys
{
	my ($o_param, $indir,  $rh_files) = @_;

	$o_param->AssertDefined('evidence_list');
	my $evidence_list =  $o_param->Get('evidence_list');
	my @a_evidence_list = split(/\s+/, $evidence_list);
	
	# Assert the keys
	foreach my $nb (@a_evidence_list) # '2', '3'
	{
		my $root       = "evidence_" . $nb;
		my $typekey    = $root . "_type";
		my $patternkey = $root . "_pat";
		my $formatkey  = $root . "_format";

		$o_param->AssertAllowedValue($formatkey, 'fastq', 'bed', 'bedpe', 'fasta', 
									 'wig', 'sam', 'bam', 'pair');
		$o_param->AssertDefined($patternkey, $formatkey);

		my $format = $o_param->Get($formatkey);

		if ($format =~ /fast[qa]/)
		{
			$o_param->AssertDefined($typekey);
			$o_param->AssertAllowedValue( $typekey, 'ose', 'ope');
		}
		
		if ($format eq "bed")
		{
			$o_param->Set($typekey, 'ose');
		}
		if ($format eq "bedpe")
		{
			$o_param->Set($typekey, 'ope');
		}

		# Need to know the orientation for all format except wig
		if ($format eq "bam"      || $format eq "sam" || 
			$format =~ /fast[qa]/ || $format =~ /^bed/ )
		{
			$o_param->AssertAllowedValue($root."_small", 0, 1);
			$o_param->AssertAllowedValue($root."_oriented_end", 1, 2);
		}

		if ($format eq "wig")
		{
			$o_param->AssertAllowedValue( $root . "_strand", 'forward', 'reverse');
		}
		if ($format eq "pair")
		{
			my $ndf_filename = $o_param->Get($root. "_ndf");
			my $ndf_file     = $indir . "/" . $ndf_filename;
			$o_param->Set($root. "_ndf", $ndf_file);
			$o_param->AssertFileExists($root. "_ndf");
			$o_param->AssertDefined('prg_glint_tiling_param');
		}
		
		
		my $pattern = $o_param->Get($patternkey);
		$o_param->Set('pattern_path', "$indir/$pattern");
		my $ra_files = $o_param->GetListOfFiles('pattern_path');
		
		if (scalar(@$ra_files) == 0)
		{
			print "No file $indir/$pattern found for RNAseq data number $nb --> IGNORED\n";
			next;
		}

		if ($format =~ /fast[aq]/ && $o_param->Get($typekey) =~ /ope/)
		{
			# Check the number of files is modulo2
			if ( (scalar(@$ra_files)%2) != 0 )
			{
				Carp::croak "ERROR looking for paired-end files: " . $o_param->Get('pattern_path')
					. ". " . scalar(@$ra_files) 
					. " files found (not a pair number). Two files are attending per library.\n";
			}
			
			my @a_gp;
			&GroupPair($ra_files, \@a_gp);
			@$ra_files = @a_gp;
		}
		$o_param->Delete('pattern_path');


		$$rh_files{$nb} = $ra_files;
	}


	return;
}



=head2 procedure GroupPair

 Title        : GroupPair
 Usage        : 
 Prerequisite : $ra_file array of existing files
 Procedure    : concatenate paired files and fill the array $ra_gp 
 Args         : $ra_file Ref to an array of files
                $ra_gp: Ref to an array to fill with pair of files.
 Globals      : none

=cut
sub GroupPair
{
	my ($ra_file, $ra_gp) = @_;
	
	my $file1 = ""; # First file
	foreach my $file (@$ra_file)
	{
		if ($file1 ne "")
		{
			# Conatenate the two files
			push(@$ra_gp, "$file1,$file");
			$file1 = "";
		}
		else
		{
			$file1 = $file;
		}
	}
	
	return;
}


=head2 procedure BuildEGNMatrix

 Title        : BuildEGNMatrix
 Usage        : 
 Prerequisite : prg_egn-prok_train has to be defined
 Function     : Build the eugene matrix file, using the blastx result
 Args         : $o_param               ParamParser object
                $seq                   fasta file path
                $workingdir            working directory         
                $matrixfile            Result matrix file path
 
 Globals      : none

=cut
sub BuildEGNMatrix
{
	my ($o_param, $seq, $workingdir, $matrixfile, $o_logger, $logger_stamp) = @_;

	if ($DEBUG && (-e $matrixfile))
	{
		print STDERR "DEBUG MODE: $matrixfile used.\n";
		return;
	}

	my $training_blastxfile;
	my $training_db_file = $o_param->Get('training_db_file');

	my ($seqpath, $seqname) = $seq =~ /(.+)\/(.+)/; 
	# Check if the blastX against this protein database was previously done
	my $prev_blastx = &SearchBlastXResult($o_param, $workingdir, $seqname, $training_db_file);
	if ($prev_blastx ne "")
	{	
		print STDERR "Prev blastX result found : $prev_blastx\n" if ($DEBUG);
		$training_blastxfile = $prev_blastx;
	}
	else
	{
		print STDERR "No Prev BlastX result found for  $training_db_file\n" if ($DEBUG);
		# compute the BlastX
		&GetBlastX($o_param, $seq, $training_db_file, "mat", $workingdir, $o_logger, $logger_stamp);
		$training_blastxfile = "$workingdir/$seqname.blastmat";
	}
	
	if (&IsOKTrainingDatasets($training_blastxfile, $seq, $o_param) )
	{
		my $prg_train = $o_param->Get('prg_egn-prok_train');
		my $cmd_train = "$prg_train --cfg " . $o_param->Get('cfg') . " --infile $seq --workingdir $workingdir " . 
			"--blastxfile $training_blastxfile --matrixfile $matrixfile";
		&ExeCmd($cmd_train, $o_logger, $logger_stamp);
		
		$o_param->SetUnlessDefined('reference_matrix', $matrixfile);
	}
	else
	{
		if ($o_param->IsDefined('reference_matrix'))
		{
			&PrintInfo("Not enough blastx hsp to build the matrix of $seq ==> Use a previous one.\n", $o_logger, $logger_stamp);
		
			my $cmd = "cp " . $o_param->Get('reference_matrix') . " $matrixfile";
			&ExeCmd($cmd, $o_logger, $logger_stamp);
		}
		else
		{
			# If the blastx file is not empty
			if (-s $training_blastxfile)
			{
				&PrintInfo("WARNING: build the matrix for $seq but not enough blastx hsp (no previous matrix found)\n", 
						   $o_logger, $logger_stamp);
				
				my $prg_train = $o_param->Get('prg_egn-prok_train');
				my $cmd_train = "$prg_train --cfg " . $o_param->Get('cfg') . " --infile $seq --workingdir $workingdir " . 
					"--blastxfile $training_blastxfile --matrixfile $matrixfile";
				&ExeCmd($cmd_train, $o_logger, $logger_stamp);
			}
			else
			{
				my $info = "ERROR: impossible to train EuGene on $seq. " . 
					"BlastX results are mandatory but file $training_blastxfile is empty. " . 
					" You should try an other training protein database 'training_db_file'\n";
				$o_logger->Trace("$info", $logger_stamp, $LOGGER_LEVEL_EXEC);
				Carp::croak ($info);
			}
		}	
	}

	return;
}

=head2 function IsOKTrainingDatasets

 Title        : IsOKTrainingDatasets
 Usage        : 
 Prerequisite : -
 Function     : Count the number of HSP after filtering (len, pci, evalue)
                Return 1 if its upper than the threshold
 Args         : $blastxfile M8 blastx result file
                $seqfile fasta file
                $o_param
 Globals      : none

=cut
sub IsOKTrainingDatasets
{
	my ($blastxfile, $seqfile, $o_param) = @_;
	
	my %h_seq;
	&FastaToHash($seqfile,\%h_seq);
	my %h_hit  = ();
	my $hsp_nb = 0;
	my $hsp_pci    = $o_param->Get('prg_egn-prok_train_hsp_training_pci');
	my $hsp_evalue = $o_param->Get('prg_egn-prok_train_hsp_training_evalue');
	my $hsp_length = $o_param->Get('prg_egn-prok_train_hsp_training_length');

	my $f_blast = &GetStreamIn($blastxfile);
	my %h_deja=();
	while(my $line=<$f_blast>)
	{
		next if ( $line =~ /^#/ );

		&ParseM8Line(\$line,1,\%h_hit);
		next if ( ! ( ($h_hit{pci} >= $hsp_pci) && 
					  ($h_hit{"e-value"} <= $hsp_evalue) && 
					  ($h_hit{len} >= $hsp_length))  );	# for blastx the length is in aa
		next if ( defined($h_deja{$h_hit{target}}) );
		$h_deja{$h_hit{target}}= 1 ;

		my($start,$end) = &MinMax($h_hit{qstart},$h_hit{qend});
		next if ( ($end-$start+1) % 3 != 0 );
		my ($query) = $h_hit{query};
		my $sseq = &ExtractSeq(\$h_seq{$h_hit{query}}{sequence},$start,$end,$h_hit{strand});
		next if ( $sseq =~ /N/ );
		$sseq =~ s/\w\w\w$//; # j'enleve le dernier codon
		my $ok = &TRUE;
		while ( $sseq =~ /(...)/g )
		{	
			my $codon = $1;
			if ( $codon =~ /TAA|TGA|TAG/ )
			{
				$ok = &FALSE;
				last;
			}
		}
		$hsp_nb++ if ( $ok );
	}
	$f_blast->close;

	return 1 if ($hsp_nb >= $o_param->Get('prg_egn-prok_train_hsp_training_min_nb'));
	return 0;
}


=head2 procedure BuildEGNProkParameterFile

 Title        : BuildEGNProkParameterFile
 Usage        : 
 Prerequisite : $o_param has the following keys: 'organism' 'prg_eugene_cfg_model' 
                'prg_eugene_cfg_gc' 'prg_eugene_cfg_blastx' and 'gc_percent'
 Function     : Create the complete eugene configuration file
 Args         : $o_param                       ParamParser object
                $cfg_file                      File to create
                $rh_db_level           hashref Ref to a hash : key=blastx number, 
                                               value=level of confidence of the blastx result
                $annotastruct_content  string  Annotastruct instances description
                $annotastruct_cpt      int     Index of the last AnnotaStruct instance        
 Globals      : none

=cut
sub BuildEGNProkParameterFile
{
	my ($o_param, $cfg_file, $rh_db_level, $annotastruct_content, $annotastruct_cpt) = @_;
   
	# Result file
	my $f_cfg = &General::GetStreamOut($cfg_file);

	# Copy the conf model file
	my $f_model = $o_param->GetStreamIn('prg_eugene_cfg_model');
	while (my $cfg_line = <$f_model>)
	{
		$cfg_line = "EuGene.organism ". $o_param->Get('organism')."\n" if ($cfg_line =~ /^EuGene.organism/);
		print $f_cfg $cfg_line;
	}	
	$f_model->close;

    # add to the cfg file the part depending on the GC% of the organism
	my $cfg_gc_model = ($o_param->Get('gc_percent') eq "high") ? $o_param->Get('prg_eugene_cfg_high_gc'):$o_param->Get('prg_eugene_cfg_low_gc');
	my $f_gc_model = &General::GetStreamIn($cfg_gc_model);
	while (my $cfg_line = <$f_gc_model>)
	{
		print $f_cfg $cfg_line;
	}	
	$f_gc_model->close;


	# Add the instances of AnnotaStruct
	print $f_cfg $annotastruct_content;
	print $f_cfg "Sensor.AnnotaStruct.use ". ($annotastruct_cpt+1) . "\n##\n";

	
	# Create the blastX instance
	if ($o_param->Get('blastx_db_list') ne "")
	{
		my $blastx_line;
		my $f_blastx = $o_param->GetStreamIn('prg_eugene_cfg_blastx');
		while (my $line = <$f_blastx>)
		{
			$line =~ s/\s+$//;
			$line =~ s/\s+/\t/g;
			$line .= "\n";
			$blastx_line .= $line;
		}
		$f_blastx->close;
		my $levels;
		foreach my $nb (keys %$rh_db_level)
		{
			my $val = $$rh_db_level{$nb};
			$blastx_line =~ s/BlastX\.level$nb\*\t0\.0/BlastX\.level$nb\*\t$val/;
			$levels .= $nb;
		}
		$blastx_line =~ s/BlastX\.levels\t\d/BlastX\.levels\t$levels/;
		print $f_cfg $blastx_line;
		print $f_cfg "Sensor.BlastX.use	1";
	}
	else
	{
		print $f_cfg "Sensor.BlastX.use	0";
	}


	$f_cfg->close;

	return;
}
	



=head2 procedure GetTemplate

 Title        : GetTemplate
 Procedure     :fill $rh_template
 Args         : $o_param   
                $rh_template key=position/region/ncrna/tstart/blastx, 
                             value=template of the cfg part for this plugin
 Globals      : none

=cut
sub GetTemplate
{
	my ($o_param, $rh_template) = @_;

	# configuration file part for blastX and AnnotaStruct
	$$rh_template{position} = &ReadFile($o_param->Get('prg_eugene_cfg_rnaseq_annotastruct'));
	$$rh_template{region}   = &ReadFile($o_param->Get('prg_eugene_cfg_region_annotastruct'));
	$$rh_template{ncrna}    = &ReadFile($o_param->Get('prg_eugene_cfg_ncrna_annotastruct'));
	$$rh_template{tstart}   = &ReadFile($o_param->Get('prg_eugene_cfg_tstart_annotastruct'));
	$$rh_template{start}    = &ReadFile($o_param->Get('prg_eugene_cfg_start_annotastruct'));
	
	return;
}




=head2 procedure GetGlobalEvidence

 Title        : GetGlobalEvidence
 Usage        : 
 Prerequisite : 
 Function     : Launch the compute which have to be done together on all the input sequences 
                * Case FASTQ: mapping and mapping filtering
                * Case wig: 
                Then split the results in the appropriate directory  
 Args         : $o_param 
                $rh_evidence_type ref to a hash  key=evidence nb, value=ref to an array of files
                $workingdir       dir path       Output directory
                $seqfile          file           Input multifasta sequence file
                $rootname         String         Prefix of the output files 
 Globals      : none

=cut
sub GetGlobalEvidence
{
	
	my ($o_param, $rh_evidence_type, $workingdir, $seqfile, 
		$rootname, $o_logger, $logger_stamp, $uniq) = @_;
	
	chdir($workingdir);

	# For each type of evidence
	foreach my $nb (sort keys(%$rh_evidence_type))
	{
		my $prefix        = "evidence_" . $nb;
		my $orientation   = $o_param->Get($prefix . "_type");
		my $format        = $o_param->Get($prefix . "_format");
		my $small         = $o_param->Get($prefix . "_small");

		&PreProcessing($prefix, $seqfile, $workingdir, 
					   $o_param, $o_logger, $logger_stamp);

		# for each file of evidence
		foreach my $evidence (@{$$rh_evidence_type{$nb}})
		{
			my $evidence_name = &GetEvidenceName($evidence);
			
			my $tomap_fastq;
			my $tomap_format;

			# Conversion SAM|BAM to FASTQ
			if ($format =~ /[sb]am/)
			{
				my $bam_file;
				if ($format eq "sam")
				{
					$bam_file = "$workingdir/$rootname.$evidence_name.bam";  
					&Sam2Bam($evidence, $bam_file, $o_param, $o_logger, $logger_stamp);
				}
				else
				{
					$bam_file = $evidence;
				}
				my $fastq_file_root = "$workingdir/$rootname.$evidence_name";  
				($tomap_fastq, $orientation) = 
					&Bam2Fastq($bam_file, $fastq_file_root, $o_param, $o_logger, $logger_stamp);
				$tomap_format = "fastq";
				&SetBamOrientation($prefix, $orientation, $o_param);
			}
			
			if ($format =~ /fast[aq]/)
			{
				$tomap_fastq     = $evidence;
				$tomap_format    = $format;
			}

			
			# MAPPING
			my $mapping_bed_file;
			if ($format =~ /fast[aq]/ || $format=~ /[sb]am/) 
			{
				$mapping_bed_file = "$workingdir/$rootname.$evidence_name.bed";  
				&Mapping($seqfile, $tomap_fastq, $tomap_format, 
						 $orientation, $mapping_bed_file, $small, 
						 $o_param, $o_logger, $logger_stamp);
				$mapping_bed_file .= ".gz" if ( -e "$mapping_bed_file.gz" );
			}
			elsif ($format eq "bed" || $format eq "bedpe")
			{
				$mapping_bed_file = $evidence;
			}
			

			# FILTERING OF MAPPING RESULT
			if ($format =~ /fast[aq]/ || $format eq "bed" || $format eq "bedpe" || $format =~ /[sb]am/)
			{
				my $filter_bed_file = "$workingdir/$rootname.$evidence_name.$MAPPING_EXT";  
				&MappingFiltering($mapping_bed_file, $orientation, $small,  
								  $filter_bed_file, $prefix, $o_param, 
								  $o_logger, $logger_stamp);
				
				# Split the bed result by sequence
				&SplitBed($filter_bed_file, $evidence, $uniq, $o_param, $o_logger, $logger_stamp);
			}

			# Tiling array
			if ($format eq "pair")
			{
				&SplitPair2Expr($evidence,  $seqfile, $workingdir, $prefix, 
								$rootname, $o_param, $o_logger, $logger_stamp);
			}

			# wig file
			if ($format eq "wig")
			{
				&SplitWig2Expr($o_param, $evidence, $o_param->Get($prefix . "_strand"), 
							   $workingdir, $rootname, $o_logger, $logger_stamp);	
			}
		}
	}

	return;
}




=head2 procedure PreProcessing

 Title        : PreProcessing
 Procedure    : Do primary computing needed for all files of the evidence type
                If format = pair, save the coordinates of the sequences included in the ndf file
 Prerequisite : None
 Args         : $prefix     Root of the key of the current evidence type. Ex evidence_2 
                $seqfile    Multifasta file of the genomic sequènces
                $workingdir Temp directory
                $o_param
                $o_logger
                $logger_stamp
 Globals      : none

=cut
sub PreProcessing
{
	my ($prefix, $seqfile, $workingdir, $o_param, $o_logger, $logger_stamp) = @_;
	
	if ($o_param->Get($prefix . "_format") eq "pair")
	{
		&LoadNdfSequences($prefix, $seqfile, $workingdir, $o_param, $o_logger, $logger_stamp);
	}

	return;
}






=head2 procedure SetBamOrientation

 Title        : SetBamOrientation
 Procedure     :Check that all the sam/bam files have the same orientation. 
                Save the orientation in the o_param object
 Prerequisite : 
 Args         : $prefix
                $orientation
                $o_param
 Globals      : none

=cut
sub SetBamOrientation
{
	my ($prefix, $orientation, $o_param) = @_;
	
	if ($o_param->IsDefined($prefix . "_type") && 
		$o_param->Get($prefix . "_type") ne $orientation)
	{
		Carp::croak("Problem with the bam data $prefix. All the files " . 
					"have to be paired-end or single end, not both.\n " . 
					"You should create two evidence families to fix that.\n");
	}
	$o_param->Set($prefix . "_type", $orientation);
	
	return;
}

	
=head2 procedure Sam2Bam

 Title        : Sam2Bam
 Procedure     :Convert sam file to bam
 Prerequisite : prg_samtools has to be defined. gzip has to be in the path
 Args         : $sam_file Sam file
                $bam_file result bam file
                $o_param
                $o_logger
                $logger_stamp
 Globals      : none

=cut
sub Sam2Bam
{
	my ($sam_file, $bam_file, $o_param, $o_logger, $logger_stamp) = @_;

	if ($DEBUG && (-e $bam_file))
	{
		print STDERR "DEBUG MODE: $bam_file already computed.\n";
		return;
	}
	my $in_exe_path = &GetCompressPath($sam_file);    # "/bin/gzip" for example
    if ($in_exe_path ne "")
    {
        $in_exe_path .= " -cd ";             # Ex "/bin/gzip -cd "
    }
	else
	{
		$in_exe_path = "cat ";
	}
	# -S entry=sam; -b to convert in bam
    my $cmd = "$in_exe_path $sam_file | " . $o_param->Get('prg_samtools') . " view -b -S - > $bam_file ";
	&ExeCmd($cmd, $o_logger, $logger_stamp);

	return;

}

=head2 function Bam2Fastq

 Title        : Bam2Fastq
 Procedure     :Convert bam file to fastq and return the fatsq result files(s) 
                and the type (ope or ose)
 Prerequisite : prg_bam2fastq has to be defined. gzip is in the path
 Args         : $bam_file        Input bam file
                $fastq_file_root Root of the result fastq file(s)
                $o_param
                $o_logger
                $logger_stamp
 Globals      : none

=cut
sub Bam2Fastq
{
	my ($bam_file, $fastq_file_root, $o_param, $o_logger, $logger_stamp) = @_;

	my $fastqfile;   # name of the result fastq file(s)
	my $orientation; # ope or ose
	
	if ($DEBUG)
	{
		if ( (-e "$fastq_file_root\_1.fastq.gz") && (-e "$fastq_file_root\_2.fastq.gz"))
		{
			print STDERR "DEBUG MODE: $fastq_file_root\_1.fastq.gz and $fastq_file_root\_2.fastq.gz already computed.\n";
			$fastqfile   = "$fastq_file_root\_1.fastq.gz,$fastq_file_root\_2.fastq.gz";
			$orientation = "ope";
			return ($fastqfile, $orientation);
		}
		elsif  (-e "$fastq_file_root.fastq.gz")
		{
			print STDERR "DEBUG MODE: $fastq_file_root.fastq.gz already computed\n";
			$fastqfile   = "$fastq_file_root.fastq.gz";
			$orientation = "ose";
			return ($fastqfile, $orientation);
		}
	}

	my $cmd = $o_param->Get('prg_bam2fastq') . " --output $fastq_file_root#.fastq $bam_file > /dev/null 2>&1";
   	
	&ExeCmd($cmd, $o_logger, $logger_stamp);
	
	if (-e "$fastq_file_root\_1.fastq" && "$fastq_file_root\_2.fastq")
	{
		$orientation = "ope";
		$cmd         = "gzip -f $fastq_file_root\_1.fastq  $fastq_file_root\_2.fastq";
		$fastqfile   = "$fastq_file_root\_1.fastq.gz,$fastq_file_root\_2.fastq.gz";
	}
	else
	{
		$orientation = "ose";
		$cmd         = "gzip -f $fastq_file_root.fastq";
		$fastqfile   = "$fastq_file_root.fastq.gz";
	}
	# Zip files
	&ExeCmd($cmd, $o_logger, $logger_stamp);
	
	#print "Return $fastqfile, $orientation\n";
	return ($fastqfile, $orientation);
}



=head2 procedure SplitWig2Expr

 Title        : SplitWig2Expr
 Procedure    : Read a wig file. Convert a wig file in GFF3 files
                One containing expression depth with filtering an other without (used to search tstart)
                Split thes files per sequence and save them in workingdir
 Prerequisite : prg_wig_to_expr , prg_wig_to_expr_annot_param
                prg_bed_to_expr_annot_scoremax  prg_bed_to_expr_annot_scalingfactor
                prg_bed_to_expr_annot_min_coverage_percent has to be defined
 Args         : $o_param   
                $wigfile
                $strand
                $workingdir
                $rootname
 Globals      : none

=cut

sub SplitWig2Expr
{
	my ($o_param, $wigfile, $strand, $workingdir, $rootname, $o_logger, $logger_stamp) = @_;

	my $ra_1 = &Wig2Expr4EGN($o_param, $wigfile, $strand, $workingdir, $rootname,
							 $o_logger, $logger_stamp);
	my $ra_2 = &Wig2Expr4tStart($o_param, $wigfile, $strand, $workingdir, $rootname, 
								$o_logger, $logger_stamp);			
	
	my @a_created_files = (@$ra_1, @$ra_2);

	return \@a_created_files;
}


=head2 procedure Wig2Expr4EGN

 Title        : Wig2Expr4EGN
 Procedure    : Convert a wig file in expression files (.region and .position gff3 files). 
                Split by seq and move it in the appropriated sequence directory
 Prerequisite : prg_wig_to_expr , prg_wig_to_expr_annot_param
                prg_bed_to_expr_annot_scoremax  prg_bed_to_expr_annot_scalingfactor
                prg_bed_to_expr_annot_min_coverage_percent has to be defined
 Args         : $o_param   
                $wigfile
                $strand
                $workingdir
                $rootname
 Globals      : none

=cut
sub Wig2Expr4EGN
{
	my ($o_param, $wigfile, $strand, $workingdir, $rootname, $o_logger, $logger_stamp) = @_;

	my $evidence_name    = &GetEvidenceName($wigfile);
	my $tmpoutfilename   = "$rootname.$evidence_name.$WIG_EXT";
	my $tmpoutfile       = "$workingdir/$tmpoutfilename";  	

	my $cde;

	$cde = $o_param->Get('prg_wig_to_expr'). " " . $o_param->Get('prg_wig_to_expr_annot_param') ;
	$cde .= " --wigfile $wigfile --outprefix $tmpoutfile --strand $strand";
	$cde .=	" --tag WIG --split ";
	$cde .= " --scoremax "      . $o_param->Get('prg_bed_to_expr_annot_scoremax');
	$cde .= " --scalingfactor " . $o_param->Get('prg_bed_to_expr_annot_scalingfactor');
	$cde .= " --min_coverage_percent "  . $o_param->Get('prg_bed_to_expr_annot_min_coverage_percent');

	&ExeCmd($cde, $o_logger, $logger_stamp);
	
	# For each sequence, copy region.gff3 and position.gff3 file
	my $ra_created_files;

	my @a_ = glob("$workingdir/*.$tmpoutfilename.$EXPR_POS_EXT.gff3");
	foreach my $file (@a_)
	{
		my ($seqname) = $file =~ /$workingdir\/(.+)\.$tmpoutfilename/;
		my $seqdir    = &GetSeqDir($seqname, $o_param);
		my $reg_file  = "$workingdir/$seqname.$tmpoutfilename.$EXPR_REG_EXT.gff3";
		# If the sequence directory exists, copy in 
		if (-d $seqdir)
		{
		    # Copy position file
			$cde = "mv $file $seqdir/$seqname.$evidence_name.$EXPR_POS_EXT.gff3";
			&ExeCmd($cde, $o_logger, $logger_stamp);
			push (@$ra_created_files, "$seqdir/$seqname.$evidence_name.$EXPR_POS_EXT.gff3");
			# Copy region file
			$cde = "mv $reg_file $seqdir/$seqname.$evidence_name.$EXPR_REG_EXT.gff3";
			&ExeCmd($cde, $o_logger, $logger_stamp);
			push (@$ra_created_files, "$seqdir/$seqname.$evidence_name.$EXPR_REG_EXT.gff3");
			
		}
		else
		{
			&PrintInfo( "Warning: $wigfile contains data from $seqname but this sequence ". 
						"is not in the input/genome directory. Data ignored\n", $o_logger, $logger_stamp);
			&ExeCmd("gzip $file $reg_file", $o_logger, $logger_stamp);
		}
	}

	return $ra_created_files ;
}

=head2 procedure Wig2Expr4tStart

 Title        : Wig2Expr4tStart
 Procedure     :Convert wig file in expression file for tstart detection. 
                Split by seq and move it in the working sequence directory
 Args         : $o_param   
                $wigfile
                $strand
                $workingdir
                $rootname
 Globals      : none

=cut
sub Wig2Expr4tStart
{
	my ($o_param, $wigfile, $strand, $workingdir, $rootname, $o_logger, $logger_stamp) = @_;

	my @a_created_files;

	my $evidence_name = &GetEvidenceName($wigfile);
	my $outfilename   = "$rootname.$evidence_name.$TSTART_POS_EXT";
	my $outfile       = "$workingdir/$outfilename";  

	my $cde;
	$cde = $o_param->Get('prg_wig_to_expr'). " " . $o_param->Get('prg_wig_to_expr_annot_param') ;
	$cde .= " --wigfile $wigfile --outprefix $outfile --strand $strand";
	$cde .=	" --tag WIG --split ";
	$cde .= " --scoremax "      . $o_param->Get('prg_bed_to_expr_tstart_scoremax');
	$cde .= " --scalingfactor " . $o_param->Get('prg_bed_to_expr_tstart_scalingfactor');
	$cde .= " --min_coverage "  . $o_param->Get('prg_bed_to_expr_tstart_min_coverage');

	&ExeCmd($cde, $o_logger, $logger_stamp);

	my @a_ = glob("$workingdir/*.$outfilename.position.gff3");
	foreach my $file (@a_)
	{
		my ($seqname) = $file =~ /$workingdir\/(.+)\.$outfilename.position.gff3/;
		my $reg_file  = "$workingdir/$seqname.$outfilename.region.gff3";
		my $seqdir    = &GetWorkSeqDir($seqname, $o_param);
		if (-d $seqdir)
		{
			$cde = "mv $file $seqdir/$seqname.$evidence_name.$TSTART_POS_EXT.gff3";
			&ExeCmd($cde, $o_logger, $logger_stamp);
			push(@a_created_files, "$seqdir/$seqname.$evidence_name.$TSTART_POS_EXT.gff3");
			unlink ($reg_file);
		}
		else
		{
			&PrintInfo("Warning: $wigfile contains data from $seqname but this " .
					   "sequence is not in the input/genome directory. Data ignored\n", 
					   $o_logger,$logger_stamp);
			&ExeCmd("gzip $file $reg_file", $o_logger, $logger_stamp);
		}
	}

	return \@a_created_files;
}

=head2 procedure SplitBed

 Title        : SplitBed
 Procedure     :Split bed file per sequence and move files in sequence specific directories
 Args         : $bed_file Bed file  
                $evidence 
                $uniq
 Globals      : none

=cut

sub SplitBed
{
	my ($bed_file, $evidence, $uniq, $o_param, $o_logger, $logger_stamp) = @_;
	
	my ($workingdir, $bedname) =  $bed_file =~ /(.+)\/(.+)/;
	my $prg_split_bed = $o_param->Get('prg_split_bed');
	my $seqdir        = $o_param->Get('rootseqdir');

	# split the results and move it in the sequence directory
	chdir($workingdir);

	my $cmd = "$prg_split_bed --in $bed_file";
	&ExeCmd($cmd, $o_logger, $logger_stamp);
	
	my @a_ = glob("$workingdir/*.$bedname");
	foreach my $file (@a_)
	{
		my ($seqname) = $file =~ /$workingdir\/(.+)\.$bedname/;
		
		my $cmd = "mv $file $seqdir/$seqname/work.$uniq/" . &GetMappingFileName($seqname, $evidence);
		&ExeCmd($cmd, $o_logger, $logger_stamp);
	}

	return;
}	

=head2 procedure GetEvidences

 Title        : GetEvidences
 Usage        : 
 Prerequisite : $prefix . "_pat" $prefix . "_type"  $prefix . "_format" $prefix . "_small" exists
 Function     : Manage one type of RNAseq data
 Args         : $o_param 
                $seqfile file     Input fasta sequence file
                $seqid  string    Name of the sequence
                $seqlen integer   Length of the sequence  
                $seqname string   Name of the sequence
                $workdir dir path Output directory
                $prefix           Prefix of the type of evidence ex: "evidence_2"
                $annotastruct_cpt Use to count the number of rnaseq data computed. 
                                  This value is update by this procedure.
                $annotastruct_cpt,
                $rh_template
                $cfg_content 
 Globals      : none

=cut
sub GetEvidences
{
	my ($o_param, $seqfile, $seqid, $seqlen, $seqname, $workdir, $prefix, $ra_evidences_files, 
		$r_annotastruct_cpt, $rh_template, $r_cfg_content, $o_logger, $logger_stamp, $uniq) = @_;

	my $orientation = $o_param->Get($prefix . "_type");
	my $format      = $o_param->Get($prefix . "_format");

	&PrintInfo("Load $prefix RNAseq data\n", $o_logger, $logger_stamp);
	
	my $data_type = "";
	$data_type = "MAPPED" if ($format =~ /fast[aq]/ || $format eq "bed" || 
							  $format eq "bedpe" || $format =~ /[sb]am/);
	
	foreach my $evidence (@$ra_evidences_files)
	{
		&PrintInfo("File $evidence..\n", $o_logger, $logger_stamp);

		# Name of this evidence			
		my $evidence_name = &GetEvidenceName($evidence);
		my $expr_position_file = "$seqfile.$evidence_name.$EXPR_POS_EXT.gff3";
		my $expr_region_file   = "$seqfile.$evidence_name.$EXPR_REG_EXT.gff3";
		my $tstart_file_gff3   = "$seqfile.$evidence_name.$TSTART_EXT.gff3"; 
		
		if (!$DEBUG || 
			($DEBUG && !(-e $expr_position_file && -e $expr_region_file && -e $tstart_file_gff3)))
		{
			my $expr_4tstart_file = "$workdir/$seqname.$evidence_name.$TSTART_POS_EXT.gff3";
			if ($data_type eq "MAPPED")
			{
				# 0 Get the already compute mapping file
				my $filtering_bed_file = "$workdir/" . &GetMappingFileName($seqname, $evidence); 
				if (-e $filtering_bed_file)
				{
					# 1 Build the expression files to integrate in EuGene			
					&Bed2Expr4EGN($filtering_bed_file, $orientation, $expr_position_file, 
							  $expr_region_file, $o_param, $o_logger, $logger_stamp, $uniq );
					
					# 2 Build the expression files to detect start of transcription
					&Bed2Expr4tStart($filtering_bed_file, $orientation, $expr_4tstart_file, 
									 $o_param, $o_logger, $logger_stamp, $uniq );
				}
				else
				{
					&PrintInfo("No mapping result for $seqfile $evidence_name\n", 
							   $o_logger, $logger_stamp);
				}
			}
			if (-e $expr_4tstart_file)
			{
				# 4 Detect start of transcription
				&TStartDetection($expr_4tstart_file, $tstart_file_gff3, $seqid, $seqlen, $seqname, 
								 $workdir, $evidence_name, $o_param, $o_logger, $logger_stamp);
				&Gzip( $o_logger, $logger_stamp, $expr_4tstart_file);
			}
			else
			{
				&PrintInfo("No expression result for $seqfile $evidence_name\n", 
						   $o_logger, $logger_stamp);
			}
		}
		else
		{
			print STDERR "DEBUG MODE: use region, position and tstart files for $evidence_name.\n";
		}
		
		&CompleteCfg($r_cfg_content, $r_annotastruct_cpt, "$evidence_name.$EXPR_POS_EXT", 
					 $$rh_template{position}, $expr_position_file) if (-e $expr_position_file);
		&CompleteCfg($r_cfg_content, $r_annotastruct_cpt, "$evidence_name.$EXPR_REG_EXT", 
					 $$rh_template{region}, $expr_region_file)  if (-e $expr_region_file);
		&CompleteCfg($r_cfg_content, $r_annotastruct_cpt, "$evidence_name.$TSTART_EXT", 
					 $$rh_template{tstart}, $tstart_file_gff3) if (-e $tstart_file_gff3);
		
	}
 

	return;
}


=head2 procedure CompleteCfg

 Title        : CompleteCfg
 Usage        : 
 Prerequisite : 
 Function     : Complete $cfg_content string. If $evidence_file exists, 
                add an instance of a plugin. substitute CPT by $cpt 
                and EXT by $ext in the template and concatenate this string to cfg_content
 Args         : $r_cfg_content   String ref  Parameter file part
                $r_cpt           Ref to int  Cpt to create new instance of the plugin
                $ext             string      Extension of the instance
                $template        string      Template of a plugin instance
 Globals      : none

=cut
sub CompleteCfg
{
	my ($r_cfg_content, $r_cpt, $ext, $template, $evidence_file) = @_;
	
	if (-e $evidence_file)
	{
		$$r_cpt++;
		my $content = $template;
		$content =~ s/CPT/$$r_cpt/g;
		$content =~ s/EXT/$ext/g;
		$content .= "##\n";
		
		$$r_cfg_content .= $content;
	}

	return;
}


=head2 procedure TStartDetection

 Title        : TStartDetection
 Usage        : 
 Prerequisite : prg_tstart_detection and prg_tstart_detection_param prg_tstart_detection_derivative_threshold
                prg_tstart_detection_log_threshold prg_tstart_detection_boostscore_factor
                prg_tstart_detection_boost_ig, prg_tstart_detection_tstart_ratio  are defined in $o_param
 Procedure     : 
 Args         : $infile input fasta file
                $seqid  name of the sequence
                $seqlen Len of the sequence
                $outfile 
                $format input format gff3 or wig
                $o_param 
 Globals      : none

=cut
sub TStartDetection
{
	my ($infile, $outfile, $seqid, $seqlen, $seqname, $workdir, $evidence_name,
		$o_param, $o_logger, $logger_stamp) = @_;

	# Test to catch case where no 
	if (-e $infile)
	{
		my $tmpfile = "$workdir/$seqname.$evidence_name.$TSTART_EXT.tmp"; 
		
		my $cmd;
		$cmd = $o_param->Get('prg_tstart_detection') . " ".  $o_param->Get('prg_tstart_detection_param');
		$cmd .= " --derivative_threshold " . $o_param->Get('prg_tstart_detection_derivative_threshold'); 
		$cmd .= " --log_threshold "        . $o_param->Get('prg_tstart_detection_log_threshold');
		$cmd .= " --boostscore_factor "    . $o_param->Get('prg_tstart_detection_boostscore_factor');
		$cmd .= " --tstart_ratio "         . $o_param->Get('prg_tstart_detection_tstart_ratio');
		$cmd .= " --boost_ig "             if ($o_param->Get('prg_tstart_detection_boost_ig') eq "TRUE");
		$cmd .= " --in $infile --out $tmpfile --seq_len $seqlen --seq_name \"$seqid\" 2> /dev/null";
		
		&ExeCmd($cmd, $o_logger, $logger_stamp);
		
		$cmd = "mv $tmpfile.gff3 $outfile";
		&ExeCmd($cmd, $o_logger, $logger_stamp);
	}
	else
	{
		# No expression file for this sequence, for this evidence
		# Possible if evidence is a wig file which contain no data about the sequence
		&PrintInfo("INFO TStartDetection: $seqid $evidence_name No data found in $evidence_name for $seqid.\n", 
				   $o_logger, $logger_stamp);
	}

	return;
}


=head2 function SearchBlastXResult

 Title        : SearchBlastXResult
 Usage        : 
 Prerequisite : blastx_db_list and training_db_file are defined in o_param
 Function     : Search if a previous BlastX result
 Args         : $o_param 
                $workingdir
                $seqname
                $training_db_file
              
 Globals      : none

=cut
sub SearchBlastXResult
{
	my ($o_param, $workingdir, $seqname, $training_db_file) = @_;

	my $list    = $o_param->Get('blastx_db_list'); # "2 5 6"
	my @a_db_nb = split(/\s+/, $list); 
	
	foreach my $nb (@a_db_nb)
	{
		if ($o_param->Get("blastx_db_$nb\_file") eq $training_db_file)
		{
			if ( -f $workingdir ."/". $seqname. ".blast$nb")
			{
				return $workingdir ."/". $seqname. ".blast$nb";
			}
		}
	}

	return "";
}

=head2 procedure GetBlastX

 Title        : GetBlastX
 Usage        : 
 Prerequisite : prg_get_blastx has to be defined
 Function     : Launch blastx. Result file is seq.blast.$nb
 Args         : $o_param  ParamParser object
                $seq      fasta file path
                $db       protein fasta file path
                $nb       number of the database. Use to build the blastx result file
                $workingdir working directory
 Globals      : none

=cut
sub GetBlastX
{
	my ($o_param, $seq, $db, $nb, $workingdir, $o_logger, $logger_stamp) = @_;


	my $outfile = "$seq.blast$nb";
	
	if ( $DEBUG && (-e "$outfile.gff3") )
	{
		print STDERR "DEBUG MODE: $outfile.gff3 used.\n";
		return;
	}
	
	my $cmd = $o_param->Get('prg_get_blastx') . " --sequence $seq --cfg " . 
		$o_param->Get('cfg') . 
		" --db $db --outfile $outfile --workingdir $workingdir";
	$cmd .= " --verbose " if ($DEBUG);
	&ExeCmd($cmd, $o_logger, $logger_stamp);

	# Mv $seq.blast$nb in $workingdir (we only keep the .gff3)
	$cmd = "mv $outfile $workingdir/";
	&ExeCmd($cmd, $o_logger, $logger_stamp);

	return;	
}


=head2 procedure GettRNAscanSE

 Title        : GettRNAscanSE
 Usage        : 
 Prerequisite : prg_get_trnascanse has to be defined
 Function     : Launch trnascanse. Result file is seq.trnascanse.gff3
 Args         : $o_param  ParamParser object
                $seq      fasta file path
 Globals      : none

=cut
sub GettRNAscanSE
{	
	my ($o_param, $seq, $r_annotastruct_cpt, $rh_cfg_template, $r_cfg_content, $o_logger, $logger_stamp) = @_;

	return if ($o_param->IsDefined('no_ncrna_detection') || $o_param->IsDefined('no_trnascan'));

	my $trnascan_file = "$seq.$TRNASCAN_EXT.gff3";
	if (!$DEBUG || ($DEBUG && !(-e $trnascan_file)) )
	{
		my $cmd = $o_param->Get('prg_get_trnascanse') . " --sequence $seq --cfg " . 
			$o_param->Get('cfg') . 
			" --outfile $trnascan_file --workingdir " . $o_param->Get('workingdir');
		&ExeCmd($cmd, $o_logger, $logger_stamp);
	}
	else
	{
		print STDERR "DEBUG MODE: use previous $seq.$TRNASCAN_EXT.gff3\n";
	}

	&CompleteCfg($r_cfg_content, $r_annotastruct_cpt, $TRNASCAN_EXT, $$rh_cfg_template{ncrna}, $trnascan_file);

	return;	
}

=head2 procedure GetRfamScan

 Title        : GetRfamScan
 Usage        : 
 Prerequisite : prg_get_rfamscan has to exist
 Function     : Launch rfamscan. Result file is seq.rfamscan.gff3
 Args         : $o_param  ParamParser object
                $seq      fasta file path
 Globals      : none

=cut
sub GetRfamScan
{	
	my ($o_param, $seq, $r_annotastruct_cpt, $rh_cfg_template, $r_cfg_content, $o_logger, $logger_stamp) = @_;

	return if ($o_param->IsDefined('no_ncrna_detection') || $o_param->IsDefined('no_rfamscan') );

	my $rfam_file = "$seq.$RFAMSCAN_EXT.gff3";
	if (!$DEBUG || ($DEBUG && !(-e $rfam_file)) )
	{
		
		my $cmd = $o_param->Get('prg_get_rfam') . " --fasta_file $seq " .
			" --output_file $rfam_file --working_dir " . $o_param->Get('workingdir') . 
                        " --cm_file " . $o_param->Get('rfam_cm_file');
		$cmd .= " --verbose " if ($DEBUG);
		&ExeCmd($cmd, $o_logger, $logger_stamp);
	}
	else
	{
		print STDERR "DEBUG MODE: use previous $rfam_file\n";
	}

	&CompleteCfg($r_cfg_content, $r_annotastruct_cpt, $RFAMSCAN_EXT, $$rh_cfg_template{ncrna}, $rfam_file);

	return;	
}


=head2 procedure GetRnammer

 Title        : GetRnammer
 Usage        : 
 Prerequisite : prg_get_Rnammer has to be defined
 Function     : Launch rnammer. Result file is seq.rnammer.gff3
 Args         : $o_param  ParamParser object
                $seq      fasta file path
 Globals      : none

=cut
sub GetRnammer
{	
	my ($o_param, $seq, $r_annotastruct_cpt, $rh_cfg_template, $r_cfg_content, $o_logger, $logger_stamp) = @_;

	return if ($o_param->IsDefined('no_ncrna_detection') || $o_param->IsDefined('no_rnammer'));
	
	my $rnammer_file = "$seq.$RNAMMER_EXT.gff3";
	if (!$DEBUG || ($DEBUG && !(-e $rnammer_file)) )
	{	
		my $cmd = $o_param->Get('prg_get_rnammer') . " --sequence $seq --cfg " . 
			$o_param->Get('cfg') . 
			" --outfile $rnammer_file --workingdir " . $o_param->Get('workingdir');
		&ExeCmd($cmd, $o_logger, $logger_stamp);
	}
	else
	{
		print STDERR "DEBUG MODE: use previous $rnammer_file\n";
	}

	&CompleteCfg($r_cfg_content, $r_annotastruct_cpt, $RNAMMER_EXT, $$rh_cfg_template{ncrna}, $rnammer_file);

	return;	
}

=head2 procedure GetProdigal

 Title        : GetProdigal
 Usage        : 
 Prerequisite :
 Function     : Launch prodigal. Result file is seq.prodigal.gff3
 Args         : $o_param  ParamParser object
                $seq      fasta file path
                $seqlen    integer    Sequence length
                $r_annotastruct_cpt
                $rh_cfg_template
                $r_cfg_content
 Globals      : none

=cut
sub GetProdigal
{	
	my ($o_param, $seq, $seqlen, $r_annotastruct_cpt, $rh_cfg_template, $r_cfg_content, $o_logger, $logger_stamp) = @_;
	
	return if ($o_param->IsDefined('no_prodigal'));
	
	my $cmd;
	my $prg_prodigal    = $o_param->Get('prg_prodigal');
	my $prodigal_file   = "$seq.$PRODIGAL_EXT.gff3";
	my $prodigal_minlen = $o_param->Get('prg_prodigal_minlen');

	if ($seqlen >= $prodigal_minlen)
	{
		$cmd = "$prg_prodigal -f gff -i $seq -o $seq.$PRODIGAL_EXT.gff > /dev/null 2>&1";
		&ExeCmd($cmd, $o_logger, $logger_stamp);

		# Create a reference training set for prodigal
		if (!$o_param->IsDefined('reference_prodigal_train'))
		{
			$cmd = $o_param->Get('prg_prodigal') . " -f gff -i $seq -t $seq.$PRODIGAL_EXT.train > /dev/null 2>&1";
			&ExeCmd($cmd, $o_logger, $logger_stamp);
			$o_param->Set('reference_prodigal_train', "$seq.$PRODIGAL_EXT.train");
		}
	}
	else
	{
		if ( $o_param->IsDefined('reference_prodigal_train') )
		{
			my $trainset = $o_param->Get('reference_prodigal_train');
			&PrintInfo("Sequence length smaller than prodigal minlen ($prodigal_minlen): use the prodigal training set $trainset\n", 
					   $o_logger, $logger_stamp);

			$cmd = "$prg_prodigal -f gff -i $seq -o $seq.$PRODIGAL_EXT.gff -t $trainset > /dev/null 2>&1";
			&ExeCmd($cmd, $o_logger, $logger_stamp);
		}
		else
		{
			&PrintInfo("Sequence size smaller than prodigal minlen ($prodigal_minlen): ". 
					   "no prodigal training set available so use -p meta option\n",
					   $o_logger, $logger_stamp);
			
			$cmd = "$prg_prodigal -f gff -i $seq -o $seq.$PRODIGAL_EXT.gff -p meta > /dev/null 2>&1";
			&ExeCmd($cmd, $o_logger, $logger_stamp);
		}
	}
	
	# Create a eugene formated GFF3 file (Add the SO ontology to the CDS lines)
	my $f_in  = &General::GetStreamIn("$seq.$PRODIGAL_EXT.gff");
	my $f_out = &General::GetStreamOut($prodigal_file);
	while (my $l = <$f_in>)
	{
		if ($l !~ /^#/ && $l =~ /\tCDS\t/)
		{
			chomp($l);
			$l .= ";Ontology_term=SO:0005845\n";
		}
		print $f_out $l;
	}
	$f_out->close;
	$f_in->close;



	$o_logger->Trace("Create $prodigal_file\n", $logger_stamp, 
					 $LOGGER_LEVEL_EXEC);

	&CompleteCfg($r_cfg_content, $r_annotastruct_cpt, $PRODIGAL_EXT, 
				 $$rh_cfg_template{start}, $prodigal_file);

	return;	
}






=head2 procedure LoadNdfSequences

 Title        : LoadNdfSequences
 Procedure    : Extract from the ndf file the sequences. Map them to the genomic sequences
                Save the coordinate in data structure.
 Prerequisite : None
 Args         : $prefix     Root of the key of the current evidence type. Ex evidence_2 
                $seqfile    Multifasta file of the genomic sequènces
                $workingdir Temp directory
                $o_param
                $o_logger
                $logger_stamp
 Globals      : none

=cut

sub LoadNdfSequences
{
	my ($prefix, $seqfile, $workingdir, $o_param, $o_logger, $logger_stamp) = @_;

	my $ndf_file = $o_param->Get($prefix . "_ndf");
	my ($ndf_path, $ndf_name) =  $ndf_file =~ /(.+)\/(.+)/;
	&PrintInfo("Extract data from $ndf_file\n", $o_logger, $logger_stamp);

	# 1) Extract from ndf file the probes and map them to the genomc loci	
	my $fasta_file = $workingdir . "/" .  $ndf_name. ".fna.gz";
	my $glintres   = $workingdir . "/" .  $ndf_name. ".fna.glint";
	&Ndf2Fasta($ndf_file, $fasta_file);
	
	
	my $cmd;
	if (!$DEBUG || ($DEBUG && !-e $glintres.".gz"))
	{
		my $cmd =  $o_param->Get('prg_glint') . 
			" map $seqfile $fasta_file --input-format fasta -o $glintres " . 
			$o_param->Get('prg_glint_tiling_param');
		&ExeCmd($cmd, $o_logger, $logger_stamp);
	}
	$glintres .= ".gz";

	print STDERR "Load $ndf_file...\n" if ($DEBUG);
	#Parse results

	my %h_chr;
	my %h_start;
	my %h_end;
	my %h_strand;
	my %h_seqid;
	
	my %h_m8_res;	
	my $f_map_res = &General::GetStreamIn($glintres);
	while (my $m8_line = <$f_map_res>)
	{
		chomp($m8_line);
		&GeneralBioinfo::ParseM8Line(\$m8_line, 1, \%h_m8_res);

		my $seqid = $h_m8_res{query};		
		# If already view, remove the probe because not specific
		if (defined ($h_seqid{$seqid}))
		{
			delete ($h_chr{$seqid});
			delete ($h_start{$seqid} );
			delete ($h_end{$seqid});
			delete ($h_strand{$seqid});
			next;
		}
		$h_seqid{$seqid}  = $seqid;
		$h_chr{$seqid}    = $h_m8_res{target};
		$h_start{$seqid}  = $h_m8_res{tstart};
		$h_end{$seqid}    = $h_m8_res{tend};
		$h_strand{$seqid} = $h_m8_res{strand};
	}

	$o_param->Set("$prefix\_chrs",    \%h_chr);
	$o_param->Set("$prefix\_starts",  \%h_start);
	$o_param->Set("$prefix\_ends",    \%h_end);
	$o_param->Set("$prefix\_strands", \%h_strand);
	

	return;
}


=head2 procedure SplitPair2Expr

 Title        : SplitPair2Expr
 Procedure    : Convert pair file in wig, then convert wig to expr files and 
                split them per sequence. Save results in the sequence directories
 Prerequisite : 
 Args         : $pair_file     Pair file from nimblescan
                $seqfile       multifasta of all the genomic sequences
                $workingdir    Temporary directory where create the temp files
                $evidence_key  Exe: evidence_5. Used to get sup parameters
                $rootname      Root name of the temporary files
 Globals      : none

=cut
sub SplitPair2Expr
{
	my ($pair_file, $seqfile, $workingdir, $evidence_key, $rootname,
		$o_param, $o_logger, $logger_stamp) = @_;
	
	my $evidence_name = &GetEvidenceName($pair_file);
	my ($pair_path, $pair_name) =  $pair_file =~ /(.+)\/(.+)/;

	if ($DEBUG)
	{
		my $globseqdir = $o_param->Get('rootseqdir');
		my @a_pos    = glob( "$globseqdir/*/*$evidence_name.$EXPR_POS_EXT.gff3");
		my @a_reg    = glob( "$globseqdir/*/*$evidence_name.$EXPR_REG_EXT.gff3");
		my @a_tstart = glob( "$globseqdir/*/*$evidence_name.$TSTART_EXT.gff3");
		if (scalar(@a_pos) > 0 && scalar(@a_reg) > 0 && scalar(@a_tstart) > 0)
		{
			my $text = "DEBUG MODE: use already computed file for tiling pair evidence $evidence_name:\n";
			$text .= join(' ', @a_pos) ." ". join(' ', @a_reg) . " ". join(' ', @a_tstart) . "\n";
			&PrintInfo($text, $o_logger, $logger_stamp);
			
			return;	
		}
	}
	

	my $wig_file_plus  = "$workingdir/$evidence_name\_FWD.$WIG_EXT.gz";  
	my $wig_file_minus = "$workingdir/$evidence_name\_REV.$WIG_EXT.gz"; 
	&Pair2Wig($pair_file, $wig_file_plus, $wig_file_minus, $evidence_key, $o_param);

	# Convert wig file (forward strand) to expression files
	my $ra_created_files = 
		&SplitWig2Expr($o_param, $wig_file_plus, "forward", 
					   $workingdir, $rootname, $o_logger, $logger_stamp);
	# Convert wig file (reverse strand) to expression files
	&SplitWig2Expr($o_param, $wig_file_minus, "reverse", 
				   $workingdir, $rootname, $o_logger, $logger_stamp);


	# Concatenate forward and reverse expression files
	foreach my $forward_f (@$ra_created_files)
	{
		my $reverse_f = $forward_f;
		my $cat_f     = $forward_f;
		$reverse_f =~ s/_FWD\.([^\.]+)\.gff3/_REV\.$1\.gff3/;
		$cat_f     =~ s/_FWD\.([^\.]+)\.gff3/\.$1\.gff3/;

		my $cmd = "cat $forward_f $reverse_f > $cat_f ";
		&ExeCmd($cmd, $o_logger, $logger_stamp);
		unlink($forward_f, $reverse_f);
	}

	return;
}

=head2 procedure Pair2Wig

 Title        : Pair2Wig
 Procedure    : Convert pair file in wig. create two wig files, one per strand
 Prerequisite : 
 Args         : $pair_file      Pair file from nimblescan
                $wig_file_plus  wig file for strand +
                $wig_file_minus wig file for strand -
                $prefix
 Globals      : none

=cut
sub Pair2Wig
{
	my ($pair_file, $wig_file_plus, $wig_file_minus, $prefix, $o_param) = @_;


	if ($DEBUG && (-e $wig_file_plus) && (-e $wig_file_minus))
	{
		print STDERR "DEBUG MODE: $wig_file_plus and $wig_file_minus already computed.\n";
		return;
	}

	my $rh_chr    = $o_param->Get("$prefix\_chrs");
	my $rh_start  = $o_param->Get("$prefix\_starts");
	my $rh_end    = $o_param->Get("$prefix\_ends");
	my $rh_strand = $o_param->Get("$prefix\_strands");

	my %h_probe_intensity; # key   = "chr"_"strand", 
                           # value = array of intensity per position
	my %h_probe_nb;  # key   = "chr"_"strand", 
                     # value = array containing the number of probe matchs per position

	my $key;
	# Parse  pair files
	my $f_pair = &General::GetStreamIn($pair_file);
	while  (my $line = <$f_pair>)
	{
		chomp($line);		
		my @a_line = split(/\t/, $line);
		my $probid = $a_line[3];
		my $signal = $a_line[9];

		if (defined($$rh_chr{$probid}))
		{
			$key = $$rh_chr{$probid} ."_" . $$rh_strand{$probid};

			$h_probe_intensity{$key} = [] if (!defined $h_probe_intensity{$key});
			$h_probe_nb{$key}        = [] if (!defined $h_probe_nb{$key});
			my $ra_intensity    = $h_probe_intensity{$key};
			my $ra_nb           = $h_probe_nb{$key} ;
			my $starti = $$rh_start{$probid} -1;
			my $endi   = $$rh_end{$probid}-1;

			for (my $i= $starti; $i <= $endi ; $i++)
			{
				$$ra_intensity[$i] += $signal;
				$$ra_nb[$i]        += 1;				   
			}
		}
	}

	print STDERR "end of read pair file\n" if ($DEBUG);

	# Save in wig files
	# create one wig file per strand
	my $f_wig_plus  = &General::GetStreamOut($wig_file_plus);
	my $f_wig_minus = &General::GetStreamOut($wig_file_minus);

	my $f_current;
	foreach my $seq (keys %h_probe_intensity)
	{	
		my ($chr, $strand) = $seq =~ /(.+)_([-\+])$/;
		$f_current = ($strand eq "+") ? $f_wig_plus : $f_wig_minus;
 
		print $f_current "track type=wiggle_0 name=\"tiling $chr $strand\"\n"; 
		print $f_current "fixedStep chrom=$chr start=1 step=1 span=1\n";
		
		my $ra_intensity = $h_probe_intensity{$seq};
		my $ra_nb        = $h_probe_nb{$seq};
		my $l   = scalar(@$ra_intensity);
		#print STDERR "length array ! $l\n";
		for (my $i=0; $i < $l; $i++)
		{
			if (defined($$ra_intensity[$i]))
			{
				my $mean_intensity = $$ra_intensity[$i]/$$ra_nb[$i];
				printf $f_current "%.2f\n", $mean_intensity ;
			}
			else
			{
				print $f_current "0\n";
			}
		}
	}
	$f_wig_plus ->close;
	$f_wig_minus->close;

	
	return;
}




=head2 procedure Ndf2Fasta

 Title        : Ndf2Fasta
 Procedure    : Convert ndf to multifasta
 Prerequisite : 
 Args         : $ndf    Ndf file generated by nimblescan
                $fasta  Result fasta file
 Globals      : none

=cut
sub Ndf2Fasta
{
	my ($ndf, $fasta) = @_;
	
	my $f_      = &General::GetStreamIn($ndf);
	my $f_fasta = &General::GetStreamOut($fasta);
	my @a_line;
	while  (my $line = <$f_>)
	{
		chomp($line);
		next if ($line =~ /^PROBE_DESIGN_ID/);
		@a_line = split(/\t/, $line);
		if ( $a_line[1] eq "BLOCK1")
		{
			print $f_fasta ">$a_line[12]\n$a_line[5]\n";	
		}
	}

	$f_->close;
	$f_fasta->close;

	return;
}



=head2 procedure Consolidate
	
 Title        : Consolidate
 Usage        : 
 Prerequisite : prg_biofileconverter and prg_genome_statistics in o_param correspond to existing programs.
 Function     : Compute protein sequences file, and statistics about structural annotation  
 Args         : $o_param 
                $outdir      Output directory
                $ra_res_gff3 Ref to an array of gff3 files
                $allseq_file Mutifastafile of genomic sequences in working directory
 Globals      : none

=cut

sub Consolidate
{
	my ($o_param, $outdir, $ra_res_gff3, $allseq_file, $o_logger, 
		$logger_stamp) = @_;
	
	my ($allseq_dir, $allseq_name) = $allseq_file =~ /(.+)\/(.+)/;

	my $cmd;
	# copy the multifasta file in the result directory if workdir and outdir are different
	if ($allseq_dir ne $outdir)
	{
		$cmd = "cp $allseq_file $outdir/$allseq_name";
		&ExeCmd($cmd, $o_logger, $logger_stamp);
		# zip the multifasta file in the working dir
		&Gzip($o_logger, $logger_stamp, $allseq_file);
	}

	# Cat all the gff3 results in a uniq file in the result directory
	$cmd = "cat " . join(' ', @$ra_res_gff3) . " > $outdir/$allseq_name.gff3";
	&ExeCmd($cmd, $o_logger, $logger_stamp);
	
	# Compute protein fasta file
	$cmd = $o_param->Get('prg_biofileconverter');
	$cmd .= " --input-ann-file $outdir/$allseq_name.gff3 --input-seq-file $outdir/$allseq_name --all-contigs " .
		"--output-ann-file $outdir/$allseq_name.fna --kingdom bacteria --input-ann-format GFF3  " . 
		"--output-ann-format FASTA 2> /dev/null";
	&ExeCmd($cmd, $o_logger, $logger_stamp);


	# Compute basic statistics
	$cmd = $o_param->Get('prg_genome_statistics');
	$cmd .= " --kingdom prokaryote --in  $outdir/$allseq_name.gff3 --out  $outdir/$allseq_name --scaffolds  $outdir/$allseq_name";
	&ExeCmd($cmd, $o_logger, $logger_stamp);

	return;
}


=head2 procedure CreateWorkSeqTree

 Title        : CreateWorkSeqTree
 Usage        : 
 Prerequisite : keys 'prg_fasta2tree' in o_param is an existing program
 Function     : Search fasta files in $indir, Create a ordered tree of files
                in the working directory. (from the longer sequences to the smaller)
                Compute global GC percent
 Args         : $o_param     Param object
                $indir       Input directory containing fasta sequences
                $ra_seq      Ref to an array  The method fills it with the path of the FASTA files
                $workingdir  directory where create the ordered fiel tree                
                $root        Name of the concatenation of fasta files
 Globals      : none

=cut
sub CreateWorkSeqTree
{
	my ($o_param, $indir, $ra_seq, $workingdir, $root, $o_logger, $logger_stamp, $uniq) = @_;

	my $seqrootdir = "$workingdir/0001/";
	my $f_all = &General::GetStreamOut("$workingdir/$root.tmp");
	my $files;

	# Create a multifasta file with the fasta sequences
	foreach my $f (glob("$indir/*"))
	{
		if (&IsFasta($f))
		{
			printf STDERR "Load fasta file $f\n" if ($DEBUG);	
			$o_logger->Trace("Load fasta file $f\n", $logger_stamp, $LOGGER_LEVEL_EXEC);
			
			my $f_in = &General::GetStreamIn($f);
			while (my $l=<$f_in>)
			{
				print $f_all $l;
			}
			$files .= "$f ";
		}
		else
		{
			printf STDERR "WARNING: $f ignored. Dont look like a FASTA file\n";
			$o_logger->Trace("WARNING: $f ignored. Dont look like a FASTA file\n");
		}
	}
	$f_all->close;
	

	if (!$DEBUG && -d $seqrootdir)
	{
		$o_logger->Trace("ERROR: $seqrootdir exists: remove it before launch program.\n");
		Carp::croak "$seqrootdir exists: remove it before launch program.\n";
	}
	
	if (!-d $seqrootdir)
	{
		my $cmd = $o_param->Get('prg_fasta2tree');
		$cmd .= " --sequence $workingdir/$root.tmp --outdir  $workingdir --type dna --sort_by_len ";
		&ExeCmd($cmd, $o_logger, $logger_stamp);
	}
	
	# Remove temporary full fasta file
	my $cmd = "rm $workingdir/$root.tmp";
	&ExeCmd($cmd, $o_logger, $logger_stamp);
		
	# For each sequence, create a workingdirectory
	# And save it in a full sorted FASTA file
	my $f_sorted_all = &General::GetStreamOut("$workingdir/$root");
	my $f_fof = &General::GetStreamIn("$workingdir/entries.fof");
	while (my $f = <$f_fof>)
	{
		chomp($f);
		push (@$ra_seq, $f);
		my ($seqpath, $seqname) = $f =~ /(.+)\/(.+)/;
 
		# Create the workingdir for this sequence
		chdir ($seqpath);
		mkdir ("work.$uniq");		
		my $f_in = &General::GetStreamIn($f);
		while (my $l=<$f_in>)
		{
			print $f_sorted_all $l;
		}
	}
	$f_sorted_all->close;
	print STDERR "Cat $files in $workingdir/$root\n" if ($DEBUG);
	$o_logger->Trace("Cat $files in $workingdir/$root\n", $logger_stamp, $LOGGER_LEVEL_EXEC);


	&ComputeGCPercent($o_param, "$workingdir/$root");

	$o_param->Set('rootseqdir', $seqrootdir);

	return $seqrootdir;
}


=head2 function IsFasta

 Title        : IsFasta
 Usage        : IsFasta($file)
 Prerequisite : none
 Function     : return true if its the file looks like a fasta. Just assert the first line start by '>'
 Returns      : ture/false
 Args         : $seq  a file
 Globals      : none

=cut
sub IsFasta
{
	my ($seq) = @_;

	my $f_seq = &General::GetStreamIn($seq);
	my $line = <$f_seq>;
	return 0 if (defined($line) && $line !~ /^>\S+/);
	return 1;
}


=head2 function Mapping

 Title        : Mapping
 Usage        : 
 Prerequisite : $o_param has the keys: prg_glint_mate_dist, prg_glint, 
                prg_glint_mm_small_pe, prg_glint_mm_pe, prg_glint_small_param, prg_glint_param
 Function     : Launch glint. Adapt parameters of the type of data (ose/ope, small)
 Args         : $seq  a fasta file
                $evidence fastq file
                $format  fast or fasta
                $orientation ope/ose
                $outfile
                $small 1 if it is small rna, 0 else
                $o_param
 Globals      : none

=cut
sub Mapping
{
	my ($seq, $evidence, $format, $orientation, $outfile, $small, $o_param, $o_logger, $logger_stamp) = @_;


	if ($DEBUG && (-e "$outfile.gz"))
	{
		print STDERR "DEBUG MODE: use $outfile.gz.\n";
		return;
	}

	my $cmd;
	my $mm   = 0; 
	

	if ( $orientation =~ /ope/ )
	{
		#Carp::croak("An even number of filenames is expected for paired read files\n") if (($#$ra_lof+1) %2 != 0 );
		my $matedist = " --mate-maxdist ". $o_param->Get('prg_glint_mate_dist');
		$cmd = $o_param->Get('prg_glint') . " mappe $matedist ";

		# lorsque les sequences recherches sont tres courtes, je veux 0 mm, 
        # sinon 2 reste accceptable car c'est le meilleur
		# et que je vais jouer sur la paire pour retrouver la specificite
		$mm = ($small) ? $o_param->Get('prg_glint_mm_small_pe') : $o_param->Get('prg_glint_mm_pe');
	}
	else
	{
		$cmd =  $o_param->Get('prg_glint') . " map ";
		# lorsque les sequences recherches sont tres courtes, 
		# je veux 0 mm, sinon 1 reste accceptable car c'est le meilleur
		$mm = ($small) ? $o_param->Get('prg_glint_mm_small_se') : $o_param->Get('prg_glint_mm_se');
	}
	my $glint_params = " --mmis $mm --input-format $format ";
	if ($small)
	{
		$glint_params .=  $o_param->Get('prg_glint_small_param') . " ";
	}
	else
	{
		$glint_params .=  $o_param->Get('prg_glint_param') . " ";
	}
	#my $outfile = ( $orientation =~ /pe/ ) ? "$outdir/$ref.vs.$outprefix.bedpe" : "$outdir/$ref.vs.$outprefix.bed";

	# 1. mapping
	$cmd .= " $seq  $evidence $glint_params -o $outfile 2> /dev/null";
	
	print STDERR "MAPPING... $cmd\n" if ($DEBUG);
	&ExeCmd($cmd, $o_logger, $logger_stamp);
	print STDERR "Mapping done.\n" if ($DEBUG);

	return;

}


=head2 procedure MappingFiltering

 Title        : MappingFiltering
 Usage        : MappingFiltering
 Prerequisite : prg_bed_filter and prg_bed_filter_param have to be defined in o_param
 Function     : 
 Returns      : none
 Args         : $mapping_file file    bed/bedpe file
                $orientation  string  ose/ope
                $small        integer [0 or 1] 
                $outfile      file    bed result file
                $pat          string  pattern, for instance 'evidence_1'
                $o_param
               
 Globals      : none

=cut
sub MappingFiltering
{
	my ($mapping_file, $orientation, $small, $outfile, $pat, $o_param, $o_logger, $logger_stamp) = @_;


	if ($DEBUG && (-e $outfile) )
	{
		print STDERR "DEBUG MODE: use $outfile file.\n";
		return;
	}

	my $cde;
	my $prg_filter       = $o_param->Get('prg_bed_filter');
	my $prg_filter_param = $o_param->Get('prg_bed_filter_param');

	if ( $orientation  =~ /ope/ )
	{
		$prg_filter .= " --bedpe";
		if ($small)
		{
			$prg_filter .= " " . $o_param->Get('prg_bed_filter_pe_small_param') ;
		}
		else
		{
			$prg_filter .= " " . $o_param->Get('prg_bed_filter_pe_param') ;
		}
	}
	$prg_filter .= " --oriented_end ".$o_param->Get($pat."_oriented_end") . " "; # evidence_x_oriented_end
	

	$cde = "$prg_filter --in $mapping_file --out $outfile $prg_filter_param";

	print STDERR "FILTERING... $cde\n" if ($DEBUG);
	&ExeCmd($cde, $o_logger, $logger_stamp);
	print STDERR "Filtering done.\n" if ($DEBUG);

	return;
}

=head2 procedure Bed2Expr4EGN

 Title        : Bed2Expr4EGN
 Usage        : Bed2Expr4EGN()
 Prerequisite : prg_bed_to_expr prg_bed_to_expr_annot_param workingdir prg_bed_to_expr_annot_scoremax
                prg_bed_to_expr_annot_min_coverage_percent prg_bed_to_expr_annot_scalingfactor
 Function     : Convert bed[pe] file to expression files for the annotation (position.gff3 and region.gff3) 
 Returns      : none
 Args         : $filtering_bed_file   file    bed/bedpe file
                $orientation  string  ose/ope
                $outpositionfile      file      
                $outregionfile        file 
                $o_param
               
 Globals      : none

=cut
sub Bed2Expr4EGN
{
	my ($filtering_bed_file, $orientation, $outpositionfile, $outregionfile, 
		$o_param, $o_logger, $logger_stamp, $uniq ) = @_;

	my $cde;

	my $out = $o_param ->Get('workingdir'). "/expr_gff3.".$uniq;
	my $tag = "glint"; #TODO: mettre un flag dependant de la banque

	$cde = $o_param->Get('prg_bed_to_expr'). " " . $o_param->Get('prg_bed_to_expr_annot_param') ;
	
	$cde .=	" --tag $tag ";
	$cde .= " --scoremax "      . $o_param->Get('prg_bed_to_expr_annot_scoremax');
	$cde .= " --scalingfactor " . $o_param->Get('prg_bed_to_expr_annot_scalingfactor');
	$cde .= " --min_coverage_percent "  . $o_param->Get('prg_bed_to_expr_annot_min_coverage_percent');
	$cde .= " --interval --bedpe " if ($orientation =~ /ope/);
	$cde .= " --bedfile $filtering_bed_file --outprefix $out";

	&ExeCmd($cde, $o_logger, $logger_stamp);
	
	# Rename files *position.gff3 and *region.gff3
	$cde = "mv $out.position.gff3 $outpositionfile";
	&ExeCmd($cde, $o_logger, $logger_stamp);
	
	$cde = "mv $out.region.gff3 $outregionfile";
	&ExeCmd($cde, $o_logger, $logger_stamp);
	unlink("$out.non_intergenic.gff3");

	return;
}

=head2 procedure Bed2Expr4tStart

 Title        : Bed2Expr4tStart
 Usage        : Bed2Expr4tStart()
 Prerequisite : prg_bed_to_expr prg_bed_to_expr_annot_param workingdir 
                prg_bed_to_expr_tstart_scoremax prg_bed_to_expr_tstart_scalingfactor
                prg_bed_to_expr_tstart_min_coverage,
                have to be keys of o_param
 Function     : Convert bed[pe] file to expression files for the tStart detection (position.gff3) 
 Returns      : none
 Args         : $bed_file             file    bed/bedpe file
                $orientation          string  ose/ope
                $outpositionfile      file      
                $o_param
               
 Globals      : none

=cut
sub Bed2Expr4tStart
{
	my ($bed_file, $orientation, $outpositionfile, $o_param, $o_logger, $logger_stamp, $uniq ) = @_;

	my $cde;

	my $out = $o_param ->Get('workingdir'). "/expr_tStart_gff3.".$uniq;
	my $tag = "glint"; #TODO: mettre un flag dependant de la banque

	$cde = $o_param->Get('prg_bed_to_expr'). " " . $o_param->Get('prg_bed_to_expr_tstart_param');
	 
	$cde .=	" --tag $tag ";
	$cde .=	" --scoremax "      . $o_param->Get('prg_bed_to_expr_tstart_scoremax');
	$cde .=	" --scalingfactor " . $o_param->Get('prg_bed_to_expr_tstart_scalingfactor');
	$cde .=	" --min_coverage "  . $o_param->Get('prg_bed_to_expr_tstart_min_coverage');
	$cde .= " --interval --bedpe " if ($orientation =~ /ope/);
	$cde .= " --bedfile $bed_file --outprefix $out";

	&ExeCmd($cde, $o_logger, $logger_stamp);

	# Rename file *position.gff3 
	$cde = "mv $out.position.gff3 $outpositionfile";
	&ExeCmd($cde, $o_logger, $logger_stamp);

	unlink("$out.region.gff3", "$out.non_intergenic.gff3");	

	return;
}



sub LaunchEGN
{
	my ($o_param, $seq, $cfg, $matrix, $outdir, $o_logger, $logger_stamp) = @_;

	my $pgr_egn = $o_param->Get('prg_eugene');
	my $cmd     = "";
	$cmd = "$pgr_egn -A $cfg -m $matrix -pg -O $outdir/ $seq";

	&ExeCmd($cmd, $o_logger, $logger_stamp);

	return;
}




=head2 function GetEvidenceName

 Title        : GetEvidenceName
 Usage        : 
 Prerequisite : none 
 Function     : Get the name of the evidence
 Args         : $file Path of a evidence file (fastq for instance)
 Globals      : none

=cut
sub GetEvidenceName
{
	my ($file) = @_;
	
	my ($evidence_name) = $file =~ /[^;]+\/([^;]+)/;
	$evidence_name =~ s/\.(lzma|xz|gz|zip|fna)$//; # Remove extension
	$evidence_name =~ s/\.[^.]+$//; # Remove the type of the file (fastq, wig, ...)
	
	return $evidence_name;
}



sub ExeCmd
{
	my ($cmd, $o_logger, $logger_stamp) = @_;

	print STDERR $cmd . "\n" if ($DEBUG);
	$o_logger->Trace("$cmd\n", $logger_stamp, $LOGGER_LEVEL_EXEC);
	system($cmd) && Carp::croak "ERROR >$cmd<\n";

	return;
}


sub GetMappingFileName
{
	my ($seqname, $evidencefile ) = @_;
	return "$seqname." . &GetEvidenceName($evidencefile). "." . $MAPPING_EXT; 
}

sub GetWorkSeqDir
{
	my ($seqname, $o_param) = @_;

	my $rootseqdir = $o_param->Get('rootseqdir');
	my $uniq       = $o_param->Get('uniq');

	return "$rootseqdir/$seqname/work.$uniq/";	
}

sub GetSeqDir
{
	my ($seqname, $o_param) = @_;

	my $rootseqdir = $o_param->Get('rootseqdir');

	return "$rootseqdir/$seqname/";	
}


=head2 function ComputeGCPercent

 Title        : ComputeGCPercent
 Usage        : &ComputeGCPercent($seqfile)
 Function     : Compute the gc percent (nb of G|C) vs length of the sequence except 'N'
                Set variable gc_percent to "high" or "low" 
 Args         : $o_param
                $seqfile Multifasta file
 Return       : gc_percent
 Globals      : none

=cut
sub ComputeGCPercent
{
	my($o_param, $seqfile) = @_;
	
	my $gccount     = 0;
	my $lenwithoutN = 0;
	my $f_ = &General::GetStreamIn($seqfile);
	while (my $line = <$f_>)
	{
		chomp($line);
		next if $line =~ /^>/;
		($gccount) += $line =~ tr/GCgc/GCgc/;
		$line =~ s/N//gi; # remove N
		$lenwithoutN += length($line);
	} 
	$f_->close;

	my $gc_percent = $gccount/$lenwithoutN*100;
	print STDERR "GC% : $gc_percent\n" if ($DEBUG);
	if ($gc_percent >= 50)
	{
		$o_param->Set('gc_percent', "high");
	}
	else
	{
		$o_param->Set('gc_percent', "low");
	}
	return $gc_percent;
}


=head2 function GetSeqInfo
	
 Title        : GetSeqInfo
 Usage        : ($p, $n, id, $len) = &GetSeqInfo($o_param)
 Function     : Return the seq directory, the seq file name, 
                the seqid and the seq len
 Args         : $seq : Path of a fasta file
 Globals      : none

=cut
sub GetSeqInfo
{
	my ($seq) = @_;

	my ($seqpath, $seqname) = $seq =~ /(.+)\/(.+)/; 

	# Get name and length of the sequence
	my %h_seq;
	&FastaToHash($seq, \%h_seq, 1);
	my @a_keys = keys(%h_seq);
	my $seqid  = $a_keys[0];
	my $seqlen = $h_seq{$seqid}{len};

	return ($seqpath, $seqname, $seqid, $seqlen);	
}


=head2 function ReadFile

 Title        : ReadFile
 Function     : Return the content of the file
 Args         : $file 
 Globals      : none

=cut
sub ReadFile
{	
	my ($file) = @_;

	my $content = "";
	my $f_      = &General::GetStreamIn($file);
	while ( my $line = <$f_> )
	{
		$content .= $line ;
	}
	$f_->close;
	
	return $content;
}


sub Gzip
{
	my ($o_logger, $logger_stamp, @a_files) = @_;

	my $cmd = "gzip -f ";
	foreach my $file (@a_files)
	{	
		$cmd .= " $file ";
	}
	
	&ExeCmd($cmd, $o_logger, $logger_stamp);

	return;

}



sub CatSeq
{
	my ($ra_seq, $newseq) = @_;

	my $f_fullseq = &General::GetStreamOut($newseq);
	foreach my $seq (@$ra_seq)
	{
		my $f_ = &General::GetStreamIn($seq);
		while (my $line = <$f_>)
		{
			print $f_fullseq $line;
		}
		$f_->close;
	}
	$f_fullseq->close;
	
	return;
}

sub PrintInfo
{
	my ($text, $o_logger, $logger_stamp) = @_;

	print STDERR $text if ($DEBUG);
	$o_logger->Trace("$text", $logger_stamp, $LOGGER_LEVEL_EXEC);

	return;
}



sub Usage
{
    print STDERR<<END
$0
[Mandatory]
    --indir      dirname        Directory containing two directories:
                                  - genome/*.[fna|fa|fasta]* : genomic sequence fasta files
                                  - data/*                   : data files(fastq/fasta/deb/bedpe/) + (wig/bam)
    --outdir     dirname        Existing directory where will be stored the EuGene predictions
    --cfg        filename       Configuration file

[Optional]
    --workingdir dirname        [Default outdir] directory where will be run the computations

[Advanced optional]
    --no_ncrna_detection        Flag If activate, dont compute ncrna program detection (tRNAscan-SE, rfam_scam and Rnammer)
    --no_rfamscan               Flag If activate, don't launch rfam_scan
    --no_trnascan               Flag If activate, don't launch tRNAscan-SE
    --no_rnammer                Flag If activate, don't launch Rnammer
    --no_prodigal               Flag If activate, don't lauch Prodigal
    --help
    --debug      integer        [Default 0] 0: don't display message, 
                                            1: display command line 
END
}
