#!/usr/bin/perl
#
# $Id
#

# Copyright INRA

# Jerome.Gouzy@toulouse.inra.fr
# Erika.Sallet@toulouse.inra.fr
# Thomas.Schiex@toulouse.inra.fr

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


=pod

=head1 NAME

	get_Rfam.pl 

=head1 SYNOPSIS

 $EGNP/bin/int/get_Rfam.pl --fasta_file seq.fna  --output_file seq.Rfam.gff3 [--cm_file /databases/rfam12/Rfam.cm]

=head1 DESCRIPTION

Search ncRNAs against Rfam database using cmsearch program from Infernal.
Tested with Rfam 12.0 (ftp://ftp.ebi.ac.uk/pub/databases/Rfam/) 
and Infernal 1.1.1 (http://infernal.janelia.org/)

=cut

BEGIN
{
    $ENV{LANG} = "C";
}

use FindBin;
use lib "$FindBin::Bin/../../corelib/";
use lib "$FindBin::Bin/../../lib/"; 

use strict;
use warnings;
use ParamParser;
use General;
use Carp;

MAIN:
{
    my $o_param = New ParamParser('GETOPTLONG', \&Usage, "cm_file=s", "fasta_file=s", "output_file=s", "score_threshold=s", 
								  'evalue_threshold=s', "cpu=i", "add_parent=s", "working_dir=s", "global", "verbose", "noclean", "help");

	$o_param->Update('ENV', 'A');    # trying to get RFAM_DIR
	
	
	$o_param->AssertFileExists('fasta_file');
	$o_param->AssertDefined('output_file');
	
	$o_param->SetUnlessDefined('cpu', 2);
	$o_param->SetUnlessDefined('working_dir', "/tmp/");
	
	$o_param->AssertDirExists('working_dir');
	
	if ($o_param->IsDefined('cm_file'))
	{
		$o_param->AssertFileExists('cm_file');
	}
	else
	{
		if ($o_param->IsDefined('RFAM_DIR'))
		{
			$o_param->Set('cm_file', $o_param->Get('RFAM_DIR')."/Rfam.cm");
			$o_param->AssertFileExists('cm_file');
		}
		else
		{
			print STDERR "The parameter cm_file must be provided. 'get_Rfam.pl --help' for more details.\n";
			exit;
		}
	}
	
	&GetExePath("cmsearch");

	
	my $verbose = 0;
	$verbose    = 1 if ($o_param->IsDefined('verbose'));

	my $noclean = 0;
	$noclean    = 1 if ($o_param->IsDefined('noclean'));

	my $parent = "";
	$parent    =  $o_param->Get('add_parent') if ($o_param->IsDefined('add_parent'));
	
	my $scoreThresh;  # -T de cmsearch 
	my $evalueThresh; # -E de cmsearch
	$scoreThresh  = $o_param->Get('score_threshold') if ($o_param->IsDefined('score_threshold'));
	$evalueThresh = $o_param->Get('evalue_threshold') if ($o_param->IsDefined('evalue_threshold'));
	
	my $global = 0;
	$global    = 1 if ($o_param->IsDefined('global'));
	
	my $cmfile     = $o_param->Get('cm_file');
	my $fafile     = $o_param->Get('fasta_file');
	my $cpu        = $o_param->Get('cpu');
	my $outfile    = $o_param->Get('output_file');
	my $workingdir = $o_param->Get('working_dir');
	
	my ($resfile, $cmsfile) = &RunInfernalSearch($cmfile, $fafile, $scoreThresh, $evalueThresh, $global, $cpu, $workingdir, $verbose);
	
    &ParseInfernalRes($resfile, $outfile, $parent, $noclean);

	if (!$noclean)
	{
		unlink($resfile); 
		unlink ($cmsfile);
	}
}

=head2 function RunInfernalSearch

 Title        : RunInfernalSearch
 Usage        : my ($res, $res2) = &RunInfernalSearch($cmfile, $fafile, $thresh, $eThresh, $global, $cpu, $workingdir, $verbose)
 Prerequisite : none
 Function     : Launch cmsearch with options used by rfam curators (except if $thresh  or $eThresh are defined)
                IMPORTANT : if $thresh  or $eThresh are defined, the default Rfam options are not used!!
 Returns      : tabulated file of cmsearch, alignment file of cmsearch
 Args         : $cmfile :    covariance model file (ex Rfam.cm)
                $fafile :    fastafile
                $thres       Score threshold (Equivalent to -T param of cmsearch)
                $evalueThres Evalue threshold (Equivalent to -E param of cmsearch)
                $global      Launch in global mode (Equivalent to -g flag of cmsearch)
                $cpu         
                $workingdir  working directory
                $verbose     Print cmsearch command line
 Globals      : none

=cut

sub RunInfernalSearch 
{
	my ($cmfile, $fafile, $thresh, $evalueThresh, $global, $cpu, $workingdir, $verbose) = @_;

	print STDERR "Run infernal search\n" if ($verbose);

	my $prefix_file    = "$workingdir/cmsearch.$$.".time;
	my $cmsearch_tab   = $prefix_file.".tab";
	my $cmsearch_align = $prefix_file.".align";

    my $options = "";
	
    if (defined $thresh) 
	{
		$options .= " -T $thresh ";
    }
    elsif (defined $evalueThresh)
	{
	    $options .= " -E $evalueThresh ";
    }
    else 
	{
	    # ES 20150121 Options recommanded to reproduce Rfam curator method
        # ref "Rfam 12.0: updates to the RNA families database" NAR 2014 
		$options = " --cut_ga --rfam --nohmmonly ";
    }

    if ($global) 
	{
		$options .= " -g ";
    }	
	$options .= " --cpu $cpu ";
	
    my $cmd = "cmsearch $options --tblout $cmsearch_tab $cmfile $fafile > $cmsearch_align";
	
	print STDERR "$cmd\n" if ($verbose);
	system($cmd);

    return ($cmsearch_tab, $cmsearch_align);
}


=head2 procedure ParseInfernalRes

 Title        : ParseInfernalRes
 Usage        : &ParseInfernalRes($resfile, $outfile, $parent)
 Prerequisite : none
 Function     : Converse cmsearch tabulated file to gff3
 Returns      : -
 Args         : $file : tabulated file generated by cmsearch 
                $gff3 : GFF3 to create
                $parent : if defined, create a parent line in gff3
 Globals      : none

=cut

sub ParseInfernalRes 
{
    my ($file, $gff3, $parent, $noclean) = @_;

	my $tab_file    = $file.".tab";	
	my $sorted_file = $file.".sorted";

	my $f_in  = &General::GetStreamIn($file);
	my $f_out = &General::GetStreamOut($tab_file);
	# Create a tabulated file
	while (my $line = <$f_in>)
	{
		chomp($line);
		if( $line !~ /^\#/ )
		{
			$line =~ s/\s+/\t/g;
		}
		print $f_out $line."\n";
	}
	$f_in->close;
	$f_out->close;
	
	# sort lines by genomic position
	&UnixSort($tab_file, $sorted_file, ["8n"]);
	
	
    $f_in  = &General::GetStreamIn($sorted_file);
	$f_out = &General::GetStreamOut($gff3);
	
	my($seqid, $rfamid, $acc, $start, $end, $modst, $moden, $score, $evalue, $gc, $strand, $id, $trunc, $att);
	my $cpt = 1;
	# convert in GFF3
	print $f_out "##gff-version 3\n";
	
    while (my $line = <$f_in>)
	{
		next if( $line =~ /^\#/ );
		chomp($line);
		$line =~ s/\s+/\t/g;
		
		my @a_fields = split(/\t/, $line);
		
		#GMI11495-Rm2011G.c   -         5S_rRNA              RF00001    cm        1      119    87488    87602      +    no    1 0.64   0.0   74.0   7.6e-17 !   len=3657276
		$seqid   = $a_fields[0];
		$rfamid  = $a_fields[2];
		$acc     = $a_fields[3];
		$modst   = $a_fields[5];
		$moden   = $a_fields[6];
		$start   = $a_fields[7];
		$end     = $a_fields[8];
		$strand  = $a_fields[9];
		$trunc   = $a_fields[10];
		$gc      = $a_fields[12];
		$score   = $a_fields[14];
		$evalue  = $a_fields[15];

		$id = "$seqid-$acc.$cpt";
		
	    ($start, $end) = ($end, $start) if ($end < $start);
		
		$att = "Name=$id;Ontology_term=SO:0000655;evalue=$evalue;gc_content=$gc;model_start=$modst;model_end=$moden;rfam_acc=$acc;rfam_id=$rfamid";
		$att .= ";truncate=true" if ($trunc ne "no");

		if ($parent ne "")
		{
			print $f_out "$seqid\tRfam\t$parent\t$start\t$end\t$score\t$strand\t.\tID=gene:$id;$att\n";
			print $f_out "$seqid\tRfam\tncRNA\t$start\t$end\t$score\t$strand\t.\tID=ncRNA:$id;Parent=gene:$id;$att\n";
		}
		else
		{
			print $f_out "$seqid\tRfam\tncRNA\t$start\t$end\t$score\t$strand\t.\tID=$id;$att\n";
		}
		
		$cpt++;
	}
	$f_in->close;
	$f_out->close;

	if (!$noclean)
	{
		unlink ($file.".sorted");
		unlink($file.".tab");
	}
    
    return;
}





sub Usage
{
    print STDERR<<END
Usage:$0	

[Mandatory]
    --fasta_file          filename    Sequence file path
    --output_file         filename    GFF3 output file

[Optional] 
    --cm_file             filename   Covariance model file
                                     If not defined \$RFAM_DIR/Rfam.cm has to exist    
    --working_dir         directory  [/tmp] Temporary file directory 
    --add_parent          string     Add systematically a parent element from this type. 
                                     (column 3 in GFF3) Examples: 'gene', 'ncRNA'                         
    --cpu                 integer    [2] Number of parallel CPU workers to use for multithreads
    --global                         Configure CM for glocal alignment (equivalent to -g of cmsearch)
    --evalue_threshold    float      Evalue threshold (equivalent to -E parameter of cmsearch) (*)
    --score_threshold     float      Score threshold (equivalent  to -T parameter of cmsearch) (*)
(*) The two options disable the default parameters used to reproduce Rfam curator method. 
    Execution time can be very long. 

    --noclean                        No cleaning temporary files
    --help                           This message
    --verbose

END
}

