#!/usr/bin/perl

#
# $Id $
#

# Copyright INRA

# Jerome.Gouzy@toulouse.inra.fr
# Erika.Sallet@toulouse.inra.fr
# Thomas.Schiex@toulouse.inra.fr
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

=pod 

=head1 NAME

egn_annotation_transfer.pl

=head1 SYNOPSIS

If --transcripts not defined, launch Biofileconverter to extract mrna
Map transcripts with gmap to the scaffolds
Filter results stringently 
Extract unmapped mrna and lauch SIBsim4
Filter results (pci=95, pcs=70)
Extract start region of the reference and map to the scaffolds
Launch EuGene with these results (AnnotaStruct + Est + Start) to adjust splice sites and to provide correct gene models.
Get the name of the ortholog genes
Keep only genes whose all exons are conserved at 80% before and after EuGene run if flag --exon_filter.


=head1 DESCRIPTION


=cut


BEGIN
{   
	$ENV{LANG} = "C";

    use Cwd 'abs_path';
    use File::Basename;
    my ($dirprg) = dirname(abs_path($0));
    unshift(@INC, "$dirprg/../../../lib");
    unshift(@INC, "$dirprg/../../../corelib");
}

use ParamParser;
use strict;
use General;
use GeneralBioinfo;
use Global;
use File::Which;

use constant SEP  => "%%%%";

my $DEFAULT_INTERSECTBED_PARAM = " -f 0.6 -r -s ";

my $DEFAULT_GMAP_FILTER_PCI    = 100;
my $DEFAULT_GMAP_FILTER_PCS    = 100;
my $DEFAULT_GMAP_FILTER_SPAN   = 100;

my $DEFAULT_FILTER_PCI         = 95;
my $DEFAULT_FILTER_PCS         = 70;
my $DEFAULT_FILTER_SPAN        = 90;
my $START_PCI                  = 100;
my $START_CONTEXT_LEN          = 200;
my $DEFAULT_LEN_GENE_NUMBER    =  7;

my $DEBUG = 0;
my ($EGNEP, $EUGENEDIR);

MAIN:
{
    my $o_param = New ParamParser('GETOPTLONG', "scaffolds=s", "ref_transcripts=s", 
								  "ref_scaffolds=s", 'ref_gff3=s',
								  "outfile=s", "workingdir=s",
								  "egn_transfer_par=s", 
								  "gmap_filter_pci=i", 
								  "gmap_filter_pcs=i", 
								  "gmap_filter_span=i",
								  "filter_pci=i", "filter_pcs=i", "filter_span=i", 
								  "exon_filter",  "intersectbed_param=s", 
								  "gmap_out=s",
								  "mapping_out=s", 
								  "eugene_out=s", 
								  "len_gene_number=i", "start_ok", "debug", "cfg=s");
	
	$o_param->SetBehaviour('ignore_space');
    $o_param->Update('ENV', "A");
    $o_param->SetUsage(my $usage = sub {&Usage();});

    $o_param->AssertDirExists('EGNEP');
    $o_param->AssertDirExists('EUGENEDIR');
    $o_param->AssertFileExists('cfg');
    $o_param->AssertFullPath('cfg');
    $o_param->Update($o_param->Get('cfg'), 'A');
    $o_param->SetBehaviour('use_substitution_table');
    $o_param->SetSubstitution('%i', $o_param->Get('EGNEP'));
    $o_param->SetSubstitution('%e', $o_param->Get('EUGENEDIR'));
    $o_param->Update($o_param->Get('cfg'), 'U');

	$EGNEP     = $o_param->Get('EGNEP');
	
	$DEBUG = 1 if ($o_param->IsDefined('debug'));
    $o_param->AssertFileExists('scaffolds', 'ref_scaffolds', 'ref_gff3');
	$o_param->AssertDefined('workingdir', 'outfile', 'prg_gmap_param');
	
	my $outfile = $o_param->Get('outfile');
	if ($outfile !~ /\.gff3$/i)
	{
		print STDERR "ERROR: the GFF3 result file '--outfile' must end with \".gff3\". Please change it.\n";exit;
	}
	$o_param->AssertFullPath('scaffolds', 'outfile');
	
	$o_param->AssertDefined('gmap_dir', 'prg_blastn_param', 'prg_makeblastdb');
	$o_param->AssertFileExists('prg_blastn', 'prg_intersectbed', 'prg_gmap_filter', 'prg_fasta2tree' , 'prg_SIBsim4', 'prg_fasta_filter', 'prg_biofileconverter');
	
	$o_param->SetUnlessDefined('egn_transfer_par',   "$ENV{EGNEP}/cfg/euk/egn_annotation_transfer.par");
	$o_param->AssertFileExists('egn_transfer_par');
	$o_param->AssertFullPath('egn_transfer_par');
	$o_param->SetUnlessDefined('intersectbed_param', $DEFAULT_INTERSECTBED_PARAM);
	$o_param->SetUnlessDefined('gmap_filter_pci',    $DEFAULT_GMAP_FILTER_PCI);
	$o_param->SetUnlessDefined('gmap_filter_pcs',    $DEFAULT_GMAP_FILTER_PCS);
	$o_param->SetUnlessDefined('gmap_filter_span',   $DEFAULT_GMAP_FILTER_SPAN);
	$o_param->SetUnlessDefined('filter_pci',         $DEFAULT_FILTER_PCI);
	$o_param->SetUnlessDefined('filter_pcs',         $DEFAULT_FILTER_PCS);
	$o_param->SetUnlessDefined('filter_span',        $DEFAULT_FILTER_SPAN);
	$o_param->SetUnlessDefined('len_gene_number',    $DEFAULT_LEN_GENE_NUMBER);


	my $scaffolds      = $o_param->Get('scaffolds');
	my ($scadir, $scaname) = $scaffolds =~ /\/(.+)\/(.+)/;
	my $annotationfile = $o_param->Get('outfile');
	my $workingdir     = $o_param->Get('workingdir');
	my $uniq           = $$ . time;
	# Create the temporary directory if it doesnt exist
	my $cmd ;

	my $current_dir = `pwd`;
	if (! -d $workingdir)
	{
		mkdir ($workingdir);
	}
	chdir ($workingdir);
	$o_param->AssertDirExists('workingdir');
	
	my $ref_gff3      = $o_param->Get('ref_gff3');
	my $ref_scaffolds = $o_param->Get('ref_scaffolds');

	my $only_eval = 0;
	if (-e $annotationfile )
	{
		print "INFO: $annotationfile file already exists so only evaluate results.\n";
		$only_eval = 1;
	}


	if (!$only_eval)
	{
		if (!$o_param->IsDefined('ref_transcripts'))
		{
			my $transcripts_root = $workingdir ."/reference_transcripts_$uniq";
			# Compute the mRNA
			&Convert2Fasta( $ref_gff3, $ref_scaffolds, $transcripts_root, $o_param); 
			unlink($transcripts_root."_prot", $transcripts_root."_ncrna", $transcripts_root."_gene", $transcripts_root."_cds");
			$o_param->Set('ref_transcripts', $transcripts_root."_mrna");
		}
		$o_param->AssertFileExists('ref_transcripts');
		
		if (!defined(which("SIBsim4")))
		{
			print  "INFO : SIBsim4 not in the path. Skip the SIBsim4 mapping step.\n";
		}
	}
	
	my $transcripts = $o_param->Get('ref_transcripts');

	my $entries_file = &CreateArbo($o_param, $scaffolds, $workingdir);
	my $seqrootdir   = $workingdir."/0001/";
	
	my $eugene_out             = $workingdir . "/eugene.$uniq.GFF3";	
	my $eugenedir              = $workingdir . "/EUGENE.$uniq/";
	my $gmap_out               = $workingdir . "/gmap.$uniq.GFF3";
	my $sim4_out               = $workingdir . "/sim4.$uniq.GFF3";
	my $mapping_out            = $workingdir . "/$scaname.$uniq.GFF3";

	my $gmap_est_file             = $gmap_out . ".filter_est";
	my $mapping_est_file          = $mapping_out . ".filter_est";
	my $mapping_annotastruct_file = $mapping_out . ".filter_annotastruct";


	my $startregion     = $workingdir . "/start.$uniq.fna";
	my $codonstart_gff3 = $workingdir . "/start.$uniq.gff3";

	if (!$only_eval && !$o_param->IsDefined('eugene_out'))
	{
		if ($o_param->IsDefined('start_ok') && (-f &GetFirstSeqPath($entries_file).".start.gff3"))
		{	
			print "Skip start regions extraction. Already done\n";
		}
		else
		{
			# Extract start position
			&GetStart($o_param, $startregion, $codonstart_gff3);
			&SplitBySeq ($entries_file, $codonstart_gff3, "start.gff3", $seqrootdir);
			my $cmd = "gzip $startregion $codonstart_gff3"; system($cmd);
		}
	}

	if (!$only_eval  && !$o_param->IsDefined('mapping_out'))
	{
		# Map the transcripts to the scaffold using gmap
		if (!$o_param->IsDefined('gmap_out'))
		{
			&Gmap($o_param, $scaffolds, $transcripts, $gmap_out, $workingdir);
		}
		else
		{
			$gmap_out = $o_param->Get('gmap_out');
			if ($gmap_out =~ /\.gz$/)
			{
				my $cc = "gunzip $gmap_out ";
				system($cc);
				$gmap_out =~ s/\.gz$//;
			}
			print "INFO: Dont launch gmap, use $gmap_out\n";	
		}
		
		# Filter gmap results
		&GmapFilter($o_param, $gmap_out, $gmap_est_file, "est", 
					$o_param->Get('gmap_filter_pcs'), 
					$o_param->Get('gmap_filter_pci'), 
					$o_param->Get('gmap_filter_span'));
		
		# Extract the unmapped mrna and launch SIBsim4
		&RescueSIBsim4($gmap_est_file, $transcripts, $workingdir, $entries_file, $sim4_out, $o_param);
		
		# Concat both gmap and sim4 rescue results
		$cmd = "cat $gmap_out $sim4_out > $mapping_out";
		print "$cmd\n" if ($DEBUG); system ($cmd);
		
		# compress results
		$cmd = "gzip $gmap_out $sim4_out $mapping_out";
		print "$cmd\n" if ($DEBUG); system ($cmd);
		$mapping_out = $mapping_out.".gz";
	}
	else
	{
		$mapping_out = $o_param->Get('mapping_out');
	}

	if (!$only_eval )
	{
		# Filter the mapping results. 
		# Create two files: one for annotastruct, one for est plugin
		my $filter_pcs  = $o_param->Get('filter_pcs');
		my $filter_pci  = $o_param->Get('filter_pci');
		my $filter_span = $o_param->Get('filter_span');
		&GmapFilter($o_param, $mapping_out, $mapping_est_file, "est", $filter_pcs, $filter_pci, $filter_span);
		&GmapFilter($o_param, $mapping_out, $mapping_annotastruct_file, "annotastruct", $filter_pcs, $filter_pci, $filter_span);
		
		
		# Split results by seq and save it in the file tree (needed to launch eugene)
		&SplitBySeq ($entries_file, $mapping_est_file,          "gmap.est.gff3",          $seqrootdir);
		&SplitBySeq ($entries_file, $mapping_annotastruct_file, "gmap.annotastruct.gff3", $seqrootdir);
	}

	if (!$only_eval)
	{
		if (!$o_param->IsDefined('eugene_out'))
		{
			# Launch eugene to adjust predictions (splice site, start, etc..)
			&EuGene($entries_file, $eugenedir, $eugene_out, $o_param);
			$eugene_out = "$eugene_out.gz";
			$o_param->Set('eugene_out', $eugene_out);
		}
		else
		{ 
			$o_param->AssertFileExists('eugene_out');
			$eugene_out = $o_param->Get('eugene_out');
		}
	}
	# Create final annotation file
	# compare gmap result and eugene result and get ther name of the orthologs
	# If option filter, extract only genes whose structure does not change
	&CompleteAnnotation($o_param, $eugene_out, $mapping_annotastruct_file, $annotationfile, $ref_gff3) if (!$only_eval);
	
	if ( defined(which("nrdb") ))
	{
		$o_param->Set('prg_nrdb', "nrdb");
		&Eval($scaname, $workingdir, $annotationfile, $o_param);
	}
	else
	{
		print  "INFO : no evaluation because the mandatory program \"nrdb\" is not in the path (See help for more details)\n";
	}

}


sub GetFirstSeqPath
{
	my ($fof) = @_;

	my $f_in    = &General::GetStreamIn($fof);
    my $l       = <$f_in>;
	chomp($l);

    return $l;
}

# Detect start codon  on new scaffold sequences
sub GetStart
{
	my ($o_param, $startfasta, $resgff3) = @_;

	my $ref_gff3      = $o_param->Get('ref_gff3');
	my $ref_scaffolds = $o_param->Get('ref_scaffolds');
	
	my %h_seqs ;
	&GeneralBioinfo::FastaToHash($o_param->Get('ref_scaffolds'), \%h_seqs);

	my %h_starts = ();
	# everything is loaded to avoid sorting constraints
	&LoadStart($ref_gff3, \%h_starts);
	# Search the strat position on the new sequence
	&ExtractSites($o_param, $startfasta, $START_CONTEXT_LEN, \%h_starts,  \%h_seqs, $resgff3);

	return;
}


# Extract from the gff3 file the start position and the strand 
# Only use CDS information
# Prerequisite: file has to be sorted according to the start position
# Fill the hashtable referenced by $rh_starts. Keys: parent gene id, value: "start-end-strand-$cds_len"
sub LoadStart
{
    my($gfffile, $rh_starts) = @_;

    my $f_gff = &General::GetStreamIn($gfffile);

	my %h_res; # to parse a gff3 line
	
	while(my $line = <$f_gff>)
	{
		&ParseGFF3Line(\$line, 1, \%h_res);
		
		next if ($h_res{'type'} ne "CDS"); # Only CDS are read
		
		my $key =  $h_res{'Parent'}.&SEP.$h_res{'strand'};
		if ( !exists $$rh_starts{$key} ) # New gene
		{
			my $min    = $h_res{'start'};
			my $max    = $h_res{'end'};
			$$rh_starts{$key} = "$min\t$max\t".$h_res{'strand'}."\t".$h_res{'len'}."\t".$h_res{'seqid'};
		}
		else
		{
			# Update previous values
			my ($prev_min, $prev_max,  $prev_strand, $prev_len) = ( $$rh_starts{$key} =~ /(\d+)\t(\d+)\t(.+)\t(\d+)/);
			my $new_min = &General::Min($prev_min, $h_res{'start'});
			my $new_max = &General::Max($prev_max, $h_res{'end'});
			my $new_len = $prev_len + $h_res{'len'};
			$$rh_starts{$key} = "$new_min\t$new_max\t".$h_res{'strand'}."\t$new_len\t".$h_res{'seqid'};	
		}	
    }
    $f_gff->close;
	
    return;
}

# Detect the START codon. 
# 1) Extract the start region from the reference
# 2) Map to the new sequence
#3) Extract the start positions on the the new sequence and generate a GFF3 file
sub ExtractSites
{
    my($o_param, $tpfile_start, $contextlen,  $rh_starts,  $rh_seq, $res_gff3) = @_;

	my $f_tp_start = &General::GetStreamOut($tpfile_start);
	my $slen = 3;
	my $seq;
	my $id;

	my %h_start_pos;
	foreach my $idparent (keys(%$rh_starts))
	{
		my @a_curr  = split(/\t/, $$rh_starts{$idparent});
		my $strand  = $a_curr[2];
		my $cds_len = $a_curr[3];
		my $seqid   = $a_curr[4];
		my $genome_seq = $$rh_seq{$seqid}{sequence};
		my $genome_len = $$rh_seq{$seqid}{len};

		if ( $strand eq '+' )
		{
			my $leftlen;
			my $min_pos = $a_curr[0];
			if ( substr($genome_seq, $min_pos-1, 3) !~ /ATG/i )
			{
				print STDERR "WARNING IGNORE $idparent because not ATG $seqid $min_pos + " . substr($genome_seq, $min_pos-1, 3) ."\n" if ($DEBUG);
				next;
			}
			
			my $region_min;
			if ( ($min_pos-$contextlen) >= 1)
			{
				$region_min = $min_pos - $contextlen;
				$leftlen    = $contextlen;
			}
			else
			{
				$region_min = 1;
				$leftlen    = $min_pos-1;
			}
			my $region_length = ( ($region_min + $slen + $leftlen + $contextlen) <= $genome_len) ? ($leftlen + $contextlen + $slen) : ($genome_len-$region_min);
			$seq = substr($genome_seq, $region_min-1, $region_length);
			$id  = "$seqid:$region_min:$strand;";
			$h_start_pos{$idparent} = $leftlen+1;
		}
		else # strand minus
		{
			my $max_pos = $a_curr[1];
			if ( &ReverseComplement(substr($genome_seq, $max_pos-3, 3)) !~ /ATG/i )
			{
				print STDERR "WARNING IGNORE $idparent because not ATG $seqid $max_pos - " . substr($genome_seq, $max_pos-3, 3) ."\n" if ($DEBUG);
				next;
			}
			
			my $region_min;			
			my $leftlen;
			my $rightlen;
			if ( ($max_pos-$slen-$contextlen+1) >= 1)
			{
				$region_min = $max_pos-$slen-$contextlen+1;
				$leftlen    = $contextlen;
			}
			else
			{
				$region_min  = 1;
				$leftlen     = $max_pos-3;
			}
			
			my $length;
			if ( ($region_min + $leftlen + $contextlen + $slen) <= $genome_len)
			{
				$length   = $leftlen + $contextlen + $slen;
				$rightlen = $contextlen;
			}
			else
			{
				$length   = ($genome_len-$region_min);
				$rightlen = $genome_len - $leftlen - $region_min - $slen;
			}
			$seq = substr($genome_seq, $region_min-1, $length);
			$seq = &ReverseComplement($seq);
			$id  = "$seqid:$region_min:$strand";
			$h_start_pos{$idparent} = $rightlen+1;	
		}
		# assert position of the start on the new fasta sequence
			
		my $codon = substr($seq, ($h_start_pos{$idparent}-1 ), 3);
		if ($codon !~ /ATG/i )
		{
			print "NOT ATG !! $codon " . $h_start_pos{$idparent}. " $idparent\n$seq\n";
			next;
		}
		print $f_tp_start ">$idparent\n\U$seq\n";
	}
	$f_tp_start->close;

	# formater scaffolds
	my $sca = $o_param->Get('scaffolds');
	if (! -e $sca.".nin")
	{
		#my $cmd = prg_makeblastdb"formatdb -p F -i $sca";
		my $cmd = $o_param->Get('prg_makeblastdb') . " -in $sca -dbtype nucl -parse_seqids";
		system($cmd);
	}

	# Blast Start region to the genomic sequence
	my $cmd = $o_param->Get('prg_blastn') . " -query $tpfile_start -db " . $o_param->Get('scaffolds') . 
		" -out $tpfile_start.blast.m8 ". $o_param->Get('prg_blastn_param')." ";
	system($cmd);

	my $cpt   = 1;
	my $f_m8  = &General::GetStreamIn("$tpfile_start.blast.m8");
	my $f_res = &General::GetStreamOut($res_gff3);

	my %h_seqs ;
	&GeneralBioinfo::FastaToHash($o_param->Get('scaffolds'), \%h_seqs);

	# Extract from the m8 blast result the starts positions
	while (my $line = <$f_m8>)
	{
		chomp($line);
		my %h_m8;
		&GeneralBioinfo::ParseM8Line(\$line, 1, \%h_m8);
		# hit minimum len should be contextlen*2,  except for the case where the hit 
        # is a the begin or at the end of the genomic sequence
		# so test only len > contextlen
		if ($h_m8{pci} == $START_PCI && $h_m8{qstart} == 1 && $h_m8{len} > $contextlen)
		{	
			my $start_begin;
			if ($h_m8{strand} eq "+")
			{
				$start_begin = $h_m8{tstart} + $h_start_pos{$h_m8{query}} -1;	
			}
			else
			{
				$start_begin = $h_m8{tend} - $h_start_pos{$h_m8{query}} - 1;
			}

			my $codon = substr( $h_seqs{$h_m8{target}}{sequence}, $start_begin-1, 3);
			if  ($codon ne "ATG" && $codon ne "CAT")
			{
				print STDERR "FALSE : \"$codon\" $start_begin : $h_m8{target}\tEGN_ANNOT_TRANSFER\tstart_codon\t$start_begin\t".($start_begin+2)."\t.\t$h_m8{strand}\t.\tID=$h_m8{target}.start.$cpt;Name=$h_m8{query}\n" if($DEBUG);
				next;
			}

			print $f_res "$h_m8{target}\tEGN_ANNOT_TRANSFER\tstart_codon\t$start_begin\t".($start_begin+2)."\t.\t$h_m8{strand}\t.\tID=$h_m8{target}.start.$cpt;Name=$h_m8{query}\n";
			$cpt++;
		}
	}
	$f_res->close;
	$f_m8->close;

	my $cde = " gzip $tpfile_start.blast.m8"; system($cde);
	
	return;
}



sub ReverseComplement
{
	my($seq) = @_;

	$seq =~ tr/ATGCatgc/TACGtacg/;
	return reverse($seq);
}




=head2 procedure RescueSIBsim4
	
 Title        : RescueSIBsim4
 Usage        : 
 Prerequisite : 
 Function     : Extract the mrna not previously mapped. Launch SIBsim4 to these sequences
 Args         : $gmap_est_file GFF3       Mapping results
                $transcripts   Fasta file Reference transcripts
                $workingdir    Dir
                $entries_file  FoF        Path of the news genomic sequences
                $outfile       File       SIBSim4 result file   
                $o_param 
 Globals      : none

=cut
sub RescueSIBsim4
{
	my ($gmap_est_file, $transcripts, $workingdir, $entries_file, $outfile, $o_param) = @_;

	my $cmd;
	my ($transcriptdir, $transcriptname) = $transcripts =~ /\/(.+)\/(.+)/;
	my $idfile            = "$workingdir/$transcriptname.ID";
    my $transcriptsfilter = "$workingdir/$transcriptname.unmapped";

	my $f_    = &General::GetStreamIn($gmap_est_file);
	my $f_out = &General::GetStreamOut($idfile);
	my %h_view;
	while (my $line = <$f_>)
	{
		chomp($line);
		if ($line =~ /\tEST_match\t/)
		{
			my ($name) = $line =~ /Name=([^;]+)/;
			if (!exists($h_view{$name}))
			{
				print $f_out "$name\n";
				$h_view{$name} = 1;
			}
		}	
	}
	$f_->close;
	$f_out->close;

	# Extract the transcripts not previously found
	$cmd = $o_param->Get('prg_fasta_filter') . " --in $transcripts --out $transcriptsfilter --db_filename $idfile --remove";
	print $cmd."\n" if ($DEBUG);
	system($cmd);

	# If the non mapped sequence file is not empty, launch SIBsim4
	if (-f $transcriptsfilter && defined(which("SIBsim4")) )
	{
		# Launch  SIBsim4
		$cmd = $o_param->Get('prg_SIBsim4') . " --entries $entries_file --mrnadb $transcriptsfilter --outfile $outfile --workingdir $workingdir";
		print $cmd."\n" if ($DEBUG);
		system($cmd);
	}
	else
	{
		# create an empty file
		my $f_o = &General::GetStreamOut($outfile);
		$f_o->close;
	}

	$cmd = "gzip -f $idfile $transcriptsfilter";
	print $cmd."\n" if ($DEBUG);
	system($cmd);

	return;
	

}
=head2 procedure EuGene
	
 Title        : EuGene
 Usage        : 
 Prerequisite : egn_transfer_par exists 
 Function     : Launch EuGene and concat all results in $eugene_out file
                Compress all the temporary files in $eugenedir
 Args         : $entries_file FoF   File of fasta file path
                $eugenedir    dir   Directory where save egn results
                $eugene_out   file  Uniq final eugene result file
                $o_param 
 Globals      : none

=cut

sub EuGene
{
	my ($entries_file, $eugenedir, $eugene_out, $o_param) = @_;
	my $cmd;
	my $cfgfile = $o_param->Get('egn_transfer_par');
	
	system("rm $eugene_out") if (-e $eugene_out);
	system("rm $eugene_out.gz") if (-e "$eugene_out.gz");
	
	mkdir($eugenedir)  if (! -d $eugenedir);	
	
	my $eugene_cmd = $ENV{'EUGENEDIR'} . "/bin/eugene -O $eugenedir/ -A $cfgfile -pg ";
	
	my $cat_cmd    = "cat ";
	
	my $f_ = &General::GetStreamIn($entries_file);
	while(my $file = <$f_>)
	{
		chomp($file);
		my ($seqpath, $seqname) =  $file =~ /(.+)\/(.+)/;
		# eugene
		$cmd = $eugene_cmd . " $file";	
		print $cmd."\n" if ($DEBUG); system($cmd);
	}
	$f_->close;

	$f_ = &General::GetStreamIn($entries_file);
	while(my $file = <$f_>)
	{
		chomp($file);
		my ($seqpath, $seqname) =  $file =~ /(.+)\/(.+)/;
		$cmd = "cat $eugenedir/$seqname.gff3 >> $eugene_out";
		print $cmd."\n" if ($DEBUG); system($cmd);
		
	}
	$f_->close;

	# compress
	$cmd = "gzip $eugene_out";
	system($cmd) ;print $cmd."\n" if ($DEBUG);

	# compress eugene result files
    #my $cmd = "gzip $eugenedir/*";
	#print $cat_cmd."\n" if ($DEBUG);
	
	return;
}


=head2 procedure SplitBySeq
	
 Title        : SplitBySeq
 Usage        : 
 Prerequisite : SplitBySeq 
 Function     : Split data from gff3_file per sequence. 
                Create new files in $seqrootdir/seqX/seqX.$ext
 Args         : $entries_file FoF     File of fasta files
                $gff3_file     File   GFF3 file to split
                $ext          string  Extension of the files to create
                $seqrootdir   dir     Root directory of the sequences
 Globals      : none

=cut
sub SplitBySeq
{
	my ($entries_file, $gff3_file, $ext, $seqrootdir ) = @_;

	# Open FoF and create one file per sequence in the appropriate directory
	my %h_desc;
	my $f_ = &General::GetStreamIn($entries_file);
	while(my $file = <$f_>)
	{
		chomp($file);
		my ($seqpath, $seqname) =  $file =~ /(.+)\/(.+)/;
		my $f_desc = &General::GetStreamOut("$seqrootdir/$seqname/$seqname.$ext");
		$h_desc{$seqname} = $f_desc;
	}
	$f_->close;

	# Split by seq
	my $f_current;
	$f_ = &General::GetStreamIn($gff3_file);
	while(my $line = <$f_>)
	{
		chomp($line);
		my ($seq) = $line =~ /^(\S+)\s+/;
		$f_current = $h_desc{$seq};
		if (defined ($f_current))
		{
			print $f_current "$line\n";
		}
	}
	$f_->close;

	# close all the files
	foreach my $k (keys %h_desc)
	{
		$h_desc{$k}->close;
	}

	return;
}



#'prg_fasta2tree' as to exists in o_param
sub CreateArbo
{
	my ($o_param, $seq, $workingdir) = @_;
	
	if (! (-d "$workingdir/0001"))
	{
		my $cmd = $o_param->Get('prg_fasta2tree');
		$cmd .= " --sequence $seq --outdir $workingdir --type dna";
		print "$cmd\n" if ($DEBUG); system ($cmd);
	}

	return "$workingdir/entries.fof";
}

	


=head2 procedure GmapFilter
	
 Title        : GmapFilter
 Usage        : 
 Prerequisite : prg_gmap_filter gmap_filter_pcs gmap_filter_pci have to be defined 
 Function     : Filter gmap result $in
 Args         : $o_param
                $in gmap gff3 file
                $out gmap gff3 result file
                $format Est|AnnotaStruct
                $gmap_filter_pcs
                $gmap_filter_pci
                $gmap_filter_span
 Globals      : none

=cut
sub GmapFilter
{
	my ($o_param, $in, $out, $format, $gmap_filter_pcs, $gmap_filter_pci, $gmap_filter_span) = @_;
	
	# conversion du res de gmap pour eugene
	my $gmap_filter_pg  = $o_param->Get('prg_gmap_filter');

	my $cmd = "$gmap_filter_pg --best --in $in --out $out --pci $gmap_filter_pci --pcs $gmap_filter_pcs --span $gmap_filter_span --format $format";
	print "$cmd\n" if ($DEBUG); system ($cmd);

	return;
}

=head2 procedure Gmap
	
 Title        : Gmap
 Usage        : 
 Prerequisite : gmap and gmap_build has to be in the path 
 Function     : Call gmap_build and gmap
 Args         : $o_param
                $scaffolds
                $transcripts
                $gmap_out
                $workingdir
 Globals      : none

=cut
sub Gmap
{
	my ($o_param, $scaffolds, $transcripts, $gmap_out, $workingdir) = @_;

	my ($scaffold_path, $scaffold_name) = $scaffolds  =~ /(.+)\/([^\/]+)/;
	my ($g_path, $g_name) = $gmap_out  =~ /(.+)\/([^\/]+)/;
	my $gmapdir       = $workingdir . "/GMAP"; # directory for gmap indexes
	my $gmap_err               = $gmap_out . ".stderr";
	my $gmapbuild_log          = $gmap_out . ".build.stdout";
	my $gmapbuild_err          = $gmap_out . ".build.stderr";

	my $gmap_param = $o_param->Get('prg_gmap_param');

	my $cmd;
	# call gmap	
	my $cpt    = 0;
	my $suffix = time.$$;
	
	my @a_gmap_pg_dir = split(";", $o_param->Get('gmap_dir'));
	foreach my $gmap_dir (@a_gmap_pg_dir)
	{
		my $gmap_pg       = $gmap_dir."/gmap";
		my $gmap_build_pg = $gmap_dir."/gmap_build";
		if (-x $gmap_pg && -x $gmap_build_pg)
		{
			$cpt++;
			
			$cmd = "$gmap_build_pg -d $scaffold_name -D $gmapdir\_$cpt $scaffolds > $gmapbuild_log.$cpt 2> $gmapbuild_err.$cpt";
			print "$cmd\n" if ($DEBUG); 
			system ($cmd);
			
			$cmd = "$gmap_pg -f gff3_gene -D $gmapdir\_$cpt -d $scaffold_name $gmap_param $transcripts > $gmap_out.$suffix.$cpt  2> $gmap_err.$cpt";
			print "$cmd\n" if ($DEBUG); system ($cmd);
		}
	}	
	
	#$cmd = "find $g_path -name $g_name.$suffix.* -exec cat \{}\ \; > $gmap_out";
	$cmd = "cat $gmap_out.$suffix.* > $gmap_out";
	print "$cmd\n" if ($DEBUG); system ($cmd);
	
	# compress temporary files
	$cmd = "gzip $gmap_err.* $gmap_out.$suffix.* $gmapbuild_log* $gmapbuild_err*";
	print "$cmd\n" if ($DEBUG); system ($cmd);


	return;
}

=head2 procedure FilterAnnotation
	
 Title        : FilterAnnotation
 Usage        : 
 Prerequisite : 
 Function     : Compute intersection between exon of gmap results 
                and exons from eugene prediction
                Keep only gene where all the exons are conserved at 90%
                Keep the name of the original transcript sequence
 Args         : $o_param
                $eugene_file  GFF3 eugene result file
                $gmap_file   GFF3 gmap result file
                $res GFF3 result file
 Globals      : none

=cut
sub CompleteAnnotation
{
	my ($o_param, $eugene_file, $gmap_file, $res, $ref_gff3) = @_;

	my %h_functional;
	&GetFunctionalAnnot($ref_gff3, \%h_functional);

	my $filter = 0;
	$filter = 1 if ($o_param->IsDefined('exon_filter'));
	my $len_gene_nb = $o_param->Get('len_gene_number');
	my %h_intersect_exon_nb;
	my %h_homologs;
	# Search the homolog reference mRNA(s) 
	&Intersection($eugene_file, $gmap_file, \%h_intersect_exon_nb, \%h_homologs, $o_param);

	my %h_exon_nb;	
	my %h_gene_desc;	
	my @a_sorted_gene_id;
	&ParseGff3($eugene_file, \%h_exon_nb, \%h_gene_desc, \@a_sorted_gene_id);

	my $f_res = &General::GetStreamOut($res);
	print $f_res "##gff-version 3\n";

	my %h_cpt;
	foreach my $gene (@a_sorted_gene_id)
	{
		my $homolog_transcript = $h_homologs{$gene};
		
		# Keep the gene if an homolog mrna (or many) was found and 
        # if the number of exons is kept (cas filter_exon activated)
		if ( 
			defined($homolog_transcript) && 
			( ($filter == 0) || 
			  ( ($filter == 1) && ($h_exon_nb{$gene} == $h_intersect_exon_nb {$gene}) ))
		   )
		{
			$homolog_transcript =~ s/mrna:mRNA://g;
			my $desc            = $h_gene_desc{$gene};
			my $new_name        = "";
			my @a_lines = split(/\n/, $desc);
			foreach my $line (@a_lines)
			{
				if ( $line !~ /^#/ )
				{
					if ($new_name eq "")
					{
						# Build a name as SCAFFOLDNAMEg000001
						my ($sca) = $line =~ /^(\S+)/;
						if (exists $h_cpt{$sca})
						{
							$h_cpt{$sca}++;
						}
						else
						{
							$h_cpt{$sca} = 1;
						}
						my $gene_nb = $h_cpt{$sca};
						$new_name   = sprintf ($sca."g%0".$len_gene_nb."d",$gene_nb);
					}
					
					$line =~ s/Name=[^;]+/Name=$new_name/;
					# Add Name attributes at each line
					$line = $line . ";Name=$new_name" if ($line !~ /Name=/);  
					# Add transfer mrna name
					$line = $line . ";Alias=$homolog_transcript";
					&AddFunctionalAnnot(\$line, \%h_functional, $homolog_transcript);
				}
				print $f_res "$line\n";
			}
		}	
	}
	$f_res->close;
	

	return;
}





# Extract exon line from a gff3 file
sub GetExon
{
	my ($in, $out) = @_;
	
	my $f_in  = &General::GetStreamIn($in);
	my $f_out = &General::GetStreamOut($out);

	while (my $line = <$f_in>)
	{
		chomp($line);
		print $f_out $line . "\n" if ($line =~ /\texon\t/ || $line =~ /\ttranscript_region\t/);
	}
	$f_in->close;
	$f_out->close;
	
	return;
	
}



# Fill $rh_desc  : key=name gene, value= GFF3 gene lines,
# Fill $rh_exon_nb: key = ngene name, value = exon number
# Fill $ra_sort : sorted gene name
sub  ParseGff3
{
	my ($in, $rh_exon_nb, $rh_desc, $ra_sort) = @_;

	my $f_in  = &General::GetStreamIn($in);

	my $name = "";
	my $nb_exon;
	my $one_gene_lines;

	while (my $line = <$f_in>)
	{
		chomp($line);
		$line =~ s/[ \t]+$//;
		if ($line =~ /\tgene\t/)
		{
			if ($name ne "")
			{
				# on traite le cas du gene!
				$name =~ s/\./\\\./g;
				$$rh_desc{$name}    = $one_gene_lines . "###\n";
				$$rh_exon_nb{$name} = $nb_exon;
				push(@$ra_sort, $name);
			}
			# Reinit
			$one_gene_lines = $line."\n";
			$nb_exon        = 0;
			$name           = "";
		}
		else
		{
			($name) = $line =~ /ID=mRNA:([^;]+);/ if ($line =~ /\tmRNA\t/);
			$nb_exon++ if ($line =~ /\texon\t/);
			$one_gene_lines .= $line."\n";
		}	
	}

	# Last gene
	if ($name ne "")
	{
		# on traite le cas du gene!
		$name =~ s/\./\\\./g;
		$$rh_desc{$name}    = $one_gene_lines . "###\n";
		$$rh_exon_nb{$name} = $nb_exon;
		push(@$ra_sort, $name);
	}

	$f_in->close;

	return;
}


=head2 procedure Intersection
	
 Title        : Intersection
 Usage        : 
 Prerequisite : intersectBed is in the path
 Function     : Compute intersection between exon of gmap results 
                and exons from eugene prediction
                Fill $rh_exon_nb (For each gene, contains the number of exons found in the intersection)
                and $rh_homolog (For each gene, get the name of the homolog transcript(s))
 Args         : $eugene_file  GFF3 eugene result file
                $gmap_file   GFF3 gmap result file
                $rh_exon_nb  Ref to a has  Key=gene name, Value=interestion exon number
                $rh_homolog  Ref to a has  Key=gene name, Value=Transcript name
                $o_param
 Globals      : none

=cut
sub Intersection
{
	my ($eugene_file, $gmap_file, $rh_exon_nb, $rh_homolog, $o_param) = @_;
		
	my $eugene_exon = $eugene_file . ".exon";
	my $gmap_exon   = $gmap_file   . ".exon";
	my $intersecout = $eugene_exon . ".intersec";
	
	&GetExon($eugene_file, $eugene_exon);
	&GetExon($gmap_file,   $gmap_exon);
	my $interesectbed_param = $o_param->Get('intersectbed_param');
	my $prg_intersectBed    = $o_param->Get('prg_intersectbed');
	
	my $cmd = $prg_intersectBed . " -a $eugene_exon -b $gmap_exon $interesectbed_param -wo > $intersecout ; gzip -f $intersecout";
	system($cmd); print $cmd."\n" if ($DEBUG);
	$intersecout .= ".gz";

	# Compress temporary file
    my $cmd = "gzip -f $eugene_exon $gmap_exon";
	system($cmd); print $cmd."\n" if ($DEBUG);

	my $f_in  = &General::GetStreamIn($intersecout);

	my %h_exon_view;
	my $parent_name = "";

	my %h_similarity;

	# For each egn mrna, save the length of the intersection (sum of exon intersections) 
    # with the different reference mrna
	while (my $line = <$f_in>)
	{
		chomp($line);

		my @a_ = split(/\t/, $line);
		my %h_1;		
		my %h_2;
		&GeneralBioinfo::ParseGFF3Line(\join("\t", @a_[0..8]), 1, \%h_1);
		&GeneralBioinfo::ParseGFF3Line(\join("\t", @a_[9..17]), 1, \%h_2);
		my $intersec_len = $a_[18];

		if ($h_1{type} eq "exon")
		{
			my $mrna = $h_1{Parent};
			$mrna =~ s/mRNA://;
			$mrna =~ s/\./\\\./g; # ???

			my $exonkey = $mrna . $h_1{start} . "_" . $h_1{end} . "_" . $h_1{strand};
			if (!defined $h_exon_view{$exonkey})
			{
				$$rh_exon_nb{$mrna}++;
				$h_exon_view{$exonkey} = 1;
			}
			
			my $transcript_name = $h_2{Name};
			$h_similarity{$mrna}{$transcript_name} += $intersec_len;
		}
	}

	# For each egn mrna, save the ref mRNA homolog(s) 
    # (= mRNA(s) with the longer exon intersection)
	foreach my $egn_mrna (keys %h_similarity)
	{
		my $intersec = 0;
		
		foreach my $mrna ( keys %{$h_similarity{$egn_mrna}} )
		{
			if ($h_similarity{$egn_mrna}{$mrna} > $intersec)
			{
				$$rh_homolog{$egn_mrna} = $mrna;
				$intersec = $h_similarity{$egn_mrna}{$mrna};
			}
			elsif ($h_similarity{$egn_mrna}{$mrna} == $intersec)
			{
				if (defined ($$rh_homolog{$egn_mrna}))
				{
					$$rh_homolog{$egn_mrna} .= ",$mrna";
				}
				else
				{
					$$rh_homolog{$egn_mrna} = $mrna;
				}
			}
		}
	}
	
	$f_in->close;

	return;
}


#Check if dbxref and/or product was defined for the alias. If true, add this information of the gff3 line
# r_line String ref.
sub AddFunctionalAnnot
{
	my ($r_line, $rh_functional, $alias_list) = @_;
	my @a_split = split(/,/, $alias_list);

	my $product = "";	
	my $dbxref  = "";
	foreach my $alias (@a_split)
	{
		$product .= $$rh_functional{$alias}{product} ."," if (defined ($$rh_functional{$alias}{product}));
		$dbxref  .= $$rh_functional{$alias}{Dbxref} .","  if (defined ($$rh_functional{$alias}{Dbxref}));
	}
	$product =~ s/,$//;
	$dbxref =~ s/,$//;
	$$r_line .= ";product=$product" if ($product ne "");
	$$r_line .= ";Dbxref=$dbxref"   if ($dbxref ne "");

	return;
}


sub GetFunctionalAnnot
{
	my ($ref_gff3, $rh_functional) = @_;

	print "Load functional annotation..." if ($DEBUG);

	my $f_ = &General::GetStreamIn($ref_gff3);
	my %h_res;
	while (my $line = <$f_>)
	{
		next if ($line !~ /product=/ || $line !~ /Dbxref=/);
		&ParseGFF3Line(\$line, 1, \%h_res);
		my $name = "";
		$name  = $h_res{Name} if (defined($h_res{Name}));
		$name  = $h_res{ID} if ($name eq "" && defined($h_res{ID}));
		next if ($name eq "");
		if (defined $h_res{product})
		{
			$$rh_functional{$name}{product} = $h_res{product};
		}
		if (defined $h_res{Dbxref})
		{
			$$rh_functional{$name}{Dbxref} = $h_res{Dbxref};
		}
		
	}
	$f_->close;

	
	print "Done" if ($DEBUG);

	return;
}



sub CountSeq
{
	my ($fasta) = @_;
	
	my $cpt = 0;
	my $f_ = &General::GetStreamIn($fasta);

	while (my $line = <$f_>)
	{
		chomp($line);
		$cpt++ if ($line =~ /^>/);
	}
	$f_->close;

	return $cpt;
}



# Prerquisite : nrdb in the path
# Launch nrdb 
sub Nrdb
{
	my ($ra_files, $sep, $outfile, $prg_nrdb) = @_;

	my $cmd = "($prg_nrdb -d$sep -o $outfile";
	
	foreach my $f (@$ra_files)
	{
		$cmd .= " $f";
	}
	$cmd .= " > $outfile.out ) 2> $outfile.err";

	system($cmd);

	return;
}




# Create a file containing transfer information
sub Eval
{
	my ($scaname, $workingdir, $resultfile, $o_param) = @_;

	print "Start evaluation..\n" if ($DEBUG);
	my $eval_file  = "$workingdir/egn_annotation_transfer.evaluation.txt";
	my $f_eval  = &General::GetStreamOut($eval_file);
	my $evaldir = "$workingdir/EVAL/";
	mkdir ($evaldir) if (! -d $evaldir);

	my $ref_transcript ;
	if ($o_param->IsDefined('ref_transcripts'))
	{
		$ref_transcript = $o_param->Get('ref_transcripts');
	}
	else
	{	
		my @a_reffile = glob("$workingdir/reference_transcripts_*mrna");
		$ref_transcript = $a_reffile[0];
	}
	print $f_eval "*****************************************************************\n";
	print $f_eval &CountSeq($ref_transcript) . " reference mRNA\n";


	my @a_estfile    = glob("$workingdir/$scaname*.GFF3.filter_est*");
	my $mappedIDfile = "$evaldir/EVAL_mapped_mrna_id";
	
	my $f_    = &General::GetStreamIn($a_estfile[0]);
	my $f_out = &General::GetStreamOut( $mappedIDfile);
	my $nb_estmatch = 0;
	my %h_pos;
	my %h_name;
	while (my $line = <$f_>)
	{
		if ($line =~ /\tEST_match/)
		{
			my %h_gff3;
			&GeneralBioinfo::ParseGFF3Line(\$line, 1, \%h_gff3);
			$h_pos{"$h_gff3{start}_$h_gff3{end}_$h_gff3{strand}"}++;
			$h_name{"$h_gff3{Name}"}++;
			$nb_estmatch++;
			print $f_out "$h_gff3{Name}\n";

		}
	}
	$f_->close;
	$f_out->close;

	print $f_eval "*****************************************************************\n";
	print $f_eval "***********   Mapping results (gmap + SIBsim4)   ****************\n";	
    print $f_eval "- " . keys(%h_pos).   " different loci on new scaffolds\n";
	print $f_eval "- " . keys(%h_name) . " different reference mRNA mapped\n";	
	
	&GetNotMappedMRNA($ref_transcript, $mappedIDfile, $evaldir,  $f_eval);

	&EvalEuGeneOut($workingdir, $o_param, $f_eval);

	&ParseFinalResult($resultfile, $evaldir,  $ref_transcript, $f_eval, $o_param);

	&CompareProt($o_param->Get('ref_gff3'), $o_param->Get('ref_scaffolds'), 
				 $resultfile, $o_param->Get('scaffolds'), $evaldir, $f_eval, $o_param);
	$f_eval->close;

	print "INFO: transfer information available in $eval_file\n";
	return;
}






sub EvalEuGeneOut
{
	my ($workingdir, $o_param, $f_eval) = @_;


	my $eugene_res ;
	if ($o_param->IsDefined('eugene_out'))
	{
		$eugene_res = $o_param->Get('eugene_out');
	}
	else
	{
		my @a_eugene  = glob("$workingdir/eugene.*.GFF3.gz");
		$eugene_res = $a_eugene[0];
	}

	print $f_eval "*****************************************************************\n";
	print $f_eval "*********************   EuGene result   *************************\n";
	print $f_eval "- " . &CountGene($eugene_res). " genes predicted by EuGene\n";

	return;
}


sub CountGene
{
	my ($gff3) = @_;
	my $nb_gene = 0;
	my $f_ = &GetStreamIn($gff3);
	while (my $l =  <$f_>)
	{
		$nb_gene++ if ($l=~ /\tgene\t/);
	}
	return $nb_gene;
	
}



sub CompareProt
{
	my ($refgff3, $refgenome, $newgff3, $newgenome, $workdir, $f_eval, $o_param) = @_;

	my $prg_nrdb = $o_param->Get('prg_nrdb');
	
	my $sep1 = "#";

	# Get the proteome of the transfer annotation
	my ($newdir, $newfile) = $newgff3 =~ /(.+)\/(.+)/; 
	my $new_seq  = "$workdir/$newfile\_biofileconverter.FASTA";	
	my $new_prot = "$workdir/$newfile\_biofileconverter_prot.FASTA";
	&Convert2Fasta( $newgff3, $newgenome, $new_seq, $o_param ); 

	# Get the proteome of the ref annotation
	my ($refdir, $reffile) = $refgff3 =~ /(.+)\/(.+)/; 
	my $ref_seq  = "$workdir/$reffile\_biofileconverter.FASTA";	
	my $ref_prot = "$workdir/$reffile\_biofileconverter_prot.FASTA";
	&Convert2Fasta( $refgff3, $refgenome, $ref_seq, $o_param); 

	
	# nrdb about new proteome
	my $new_prot_nr = $new_prot.".nr";
	&Nrdb([$new_prot], $sep1, $new_prot_nr, $prg_nrdb);
	my $new_prot_nb    = &CountSeq($new_prot);
	my $new_prot_nr_nb = &CountSeq($new_prot_nr);
	
	# nrdb about reference proteome
	my $ref_prot_nr = $ref_prot.".nr";
	&Nrdb([$ref_prot], $sep1, $ref_prot_nr,  $prg_nrdb);
	my $ref_prot_nb    = &CountSeq($ref_prot);
	my $ref_prot_nr_nb = &CountSeq($ref_prot_nr);


	my $sep;
	my @a_list_sep = ("#", "!", "~");
	foreach my $test_sep(@a_list_sep)
	{
		my $cmd = "grep $test_sep  $ref_prot_nr $new_prot_nr";
		my $res = `$cmd`;
		if ($res eq "")
		{
			$sep = $test_sep;
			last;
		}
	}
	if ($sep eq "")
	{
		print STDERR "PROBLEM: cant launch nrdb because Fasta headers contain nrdb separator characters\n";
		exit;
	}

	# NRBD between ref proteome and new proteome
	my $nrdb_file = "$workdir/$newfile\_biofileconverter_prot.FASTA+ref.nr";
	&Nrdb([$ref_prot_nr,$new_prot_nr], $sep, $nrdb_file,  $prg_nrdb);


	my $f_ = &General::GetStreamIn($nrdb_file);
	my $conserved_prot_nb = 0;
	while (my $line = <$f_>)
	{
		chomp($line);
		if ($line =~ /^>/)
		{
			#print "comp ". $line."\n";
			 $conserved_prot_nb++ if ($line =~ /$sep/); 
			#print "$line\n" if (!($line =~ /Proq\d\dg\d+/ && $line =~ /ProqFM164S/));
		}
	}
	print $f_eval "*****************************************************************\n";
	print $f_eval "******** Comparison reference proteome and final proteome *******\n";
	print $f_eval "- Ref proteome: $ref_prot_nb proteins, $ref_prot_nr_nb unredundant proteins\n";
	print $f_eval "- New proteome: $new_prot_nb proteins, $new_prot_nr_nb unredundant proteins\n";
	print $f_eval "- $conserved_prot_nb nr proteins exactly conserved with the reference.\n";

	return;
}



sub ParseFinalResult
{
	my ($resultfile, $workingdir,  $ref_transcript, $f_eval, $o_param) = @_;
	my $f_ = &General::GetStreamIn($resultfile);
	my $gene_nb                     = 0;	
	my $gene_without_alias_nb       = 0;
	my $gene_with_multiple_alias_nb = 0;
	my %h_ref_mrna;
	while (my $line = <$f_>)
	{
		my %h_gff3;
		&GeneralBioinfo::ParseGFF3Line(\$line, 1, \%h_gff3);
		if ($h_gff3{type} eq "gene")
		{
			$gene_nb++;
			my $alias;
			$alias = $h_gff3{Alias} if (defined $h_gff3{Alias});
			if ($alias eq "")
			{
				$gene_without_alias_nb++;
			}
			else
			{
				$gene_with_multiple_alias_nb++ if ($alias =~ /,/);
				my @a_alias =split(/,/, $alias);
				foreach my $a (@a_alias)
				{
					$h_ref_mrna{$a}++;
				}
			}	
		}
	}
	$f_->close;


	my $transfered_mrna_id  = "$workingdir/EVAL_transfered_mrna_id";
	my $not_transfered_mrna = "$workingdir/EVAL_untransfered_mrna.fna";
	my $f_found = &General::GetStreamOut($transfered_mrna_id);
	foreach  my $mrna (keys %h_ref_mrna)
	{
		#print $f_found "mrna:mRNA:$mrna\n";
		print $f_found "$mrna\n";
	}
	$f_found->close;
	my $cmd = $o_param->Get("prg_fasta_filter").  " --in $ref_transcript --out $not_transfered_mrna --db_filename  $transfered_mrna_id --db_column 1 --remove";
	system($cmd);
	#print $cmd;

	print $f_eval "*****************************************************************\n";
	print $f_eval "************   EuGene result filtering (Homolog found)  *********\n";
	print $f_eval "- $gene_nb genes\n";
	print $f_eval "- $gene_with_multiple_alias_nb genes with multiple aliases\n";
	print $f_eval "- " . keys (%h_ref_mrna) . " reference mRNA transfered\n";
	print $f_eval "- " . &CountSeq($not_transfered_mrna) ." reference mRNA not transfered ($not_transfered_mrna)\n";

	return;
}




sub GetNotMappedMRNA
{
	my ($ref_transcript, $mappedIDfile, $workdir, $f_eval) = @_;
	
	my $unmappedmrna = "$workdir/EVAL_unmapped_mrna.fna";
	my $cmd = "lipm_fastafilter.pl --in $ref_transcript --out $unmappedmrna  --db_filename  $mappedIDfile --remove --db_column 1";	
	print "$cmd\n" if ($DEBUG);
	system($cmd);
	
	print $f_eval "- " . &CountSeq($unmappedmrna) . " reference mRNA not mapped ($unmappedmrna)\n";

	return;

}




sub Convert2Fasta
{
	my ($gff3, $fasta, $out, $o_param) = @_;
	
	my $cmd = "(". $o_param->Get('prg_biofileconverter') ." --input-ann-file $gff3 --output-ann-file $out --input-seq-file $fasta --all-contigs --input-ann-format GFF3 --output-ann-format FASTA > $out.out ) 2> $out.errr";
	print $cmd if ($DEBUG);
	system($cmd);
	
	return;
}

sub Usage
{
    print STDERR<<END
[INSTALLATION REQUIREMENTS]
    (1) Make sure SIBsim4 is in the path (SIBsim4 is available as a debian package).
        If not, gmap mapping only is performed.
    (2) Make sur nrdb is in the path. 
        If not, the evaluation of the annotation transfer is not performed.

[USAGE]
$0
    --scaffolds     file   multifasta file. Genomic sequence to annotate [Not compressed]
    --ref_scaffolds file   multifasta file. Genomic sequence of reference
    --ref_gff3      file   GFF3 file. Reference annotation
    --outfile       file   GFF3 result file   
    --workingdir    dir    working directory, where temporary files will be created
    --cfg           file   Configuration file. Example in \$EGNEP/cfg/egn_annotation_transfer.cfg 
    
    --ref_transcripts file   multifasta file. Reference mRNA [Not compressed] OBSOLETE OPTION

[Optional]
    --exon_filter   Flag    Keep only genes whose all exons are conserved before and after EuGene run 
                            (Use values of 'intersectbed_param')

[Advanced Options]
    --gmap_filter_pci    [$DEFAULT_GMAP_FILTER_PCI] Integer Minimum percentage identity used to filter gmap results
    --gmap_filter_pcs    [$DEFAULT_GMAP_FILTER_PCS] Integer Minimum percentage coverage used to filter gmap results
    --gmap_filter_span   [$DEFAULT_GMAP_FILTER_SPAN] Integer Minimum length used to filter gmap results
    --filter_pci    [$DEFAULT_FILTER_PCI] Integer Minimum percentage identity used to filter mapping results (SIBsim4+gmap)
    --filter_pcs    [$DEFAULT_FILTER_PCS] Integer Minimum percentage coverage used to filter mapping results (SIBsim4+gmap)
    --filter_span   [$DEFAULT_FILTER_SPAN] Integer Minimum length used to filter mapping results (SIBsim4+gmap)
    --egn_transfer_par   [$EGNEP/cfg/euk/egn_annotation_transfer.par] File EuGene parameter file used to adjust gmap results
    --intersectbed_param [$DEFAULT_INTERSECTBED_PARAM] String Intersectbed parameters used to compare exons 
                                                       before and after eugene adjust. If exon_filter is activated
    --len_gene_number   [$DEFAULT_LEN_GENE_NUMBER] Integer   Length of the number in the gene names SCA1g000001

    --mapping_out
    --gmap_out 
    --eugene_out
    --start_ok

[TODO]
The program maps the regions around start codons to improve the start prediction. This step checks the codon is equal to "ATG"
Have to manage the other start codons!!
Remove nrdb dependancy. (need to get our own nrdb program which manages the -d option)

END
;
}
