#!/usr/bin/perl
#
# $Id
#

# Copyright INRA

# Jerome.Gouzy@toulouse.inra.fr
# Erika.Sallet@toulouse.inra.fr
# Thomas.Schiex@toulouse.inra.fr

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=pod 

=head1 NAME


=head1 SYNOPSIS

get_Est_SIBsim4.pl --sequence --mrna --outfile  --workingdir
Launch SIBsim4 on each genomic sequence
Convert result in GFF3 (same format than gmap) anc concat all th results in outfile

Note that it can be quite long

=head1 DESCRIPTION


=cut
BEGIN
{
	$ENV{LANG} = "C";

	use Cwd 'abs_path';
	use File::Basename;
	my ($dirprg) =  dirname(abs_path($0));
	unshift(@INC,"$dirprg/../../../lib");
	unshift(@INC,"$dirprg/../../../corelib");

}

use strict;
use warnings;
use Carp;
use ParamParser;
use General;
use GeneralBioinfo;

MAIN:
{
    my $o_param = New ParamParser('GETOPTLONG',\&Usage, 'sequence=s', 'entries=s', 
								  'mrnadb=s', 'workingdir=s', 'outfile=s');

	$o_param->SetBehaviour('ignore_space');
    $o_param->Update('ENV', "A");
    $o_param->SetUsage(my $usage = sub {&Usage();});

	$o_param->AssertDirExists('EGNEP');
	$o_param->AssertFileExists('mrnadb');	
	$o_param->AssertDefined('outfile');
	$o_param->AssertDirExists('workingdir');
	$o_param->AssertFullPath('outfile');

	my $workingdir = $o_param->Get('workingdir');
	my $mrnadb     = $o_param->Get('mrnadb');
	my $outfile    = $o_param->Get('outfile');

	if (!$o_param->IsDefined('entries'))
	{
		# Build tree
		$o_param->AssertFileExists('sequence');
		my $fof = &BuildTree($workingdir, $o_param->Get('sequence'), $o_param);
		$o_param->Set('entries', $fof);
	}
	else
	{
		$o_param->AssertFileExists('entries');	
	}
	
	&LaunchSIBSim4($mrnadb, $outfile, $o_param);
}


sub BuildTree
{
	my ($workingdir, $sequence, $o_param) = @_;
	my $cmd;

	if (-d "$workingdir/0001")
	{
		die "Can't build the file tree because directory $workingdir/0001 already exists\n";
	}

	my $fasta2tree = $o_param->Get('EGNEP')."/bin/ext/lipm_fasta2tree.pl";
	$cmd = "$fasta2tree --sequence $sequence --outdir $workingdir/";
	system($cmd);

	return  "$workingdir/entries.fof";
}



sub LaunchSIBSim4
{
	my ($mrna, $outfile, $o_param) = @_;

	my $cmd;
   
	my $ra_seq_files = $o_param->GetListOfFiles('entries');

	foreach my $seqfile (@$ra_seq_files)
	{
		$cmd = "SIBsim4 -A 0 $seqfile  $mrna > $seqfile.SIBsim4";
		system($cmd);
		# convert in gff3 
		&ConvertGFF3("$seqfile.SIBsim4", "$seqfile.SIBsim4.gff3");
	}

	# Collect all gff3 files, remove individual gff3 files, and compress SIBsim4 raw results
	foreach my $seqfile (@$ra_seq_files)
	{
		system("cat $seqfile.SIBsim4.gff3 >> $outfile");
		system("rm $seqfile.SIBsim4.gff3");
		system("gzip -f $seqfile.SIBsim4");
	}

	return;
}

# convert in gff3 ( gmap gff3 format)
sub ConvertGFF3
{
	my ($sim4file, $gff3file) = @_;

	my $candidate    = 1;
	my $genomename   = "";	
	my $genomestrand = "+";
	my $mrnaname     = "";
	my $mrnalen      = 0;

	my $cpt = 0;
	my $identity   ;
	my @a_exons = ();
	my $f_out = &General::GetStreamOut($gff3file);
	my $f_    = &General::GetStreamIn($sim4file);
	while (my $line = <$f_>)
	{
		chomp($line);
		next if ($line =~ /$^/); # empty line
		if ($line =~ /^>/)
		{
			my ($id) = $line =~ />(\S+)/;
			if ($candidate != 0)
			{
				if (scalar(@a_exons) > 0)
				{
					# End of the previous hsp: write it!
					&SaveHSP(\@a_exons, \$cpt, $mrnaname, $mrnalen, $genomename, $genomestrand, $f_out);
				}
				# Reinit the values
				$genomestrand = "+";
				$mrnaname     = "";
				$mrnalen      = 0;
				@a_exons       = ();

				$candidate     = 0;
				# Save the genomic sequence
				$genomename = $id;
			}
			else
			{
				# second fasta header
				# Save the mrna information
				# >mrna:mRNA:Proq01g001490a known; LEN=1745
				$mrnaname    = $id;
				($mrnalen) = $line =~ /LEN=(\d+)$/;
				$candidate = 1;
			}		
		}
		elsif ($line =~ /^(\d+)-(\d+)\s+\((\d+)-(\d+)\)\s+([\d\.]+)%/)
		{
			push(@a_exons, $line);
		}
		elsif ($line =~ /\(complement\)/)
		{
			$genomestrand = "-";
		}
	}

	&SaveHSP(\@a_exons, \$cpt, $mrnaname, $mrnalen, $genomename, $genomestrand, $f_out) if (scalar(@a_exons) > 0);
	$f_out->close;
	$f_->close;
	
	return;
}




sub SaveHSP
{
	my ($ra_exons, $r_cpt, $mrnaname, $mrnalen, $genomename, $genomestrand, $f_out) = @_;
	
	my $hsp_start = 100000000000;
	my $hsp_end   = 0;
	$$r_cpt++;
	
	# If strand minus, change the sort of the exons to generate the same output than gmap
	if ($genomestrand eq "-")
	{
		@$ra_exons = sort {my ($aa) = $a =~ /^(\d+)-/; my ($bb) = $b =~ /^(\d+)-/; $bb <=> $aa} @$ra_exons;
	}
	my $geneid = "$mrnaname.gene.$$r_cpt";
	my $mrnaid = "$mrnaname.mrna.$$r_cpt";
	
	my $exon_lines = "";
	my $cov_len       = 0;
	my $identical_len = 0;
	my $exon_cpt = 0;
	foreach my $matchline (@$ra_exons)
	{
		my ($genome_start, $genome_end, $match_start, $match_end, $ex_id) = $matchline =~ /^(\d+)-(\d+)\s+\((\d+)-(\d+)\)\s+([\d\.]+)%/;
		$hsp_start = $genome_start if ($genome_start < $hsp_start);
		$hsp_end   = $genome_end   if ($genome_end > $hsp_end );

		# Minus strand: recompute target position
		if ($genomestrand eq "-")
		{
			my $prev_match_start = $match_start;
			$match_start = $mrnalen - $match_end +1;
			$match_end   = $mrnalen - $prev_match_start +1;
		}

		my $match_len = ($match_end - $match_start + 1);
		$cov_len       += $match_len;
		$identical_len += int($match_len*$ex_id/100);

		$exon_cpt++;
		

		$exon_lines .= "$genomename\tSIBsim4\texon\t$genome_start\t$genome_end\t.\t$genomestrand\t.\tID=$mrnaid.exon.$exon_cpt;Name=$mrnaname;Parent=$mrnaid;Target=$mrnaname $match_start $match_end +\n";
	}
	
	my $coverage = sprintf ("%.01f", $cov_len/$mrnalen*100);
	my $identity = sprintf ("%.01f", $identical_len/$cov_len*100);
	my $gene_line = "$genomename\tSIBsim4\tgene\t$hsp_start\t$hsp_end\t.\t$genomestrand\t.\tID=$geneid;Name=$mrnaname\n";
	my $mrna_line = "$genomename\tSIBsim4\tmRNA\t$hsp_start\t$hsp_end\t.\t$genomestrand\t.\tID=$mrnaid;Name=$mrnaname;Parent=$geneid;coverage=$coverage;identity=$identity\n";
	
	print $f_out $gene_line.$mrna_line.$exon_lines."###\n";
	return;
}


sub Usage
{
    print STDERR<<END
$0
[Prerequisite]
program SIBsim4 in the path.

Launch SIBsim4 and convert it in GFF3 (gmap GFF3 format http://research-pub.gene.com/gmap/) 

[Mandatory]
    --sequence    file          Input multifasta file of genomic sequences
OR
    --entries     fileoffiles   File listing genomic fasta files

    --mrnadb      file          multifasta file of mRNA
    --outfile     file          GFF3 output file
    --workingdir  dir          

    --help
END
}

