#
# $Id: GTF.pm 1081 2013-03-05 13:47:04Z manu $
#

package AnnotatedSeqFileReader::GTF;
 
# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


=head1 NAME

AnnotatedSeqReader::GTF - parse a GTF formatted file and load an AnnotatedSeq object
                          See the GTF description at http://mblab.wustl.edu/GTF22.html

=head1 SYNOPSIS
	See AnnotatedSeq

    Warnings: the current version of the parser does not manage 
    see the description of other public methods in the SYNOPSIS of AnnotatedSeq.pm

=head1 DESCRIPTION

=cut

use strict;
use warnings;
use Carp;
use General;

use Error qw(:try);
use Bio::SeqIO;
use Bio::Seq;
use AnnotatedSeq;
use Data::Dumper;
use File::Spec;
use URI::Escape;

use base qw(AnnotatedSeqFileReader);


BEGIN
{
    our $VERSION = do {my @r = (q$Rev: 1081 $ =~ /\d+/g); $r[0]};
}


=head2 procedure _Init

 Title        : _Init
 Usage        : my $o_gff3 = New AnnotatedSeqFile::GFF3(...);
 Function     : Constructor

=cut

sub _Init
{
	my $self = shift;
	throw Error::Simple( 'GTF FORMAT NOT YET IMPLEMENTED - PLEASE DO NOT USE !!!' );

	$self->SUPER::_Init(@_);
	my $o_annot = $self -> { _o_annot };
}
sub GetFormat
{
	return 'GTF';
}
=head2 procedure DetectContig

 Title        : DetectContig
 Usage        : my $ctg = $o_gff3->DetectContig();
 Function     : Open the gff3 file until a tabulated line is found, the 1st col is the contig label
 Args         : none
 Globals      : none

=cut
sub DetectContig
{
	my $self    = shift;
	my $o_annot = $self -> { _o_annot };
	my $f_gtf = &GetStreamIn($o_annot->{__in_ann_file});
	my $ctg = "";
	while( my $line = <$f_gtf>)
	{
		chomp($line);
		next if ( $line =~ /^\#/ || $line eq '');
		($ctg) = split(/\t/,$line,2);
		last;
	}
	$o_annot -> { __contig } = $ctg;
}		

=head2 procedure LoadGenomicSequence

 Title        : LoadGenomicSequence
 Usage        : $o_gff3->LoadGenomicSequence();
 Function     : throw an exception, because the loader of the genomic seuqnece from a gff3 file is not yet implemented
 Returns      : none
 Args         : none
 Globals      : none

=cut
sub LoadGenomicSequence
{
	my ($self) = @_;
	throw Error::Simple( 'Loading a sequence from a gtf file is not implemented, the --input-seq-file switch is probably your friend' );
}

=head2 procedure Load

 Title        : Load
 Usage        : $o_gff3->Load();
 Prerequisite : the infile must be in GFF3 format and should contain unique IDs
 Function     : Load the protein-coding genes, the Ncrna and the repeat contained in a GFF3 formatted file
                Warning: only the features gene|RNA|repeat_region|CDS are loaded
 Returns      : none
 Args         : none
 Globals      : none
 TODO         : REFAIRE CETTE FONCTION TROP COMPLEXE, EN:
                        - EVITANT DE RECOPIER LE FICHIER gff3 CE N'EST PAS NECESSAIRE
                        - UTILISANT LES METHODES GetAllowedTypes DES OBJETS DERIVES DE Locus

=cut

sub Load
{
	my $self = shift;
	my $o_annot = $self -> { _o_annot };

	my $f_gtf = &GetStreamIn($o_annot->{__in_ann_file});
	
	# To ensure that the gene is defined before the corresponding mRNA and mRNA before CDS,
	# we split the gtf file in 4 temporary gtf files
	#my $f_gtf = new IO::File("$filename","r") or Carp::croak("unable to read >$filename<");
	
	my (%h_f_gtf,%h_fh_gtf);
	
	# NOTE The other features (start_codon, stop_codon, inter, inter_CNS, intron, intron_CNS) are ignored
	my @a_codes = qw( CDS exon 5UTR 3UTR);
	my $tmp_dir = $o_annot -> { __tmpdir };
	foreach my $code ( @a_codes )
	{
		$h_f_gtf  { $code } = File::Spec->catfile( $tmp_dir, "import." . $code . ".gtf" );
		$h_fh_gtf { $code }  = &GetStreamOut($h_f_gtf{$code});
	}
	
	# Read the gtf file, set the contig label if necessary, and create 1 gtf file / supported feature
	my $contig_label = $o_annot -> GetContigLabel();
	while( my $line = <$f_gtf>)
	{
		chomp($line);
		next if ( $line =~ /^\#/ || $line eq '');
		my ($ctg,$src,$code,$start,$end,$score,$strand,$phase,$comment) = split(/\t/,$line);
		next if ( $ctg ne $contig_label );

		next if ( $code !~ /exon|CDS|UTR/);
		$code='UTR' if ( $code =~ /UTR$/ );
		my $fh = $h_fh_gtf{$code};
		print $fh "$line\n";
	}
	foreach my $code ( @a_codes )
	{
		$h_fh_gtf { $code }->close();
	}
	$f_gtf->close;
	print STDERR "Filtering gtf files finished\n" if ( $o_annot -> GetDebug() );

	# Read the files we just wrote, in the appropriate order
	foreach my $code ( @a_codes )
	{
		#my $fh_gtf = new IO::File($h_f_gtf{$code},"r") or Carp::croak("unable to read >$h_f_gtf{$code}<");
		my $fh_gtf = GetStreamIn($h_f_gtf{$code});
		while( my $line = <$fh_gtf>)
		{
			try
			{
				chomp($line);
				my %h_param=();
				my ($ctg,$src,$type,$start,$end,$score,$strand,$phase,$comment) = split(/\t/,$line);
				
				$h_param{start}  = $start;
				$h_param{end}    = $end;
				$h_param{strand} = ( $strand eq '+' ) ? 1 : -1;
				$h_param{type}   = $type;
				$h_param{score}  = $score;
				$h_param{phase}  = $phase;
				
				my @a_com=split(';',$comment);
				for (my $i=0;$i<=$#a_com;$i++)
				{	
	                my $pair = $a_com[$i];
	                $pair =~ s/^\s+|\s+$//g ; # MANU Oct 2009 - Remove heading and trailing blanks
	                next if ($pair eq "");
	
					my ($key,$value) = split(/ +/,$pair,2);
					
					# unescape characters 
					$value = uri_unescape ( $value );
					if ( $key eq 'gene_id' )
					{
						$h_param{gene_name} = $value;
					}
					else
					{
						$h_param{$key} = $value;
					}
				}
				
				if ( $type =~ /CDS|exon|UTR/ )
				{
					my $o_region;
					my $transcript_id = $h_param{ 'transcript_id' };
					delete $h_param{ 'transcript_id' };
					
					$o_region = New Locus::CdsRegion(%h_param) if ($type =~ /CDS/);
					$o_region = New Locus::Exon(%h_param)      if ($type =~ /exon/);
					$o_region = New Locus::UtrRegion(%h_param) if ($type =~ /UTR/);

					my $o_gene = $o_annot -> AddGeneCDSorExon ( $o_region, $h_param{gene_name} );
					if ( ! defined $o_gene )
					{
						$o_gene = New Locus::Gene(%h_param,'type'=>'protein_coding_gene','locus_tag'=>$h_param{gene_name});
						$o_annot -> AddGene( $o_gene );
						$o_annot -> AddGeneCDSorExon ( $o_region, $h_param{gene_name} );
						my $o_mrna = $o_region -> GetParent() -> GetParent();
						$o_mrna -> SetTranscriptId ( $transcript_id );
						#print "$o_gene\n";
					}
					else
					{
						my $t_id = $o_region -> GetParent() -> GetParent() -> GetTranscriptId();
						if ( $t_id ne $transcript_id )
						{
							my ($s,$e) = $o_region -> GetEnds();
							#throw Error::Simple ( "Transcript Id of the mRNA is $t_id, should be $transcript_id [$s..$e]" );
							warn ( "Transcript Id of the mRNA is $t_id, should be $transcript_id [$s..$e]" );
						}
					}
				}
			}
			catch Error with
			{
				my ($o_error) = @_;
				my $msg = $o_error -> text();
				warn ( "ERROR $msg\n" );
				$o_annot -> SetFatalError();
			};
		}
		undef $f_gtf;
	}
}

=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;
