#
# $Id: UsingBioPerl.pm 1336 2013-12-04 13:35:33Z manu $
#
 
package AnnotatedSeqFileReader::UsingBioPerl;

# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=head1 NAME

AnnotatedSeqFileReader::UsingBioPerl - parse a file and load an AnnotatedSeq object, using the BioPerl API
                                       Currently a base class for genbank and embl file formats
                                       See www.bioperl.org

=head1 SYNOPSIS
	See AnnotatedSeq

=head1 DESCRIPTION

=cut

use strict;
use warnings;
use Carp;

use Error qw(:try);
use Bio::SeqIO;
use Bio::Seq;
use General;
use AnnotatedSeq;
use Data::Dumper;
use GenbankTags;
use MultiValHash;

use base qw(AnnotatedSeqFileReader);

BEGIN
{
    our $VERSION = do {my @r = (q$Rev: 1336 $ =~ /\d+/g); $r[0]};
}

=head2 procedure _Init

 Title        : _Init
 Usage        : my $o_gbk = New AnnotatedSeqFile::Embl(...);
 Function     : Constructor
 Args         : $o_annot The AnnotatedSeq object inside which we live
                $format  The file format (currently 'embl', 'genbank')

=cut

sub _Init
{
	my $self = shift;
	my ($o_annot,$format) = @_;

	$self -> SUPER::_Init($o_annot);
	$self -> { '_format' } = $format;
	my $o_genbank_tags = New GenbankTags;
	$self -> { __o_genbank_tags } = $o_genbank_tags;
}

sub GetFormat
{
	my $self = shift;
	return $self -> { '_format' };
}

sub DetectContig { 	throw LipmError::ObjectException( 'AnnotatedSeqFileReader::UsingBioPerl', 'abstract method' ); }

=head2 procedure LoadGenomicSequence

 Title        : LoadGenomicSequence
 Usage        : my $r_seq = $o_gbk->LoadGenomicSequence();
 Function     : Load the genomic sequence from the file
 Returns      : The genomic sequence
 Args         : none
 Globals      : none

=cut
sub LoadGenomicSequence
{
	my $self      = shift;
	my $o_annot   = $self -> { _o_annot };
	my $filename  = $o_annot->{__in_ann_file};
	my $format    = $self -> { _format };
	my $fh_bp     = &GetStreamIn( $filename );
	my $seqstream;
	try
	{
		$seqstream = new Bio::SeqIO(-format => $format, -fh => $fh_bp);
	}
	catch Error with
	{
		throw Error::Simple ( "ERROR - BioPerl cannot open the file: $filename, probably not a good format" );
	};
	
	my $seq       = $seqstream->next_seq();
	my $genomic_seq = $seq->seq();
	return \$genomic_seq;
}	

=head2 procedure Load

Title        : Load
Usage        : $o_gbk->Load();
Prerequisite : __in_ann_file should be an embl/gbk/... file, with only ONE LOCUS, and the file should contain unique "locus_tag"
Function     : Load the protein-coding genes, the ncRNA and the repeat contained from a file, using the BioPerl API
Returns      : none
Args         : none

=cut

sub Load
{
	my ($self) = @_;
	my $o_annot = $self -> { _o_annot };
	my $contig_label = $o_annot -> GetContigLabel();
   	my $o_genbank_tags = $self -> { __o_genbank_tags };

    # The tags declared here are reserved, other tags will be affected to the 'other' field (see Locus.pm)
	#my %h_reserved_tags = ( rpt_family => 1, locus_tag => 1, gene => 1, product => 1, note => 1, transcript_id => 1, protein_id => 1 );
    #my %h_reserved_tags = ( gene => 1, product => 1, note => 1 );

	my $filename = $o_annot->{__in_ann_file};
	my $format   = $self -> { _format };
	my $fh_bp    = &GetStreamIn( $filename );
	my $seqstream;
	try
	{
		$seqstream  = new Bio::SeqIO(-format => $format, -fh => $fh_bp);
	}
	catch Error with
	{
		throw Error::Simple ( "ERROR - BioPerl cannot open the file: $filename, probably not a good format" );
	};
	
	my $seq = $seqstream->next_seq;
	if ( ! defined $seq ) 
	{
		throw Error::Simple("ERROR - $filename is probably NOT an $format formatted file !");
	}		

	#$self -> __LoadHeaders($seq);
	
	# Some info about the genes: key = gene name (/gene), val = ref_hash with the list of mrnas, 
	# and counters for utr, exons, cds

	# @a_genes_info contains rh objects, a ref to $o_gene and counters for utr, cds, mrna
	#     o_gene => $o_gene, utr3_cnt => 0, utr5_cnt => 0, cds_cnt => 0, mrna_cnt => 0
	# Before using @a_genes_info, it must be SORTED by __SortGenesInfo
	#     After this sort: 
	#     1/ The objects are ordered as follows:
	#        The strand (-1,+1)
	#        The start position of the gene
	#        The end position of the gene
	#     2/ @a_genes_info contains the index of the FIRST STRAND + GENE
	#     3/ %h_genes_info is filled by __SortGenesInfo:
	#        key = The pk (Position Key, ie strand-strt-end)
	#        val = The index of the gene in @a_genes_info
	#        It is a MultiValHash, several genes may have the same pk (C'EST VRAI CA ???)
	my @a_genes_info;
    &MultiValHash::SetResultAsARefArray();
	my %h_genes_info = ();
	tie %h_genes_info, 'MultiValHash';
	
	# We scan several times the sequence, to get features in a correct order:
	#    1/ genes, repeats
	#    2/ RNAs (nc or m)
	#    2/ utr
	#    3/ exons
	#    4/ CDS
	#
	my $first_pass      = 1;
	
	# NOTE - Exon is still experimental !
	foreach my $short_class (qw( Gene Mrna CdsRegion MiscRegion Ncrna Repeat UtrRegion Exon ))
	{
		my $class = 'Locus::'.$short_class;
		my $rh_features = $o_genbank_tags -> GetFeaturesForClass ( $class );
		foreach my $f ( $seq->all_SeqFeatures() )
		{		
			my $feature = $f->primary_tag();
			if ( $first_pass == &TRUE )
			{
				# The feature 'source' is NOT connected to a Locus !
				if ( $feature eq 'source' )
				{
					my $sstart = $f->location->start;
					my $send   = $f->location->end;
					$o_annot -> UpdateSeqRegion($sstart,$send);
				}
				elsif ( $o_genbank_tags -> GetLocusClass( $feature ) eq "" )
				{
					# We do not use GetWarningHeader here, because we want $so to be part of the warning message, not only data
					warn ( "WARNING $feature $feature is NOT a recorded feature - Paragraph skipped !\n" );
				}
			}
			next unless exists $rh_features -> { $feature };

			# Common parameters
			my %h_param = ();

			$h_param{start} = ($f->location->start);
			$h_param{end}   = ($f->location->end);
			$h_param{strand}= $f->strand;
			$h_param{-fragment_info}=$self->__ConvertFragmentInfo($h_param{strand},$f->location);

			# The tags: 1/ control if the qualifier is legal for this feature, warn if not (but read it anyway)
			#           2/ If there are several values for a tag, keep them in an array and store the ref to this array
			#              If only one value, store just the value (NOT an array with 1 value !)
			foreach my $qualifier ( $f->all_tags() )
			{
				# SPECIAL TREATMENT FOR /pseudo => replace /pseudo by /pseudogene="unknown"
				if ( $qualifier eq 'pseudo' )
				{
					$h_param{ 'pseudogene' } = "unknown";
					next;
				}
				if ( !$o_genbank_tags -> IsTagAGenbankQualifier( $feature, $qualifier ) )
				{
					warn ( "WARNING $feature-$qualifier $qualifier is NOT a recorded qualifier for the feature $feature\n" );
				}
				
				my @a_values = $f->each_tag_value($qualifier);
				if ( $qualifier eq 'note' )
				{
					my $ra_values = $self -> __ReadNote(\%h_param, \@a_values);
					@a_values = @$ra_values;
				}
				$h_param{ $qualifier } = (@a_values==1) ? $a_values[0] : \@a_values;
			}

			my $o_added_locus;
			try
			{
				if ( $short_class eq "Repeat" )
				{
					$o_added_locus = $self -> __AddRepeat( \%h_param );
				}
				
				elsif ( $short_class eq 'Ncrna' )
				{
					my $sub_type = $feature;
					$h_param{sub_type} = $sub_type;
					$o_added_locus = $self -> __AddNcrna ( \%h_param, \%h_genes_info, \@a_genes_info, $f );
				}
			
				if ( $short_class eq 'Gene' )
				{
					$o_added_locus = $self -> __AddGene ( \%h_param, \@a_genes_info );
				}
				
				elsif ( $short_class eq 'Mrna' )
				{
					$o_added_locus = $self -> __AddMrna ( \%h_param, \%h_genes_info, \@a_genes_info, $f );
				}
				
				elsif ( $short_class eq 'UtrRegion' )
				{
					my $sub_type          = ($feature =~ /^3/)?'three_prime_UTR':'five_prime_UTR';
					$h_param { sub_type } = $sub_type;
					$o_added_locus = $self -> __AddUtr ( \%h_param, \%h_genes_info, \@a_genes_info, $f );
				}
	
				elsif ( $short_class eq "CdsRegion" )
				{
					$o_added_locus = $self -> __AddCds ( \%h_param, \%h_genes_info, \@a_genes_info, $f );
				}
				
				elsif ( $short_class eq "MiscRegion" )
				{
					# SPECIAL TREATMENT FOR THE mobile_element feature
					if ( $feature eq 'mobile_element' )
					{
						if ( defined $h_param { mobile_element_type } )
						{
							my $mobile_element_type = $h_param { mobile_element_type };
							if ( ref($mobile_element_type) eq 'ARRAY' )
							{
								warn ( "WARNING Feature mobile_element, the qualifier mobile_element_type should be specified ONLY ONE TIME\n" );
								$mobile_element_type = $mobile_element_type->[0];
							}
							if ( $mobile_element_type =~ /(.+):(.+)/ )
							{
								$feature .= '_' . $1;
								$h_param { mobile_element_name } = $2;
							}
							else
							{
								$feature .= '_' . $mobile_element_type;
							}
							delete $h_param { mobile_element_type };
						}
						else
						{
							warn ( "WARNING Feature mobile_element, the qualifier mobile_element_type is mandatory\n" );
						}
					}

					# SPECIAL TREATEMENT FOR THE misc_feature feature
					if ( $feature eq 'misc_feature' && defined $h_param{ 'pseudogene' } )
					{
						$h_param { sub_type } = 'pseudogene';
					}
					else
					{
						$h_param { sub_type } = $o_genbank_tags->Feature2So ( $feature );
					}
					$o_added_locus = $self -> __AddMiscRegion( \%h_param, \%h_genes_info, \@a_genes_info, $f );
				}

				elsif ( $short_class eq 'Exon' )
				{
					$o_added_locus = $self -> __AddExon ( \%h_param, \%h_genes_info, \@a_genes_info, $f );
				}
				#if ($f->location->isa('Bio::Location::FuzzyLocationI') ) 
				#{
				#	$self->__AddFragmentInfo($o_added_locus,$f->location);
				#}
			}
			catch AssemblyError with
			{
				# SKIP Silently the locus !
				# warn ( "WARNING - Assembly error - Locus skipped\n" );
			}
			catch Error with
			{
				my ($o_error) = @_;
				if ( $self -> { _o_annot } -> GetDebug() )
				{
					throw $o_error;
				}
				else
				{				
					my $msg = $o_error -> text();
					warn ( "ERROR $msg\n" );
					$o_annot -> SetFatalError();
				}
			};
		};
		$first_pass = 0;
		$self->__SortGenesInfo(\@a_genes_info, \%h_genes_info) if ( $short_class eq 'Gene' );
		#print "COUCOU0\n";
	}
}


# ne MARCHE PAS car bioperl n'analyse pas les annotations de debut de fichier
# => Pas le choix il FAUT passer par un fichier .cfg, ou alors mettre les annotations dans un gff3 (#bfc)
sub __LoadHeaders
{
	my $self = shift;
	my ($seq) = @_;
	
#	my $o_annot_coll = $seq->annotation();
#	foreach my $key ( $o_annot_coll->get_all_annotation_keys() ) 
#	{
#		my @a_values = $o_annot_coll->get_Annotations($key);
#		foreach my $val ( @a_values )
#		{
#			# value is an Bio::AnnotationI, and defines a "as_text" method
#			print "Annotation ",$key," stringified value ",$val->as_text,"\n";
#
#		# also defined hash_tree method, which allows data orientated
#		# access into this object
#		# $hash = $value->hash_tree();
#		}
#	}
}

=head2 function __ConvertFragmentInfo

 Title     : __ConvertFragmentInfo
 Usage     : $self -> __ConvertFragmentInfo $strand, $o_location
 Function  : Prepare the fragment information in a format suitable to feed the Locus constructor
 Args      : $strand -> The strand we are working on (+1,-1)
			 $o_location -> The location with potientially fuzzy information
 Return    : undef, OR a ref to an array, ready to be passed to the Locus constructor:
             [ 'E' ], [ 'S' ], or [ 'E', 'S' ]

=cut

sub __ConvertFragmentInfo
{
	my $self = shift;
	my ( $strand, $o_location) = @_;
	my @a_rvl;

	#print STDERR "KOUKOU " . $o_locus->GetGeneName . " start " . $o_location->start_pos_type . " end " . $o_location->end_pos_type . "\n";
	if ( $o_location->start_pos_type ne 'EXACT' )
	{
		push @a_rvl, 'S' if ($strand==1);
		push @a_rvl, 'E' if ($strand==-1);
	}
	if ( $o_location->end_pos_type ne 'EXACT' )
	{
		push @a_rvl, 'E' if ($strand==1);
		push @a_rvl, 'S' if ($strand==-1);
	}
	return \@a_rvl;
}

=head2
 1/ Sort @a_genes_info: strand, start, end
 2/ $a_genes_info[0] contains the index of the first strand 1 gene, or -1 if no gene found
 3/ Build %h_genes_info: key = the pk, val = the index of the object in @a_genes_info
=cut
sub __SortGenesInfo
{
	my $self = shift;
	my ( $ra_genes_info, $rh_genes_info ) = @_;
	@$ra_genes_info = sort { 
		if ( $a->{o_gene}->GetStrand()!=$b->{o_gene}->GetStrand() )
		{
			return $a->{o_gene}->GetStrand() <=> $b->{o_gene}->GetStrand();
		}
		elsif ( $a->{o_gene}->GetStart()!=$b->{o_gene}->GetStart() )
		{
			return $a->{o_gene}->GetStart() <=> $b->{o_gene}->GetStart();
		}
		else
		{
			return $a->{o_gene}->GetEnd() <=> $b->{o_gene}->GetEnd();
		}
	} @$ra_genes_info;

	unshift @$ra_genes_info, -1;
	
	for (my $i=1; $i<@$ra_genes_info; $i++)
	{
		my $pk = $ra_genes_info->[$i]->{o_gene}->Getpk();
		$ra_genes_info -> [0] = $i if ( $ra_genes_info -> [0] == -1 && $ra_genes_info->[$i]->{o_gene}->GetStrand() == 1 );
		$rh_genes_info->{$pk} = $i;
	}
	#$ra_genes_info -> [0] = @$ra_genes_info+1 if ($ra_genes_info -> [0] == -1);
}

=head2 function __ReadNote

 Title    : __ReadNote
 Usage    : $self -> __ReadNote(\%h_param, \@a_values)
 Function : Read the note qualifier, taking into account the possible tricks added by a previous run of BioFileConverter:
            If some note value is looking like 'some_tag -> some_value', this note is withdrawn from the note tag and 
            a new tag called 'some_tag' is created, the value is set to 'some_value'
 Args     : $rh_param (modified) The hash of parameters being built
            $ra_values           the notes - or the note being processed
 Return   : $ra_notes, the notes

=cut
sub __ReadNote
{
	my $self = shift;
	my ( $rh_param, $ra_values ) = @_;
	
	my $ra_val = ( ref ($ra_values) eq '' ) ? [ $ra_values ] : $ra_values;
	my @a_note = ();
	
	foreach my $val ( @$ra_val )
	{
		if ( $val =~ /^([^ ]+)=(.+)/ )
		{
			$rh_param -> {$1} = $2;
		}
		else
		{
			push @a_note, $val;
		}
	}
	return \@a_note;
}

=head2 function __AddRepeat

 Title        : __AddRepeat
 Usage        : $self -> __AddRepeat( $rh_param, $rh_genes_info );
 Function     : Load a repeat
 Args         : $rh_param, $rh_genes_inf
 Returns      : The added locus or undef

=cut
sub __AddRepeat
{
	my $self = shift;
	my ($rh_param) = @_;
	my $o_annot = $self -> { _o_annot };

	my $o_repeat = New Locus::Repeat(%$rh_param);
	$o_annot->AddRepeat($o_repeat);
	return $o_repeat;
}

=head2 function __AddMiscRegion

 Title        : __AddMiscRegion
 Usage        : $self -> __AddMiscRegion( $rh_param, $rh_genes_info );
 Function     : Load a repeat
 Args         : $rh_param
                $f (the feature, used for the locations)
 Returns      : The added locus or undef

=cut
sub __AddMiscRegion
{
	my $self = shift;
	my ($rh_param,$rh_genes_info,$ra_genes_info,$f) = @_;
	my $o_annot = $self -> { _o_annot };

	my $o_region = New Locus::MiscRegion(%$rh_param);

	$o_annot->AddMiscRegion($o_region);

	foreach my $loc ( $f->location->each_Location )
	{
		my %h_params = (
			start  => $loc->start,
			end    => $loc->end, 
			strand => $$rh_param{strand},
			-fragment_info  => $self->__ConvertFragmentInfo($$rh_param{strand},$loc)
			);
		my $o_r   = New Locus::MiscRegion (%h_params);
		$o_region->AddRegion($o_r);

	}
	return $o_region;
}

=head2 function __AddNcrna

 Title        : __AddNcrna
 Usage        : $self -> __AddNcrna( $rh_param, $rh_genes_info, $ra_genes_info, $f );
 Function     : Load an ncrna and try to find the gene it belongs to
                If the gene is not found, simply give the ncrna to the ancestor
 Args         : $rh_param, $rh_genes_info,$ra_genes_info
                $f (the feature, used for the ExonSet locations)
 Returns      : The added locus or undef

=cut

sub __AddNcrna
{
	my $self = shift;
	my ($rh_param,$rh_genes_info,$ra_genes_info,$f) = @_;

	my $o_ncrna = New Locus::Ncrna(%$rh_param);
	my $index  = $self -> __FindNearestLocus($o_ncrna, $rh_genes_info, $ra_genes_info);
	if ( defined $index )
	{
		my $o_gene = $ra_genes_info -> [ $index ]-> { o_gene };
		$o_gene -> AddNcrna ( $o_ncrna );

		# Read the exons, from the region decomposition of the NcRna
		foreach my $loc ( $f->location->each_Location )
		{
			my $o_region = New Locus::Exon (
				start  => $loc->start,
				end    => $loc->end, 
				strand => $$rh_param{strand}
				);
			$o_ncrna->AddExon($o_region);
		}
	}
	else
	{
		my $o_annot = $self -> { _o_annot };
		$o_annot->AddNcrna($o_ncrna);
	}
	return $o_ncrna;
}

=head2 function __AddGene

 Title        : __AddGene
 Usage        : $self -> __AddGene( $rh_param, $ra_genes_info );
 Function     : Load a gene, keeping a ref to it inside $rh_genes_info (the key is the gene's $pk)
 Args         : $rh_param, $ra_genes_info
 Returns      : The added locus or undef


=cut
sub __AddGene
{
	my $self = shift;
	my ($rh_param,$ra_genes_info) = @_;
	my $o_annot = $self -> { _o_annot };

	my $o_gene = New Locus::Gene(%$rh_param);
	$o_annot -> AddGene($o_gene);
	my %h_gene_info = ( o_gene => $o_gene, utr3_cnt => 0, utr5_cnt => 0, cds_cnt => 0, mrna_cnt => 1 );
	push @$ra_genes_info, \%h_gene_info;
	return $o_gene;
}

=head2 function __AddMrna

 Title        : __AddMrna
 Usage        : $self -> __AddMrna( $rh_param, $rh_genes_info, $ra_genes_info, $f );
 Function     : Load an mrna and try to find the gene it belongs to
                Also load the exons constituting the mrna
                increment the mrna_cnt field of $rh_genes_info
 Args         : $rh_param, $rh_genes_info,$ra_genes_info
                $f (the feature, used for the ExonSet locations)
 Returns      : The added locus or undef

=cut
sub __AddMrna
{
	my $self = shift;
	my ($rh_param,$rh_genes_info,$ra_genes_info,$f) = @_;
	#my $o_annot = $self -> { _o_annot };

	my $o_mrna = New Locus::Mrna(%$rh_param);
	my $index  = $self -> __FindNearestLocus($o_mrna, $rh_genes_info, $ra_genes_info);
	if ( defined $index )
	{
		my $o_gene = $ra_genes_info -> [ $index ]-> { o_gene };
		$o_gene -> AddmRNA ( $o_mrna );
		$ra_genes_info -> [ $index ]-> { $index } -> { mrna_cnt } += 1;

		# Read the exons, from the region decomposition of the mrna
		foreach my $loc ( $f->location->each_Location )
		{
			my $o_region = New Locus::Exon (
				start  => $loc->start,
				end    => $loc->end, 
				strand => $$rh_param{strand},
				-fragment_info  => $self->__ConvertFragmentInfo($$rh_param{strand},$loc)
			);
			$o_mrna->AddExon($o_region);
		}
		return $o_mrna;
	}
	else
	{
		my $pke = $o_mrna -> GetpkExtended();
		warn ($self -> GetWarningHeader() . "No gene (or ambiguous) found for this mRNA. Skipped $pke\n");
		return undef;
	}
}

=head2 procedure __FindNearestLocus

 Title        : __FindNearestLocus
 Prerequisite : The %h_genes_info and @a_genes_info must be initialized by __SortGenesInfo
 Usage        : my $o_gene = $self -> __FindNearestLocus ( $o_mrna, $rh_genes_info, $ra_genes_info);
 Function     : Try to find a gene locus with the SAME $pk as $o_locus
                If not found, scan $ra_genes_info, looking for the locus with highest overlap
                If not found, return undef
                If found several loci, compare with the gene name if exists, OR with the locus_tag, and return THE correct locus, or undef
 Args         : $o_locus:    The locus to work with
				$rh_genes_info,$ra_genes_info Hash and data used 
 Return       : The index in $ra_genes_info of the nearest (ie the most overlapping) locus, or undef if none or ambiguous loci found
 Access       : Private

=cut
sub __FindNearestLocus
{
	my $self = shift;
	my ($o_locus, $rh_genes_info, $ra_genes_info) = @_;
	
	# If the ONLY index found in $ra_genes_info is 0, we return undef (no genes at all)
	if ( $#$ra_genes_info == 0 )
	{
		return undef;
	}

	# The array of indexes of the nearest loci (may be several)
	my $ra_nearest = [];
	my $nearest = undef;

	# Try to find an exact match
	my $pk = $o_locus -> Getpk();
	if ( exists $rh_genes_info -> { $pk } )
	{
		# $rh_genes_info -> { $pk } returns a ra, it is a MultiValHash
		$ra_nearest = $rh_genes_info -> { $pk };
	}

	# Scan the $ra_genes_info array, looking for the most overlapping loci
	else
	{
		# Only 1 or very few genes should overlap with this locus
		# [$i_strt, $i_end] is the slice of overlapping genes in $ra_genes_info
		my ($i_strt,$i_end) = $self -> __FindOverlappingLoci($ra_genes_info, $o_locus);
		
		# Debug information (level 3)
		if ( $self -> { _o_annot } -> GetDebug() )
		{
			if ($i_strt==0)
			{
				print STDERR "===== " . $o_locus->GetpkExtended() . ": Overlapping loci NOT found\n";
			}
			else
			{
				print STDERR "===== " . $o_locus->GetpkExtended() . ": Overlapping loci found\n";
				for ( my $i=$i_strt;$i<=$i_end; $i++ )
				{
					my $ovl = $o_locus -> GetOverlapNb($ra_genes_info -> [$i] -> { o_gene });
					printf STDERR "%5i  %s\n",$ovl,$ra_genes_info -> [$i] -> { o_gene } -> GetpkExtended();
				}
			}
			print STDERR "=====\n";
		}

		if ( $i_strt != 0 )
		{
			# Only 1 overlapping gene found
			if ( $i_strt==$i_end )
			{
				$nearest = $i_strt;
			}
			
			# Several overlapping genes found
			else
			{
				my $overlap_max = 0;
				for (my $i=$i_strt; $i<=$i_end; $i++)
				{
					my $o_near = $ra_genes_info -> [$i] -> { o_gene };
					my $ovlp = $o_locus -> GetOverlapNb ( $o_near );
					#print $o_locus->GetpkExtended().','.$o_near->GetpkExtended().' -> '."$ovlp\n";
					if ( $ovlp > $overlap_max )
					{
						# Better overlap found, delete previous answers
						$ra_nearest = [];
						push @$ra_nearest, $i;
						$overlap_max = $ovlp;
					}
					elsif ( $ovlp > 0 && $ovlp == $overlap_max )
					{
						push @$ra_nearest, $i;
					}
				}
			}
		}
	}
	
	# Only 1 locus found: return this locus
	if ( @$ra_nearest == 1 )
	{
		$nearest = $ra_nearest -> [ 0 ];
	}
	
	# Several loci found: return the first (hoping, but not checking that it is the only one) having the same gene name or locus_tag
	elsif ( @$ra_nearest > 1 )
	{
		my $gene_name = $o_locus -> GetGeneNameOrLocusTag(); # The name OR the locus tag OR the getId
		foreach my $i ( @$ra_nearest )
		{
			my $near_name = $ra_genes_info->[$i]->{ o_gene }->GetGeneNameOrLocusTag();
			if ( $near_name eq $gene_name )
			{
				$nearest=$i;
				last;
			}
		}
		
		# several overlapping loci, but none with the good name: return undef
		#return undef;
	}
	
	# Debugging stuff
	if ( $self -> { _o_annot } -> GetDebug() )
	{
		if ( defined $nearest )
		{
			print STDERR $o_locus->GetpkExtended() . " -> " . $ra_genes_info->[$nearest]->{o_gene}->GetpkExtended() . "\n";	
		}
		else
		{
			print STDERR $o_locus->GetpkExtended() . " -> none\n";
		}
	}

	return $nearest;
}

=head2 procedure __FindOverlappingLoci

 Title        : __FindOverlappingLoci
 Prerequisite : @a_genes_info must be initialized by __SortGenesInfo
 Usage        : my ( $i,$j) = $self -> __FindOverlappingLoci(\@_genes_info, $o_locus);
 Function     : Compute the slice in @a_genes_info inside which we can find all loci overlapping with $o_locus
                We proceed in two stages:
                   1/ Search using dichotomy the first locus ($o_l) with $o_l->GetStart() > $o_locus -> GetEnd()
                      This is possible because $ra_genes_info is sorted in increasing Start positions
                   2/ Scan descending the array and Keep only loci overlapping with $o_locus
                      This should be very fast because there should be only very few loci in the slice
 Args         : $ra_loci_info : The array of genes
                $o_locus:    The locus to work with
 Return       : ($i,$j), the two indices defining the slice, (0,0) if none found
 Access       : Private

=cut
sub __FindOverlappingLoci
{
	my $self = shift;
	my ($ra_loci_info, $o_locus) = @_;

	my ($strt, $end) = $o_locus -> GetEnds();
	
	# dichotomy search for $j (higher index of the slice)
	# The search 
	my $jmin = ($o_locus -> GetStrand() == 1) ? $ra_loci_info -> [0] : 1;
	my $jmax = ($o_locus -> GetStrand() == 1) ? $#$ra_loci_info : $ra_loci_info -> [0] - 1;
	$jmax = $#$ra_loci_info if ( $ra_loci_info -> [0] == -1 ); # may happen if NO gene on the positive strand !

	my $j = 0;
	while ( $jmax-$jmin>1 )
	{
		#print STDERR "$jmin $j $jmax\n";
		$j = $jmin + int (($jmax - $jmin) / 2);
		if ($ra_loci_info->[$j]->{o_gene}->GetStart() > $end)
		{
			$jmax = $j;
		}
		else
		{
			$jmin = $j;
		}
	}
	
	#print STDERR $ra_loci_info->[$jmin]->{o_gene}->GetpkExtended().'  '.$o_locus->GetpkExtended().' '.$ra_loci_info->[$jmax]->{o_gene}->GetpkExtended()."\n";
	
	# scanning from $j_max
	my ( $ovr_min,$ovr_max) = (0,0);
	my $i = $jmax;
	my $imin = ($o_locus -> GetStrand() == 1) ? $ra_loci_info -> [0] : 1;
	#print STDERR "COUCOU1 $i $imin $jmax\n";
	while ( $i >= $imin )
	{
		#print STDERR "COUCOU2 " . $ra_loci_info->[$i]->{o_gene}->GetpkExtended()."\n";
		if ( $o_locus->GetOverlapNb ( $ra_loci_info->[$i]->{o_gene} ) > 0 )
		{
			$ovr_max = $i if ( $ovr_max==0 );
			$ovr_min = $i if ( $ovr_min==0 || $i<$ovr_min );
		}
		$i--;
	}
	
	# return the slice
	return ($ovr_min,$ovr_max);
}

=head2 function __AddUtr

 Title        : __AddUtr
 Usage        : $self -> __AddUtr( $rh_param, $rh_genes_info, $ra_genes_info,$f );
 Args         : $rh_param, $rh_genes_info, $ra_genes_info
                $f (the feature, used for the CDS locations)
 Returns      : The added locus or undef

=cut
sub __AddUtr
{
	my $self = shift;
	my ($rh_param,$rh_genes_info,$ra_genes_info,$f) = @_;
	my $utr_sub_type = $rh_param -> { sub_type }; 
	my $utr_cnt_lbl  = ($utr_sub_type eq 'three_prime_UTR') ? 'utr3_cnt' : 'utr5_cnt';
		
	my $o_utr = New Locus::Utr( %$rh_param);
	my @a_locations = $f->location->each_Location;
	if ( @a_locations >= 1 )
	{
		foreach my $loc ( @a_locations )
		{
			my $o_region = New Locus::UtrRegion ( %$rh_param, start => $loc->start, end => $loc->end );
			$o_utr -> AddRegion ( $o_region );
		}
	}
		
	# Look for a Gene, then an Mrna to include the Utr
	my $index = $self -> __FindNearestLocus($o_utr, $rh_genes_info, $ra_genes_info);
	if ( defined $index )
	{
		my $o_gene  = $ra_genes_info -> [ $index ] -> { o_gene };
		my $utr_cnt = $ra_genes_info -> [ $index ] -> { $utr_cnt_lbl };
		my $o_mrna  = $ra_genes_info -> [ $index ] -> { o_gene } -> GetMrna( $utr_cnt );
		
		# nth mRNA not found: create it !
		if ( !defined $o_mrna )
		{
			$o_mrna = New Locus::Mrna( %$rh_param );
			$o_gene -> AddmRNA( $o_mrna );
			$ra_genes_info -> [ $index ] -> { mrna_cnt } += 1;
		}
		
		# give the UTR to the mRNA found or created
		$o_mrna -> AddUtr ( $o_utr );
		$ra_genes_info -> [ $index ] -> { $utr_cnt_lbl } += 1;
	}
	
	# gene not found: skip the Utr
	else
	{
		my $pke = $o_utr -> GetpkExtended();
		warn ($self -> GetWarningHeader() . "UTRRegion $pke: Not found any gene for him !\n");
	}
	return $o_utr;
}

=head2 function __AddCds

 Title        : __AddCds
 Usage        : $self -> __AddCds( $rh_param, $rh_genes_info, $f );
 Function     : Load a CDS, find the gene the CDS belongs to, using __FindNearestLocus
                Then, if mrna_cnt > 0, look for the corresponding mRNA, considering that the nth CDS feature is
                related to the nth mRNA feature
                If mrna_cnt = 0, create an mRNA
 Args         : $rh_param, $rh_genes_info, $ra_genes_info
                $f (the feature, used for the CDS locations)
 Returns      : The added locus or undef

=cut
sub __AddCds
{
	my $self = shift;
	my ($rh_param,$rh_genes_info,$ra_genes_info,$f) = @_;
	#my $o_annot = $self -> { _o_annot };

	# Create the Cds with its regions
	my $o_cds = New Locus::Cds( %$rh_param, -fragment_info=>$self->__ConvertFragmentInfo($$rh_param{strand},$f->location));
	my @a_locations = $f->location->each_Location;
	if ( @a_locations >= 1 )
	{
		foreach my $loc ( @a_locations )
		{
			my $o_region = New Locus::CdsRegion ( %$rh_param, start=>$loc->start, end=>$loc->end, -fragment_info=>$self->__ConvertFragmentInfo($$rh_param{strand},$loc));
			$o_cds -> AddRegion ( $o_region );
		}
	}

	# Look for a Gene, then an Mrna to include the Cds
	my $index = $self -> __FindNearestLocus($o_cds, $rh_genes_info, $ra_genes_info);
	if ( defined $index )
	{
		my $o_gene  = $ra_genes_info -> [ $index ]-> { o_gene };
		my $cds_cnt = $ra_genes_info -> [ $index ] -> { cds_cnt };
		my $o_mrna  = $ra_genes_info -> [ $index ] -> { o_gene } -> GetMrna( $cds_cnt );
		
		# nth mRNA not found: create it !
		if ( !defined $o_mrna )
		{
			my %h_param = ();
			my $locus_tag = $o_gene -> GetLocusTag();
			$h_param{"locus_tag"} = $locus_tag if ($locus_tag ne '');
			$o_mrna = New Locus::Mrna( %h_param, $o_gene);
			$o_gene -> AddmRNA( $o_mrna );
			$ra_genes_info -> [ $index ] -> { mrna_cnt } += 1;
		}
		
		# give the CDS to the mRNA found or created
		$o_mrna -> AddCDS ( $o_cds );
		$ra_genes_info -> [ $index ] -> { cds_cnt } += 1;
		return $o_cds;
	}
	
	# gene not found: skip the CDS
	else
	{
		my $pke = $o_cds -> GetpkExtended();
		warn ($self -> GetWarningHeader() . "Skipped CDS $pke: Not found any gene for him !\n");
		return undef;
	}
}

# EXPERIMENTAL
=head2 function __AddExon

 Title        : __AddExon
 Usage        : $self -> __AddExon( $rh_param, $rh_genes_info, $f );
 Function     : Load an exon, only if no mRNA feature was described for this gene (using the mrna_cnt for detecting this)
                AND if there is no more mRNA in this gene (thus the exon feature is ignored when there is some alternative splicing)
 Args         : $rh_param, $rh_genes_info
                $f (the feature, used for the ExonSet locations)
 Returns      : The added locus or undef

=cut
sub __AddExon
{
	my $self = shift;
	my ($rh_param,$rh_genes_info,$ra_genes_info,$f) = @_;
	my $o_annot = $self -> { _o_annot };

	# Create the Exon
	my $o_exon = New Locus::Exon( %$rh_param );

	# Look for a Gene, and add the Exon to the Gene
	my $index = $self -> __FindNearestLocus($o_exon, $rh_genes_info, $ra_genes_info);
	if ( defined $index )
	{
		my $o_gene  = $ra_genes_info -> [ $index ]-> { o_gene };
		$o_gene -> AddExon ( $o_exon );
	}
	
	# gene not found: skip the Exon
	else
	{
		my $pke = $o_exon -> GetpkExtended();
		warn ($self -> GetWarningHeader() . "Skipped Exon $pke: Not found any gene for him !\n");
	}
}

=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;
