#
# $Id: Fasta.pm 1309 2013-11-18 09:03:08Z manu $
#

package AnnotatedSeqFileWriter::Eugene;

# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


=head1 NAME

AnnotatedSeqFileWriter::Eugne - Using annotations stored in a companion AnnotatedSeq object, using the genomic fasta file
                                dump files useful for Eugene training

=head1 SYNOPSIS

 my $o_dumper = New AnnotatedSeqFileWriter::Eugene( o_annot => $o_annot );
 $o_dumper -> Dump();

=head1 DESCRIPTION

 this class extends the AnnotatedSeqFileWriter::FastaLike, it dumps data from AnnotatedSeqFile to Narcisse files

=cut

use strict;
use warnings;
use Carp;

use Error qw(:try);
use LipmError::IOException;
use Bio::Seq;
use Data::Dumper;
use General;

use Locus;
use base qw( AnnotatedSeqFileWriter::FastaLike );

use constant ENVIR => 120;
use constant ENVIR2 => &AnnotatedSeqFileWriter::Eugene::ENVIR * 2 + 2;

BEGIN
{
    our $VERSION = do {my @r = (q$Rev: 1309 $ =~ /\d+/g); $r[0]};
}

sub _Init
{
	my $self = shift;

	$self->SUPER::_Init(@_);
}

sub GetFormat
{
	return 'EGN';
}

=head2 procedure Dump

 Title     : Dump
 Usage     : $o_annot_writer -> Dump ()
 Procedure : Dump the genes objects to 4 fasta-formatted files
 Args      : -append 1/0 (optional, def 0)
 Error     : Throw an Error::Simple is there is no sequence loaded

=cut

sub Dump
{
	my $self = shift;
	my (%h_params) = (-append => 0, -output_utr => 0, -output_intron => 0,  @_);
	$self->SUPER::Dump(@_);
	
	my $o_annot = $self -> { _o_annot };
	my $ra_mrnas = $self -> __SelectMrnas();
	my $ra_est_match = $self -> __SelectEST_match();
	
	my ($filename,$ext,$comp_ext) = $self->_ParseFileName();
	$self->{__filename} = $filename;
	$self->{__ext}      = $ext;
	$self->{__comp_ext} = $comp_ext;
	
	# key = $site.start, $site.donor, $site.acceptor ('ATG.start', 'AG.acceptor', GC.donor')
	# value = An opened file handler

	# GAGAGATGAGAGAGAGAGAGAGAGAGA
	# .....^^^^^^^^^^^^^^^^..... The cds
	# .....^^            ^^..... Donor and acceptor sites
	# .....^              ^..... Start and end positions
	#

	my %h_fh_out;
	$self->{__rh_fh_out} = \%h_fh_out;
	
	$self->__DumpEugeneTP($ra_mrnas,$ra_est_match);
	
	# close all the files opened by __DumpEugeneTP, but keep the list of keys
	foreach my $key ( keys %h_fh_out)
	{
		#$h_fh_out{$key} -> close();
		$h_fh_out{$key} = undef;
	}
	
	$self->__DumpEugeneTN($ra_mrnas,$ra_est_match);
}

=head2 function __SelectMrnas

 Return a $ra containing the selected Mrnas, ie:
   - those belonging to genes which DO NOT contain gaps ('N')
   - those whose cds is consistent (translation OK)

=cut
sub __SelectMrnas
{
	my $self = shift;
	my $o_annot = $self -> { _o_annot };
	my $ra_all_genes = $o_annot -> GetGenes();
	my @a_mrnas;
	foreach my $o_gene (@$ra_all_genes)
	{
		my $seq = uc $o_gene -> GetSequence();
		if ( index ( $seq, 'N' ) >= 0 )
		{
			print STDERR $self -> GetWarningHeader('') . $o_gene->GetpkExtended() . "contains some 'N' - Skipped\n";
		}
		else
		{
			my $ra_mrnas = $o_gene -> GetMrnas();
			foreach my $o_mrna (@$ra_mrnas)
			{
				my $o_cds = $o_mrna -> GetCDS();
				if ( $o_cds-> GetConsistency() ne 'C' )
				{
					my $consistency = $o_cds-> GetConsistency();
					print STDERR $self -> GetWarningHeader('') . $o_cds->GetpkExtended() . "is NOT consistent ($consistency) - mrna skipped\n";
				}
				else
				{
					push @a_mrnas,$o_mrna;
				}
			}
		}
	}
	return\@a_mrnas;
}

=head2 function __SelectEST_match

 Return a $ra containing the selected EstMatches (MisRegions with sub_type EST_match)
   - those which DO NOT contain gaps ('N')

=cut
sub __SelectEST_match
{
	my $self = shift;
	my $o_annot = $self -> { _o_annot };
	my $ra_regions = $o_annot -> GetMiscRegions();
	my @a_regions;
	foreach my $o_region (@$ra_regions)
	{
		next if ( $o_region->GetSubType() ne 'EST_match' );
		my $seq = uc $o_region -> GetSequence();
		if ( index ( $seq, 'N' ) >= 0 )
		{
			print STDERR $self -> GetWarningHeader('') . $o_region->GetpkExtended() . "contains some 'N' - Skipped\n";
		}
		else
		{
			push @a_regions,$o_region;
		}
	}
	return\@a_regions;
}

=head2 function __DumpEugeneTP

 Title     : __DumpEugeneTP
 Usage     : $self -> __DumpEugeneTP(\@a_mrnas,\@a_est_match)
 Procedure :  For each selected mrna, or est_match, dump the True Positifs, ie start/donor/acceptor sites with their environment

 Args      : $ra_mrnas = ref to the array of selected mrnas
             $ra_est_match = ref to the array of select MiscRegions/EST_match
 
=cut
sub __DumpEugeneTP
{
	my $self = shift;
	my ($ra_mrnas,$ra_est_match) = @_;
	
	my $o_annot = $self -> { _o_annot };
	
	foreach my $o_mrna (@$ra_mrnas)
	{
		my $o_cds = $o_mrna -> GetCDS();
		$self->__DumpEugeneTPStart($o_cds);

		my @a_introns = $o_mrna->GetIntrons();
		$self->__DumpEugeneTPIntrons(\@a_introns);
	}
	
	foreach my $o_region (@$ra_est_match)
	{
		my @a_introns = $o_region->GetIntrons();
		$self->__DumpEugeneTPIntrons(\@a_introns);
	}
}

=head2 function __DumpEugeneTN

 Title     : __DumpEugeneTN
 Usage     : $self -> __DumpEugeneTN(\@a_mrnas,\@a_est_match)
 Procedure :  For each selected mrna, or est_match, dump the True Negatives

 Args      : $ra_mrnas = ref to the array of selected mrnas
             $ra_est_match = ref to the array of select MiscRegions/EST_match

=cut

sub __DumpEugeneTN
{
	my $self = shift;
	my ($ra_mrnas,$ra_est_match) = @_;
	
	my $o_annot = $self -> { _o_annot };
	my $contig  = $o_annot -> GetContigLabel();
	my $filename = $self->{__filename};
	my $ext       = $self->{__ext};
	my $comp_ext = $self->{__comp_ext};
	my $rh_fh_out = $self->{__rh_fh_out};
	
	my @a_loci;
	push @a_loci,@$ra_mrnas;
	push @a_loci,@$ra_est_match;
	foreach my $o_locus (@a_loci)
	{
		my $seq = uc $o_locus->GetSequenceWithEnvironment(&AnnotatedSeqFileWriter::Eugene::ENVIR);
		my $seq = Locus::UtilRevComp($seq);
		my ($locus_start,$locus_end) = $o_locus -> GetEnds();
		my $seq_len = length $seq;

		my $strand      = - $o_locus -> GetStrand();
		
		
		# ...GAGAGAGAGA...   <- length 16
		#             ^
		#             pos_max (12)
		# $pos_max is the pos_max of the pattern relative to $seq, 0-based
		my $pos_max = $seq_len - &AnnotatedSeqFileWriter::Eugene::ENVIR - 1;
		foreach my $key (keys %$rh_fh_out)
		{
			my ($site,$type_of_site) = ($key =~ /([A-Z]+)\.([a-z]+)/);
			
			my $site_len = length $site;
			
			# $pos relative to $seq, 0-based
			my $pos = &AnnotatedSeqFileWriter::Eugene::ENVIR;
			while(&TRUE)
			{
				$pos = index ( $seq,$site,$pos );
				last if ($pos == -1 || $pos > $pos_max);

				# $pos+1 is 1-based 
				my $sub_pre = &Locus::UtilGetSeqPre(\$seq,$pos+1,&AnnotatedSeqFileWriter::Eugene::ENVIR);
				my $sub_post= &Locus::UtilGetSeqPost(\$seq,$pos+$site_len,&AnnotatedSeqFileWriter::Eugene::ENVIR);
				my $sub_seq = "";
				my $abs_pos = 0;
				if ( $strand > 0 )
				{
					$sub_seq = $sub_pre . $site . $sub_post;

					# WARNING ! $locus_start is 1-based, $pos is 0-based !
					$abs_pos = $locus_start + $pos - 2 * &AnnotatedSeqFileWriter::Eugene::ENVIR;
				}
				else
				{
					$sub_seq = $sub_pre . $site . $sub_post;
					$abs_pos = $locus_start + $seq_len - $pos - $site_len - 2 * &AnnotatedSeqFileWriter::Eugene::ENVIR;
				}
				$pos++;
				if ( $sub_seq =~ /[ N]/ )
				{
					next;
				}
				else
				{
					my $header = $contig.':'.$abs_pos.':'.($strand>0?'+':'-');
					$self->__PrepareAndPrint($rh_fh_out,{},$filename,'TN',$key,$ext,$comp_ext,$header,$sub_seq);
				}
			}
		}
	}
};	
	
	
	
=head2 function __DumpEugeneTPStart

 Title     : __DumpEugeneTPStart
 Usage     : $self -> __DumpEugeneTPStart($o_cds)
 Procedure :  Called by __DumpEugeneTP, dump the TP for the start codon

 Args      : $o_cds = The cds

=cut
sub __DumpEugeneTPStart
{
	my $self = shift;
	my ($o_cds) = @_;
	my $o_annot = $self -> { _o_annot };
	my $contig  = $o_annot -> GetContigLabel();
	
	# key = The header
	# value = 1 (to test the redundancy of the sequence)
	my %h_start_headers;
	
	my $start_seq_len= 2 * &AnnotatedSeqFileWriter::Eugene::ENVIR + 3;
	
	my $strand = $o_cds->GetStrand();
	my $o_cds_region = ($strand>0?$o_cds->GetRegion0():$o_cds->GetRegionLast());
	my $seq_env       = uc $o_cds_region -> GetSequenceWithEnvironment(&AnnotatedSeqFileWriter::Eugene::ENVIR);
	my $start_seq     = substr $seq_env,0,$start_seq_len;
	
	if ( index ($start_seq,' ') >= 0 )
	{
		print STDERR $self -> GetWarningHeader('') . "The cds " . $o_cds->Getpk() . " is located near the sequence extremities - skipped - start codon skipped\n";
		return;
	}
	if ( index ( $start_seq, 'N' ) >= 0 )
	{
		print STDERR $self -> GetWarningHeader('') . "The cds " . $o_cds->Getpk() . "has some 'N' in its environment - Skipped\n";
		return;
	}

	my $pos = 0;
	if ( $strand >0 )
	{
		$pos = $o_cds_region->GetStart()-&AnnotatedSeqFileWriter::Eugene::ENVIR;
	}
	else
	{
		$pos = $o_cds_region->GetEnd() - 2 - &AnnotatedSeqFileWriter::Eugene::ENVIR;
	}

	my $start_header    = $contig.':'.$pos.':'.$o_cds->GetStrand(&TRUE);
	my $filename = $self->{__filename};
	my $ext       = $self->{__ext};
	my $comp_ext = $self->{__comp_ext};
	my $rh_fh_out = $self->{__rh_fh_out};

	my $start_filename    = $filename . '.TP-###.start' . $ext . $comp_ext;
	$self->__PrepareAndPrint($rh_fh_out,\%h_start_headers,$filename,'TP','###.start',$ext,$comp_ext,$start_header,$start_seq);
}

=head2 function __DumpEugeneTPIntrons

 Title     : __DumpEugeneTPIntrons
 Usage     : $self -> __DumpEugeneTPIntrons(\@a_introns)
 Procedure :  Called by __DumpEugeneTP, dump the TP for the donor/acceptor sites

 Args      : $ra_introns = A ref to an array of introns

=cut
sub __DumpEugeneTPIntrons
{
	my $self = shift;
	my ($ra_introns) = @_;
	my $o_annot = $self -> { _o_annot };
	my $contig  = $o_annot -> GetContigLabel();

	my $filename = $self->{__filename};
	my $ext       = $self->{__ext};
	my $comp_ext = $self->{__comp_ext};

	# key = The header
	# value = 1 (to test the redundancy of the sequence)
	my %h_donor_headers;
	my %h_acceptor_headers;
	
	# key = $site.donor ('GT.donor') or $site.acceptor ('GT.acceptor')
	# value = An opened file handler
	# GAGAGAGAGAGAGAGAGAGAGAGAGA
	# .....^^^^^^^^^^^^^^^^..... The intron
	# .....^^            ^^..... Donor and acceptor sites
	# .....^              ^..... Start and end positions
	#
	my $rh_fh_out = $self->{__rh_fh_out};

	# The site is 2 nt length, surrounded by ENVIR
	my $donor_seq_len  = 2 * &AnnotatedSeqFileWriter::Eugene::ENVIR + 2;
	
	# The site is 2 nt length, surrounded by ENVIR, from the end
	my $acceptor_start = - ( 2 * &AnnotatedSeqFileWriter::Eugene::ENVIR + 2 );
	
	foreach my $o_intron (@$ra_introns)
	{
		my $seq_env       = uc $o_intron -> GetSequenceWithEnvironment(&AnnotatedSeqFileWriter::Eugene::ENVIR);
		my $donor_seq     = substr $seq_env,0,$donor_seq_len;
		my $acceptor_seq  = substr $seq_env,$acceptor_start;

		if ( index ($donor_seq,' ') >= 0 || index ($acceptor_seq,' ') >= 0 )
		{
			print STDERR $self -> GetWarningHeader('') . "The intron " . $o_intron->Getpk() . " near the sequence extremities - skipped - intron skipped\n";
			next;
		}
		if ( index ( $donor_seq, 'N' ) >= 0 ||  index ( $acceptor_seq, 'N' ) >= 0 )
		{
			print STDERR $self -> GetWarningHeader('') . "The intron " . $o_intron->Getpk() . "has some 'N' in its environment - Skipped\n";
			next;
		}

		my $strand = $o_intron->GetStrand(&TRUE);
		
		my ($pos1,$pos2);
		if ( $strand eq '+')
		{
			$pos1 = $o_intron->GetStart()-&AnnotatedSeqFileWriter::Eugene::ENVIR;
			$pos2 = $o_intron->GetEnd() - &AnnotatedSeqFileWriter::Eugene::ENVIR - 1;
		}
		else
		{
			$pos1 = $o_intron->GetEnd()   - 1 - &AnnotatedSeqFileWriter::Eugene::ENVIR;
			$pos2 = $o_intron->GetStart() - &AnnotatedSeqFileWriter::Eugene::ENVIR;
		}
		
		my $donor_header    = $contig.':'.$pos1.':'.$strand;
		my $acceptor_header = $contig.':'.$pos2.':'.$strand;
		
		$self->__PrepareAndPrint($rh_fh_out,\%h_donor_headers,$filename,'TP','##.donor',$ext,$comp_ext,$donor_header,$donor_seq);
		$self->__PrepareAndPrint($rh_fh_out,\%h_acceptor_headers,$filename,'TP','##.acceptor',$ext,$comp_ext,$acceptor_header,$acceptor_seq);
	}
}

=head2 Procedure __PrepareAndPrint

 Title        : __PrepareAndPrint
 Procedure    : Prepare the record and select the filehandler then call _PrintFasta
 Usage        : $self -> __PreparePrint(\%h_fh_out,\%h_acceptor_headers,$filename,$type,$tmpl,$ext,$acceptor_header,$acceptor_seq)
 Args         : $rh_fh_out  : A hash containing keys -> file handles (values)
                $rh_headers : A hash containing the headers already used, to avoid writing several times the same sequence
                $filename   : The file name (prefix only)
                $type       : TP or TN
                $tmpl       : The template: ##.acceptor, ##.donor or ###.start
                $ext        : The extension (egn)
                $comp_ext   : The compression extension
                $header     : The header
                $seq        : The sequence, will be written and is used to extract the site
 Globals      : none

=cut
sub __PrepareAndPrint
{
	my $self = shift;
	my ($rh_fh_out,$rh_headers,$filename,$type,$tmpl,$ext,$comp_ext,$header,$seq) = @_;
	
	$seq = uc $seq;
	
	# Nothing to do if already done
	return if ( exists $rh_headers->{$header});
	$rh_headers->{$header} = 1;
	
	# Compute the site and the file name
	my $key  = $tmpl;       # compute the key to use in $rh_fh_out
	my ($tmpl_site) = ($key=~/(#+)/);	# used to detect the length of the site (2 or 3)
	
	# If $tmpl_site is null, there is nothing to replace
	if ( $tmpl_site ne '' )
	{
		my $site = substr $seq, &AnnotatedSeqFileWriter::Eugene::ENVIR, length($tmpl_site);
		$key =~ s/#+/$site/;    # ###.start ==> ATG.start
	}

	my $name = $filename . '.' . $type . '.' . $key . $ext . $comp_ext;
	
	# Open the file if necessary
	if ( !defined $rh_fh_out->{$key} )
	{
		$rh_fh_out->{$key} = GetStreamOut($name);
	}
	my $fh_out = $rh_fh_out->{$key};
	
	$self->_PrintFasta($fh_out,'>'.$header,\$seq);
}

=head2 Procedure _PrintFasta

 Title        : _PrintFasta
 Procedure    : Print a sequence fasta-formatted
 Usage        : $self -> __PrinFasta ( $fh_out, $header, \$seq );
 Prerequisite : $fh_out must be an output-opened file handle
 Args         : $fh_out : A file handle
                $header : The fasta header
                $r_seq  : A ref-to-scalar, the sequence to print
 Globals      : none
 NOTE         : Remove this function to get a better fasta file format, the FastaLike function will be used instead

=cut
	
sub _PrintFasta
{
	my $self = shift;
	my ($fh_out,$header,$r_seq) = @_;
	
	print $fh_out $header . "\n" . $$r_seq . "\n";

}

=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;

