#
# $Id: GFF3.pm 1438 2015-04-20 13:33:57Z carrere $
#

package AnnotatedSeqFileWriter::GFF3;

# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


=head1 NAME

AnnotatedSeqFileWriter::GFF3 - Dumping AnnotatedSeqFile objects in gff3
                               See the GFF3 description at http://www.sequenceontology.org/gff3.shtml

=head1 SYNOPSIS

 my $o_dumper = New AnnotatedSeqFileWriter::gff3( o_annot => $o_annot );
 $o_dumper -> Dump();

=head1 DESCRIPTION

 this class extends the AnnotatedSeqFileWriter, it dumps data from AnnotatedSeq to gff3 files

=cut

use strict;
use warnings;
use Carp;
use URI::Escape;

use Error qw(:try);
#use Bio::Seq;
use Data::Dumper;
use General;
use Scalar::Util qw(looks_like_number);

use Locus;
use SeqOntology;
use base qw( AnnotatedSeqFileWriter );
use FileCompressedAppend;

BEGIN
{
    our $VERSION = do {my @r = (q$Rev: 1438 $ =~ /\d+/g); $r[0]};
}

sub _Init
{
	my $self = shift;
	$self->SUPER::_Init(@_);
	my $o_seq_ontology = New SeqOntology;
	$self -> { __o_seq_ontology } = $o_seq_ontology;
	
	# See __DumpF45678
	$self -> { __is_gff3 } = &TRUE;
}

=head2 procedure Dump

 Title     : Dump
 Usage     : $o_annot_writer -> Dump( );
 Procedure : Dump data to gff3 files
 Args      : -append 1/0 (optional, def 0)

=cut

sub Dump
{
	my $self = shift;
	my (%h_params) = (-append => 0, @_);

	# Open the output file
	my $open_mode = ($h_params{ -append } == 0) ? 'w' : 'a';
	my $o_annot = $self -> { _o_annot };

	my $f_out  = $o_annot -> GetOutputFile();
	my $o_comp   = New FileCompressedAppend ( $f_out, $open_mode );
	$f_out       = $o_comp->GetFileName();
	unlink $f_out if ( $open_mode eq 'w' );
	my $fh_out   = new IO::File( $f_out, $open_mode ) || throw LipmError::IOException($f_out,'f');

	# Dump the header
	$self -> _DumpHeader($fh_out,$o_annot);

	# Dump the annotations
	$self -> _DumpAnnotations( $fh_out );
	
	# Dump the sequence if necessary
	if ( $o_annot->GetWithSequenceEmbedded() && $o_annot -> GetLength()!=0 )
	{
		$self -> _DumpSequence( $fh_out );
	}
	$o_annot -> AddDumped( $o_comp );
}

=head2 procedure _DumpAnnotations

 Title     : _DumpAnnotations
 Usage     : $self -> _DumpAnnotations()
 Function  : Dump the annotations to the gff3 file
 Args      : $fh_out

=cut

sub _DumpAnnotations
{
	my $self = shift;
	my ($fh_out) = @_;
	
	my $o_annot = $self -> { _o_annot };

	# Sort the genes in increasing order
	#my @a_genes = sort { $a->GetStart() <=> $b->GetStart() } @{$o_annot -> { __a_genes }};
	my $ra_sorted_loci = $o_annot -> GetSortedLoci();
	my ($s,$e,$field9);
	
	# Keep the list of already used locus tags
	#my %h_locus_tags = ();
	#foreach my $o_gene ( @a_genes )
	foreach my $o_locus (@$ra_sorted_loci)
	{
		next unless $o_locus -> IsSelected();
		
		if ( $o_locus -> isa('Locus::Gene') )
		{
			# Print the gene line
			($s,$e) = $o_locus -> GetEnds();
			$self->__DumpRegions( $fh_out, $o_locus);
			
			# For coding genes, print the mRNAs, CDS, exons
			if ( $o_locus -> GetSubType() eq 'protein_coding_gene' )
			{
				my $ra_mrnas = $o_locus -> GetMrnas();
				foreach my $o_mrna ( @$ra_mrnas )
				{
					next unless $o_mrna -> IsSelected();
					$self->__DumpRegions( $fh_out, $o_mrna);

					# Printing exons and utr regions
					my @a_exons_utr=();
					my ($o_utr1,$o_exonset,$o_utr2);
					my $sub_type1 = ($o_mrna -> GetStrand() > 0) ? 'five_prime_UTR': 'three_prime_UTR';
					my $sub_type2 = ($o_mrna -> GetStrand() > 0) ? 'three_prime_UTR': 'five_prime_UTR';
					$o_utr1   = $o_mrna -> GetUtr( $sub_type1 );
					$o_utr2   = $o_mrna -> GetUtr( $sub_type2 );
					$o_exonset= $o_mrna -> GetExonSet( );
					push @a_exons_utr,$o_exonset->GetRegions() if (defined ($o_exonset) && $o_exonset -> IsSelected());;
					push @a_exons_utr,$o_utr1->GetRegions() if (defined ($o_utr1) && $o_utr1 -> IsSelected());
					$self->__DumpRegions( $fh_out, @a_exons_utr);
					
					my $o_cds = $o_mrna -> GetCDS();
					if ( !defined ( $o_cds ) )
					{
						warn $self -> GetWarningHeader() . "No cds for " . $o_mrna->GetpkExtended() . "\n";
					}
					elsif ( $o_cds -> IsSelected() )
					{
						$o_cds -> ComputePhase();
						my @a_regions = $o_cds -> GetRegions();
						if ( @a_regions == 0)
						{
							throw Error::Simple("ERROR - No region for " . $o_cds->GetpkExtended() ."\n");
						}
						else
						{
							$self->__DumpRegions( $fh_out, @a_regions );
						}
					}
					
					@a_exons_utr = ();
					push @a_exons_utr,$o_utr2->GetRegions() if (defined ($o_utr2) && $o_utr2 -> IsSelected());
					$self->__DumpRegions( $fh_out, @a_exons_utr );
				}
			}
			#... print the ncRNAs
			my $ra_ncrnas = $o_locus -> GetNcrnas();
			foreach my $o_ncrna ( @$ra_ncrnas )
			{
				$self->__DumpRegions( $fh_out, $o_ncrna);
				my $o_exonset= $o_ncrna -> GetExonSet( );
				if (defined($o_exonset))
				{
					$self -> __DumpExonSet( $fh_out, $o_exonset );
				}
			}
			
			# EXPERIMENTAL
			my $o_exonset = $o_locus -> GetExonSet();
			if (defined $o_exonset )
			{
				$self -> __DumpExonSet( $fh_out, $o_exonset, 0 );
			}
			
			# Dump the regions
			my $ra_regions = $o_locus -> GetMiscRegions();
			foreach my $o_region ( @$ra_regions )
			{
				if ( $o_region -> IsSelected() )
				{
					$self -> __DumpLocus ( $fh_out, $o_region );
				}
			}
		}
		else
		{
			$self -> __DumpLocus( $fh_out, $o_locus );
		}
		print $fh_out "###\n";
	}
}

=head2 procedure _DumpHeader

 Title     : _DumpHeader
 Usage     : $o_annot_writer -> _DumpHeader( $fh_out, $o_annot );
 Procedure : Dump header data, called by Dump
             If any, dump the data found in the header hash maintained by $o_annot
             Dump lines as (key,value) pairs

=cut

sub _DumpHeader
{
	my $self = shift;
	my ( $fh_out, $o_annot ) = @_;
	
	if ( $fh_out -> tell() == 0 )
	{
		print $fh_out "##gff-version 3\n";
	}

	# Print the sequence-region pragma
	print $fh_out "##sequence-region " . $o_annot -> GetChromosomeLabel() . " " . $o_annot -> GetSeqRegionStart() . " " . $o_annot -> GetSeqRegionEnd() . "\n";
	

	# Dump the information from the header
	my $rh_headers = $o_annot -> GetHeaders();
	
	# If nothing in headers, return
	my @a_header_keys = keys %$rh_headers;
	return if ( @a_header_keys == 0);
	
	my $contig = $o_annot -> GetChromosomeLabel();
	foreach my $key ( sort @a_header_keys )
	{
		print $fh_out "#bfc $contig $key=" . $rh_headers -> { $key } . "\n";
	}
}

sub __DumpExonSet
{
	my $self = shift;
	my ( $fh_out, $o_exonset, $min_region ) = @_;
	
	# The default is NOT to dump the exons if there is only 1 region in the ExonSet
	# But this can be changed with the parameter $min_region
	$min_region = 1 unless ( defined $min_region);
	my @a_exons = $o_exonset->GetRegions();
	if (@a_exons > $min_region)
	{
		$self->__DumpRegions( $fh_out,@a_exons);
	}
}

=head2 procedure __DumpRegions

 Title     : __DumpRegions
 Usage     : $o_annot_writer -> __DumpRegions( $fh_out, $o_region, ... );
 Procedure : Dump the list of regions passed by parameters
 Args      : $fh_out  An output opened stream
             $o_region,... The regions being dumped

=cut
sub __DumpRegions
{
	my $self = shift;
	my ( $fh_out, @a_regions) = @_;
	foreach my $o_region (@a_regions)
	{
		next unless defined($o_region);
		#print $o_region->GetSubType() . ' ' . $o_region->GetStart()."..".$o_region->GetEnd().' '.$o_region->GetStrand().' '.$o_region->GetConsistency().' '.$self->__GetPartialInfo($o_region)."\n";
		$self -> __DumpF12( $fh_out, $o_region );
		$self -> __DumpF3( $fh_out, $o_region );
		$self -> __DumpF45678( $fh_out, $o_region );
		$self -> _DumpF9( $fh_out, $o_region );
	}
}

=head2 procedure __DumpLocus

 Title     : __DumpLocus
 Usage     : $o_annot_writer -> __DumpLocus( $fh_out, $o_region, ... );
 Procedure : Dump the locus passed by parameter, calling __DumpRegions, then dump the regions, if any, calling again __DumpRegions
 Args      : $fh_out  An output opened stream
             $o_region,... The regions being dumped

=cut
sub __DumpLocus
{
	my $self = shift;
	my ( $fh_out, $o_locus ) = @_;
	$self -> __DumpRegions( $fh_out, $o_locus );
	my @a_regions = $o_locus -> GetRegions();
	if ( @a_regions > 1 )
	{
		$self -> __DumpRegions( $fh_out, @a_regions );
	}
}

=head2 procedure __DumpF12

 Title     : __DumpF12
 Usage     : $o_annot_writer -> __DumpF12( $fh_out, $o_locus );
 Procedure : Dump the columns 1 and 2
 Args      : $fh_out  An output opened stream
             $o_locus The locus being printed

=cut
sub __DumpF12
{
	my $self = shift;
	my ($fh_out,$o_locus)  = @_;

	print $fh_out $o_locus -> GetChromosomeLabel() . "\t";
	print $fh_out "EuGene\t";
}

=head2 procedure __DumpF3

 Title     : __DumpF3
 Usage     : $o_annot_writer -> __DumpF3( $fh_out, $o_locus );
 Procedure : Dump the type of locus (field 3 of gff3 files)
             If the type is allowed, print $o_locus->GetSubType(), else print the default type
 Args      : $fh_out  An output opened stream
             $o_locus The locus being printed

=cut
sub __DumpF3
{
	my $self = shift;
	my ($fh_out,$o_locus)  = @_;
	
	# For a gene, print 'gene', NOT NcRNA_gene or protein_coding_gene
	my $gff3_type = '';
	if ( ref($o_locus) eq 'Locus::Gene' )
	{
		$gff3_type = 'gene';
	}
	else
	{
		my $o_so = $self -> { __o_seq_ontology };
		$gff3_type = ($o_so -> IsTypeInSeqOntology( $o_locus )) ? $o_locus -> GetSubType() : $o_locus -> GetDefaultSubType();
		
		# If o_locus is a region, check there is a correct relationship between parent and child
		# If not, try to change the displayed type
		if ( ref ($o_locus) eq 'Locus::MiscRegion' )
		{
			my $o_parent = $o_locus -> GetParent();
			if ( $o_parent->isa('Locus') )
			{
				my $p_gff3_type = ($o_so -> IsTypeInSeqOntology( $o_parent )) ? $o_parent->GetSubType() : $o_parent->GetDefaultSubType();
				if ( $o_so -> IsADescendantOf( $gff3_type, $p_gff3_type ) == &FALSE )
				{
					my $tmp = $o_so -> GetPreferredChild( $p_gff3_type );
					if ( $tmp ne "" )
					{
						$gff3_type = $tmp;
					}
					else
					{
						warn $self -> GetWarningHeader() . "$gff3_type is not a child of $p_gff3_type, the gff3 file will be probably NOT valid " . $o_locus->GetpkExtended() . "\n";
					}
				}
			}
		}
	}		

	print $fh_out $gff3_type;
	print $fh_out "\t";
}

=head2 procedure __GetSubType

 Title     : __GetSubType
 Usage     : print $o_annot_writer -> __GetSubType ( $o_locus );
 Function  : Return the sub_type of locus (for dumping as sub_type=$sub_type on the field 9)
             If the sub_type is allowed, return "" because the field 3 is enough, else return ;sub_type=$sub_type
 Args      : $o_locus The locus being printed

=cut
sub __GetSubType
{
	my $self = shift;
	my ($o_locus)  = @_;
	my $o_so = $self -> { __o_seq_ontology };

	return ($o_so -> IsTypeInSeqOntology( $o_locus )) ? "" : ";sub_type=" . $o_locus -> GetSubType();
}

=head2 procedure __DumpF45678
 print the fields 4-8
 The flag is_gff3 is used here, if dumping to a gff3 file set to TRUE, if dumping to a tsv file set to false
 For tsv (excel), it is more convenient to represent the strand with 1/-1 rather than with -/+ (required by gff3)
=cut

sub __DumpF45678
{
	my $self = shift;
	my ($fh_out,$o_locus)  = @_;
	my $is_gff3 = $self -> { __is_gff3 };

	my ($s,$e)  = $o_locus -> GetEnds();
	print $fh_out "$s\t$e\t";

	my $score  = $o_locus -> GetScore();
	if ( $score ne '.' )
	{
		$score =~ s/ //g;    # 1.0 e -5 ==> 1.0e-5
		if ( looks_like_number($score) == 0 )
		{
			warn $self -> GetWarningHeader() . "$score should be a floating-point number. A '.', it is replaced with a '.' - " . $o_locus->GetpkExtended() . "\n";
			$score = '.';
		}
	}

	my $phase  = $o_locus -> GetPhase();
	my $strand = $o_locus -> GetStrand($is_gff3);
	print $fh_out "$score\t$strand\t$phase\t";
}

=head2 procedure _DumpF9

 Title     : __DumpF9
 Usage     : $o_annot_writer -> __DumpF9 ($fh_out, $o_locus)
 Procedure : Dump the Field 9 of gff3
 Args      : $fh_out  An output opened stream
             $o_locus The locus to dump
 Access    : PROTECTED

=cut
#our %H_IGNORED = qw( translation 1 score 1 gene 1 note 1 db_xref 1 parent 1 product 1 ID 1 gene_synonym 1 Alias 1);
sub _DumpF9
{
	my $self = shift;
	my ($fh_out,$o_locus)  = @_;
	print $fh_out $self -> _ComputeF9($o_locus)."\n";
}

=head2 function _ComputeF9

 Title     : _ComputeF9
 Usage     : $o_annot_writer -> _ComputeF9 ($o_locus)
 Function  : Compute the Field 9 of gff3, called by _DumpF9 (in this class and in derived classes)
 Args      : $fh_out  An output opened stream
             $o_locus The locus to dump
 Access    : PROTECTED

=cut

sub _ComputeF9 
{
	my $self = shift;
	my ($o_locus)  = @_;
	
    my $field9 = "";
	
	my $p_id;
	if ( $o_locus->isa('Locus::MiscRegion') )
	{
		my $o_parent = $o_locus -> GetParent();
		if ( $o_parent->isa('Locus') )
		{
			my $o_seq_ontology = $self -> { __o_seq_ontology };
			if ( $o_seq_ontology->IsADescendantOf($o_locus->GetSubType(),$o_parent->GetSubType()) )
			{
				$p_id = $o_parent -> GetId();
			}
			else
			{
				$p_id = '';
			}
		}
		else
		{
			$p_id = '';
		}
	}
	else
	{
		$p_id           = $o_locus -> GetParentId();
	}

	my $id             = $o_locus->GetId();
	my $dbxref         = $o_locus -> GetDbXref4gff3(&TRUE);
	my $ontology_term = $o_locus -> GetOntologyTerms(&TRUE);

	my ( $protein_id, $transcript_id) = ("","");
	$protein_id     = $o_locus -> GetProteinId(&TRUE)    if ( $o_locus -> can('GetProteinId'));
	$transcript_id  = $o_locus -> GetTranscriptId(&TRUE) if ( $o_locus -> can('GetTranscriptId'));

    $field9   .= "ID=${id}"           if ( $id ne "");
    $field9   .= ";Parent=${p_id}"    if ( $p_id ne "");
    $field9   .= $self -> __GetSubType ( $o_locus );
	$field9   .= ";transcript_id=${transcript_id}" if ($transcript_id ne "");
	$field9   .= ";protein_id=${protein_id}"       if ($protein_id ne "");
	$field9   .= ";Dbxref=${dbxref}"               if ( $dbxref ne "" );
	my $partial = $self->__GetPartialInfo($o_locus);
	$field9   .= ";partial=$partial" if ($partial ne "");
	$field9   .= ";Ontology_term=${ontology_term}" if ( $ontology_term ne "" );

	# Dumping an exon/CDS/UTR, or a region of MiscRegion: it's over !
	my $its_over = $o_locus->isa ( 'Locus::CdsRegion' );
	$its_over += $o_locus->isa ( 'Locus::Exon' );
	$its_over += $o_locus->isa ( 'Locus::UtrRegion' );
	$its_over += $o_locus -> GetParent()->isa('Locus::MiscRegion');
	if ( $its_over != &FALSE )
	{
		my $ra_other_keys = $o_locus -> GetOtherKeys( {'Ontology_term'=>1});
		foreach my $key (sort @$ra_other_keys)
		{
			my $value = $o_locus -> GetAsString ( $key );
			next if ( $value eq "" );
			if ($key =~ /^ec_number$/i)
			{
				$key = lc ($key);
				$value = join (',', split (' ',uri_escape ($value,",=;\t")));
			}
			$field9 .= ';' . $key . '=' . $value;
		}
		# Print the pseudogene info anyway !
		my $pseudo = $o_locus -> Get ( 'pseudogene' );
		$field9 .= ";pseudogene=$pseudo" if ($pseudo ne "");
		return $field9;
	}
	
	# WARNING - There should only ONE value for Gene name (gff3 validator dixit)
	#           So if there are several values, the FIRST only will be dedicated to the Name, the others to the alias
	my $gene_alias  = $o_locus -> GetAlias(&TRUE);
	$gene_alias     =~ s/;/,/g;	# Coming from genbank/embl
	my $gene_name   = $o_locus -> GetGeneName();
	if ( $gene_name =~ /^([^,]+),(.+)$/ )     # at least two names !
	{
		$gene_name  = $1;
		$gene_alias .= ($gene_alias eq '')?$2:",$2";
	}

	$gene_name =~ s/(^,)|(,$)//g;
	$gene_alias=~ s/(^,)|(,$)//g;
	$gene_name =~ s/,,/,/g;
	$gene_alias=~ s/,,/,/g;

	# Escape ; characters from $name, $alias
	$gene_name  = uri_escape ($gene_name,";");
	$gene_alias = uri_escape ($gene_alias,";");

	my $locus_tag   = $o_locus -> GetLocusTag(&TRUE);

	# For an mRNA, we consolidate with the children
	my $note        = $o_locus -> isa ( 'Locus::Mrna' ) ? $o_locus -> GetConsolidatedNote()    : $o_locus -> GetNote (&TRUE);
	my $product     = $o_locus -> isa ( 'Locus::Mrna' ) ? $o_locus -> GetConsolidatedProduct() : $o_locus -> GetProduct(&TRUE);
	
	# Escape ,=; and TAB characters from $note, $product
	$note    = uri_escape ($note,",=;\t");
	$product = uri_escape ($product,",=;\t");
	
	$field9   .= ";Name=${gene_name}"      if ( $gene_name ne "" );
	$field9   .= ";locus_tag=${locus_tag}" if ( $locus_tag ne "" );
	$field9   .= ";Alias=${gene_alias}"    if ( $gene_alias ne "" );
	$field9   .= ";Note=${note}"           if ( $note ne "" );
	$field9   .= ";product=$product"       if ( $product ne "" );

	# Correcting a lot of awful things which could lead to gff3 validation problems
	$field9 =~ s/=,/=/g;
	$field9 =~ s/,;/;/g;
	$field9 =~ s/,,/,/g;
	$field9 =~ s/,\s*$//g;
	
	my $f = "";
	#my $ra_other_keys = $o_locus -> GetOtherKeys(\%H_IGNORED);
	#my $ra_other_keys = $o_locus -> isa ( 'Locus::Mrna' ) ? $o_locus -> GetConsolidatedOtherKeys( {'Ontology_term'=>1} ) : $o_locus -> GetOtherKeys( {'Ontology_term'=>1} );
	my $ra_other_keys = $o_locus -> GetOtherKeys( {'Ontology_term'=>1} );
	foreach my $key (sort @$ra_other_keys)
	{
		my $value = $o_locus -> isa ( 'Locus::Mrna' ) ? $o_locus -> GetAsStringConsolidated ( $key ) : $o_locus -> GetAsString ( $key );
		next if ( $value eq "" );
		
		#SC20130711: ec_number n'est pas un mot reserve GFF3 donc en minuscule
		if ($key =~ /^ec_number$/i)
		{
			$key = lc ($key);
			# Escape ,=; and TAB characters
			$value = join (',', split (' ',uri_escape ($value,",=;\t")));
			$f .= ';' . $key . '=' . $value;
		}
	}
	
	# substitute _no_value (coming from BioPerl with genbak/embl files) with ""
	$f =~ s/_no_value/""/g;
	return "${field9}${f}";
}

sub __GetPartialInfo
{
	my $self = shift;
	my ($o_locus) = @_;
	my $o_annot = $self -> { _o_annot };
	return '' if ( $o_annot->GetWithPartial() == &FALSE );
	
	my $consistency = $o_locus -> GetConsistency();
	if ( $consistency ne 'C' )
	{
		my @a_info;
		my $strand = $o_locus -> GetStrand();
		push(@a_info,'5prime') if ($o_locus->GetStrand()>0 && $consistency=~/FS/);
		push(@a_info,'5prime') if ($o_locus->GetStrand()<0 && $consistency=~/FE/);
		push(@a_info,'3prime') if ($o_locus->GetStrand()>0 && $consistency=~/FE/);
		push(@a_info,'3prime') if ($o_locus->GetStrand()<0 && $consistency=~/FS/);
		return join(',',@a_info);
	}
	else
	{
		return '';
	}
}

=head2 procedure __DumpSequence

 Title     : __DumpSequence
 Usage     : $self -> __DumpSequence()
 Function  : Dump the sequence in fasta format to the gff3 file
 Args      : $fh_out 

=cut
sub _DumpSequence
{
	my $self = shift;
	my ($fh_out) = @_;
	
	my $o_annot = $self -> { _o_annot };
	
	# print fasta header
	print $fh_out "\n\n##FASTA\n>" . $o_annot -> GetChromosomeLabel() . "\n";
	
	# Print sequence
	my $r_sequence = $o_annot -> GetGenomicSequence();
	my $seq_length = length $$r_sequence;
	my $offset     = 0;
	while ($offset < $seq_length)
	{
		print $fh_out substr ( $$r_sequence, $offset, 60 ) . "\n" ;
		$offset += 60;
	}
}

1;


=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;


