#
# $Id: Assembly.pm 1336 2013-12-04 13:35:33Z manu $
#

package Assembly;
use base qw( LipmObject );

# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=head1 NAME

Assembly - This is an ABSTRACT CLASS, use Assembly::Agp or Assembly::Contigmd
           Correct the positions using the assembly
           My be used to create a single fasta file, using a multifasta as entry

=head1 SYNOPSIS

 my $o_assembly = New Assembly($assembly_file_name, $grouplabel, $chromosome);

 

=head1 DESCRIPTION

 
=cut

use strict;
use warnings;
use Data::Dumper;
use Error::Simple;
use Carp;
use LipmError::ParameterException;
use LipmObject;
use General;
use base qw( LipmObject );
use DataReader::Whole;
use AssemblyError;

BEGIN
{
    our $VERSION = do {my @r = (q$Rev: 1336 $ =~ /\d+/g); $r[0]};
}

=head2 procedure _Init

 Title        : Constructor
 Usage        : my $o_assembly = New Assembly($assembly_filename, $grouplabel, $chromosome, $random);
 Function     : Initialize the object, reading the assembly file
 Args         : $assembly_filename The name of the assembly file (agp or config.md file)
                $grouplabel        identifier of the required assembly 
                                   (for example human, can be Celera or Primary_Assembly or HuRef)
                $chr               The chromosome to assemble
                $random (0/1)      Consider the chromosome as random - or not              

=cut

sub _Init
{
	my $self = shift;
	my ($assembly_filename, $grouplabel, $chromosome, $random) = @_;
	
	$self->{ __assembly_filename } = $assembly_filename;
	$self->{ __grouplabel }        = $grouplabel;
	$self->{ __chromosome }        = $chromosome;
	$self->{ __random }            = $random;
}

=head2 Accessors
 Title     : Accessors
 Usage     : Several useful accessors

=cut
sub GetChromosomeLabel
{
	my $self = shift;
	return $self -> { __chromosome };
}
sub GetChromosomeLabels
{
	my $self = shift;
	throw LipmError::ObjectException($self,'abstract method');
}
sub GetGroupLabel
{
	my $self = shift;
	return $self -> { __grouplabel };
}	
sub IsRandom
{
	my $self = shift;
	return $self -> { __random } == 0 ? 0 : 1;
}

=head2 GetContigNames
 Title     : GetContigNames
 Usage     : my $ra_contigs = $o_assembly -> GetContigNames();
 Function  : Return the ordered list of contig names, removing the 'start' and 'end' contigs

=cut
sub GetContigNames
{
	my $self = shift;
	throw LipmError::ObjectException($self,'abstract method');
}

=head2 NextContig, PrevContig
 Title     : NextContig, PrevContig
 Usage     : $self -> NextContig ($step)
 Function  : Call __IterContig to move the current contig forward or backward
 Args      : $step integer, should be > 0 
 Return    : The new current contig
 

=cut
sub NextContig
{
	my $self   = shift;
	my ($step) = @_;
	$step = 1 unless defined $step;
	$self -> __IterContig( $step );
	return $self -> GetCurrentContigName();
}
sub PrevContig
{
	my $self   = shift;
	my ($step) = @_;
	$step = 1 unless defined $step;
	$self -> __IterContig( -$step );
	return $self -> GetCurrentContigName();
}
sub __IterContig
{
	my $self = shift;
	throw LipmError::ObjectException($self,'abstract method');
}

=head2 SetFirstContig
 Title     : SetFirstContig
 Usage     : $self -> SetFirstContig ()
 Procedure : Position the object to the first contig
 Args      : None

=cut		
sub SetFirstContig
{
	my $self = shift;
	throw LipmError::ObjectException($self,'abstract method');
}

=head2 SetCurrentChromosome
 Title     : SetCurrentChromosome
 Usage     : $self -> SetCurrentChromosome ()
 Procedure : Position the object to the wanted chromosome, if possible
 Args      : The chromosome label to set

=cut		
sub SetCurrentChromosome
{
	my $self = shift;
	throw LipmError::ObjectException($self,'abstract method');
}

=head2 GetAssemblyGaps
 Title     : GetAssemblyGaps
 Usage     : my $ra_assembly_gaps = $self -> GetAssemblyGaps()
 Procedure : Retrieve the assembly gaps in an array of hashes
			 The field names of the hashes is the column names of an agp files, for the U/N records
 Args      : None

=cut
sub GetAssemblyGaps
{
	my $self = shift;
	throw LipmError::ObjectException($self,'abstract method');
}
	
=head2 GetCurrentContigName, GetStart, Getend, GetStrand
 Title    : GetCurrentContigName, GetStart, Getend, GetStrand, GetFeatureStart, GetFeatureEnd
 Usage    : my $contig = $self -> GetCurrentContig ()
 Function : Get the current contig name, the start/end position of the current contig, or the strand
 Args     : None

=cut		
sub GetCurrentContigName
{
	my $self = shift;
	throw LipmError::ObjectException($self,'abstract method');
}

sub GetStrand
{
	my $self = shift;
	throw LipmError::ObjectException($self,'abstract method');
}

sub GetStart
{
	my $self = shift;
	throw LipmError::ObjectException($self,'abstract method');
}

sub GetEnd
{
	my $self = shift;
	throw LipmError::ObjectException($self,'abstract method');
}

sub GetFeatureStart
{
	my $self = shift;
	throw LipmError::ObjectException($self,'abstract method');
}

sub GetFeatureEnd
{
	my $self = shift;
	throw LipmError::ObjectException($self,'abstract method');
}

=head2 function Relative2Absolute

 Title     : Relative2Absolute
 Usage     : $pos = $self->Relative2Absolute($pos);
 function  : Compute and return the absolute position, from the relative position passed by parameter, using the current contig
 Globals   : none

=cut
sub Relative2Absolute
{
	my $self = shift;
	my ($pos) = @_;
	
	my $strand    = $self -> GetStrand();
	my $abs_pos   = 0;
	my $feat_strt = $self -> GetFeatureStart();
	my $feat_end  = $self -> GetFeatureEnd();
	if ( $pos < $feat_strt || $pos > $feat_end )
	{
		throw AssemblyError ( "ERROR - The position on the contig ($pos) should be in the contig) - IF YOU SEE THIS MESSAGE, the file AnnotatedSeqFileReader::XXXX YOU ARE USING SHOULD BE MODIFIED\n" );
	}
	if ( $strand eq '+' )
	{
		$abs_pos = $self -> GetStart() + ( $pos - $self -> GetFeatureStart() );
	}
	else
	{
		$abs_pos = $self -> GetEnd() - ( $pos - $self -> GetFeatureStart() );
	}
	return $abs_pos;
}

=head2 function Absolute2Relative

 Title     : Absolute2Relative
 Usage     : $pos = $self->Absolute2Relative($pos);
 Function  : Compute and return the relative position vs the current contig, from the absolute position passed by parameter
 Globals   : none

=cut
sub Absolute2Relative
{
	my $self = shift;
	my ($abs_pos) = @_;
	
	my $strand    = $self -> GetStrand();
	my $pos   = 0;
	my $strt = $self -> GetStart();
	my $end  = $self -> GetEnd();
	if ( $abs_pos < $strt || $abs_pos > $end )
	{
		print STDERR "ERROR - The absolute position ($abs_pos) should be in the contig ($strt..$end) \n";
		#throw AssemblyError ( "ERROR - The absolute position ($abs_pos) should be in the contig ($strt..$end) \n" );
	}
	if ( $strand eq '+' )
	{
		#$abs_pos = $self -> GetStart() + ( $pos - $self -> GetFeatureStart() );
		$pos      = $abs_pos - $self -> GetStart() + $self -> GetFeatureStart();
	}
	else
	{
		#$abs_pos = $self -> GetEnd() - ( $pos - $self -> GetFeatureStart() );
		$pos      = $self -> GetEnd() + $self -> GetFeatureStart() - $abs_pos;
	}
	return $pos;
}

=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;

