#
# $Id$
#

package Assembly::Contigmd;
use base qw( Assembly );

# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=head1 NAME

Assembly - Loads an agp or contig.md assembly file
           Corrects the positions using the assembly
           Create a single fasta file, using a multifasta as entry

=head1 SYNOPSIS

 my $o_assembly = New Assembly($assembly_file_name, $grouplabel, $chromosome);

 

=head1 DESCRIPTION

 
=cut

use strict;
use warnings;
use Data::Dumper;
use Error::Simple;
use Carp;
use LipmError::ParameterException;
use LipmObject;
use General;
use base qw( LipmObject );
use DataReader::Whole;
use AssemblyError;

BEGIN
{
    our $VERSION = do {my @r = (q$Rev: 1009 $ =~ /\d+/g); $r[0]};
}

=head2 procedure _Init

 Title        : Constructor
 Usage        : my $o_assembly = New Assembly($assembly_filename, $grouplabel, $chromosome, $random);
 Function     : Initialize the object, reading the assembly file
 Args         : $assembly_filename The name of the assembly file (agp or config.md file)
                $grouplabel        identifier of the required assembly 
                                   (for example human, can be Celera or Primary_Assembly or HuRef)
                $chr               The chromosome to assemble
                $random (0/1)      Consider the chromosome as random - or not              

=cut

sub _Init
{
	my $self = shift;
	my ($assembly_filename, $grouplabel, $chromosome, $random) = @_;
	
	$self->{ __assembly_filename } = $assembly_filename;
	$self->{ __grouplabel }        = $grouplabel;
	$self->{ __chromosome }        = $chromosome;
	$self->{ __random }            = $random;
	if ( !defined($chromosome) || $chromosome eq '' )
	{
		throw Error::Simple('Assembly::Contigmd, you MUST specify a chromosome');
	}
	
	$self->__LoadSeqContigMd();

	#use Data::Dumper;
	#print Dumper ($self -> { __rh_contigs });
	#exit 0;
}
=head2 SetCurrentChromosome
 Title     : SetCurrentChromosome
 Usage     : $self -> SetCurrentChromosome ()
 Procedure : Position the object to the wanted chromosome, if possible
             
 Args      : The chromosome label to set

=cut		
sub SetCurrentChromosome
{
	my $self = shift;
	my ( $chromosome ) = @_;
	if ( $chromosome ne $self -> { __chromosome } )
	{
		throw Error::Simple('You CANNOT change chromosome with SetCurrentChromosome for Assembly::Contigmd objects');
	}
}

=head2 Accessors
 Title     : Accessors
 Usage     : Several useful accessors

=cut
sub GetChromosomeLabels
{
	my $self = shift;
	die "GetChromosomeLabels IS NOT IMPLEMENTED FOR THIS OBJECT - SORRY !";
}

=head2 GetContigNames
 Title     : GetContigNames
 Usage     : my $ra_contigs = $o_assembly -> GetContigNames();
 Function  : Return the ordered list of contig names, removing the 'start' and 'end' contigs

=cut
sub GetContigNames
{
	my $self  = shift;
	my $chrom = $self->{ __chromosome };
	my $ra_contigs = $self -> { __rh_contigs } -> { $chrom };
	my @a_contigs = @$ra_contigs;
	
	# Remove 'start' and 'end' pseudo-contigs
	shift @a_contigs if ( $a_contigs [ 0 ]  eq 'start' );
	pop @a_contigs if ( $a_contigs [ -1 ]  eq 'end' );

	return \@a_contigs;
}

=head2 GetAssemblyGaps
 Title     : GetAssemblyGaps
 Usage     : my $ra_assembly_gaps = $self -> GetAssemblyGaps()
 Procedure : Print a warning and return a ref to an empty array
 Args      : None

=cut
sub GetAssemblyGaps
{
	my $self = shift;
	warn $self -> GetWarningHeader() . "AssemblyGaps is NOT IMPLEMENTED for Assembly::Contigmd. The gaps will not be written to the output annotation file\n";
	return [ ];
}

=head2 __GetContigNames
 Title     : __GetContigNames
 Usage     : my $ra_contigs = $self -> __GetContigNames();
 Function  : Return the ordered list of contig names, but the 'start' and 'end' contigs are NOT removed

=cut
sub __GetContigNames
{
	my $self  = shift;
	my $chrom = $self->{ __chromosome };
	my $ra_contigs = $self -> { __rh_contigs } -> { $chrom };
	return $ra_contigs;
}

=head2 NextContig, PrevContig
 Title     : NextContig, PrevContig
 Usage     : $self -> NextContig ($step)
 Function  : Call __IterContig to move the current contig forward or backward
 Args      : $step integer, should be > 0 
 Return    : The new current contig
 

=cut
sub NextContig
{
	my $self   = shift;
	my ($step) = @_;
	$step = 1 unless defined $step;
	$self -> __IterContig( $step );
	return $self -> GetCurrentContigName();
}
sub PrevContig
{
	my $self   = shift;
	my ($step) = @_;
	$step = 1 unless defined $step;
	$self -> __IterContig( -$step );
	return $self -> GetCurrentContigName();
}

=head2 __IterContig
 Title     : __IterContig
 Usage     : $self -> __Itercontig ($step)
 Procedure : Starting from the current contig, sets the current contig $step contigs forward (if $step>0) or backward (if $step<0)
 Args      : $step integer, may be > 0 or < 0
 Access    : Private
 Error     : Throw an error instead of setting the current contig before the first or after the last one

=cut
sub __IterContig
{
	my $self   = shift;
	my ($step) = @_;

	my $ra_contigs  = $self -> __GetContigNames();
	if ( ! defined ( $self -> { __current_contig_index } ) )
	{
		$self -> SetCurrentContig ( $ra_contigs -> [ 1 ] );
	}
	
	my $current_contig= $self -> GetCurrentContigName();
	my $current_index = $self -> { __current_contig_index };

	$current_index += $step;
	if ( $step > 0 && ($current_contig eq 'end' || $current_index >= @$ra_contigs ) )
	{
		throw Error::Simple ( 'Already at end' );
	}
	elsif ( $step < 0 && ($current_contig eq 'start' || $current_index < 0 ) )
	{
		throw Error::Simple ( 'Already at start' );
	}
	else
	{
		$current_contig = $ra_contigs -> [$current_index];
	}
	$self -> __SetCurrentContig ( $current_contig );
	$self -> { __current_contig_index } = $current_index;
}

=head2 SetFirstContig
 Title     : SetFirstContig
 Usage     : $self -> SetFirstContig ()
 Procedure : Position the object to the first contig
 Args      : None

=cut		
sub SetFirstContig
{
	my $self = shift;
	$self -> SetCurrentContig('start');
}

=head2 SetCurrentContig
 Title     : SetCurrentContig
 Usage     : $self -> SetCurrentContig ('contig_name')
 Procedure : Set the current contig to some value
 Args      : $current_contig The name of the wanted current contig
 Error     : Throw an error if the wanted contig is not found

=cut		
sub SetCurrentContig {
	my $self = shift;
	$self -> __SetCurrentContig ( @_ );
	$self -> __SetContigIndex();
}
	
sub __SetCurrentContig
{
	my $self = shift;
	my ( $current_contig) = @_;
	my $rh_contigs = $self -> { __rh_contigs };
	if ( ! exists ($rh_contigs -> {data } -> { $current_contig } ) )
	{
		throw LipmError::ParameterException ('current_contig','value', -text => "$current_contig is unknown\n");
	}
	$self -> { __current_contig  } = $current_contig;	
	$self -> { __rh_current_data } = $self -> { __rh_contigs } -> { data } -> { $current_contig };
}

sub __SetContigIndex
{
	my $self = shift;
	my $current_contig = $self -> GetCurrentContigName();
	my $ra_contigs     = $self -> __GetContigNames();
	my $i;
	for ( $i=0; $i < @$ra_contigs; $i++ )
	{
		last if ( $ra_contigs -> [ $i ] eq $current_contig );
	}
	$self -> { __current_contig_index } = $i;
}

=head2 GetCurrentContigName, GetStart, Getend, GetStrand
 Title    : GetCurrentContigName, GetStart, Getend, GetStrand, GetFeatureStart, GetFeatureEnd
 Usage    : my $contig = $self -> GetCurrentContigName ()
 Function : Get the current contig name, the start/end position of the current contig, or the strand
 Args     : None

=cut		
sub GetCurrentContigName
{
	my $self = shift;
	return $self -> { __current_contig  };
}

sub GetStrand
{
	my $self = shift;
	if ( defined $self -> { __rh_current_data } )
	{
		# the DEFAULT is returning '+' (cf the agp specifications)
		return ( $self -> { __rh_current_data } -> { orientation } eq '-' ) ? '-' : '+';
	}
	else
	{
		throw LipmError::ObjectException($self, '!attribute', -text => 'Please call SetCurrentConfig before calling this function');
	}
}
sub GetStart
{
	my $self = shift;
	if ( defined $self -> { __rh_current_data } )
	{
		return $self -> { __rh_current_data } -> { chr_start };
	}
	else
	{
		throw LipmError::ObjectException($self, '!attribute', -text => 'Please call SetCurrentConfig before calling this function');
	}
}
sub GetEnd
{
	my $self = shift;
	if ( defined $self -> { __rh_current_data } )
	{
		return $self -> { __rh_current_data } -> { chr_stop };
	}
	else
	{
		throw LipmError::ObjectException($self, '!attribute', -text => 'Please call SetCurrentConfig before calling this function');
	}
}
sub GetFeatureStart
{
	my $self = shift;
	if ( defined $self -> { __rh_current_data } )
	{
		return $self -> { __rh_current_data } -> { feature_start };
	}
	else
	{
		throw LipmError::ObjectException($self, '!attribute', -text => 'Please call SetCurrentConfig before calling this function');
	}
}
sub GetFeatureEnd
{
	my $self = shift;
	if ( defined $self -> { __rh_current_data } )
	{
		return $self -> { __rh_current_data } -> { feature_end };
	}
	else
	{
		throw LipmError::ObjectException($self, '!attribute', -text => 'Please call SetCurrentConfig before calling this function');
	}
}

=head2 procedure __LoadSeqContigMd

 Title     : __LoadSeqContigMd
 Usage     : $self->__LoadSeqContigMd();
 Procedure : Load the tiling/golden path from a contig.md file
 Globals   : none
 
  __rh_contigs description:
                key = $chr_start OR $chr_end OR $chr OR data
                val = 
                      1/ If data, the val is a HASH or HASHES key = feature_name, val = the line read from seq_contig.md, with some corrections in the case of a random chromosome
                         keys = chromosome (part number in agp specification),
                                chr_start,chr_stop,orientation (object beg and end, orientation in agp specification)
	                            feature_type (component type in agp specification),
	                            feature_start,feature_end (component beg and end in agp specification)
	                            chromosome (irrelevant in agp specification)
	                            feature_name (component_id in agp specification)
	                  2/ If $chr_start or $chr_end, the val is an INTEGER
	                  3/ If $chr, the val is an ARRAY, the list of ordered features is stored in it

=cut

sub __LoadSeqContigMd
{
	my $self = shift;

	my $f_assembly = $self -> { __assembly_filename };
	my $grouplabel = $self -> { __grouplabel };
	my $chromosome = $self -> { __chromosome };
	my $randomchrom= $self -> { __random };
	
	# JER 20070201 : my @intitule = ("taxid","chr","from","to","orientation","ac_version","id","type","assembly","weight");
	my @intitule = ("tax_id","chromosome","chr_start","chr_stop","orientation","feature_name","feature_id","feature_type","group_label","weight");
	
	# Used for a random chromosome
	my $artificial_offset=0;
	my $nuclbuffersize=600;       # the number of 'N' to insert between two contigs

	my $o_reader = New DataReader::Whole();
	
	# read only the lines with correct group label (if specified)
	my %h_value;
	$h_value { group_label } = $grouplabel if ( defined $grouplabel );
	
	$o_reader    -> Prepare ( -from => $f_assembly, -where => { VALUE => \%h_value } );
	my $ra_data  =  $o_reader -> FetchallArrayref( -line_type => 'HASH' );
	
	if ( @$ra_data == 0 )
	{
		throw Error::Simple " ERROR - The chromosome $chromosome is not referenced by $f_assembly";
	}
	
	my %h_contigs;
	my @a_features = ();
	$h_contigs { $chromosome } = \@a_features;
	foreach my $rh_line ( @$ra_data )
	{
		# ????
		next if ( $rh_line -> {chromosome} =~ /\|/ && !$randomchrom );
		next if ( $rh_line -> {chromosome} ne $chromosome );

		my $feature_name = $rh_line -> {feature_name};
		
		# Discovering the chromosome limits, and the default group label
		my $chr_key_start = $chromosome . "_start";
		if ( ($feature_name =~ /start/ || $rh_line -> {chr_start} == 1)  &&  !defined($h_contigs{$chr_key_start}) )
		{
			$grouplabel = $rh_line -> {group_label}; # useful only if the group label is not specified through the parameters
			$h_contigs{$chr_key_start} = $rh_line -> {chr_start};
		}
		next if ( $rh_line -> {group_label} !~ /$grouplabel/ );
		
		#if ( $feature_name =~ /end/ )
		if ( $feature_name eq 'end' )	# plus robuste que /end/
		{
			my $chr_key_end = $chromosome . "_end";
			$h_contigs{$chr_key_end} = $rh_line -> {chr_stop};
		}

		my $length = $rh_line -> {chr_stop} - $rh_line -> {chr_start} + 1;

		# Correct limits for the random chrom	
		if ($randomchrom) {
			$rh_line -> {chr_start}  += $artificial_offset;
			$rh_line -> {chr_stop}   += $artificial_offset;
			$artificial_offset += $length + $nuclbuffersize;
		}
		
		#THOMAS: adding feature_start and feature_end to be compliant with agp files (ce n'est pas très propre)
		$rh_line -> {feature_start} = 1;
		$rh_line -> {feature_end}   = $length;

		# Keep $rh_line in %h_contigs
		$h_contigs {data}->{$feature_name} = $rh_line;

		# Keeping the features order
		push @a_features, $feature_name;
	}
	
	# For a random chromosome, there should not be any start, end feature
	# TODO Le verifier tout-de-meme ?	
	if ( $randomchrom )
	{
		unshift @a_features, "start";
		push    @a_features, "end";
		my $chr_key_end = $chromosome . "_end";		
		# la JE COMPRENDS PAS !!! (manu)
		$h_contigs{$chr_key_end} = $artificial_offset+$artificial_offset;
	}
	
	# If the $chromosome_start, $chromosome_end keys are not defined, we compute them

	my ($from,$to) = (99999999999,0);
	foreach my $ctg ( @a_features )
	{
		next if ($ctg eq "start" || $ctg eq "end");
		$from = $h_contigs{data}->{$ctg}->{chr_start} if ( $h_contigs{data}->{$ctg}->{chr_start} < $from );
		$to   = $h_contigs{data}->{$ctg}->{chr_stop}  if ( $h_contigs{data}->{$ctg}->{chr_stop}  < $to );
	}
	my $key_start = $chromosome . "_start";
	my $key_end   = $chromosome . "_end";

	if ( !defined($h_contigs{$key_start}) ) { $h_contigs{$key_start} = $from; }
	if ( !defined($h_contigs{$key_end})   ) { $h_contigs{$key_end}   = $to;   }

	#print Dumper(\%h_contigs) if ( $self->GetDebug() );
	
	$self -> { __rh_contigs } = \%h_contigs;
}

=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;

