#
# $Id: GenbankTags.pm 1288 2013-10-21 08:39:57Z manu $
#

package GenbankTags;

use strict;
use warnings;

# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=head1 NAME

GenbankTags - The gateway between genbank/embl/ddbj features and qualifiers and the objects derived from Locus
              This module is used by AnnotatedSeqFileReader and AnnotatedSeqFileWriter

=head1 SYNOPSIS
 my $o_so = New GenbankTags();
 if ( $o_so->IsTypeInSeqOntology( $o_locus ) ...
 my $locus_class = GetLocusClass( $term);

=head1 DESCRIPTION

 This object may help to create a Locus:: object from a genbank feature
 It may also tell you if the type of some locus corresponds to a genbank feature
 See http://www.ebi.ac.uk/ena/WebFeat/ for the genbank tags 
 See http://www.sequenceontology.org/ for the sequence ontology terms

 
=cut

use General;
use Error qw(:try);
use base qw( LipmSingleton );
use LipmError::ParameterException;
use LipmError::ObjectException;

BEGIN
{
    our $VERSION = do {my @r = (q$Rev: 1288 $ =~ /\d+/g); $r[0]};
}

# The data are stored inside the SeqontologyData.pl file, this file is evaled
# The following global variables will be filled:
use constant MODULE_DATA_FILE => 'GenbankTagsData.pl';

# %H_CL_FEATURES connects the Genbank features and the classes used in BioFileConverter.
# All the classes should derive from Locus
# Several features may correspond to a class, thus the value is a "set of values", implemented with a perl hash
# key   = the classes derived from Locus (Locus::XXX)
# Value = sets of names
our %H_CL_FEATURES; 

# %H_QUALIFIERS gives, for each genbank feature the list of possible qualifiers
# Only the list of qualifiers is used yet, the type of each qualifier is not used (but may be used in the future)
our %H_QUALIFIERS;

# %H_SO2FEATURE connects Sequence Ontology terms with Genbank/Embl/Ddbj features
# Is is used when reading/writing genbank/embl/ddbj files
# key  = A sequence ontology term
# value= 1/ A Genbank/Embl/Ddbj feature
#        2/ mobile_element followed by _ followed by a mobile element type
our %H_SO2FEATURE;

# %H_FEATURE2SO connects Genbank/Embl/Ddbj features with Sequence Ontology terms
# Is is used when reading/writing genbank/embl/ddbj files
#  It IS NOT the reverse of %H_SO2FEATURE, because the former is a n21 relationship
# key  = A Genbank/Embl/Ddbj feature
# value= A sequence ontology term
our %H_FEATURE2SO;

# key   = some genbank features
# value = some legal types for this feature, will be integrated to /note qualifiers
our %H_TYPES;

# %$RH_CLASSES is the reverse of %H_CL_FEATURES
# It is computed by the constructor from %H_CL_FEATURES
# keys   = The genbank features
# values = The corresponding classes

our $RH_CLASSES = undef;
sub _Init
{
	my $self = shift;
	eval $self->_InitModuleData(&MODULE_DATA_FILE);

	if ( ! defined $GenbankTags::RH_CLASSES )
	{
		my %h_classes;
		foreach my $locus_class (keys %GenbankTags::H_CL_FEATURES)
		{
			my $rh_features= $GenbankTags::H_CL_FEATURES{$locus_class};
			foreach my $feature ( keys %$rh_features )
			{
				if (exists $h_classes{$feature})
				{
					throw LipmError::ParameterException('%H_CL_FEATURES','value', -text => "Duplicate value in %H_CL_FEATURES: $feature");
				}
				$h_classes{ $feature } = $locus_class;
			}
		}
		$GenbankTags::RH_CLASSES  = \%h_classes;
	}
}

=head2 function IsTypeAGenbankFeature

 Title    : IsTypeAGenbankFeature
 Usage    : if ( $o_gbk -> IsTypeAGenbankFeature( $o_locus ) ...
 Function : returns &TRUE if the $o_locus passed by parameters is not undef AND if its type returns a term
            found in the Genbank Features
            returns &FALSE in any other case
 Args     : $o_locus Locus::XXX object
 Return   : &TRUE/&FALSE
 
 Global   : %H_SEQ_ONTOLOGY_TYPE
 NOT USED YET, COMMENT OUT IF NECESSARY !!!
=cut
#sub IsTypeAGenbankFeature
#{
	#my $self = shift;
	#my ($o_locus) = @_;
	#if ( $o_locus -> isa ( 'Locus' ) && exists ($GenbankTags::H_CL_FEATURES{ref($o_locus)} -> {$self -> So2Feature($o_locus -> GetSubType())}) )
	#{
		#return &TRUE;
	#}
	## bidouille...
	#elsif ( $o_locus -> isa ( 'Locus::Gene' ) )
	#{
		#return &TRUE;
	#}
	#else
	#{
		#return &FALSE;
	#}
#}


=head2 function IsAGenbankFeature

 Title    : IsAGenbankFeature
 Usage    : if ( $o_gbk -> IsAGenbankFeature( $feature ) ...
 Function : returns &TRUE if the string passed by parameters is a known feature
 Args     : $feature a string
 Return   : &TRUE/&FALSE
 
 Global   : $RH_CLASSES
=cut
sub IsAGenbankFeature
{
	my $self = shift;
	my ($feature) = @_;
	
	if ( exists $RH_CLASSES -> { $feature } )
	{
		return &TRUE;
	}
	else
	{
		return &FALSE;
	}
}

=head2 function GetFeaturesForClass

 Title    : GetFeaturesForClass
 Usage    : if ( $o_gbk -> GetFeaturesForClass( $class ) ...
 Function : If the parameter is a Locus class (Locus::Gene etc), returns a hash whose keys are the corresponding features
 Args     : $class a string
 Return   : ref to a hash, only keys are meaningfull
 Error    : If the string does not correspond to a known class, throw an exception
 
 Global   : %H_CL_FEATURES
=cut
sub GetFeaturesForClass
{
	my $self = shift;
	my ( $class) = @_;
	if (exists $GenbankTags::H_CL_FEATURES{$class})
	{
		return \%{$GenbankTags::H_CL_FEATURES{$class}};
	}
	else
	{
		throw LipmError::ParameterException('class','value', -text => "No feature for this class: $class\n");
	}
}

=head2 function So2Feature

 Title    : So2Feature
 Usage    : my $feature = $o_gbk -> So2Feature ( $seq_ontology_term )
 Function : If the parameter is a sequence ontology term with a corresponding genbank feature (ex: three_prime_UTR => 3'UTR),
            return the corresponding feature.
            Else, return the parameter
 Args     : $seq_ontology_term, should be a sequence ontology term (but this is NOT checked) !
 Return   : The parameter, or the equivalent in genbank features
 
 Global   : %H_SO2FEATURE
=cut
sub So2Feature
{
	my $self = shift;
	my ( $so_term ) = @_;
	if ( exists $GenbankTags::H_SO2FEATURE { $so_term } )
	{
		return $GenbankTags::H_SO2FEATURE { $so_term };
	}
	else
	{
		return $so_term;
	}
}

=head2 function Feature2So

 Title    : Feature2So
 Usage    : my $so = $o_gbk -> Feature2So ( $feature )
 Function : If the parameter is a feature with a corresponding seq ontology term (ex: 3'UTR => three_prime_utr)
            return the corresponding seq ontology term.
            Else, return the parameter
 Args     : $feature, should be a valid Genbank/Embl/Ddbj feature (but this is NOT checked) !
 Return   : The parameter, or the equivalent in seq ontology terms
 
 Global   : $RH_FEATURE2SO
=cut
sub Feature2So
{
	my $self = shift;
	my ( $f ) = @_;
	if ( exists $GenbankTags::H_FEATURE2SO { $f } )
	{
		return $GenbankTags::H_FEATURE2SO { $f };
	}
	else
	{
		return $f;
	}
}
=head2 function GetLocusClass

 Title    : GetLocusClass
 Usage    : my $class = $o_gbk -> GetLocusClass( $feature );
 Function : return the type of Locus object corresponding to the genbank feature term passed by parameter
 Args     : $feature A string, should be a genbank/embl feature
 Return   : The Locus::XXX class, or "" if none
 
 Global   : $RH_CLASSES; 

=cut
sub GetLocusClass
{
	my $self = shift;
	my ($feature) = @_;
	if ( exists ( $GenbankTags::RH_CLASSES -> {$feature} ) )
	{
		return $GenbankTags::RH_CLASSES -> {$feature};
	}
	else
	{
		return "";
	}
}

=head2 IsTagAGenbankQualifier

 Title    : IsTagAGenbankQualifier
 Usage    : if ( $o_gbk -> IsTagAGenbankQualifier( $feature, $qualifier )
 Function : return &TRUE if the tag is a genbank qualifier legal for the feature
 Args     : $feature A genbank feature (if not, throw an exception)
            $qualifier A genbank qualifier
 Return   : &TRUE/&FALSE
 NOTE     : We check ONLY the existence of the qualifier, NOT its type
 
 Global   : %H_QUALIFIERS

 
=cut
sub IsTagAGenbankQualifier
{
	my $self = shift;
	my ($feature, $qualifier) = @_;
	
	if (exists $GenbankTags::H_QUALIFIERS{$feature})
	{
		if ( exists $GenbankTags::H_QUALIFIERS{$feature} -> { $qualifier } )
		{
			return &TRUE;
		}
		else
		{
			return &FALSE;
		}
	}
	else
	{
		throw LipmError::ParameterException('feature','value', -text => "The feature $feature does not exist\n");
	}
}

=head2 IsAType

 Title    : IsAType
 Usage    : if ( $o_gbk -> IsAType ( $feature, $type )
 Function : return &TRUE if the $type parameter is a recorded as a type legal for the feature
            this is used only for some features, namely misc_feature (we want to give precision to the word 'misc')
 Args     : $feature A genbank feature
            $type A string
 Return   : &TRUE/&FALSE
 
 Global   : %H_TYPES
 
=cut
sub IsAType
{
	my $self = shift;
	my ($feature, $type) = @_;
	if ( exists ($GenbankTags::H_TYPES { $feature } -> { $type } ) )
	{
		return &TRUE;
	}
	else
	{
		return &FALSE;
	}
}

=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;
