#
# $Id: Locus.pm 1347 2014-01-13 08:16:50Z manu $
#

package Locus;

use strict;
use warnings;
use Data::Dumper;
use Error::Simple;
use LipmError::ObjectException;

use base qw( LipmObject );

# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=head1 NAME

Locus - describes any locus on a sequence

=head1 SYNOPSIS

 my $o_locus = New Locus('start'=>integer,'end'=>integer,'strand'=>'-1|1|+|-','locus_tag'=>string,parent=>$o_locus);

 

=head1 DESCRIPTION

 the locus properties are classified in three classes:
     - Mandatory        : start, end, strand
     - Optional         : locus_tag, parent (the parent locus tag), etc. See _Init for the details
     - Other properties : Any other parameter passed by the constructor is stored in the hash __rh_other
     
 A locus is identified by at least a start, end positions on the genomic sequence and generally by a locus_tag
 The locus_tag should be  - but is not always uniq, so we manage the possibility to get the same locus_tag for several loci
 The parent is another locus inside which the current locus lives, ex:
 cds --> mRNA --> gene --> contig
 The last parent in this chain must be an object with following methods:
     - GetContigLabel()
     - GetSpecies()
     - GetKingdom()
 Typically any object deriving from AnnotatedSeq is OK
 A Locus may also maintain an array of "regions". Each region is itself a locus. This is well adapted to the Cds, ExonSet, Utr types

=cut

use Carp;
use Error::Simple;
use LipmError::ParameterException;
use LipmObject;
use General;
use base qw( LipmObject );

BEGIN
{
    our $VERSION = do {my @r = (q$Rev: 1347 $ =~ /\d+/g); $r[0]};
}

# used by Locus::_GetBioPerlRegions
sub _GetSplitType() { return 'join'; }

=head2 procedure InitAllowedOverlap

 Title     : InitAllowedOverlap
 Procedure : The default behaviour is : NO OVERLAP between 2 regions of a locus
             However, it may be useful to allow for some overlap between two regions
             Ex: Allowing an overlap of 2 in the CDS to allow for translational frameshifts
 NOTE      : This is initialized for EVERY Locus derived objects (ie for CdsRegions as well as for ExonRegions, etc)
 Usage     : &Locus::InitAllowedOverlap ( 2 );
 Args      : The authorized overlap: 1 or 2

=cut
# The allowed overlap between two regions in a locus
our $ALLOWED_OVERLAP = 0;
sub InitAllowedOverlap
{
	my ($overlap) = @_;
	if ( $overlap!=1 and $overlap!=2 )
	{
		throw LipmError::ParameterException( 'allowed_overlap', 'value', -text => 'allowed overlap should be 1 or 2' );
	}
	$Locus::ALLOWED_OVERLAP = $overlap;
}

=head2 procedure InitGenerateId

 Title     : InitGenerateId
 Procedure : The default behaviour is : GetId returns 'ID' if possible, NarcisseId if ID is empty
             However, it may be useful to return only NarcisseId, this may be necessary to get gff3 validated files, as NarcisseId is uniq
 NOTE      : This is initialized for EVERY Locus derived objects
 Usage     : &Locus::InitGenerateId ( 1/0 );

=cut
our $GENERATE_ID = 0;
sub InitGenerateId
{
	my ($flg) = @_;
	$Locus::GENERATE_ID = $flg;
}

=head2 procedure InitAssembly

 Title     : InitAssembly,ClearAssembly
 Procedure : Init or reinit the O_ASSEMBLY variable
 Usage     : &Locus::InitAssembly($o_assembly);
 Args      : $o_assembly: An initialized Assembly object

=cut
our $O_ASSEMBLY = undef;
sub InitAssembly
{
	my ($o_assembly) = @_;
	if ( ref ($o_assembly ne 'Assembly' ) )
	{
		throw LipmError::ParameterException( 'o_assembly', 'type', -text => 'The type of the parameter should be Assembly' );
	}
	$Locus::O_ASSEMBLY = $o_assembly;
}
sub ClearAssembly
{
	$Locus::O_ASSEMBLY = undef;
}
=head2 procedure GetAssembly

 Title     : GetAssembly

=cut
sub GetAssembly
{
	return $Locus::O_ASSEMBLY;
}

=head2 procedure InitEnzymeValidation

 Title     : InitEnzymeValidation
 Procedure : Init the O_ENZYMVAL variable, unless already initialized
 Usage     : &Locus::InitEnzymeValidation($o_enz_val);
 Args      : An initialized EnzymeValidation object

=cut
our $O_ENZYMVAL = undef;
sub InitEnzymeValidation
{
	my ($o_enz_val) = @_;
	if ( ref ($Locus::O_ENZYMVAL) ne '' )
	{
		throw Error::Simple ("Locus::O_ENZYMVAL is ALREADY INITIALIZED !");
	}
	if ( ref ($o_enz_val ne 'EnzymeValidation' ) )
	{
		throw LipmError::ParameterException( 'o_enz_val', 'type', -text => 'The type of the parameter should be EnzymeValidation' );
	}
	$Locus::O_ENZYMVAL = $o_enz_val;
}

=head2 procedure GetCount

 Title    : GetCount
 Function : Auto-incremented counter, specific to some subclass: 
 Usage    : my $count = $self -> GetCount();
            This increments the count and returns the new value
 Args     : none

=cut
our %H_COUNTS;
sub GetCount
{
	my $self  = shift;
	my $class = ref $self;
	$Locus::H_COUNTS{$class} += 1;
	return $Locus::H_COUNTS{$class};
}

=head2 procedure ResetCount

 Title    : ResetCount

=cut
sub ResetCount
{
	my $self  = shift;
	my $class = ref $self;
	delete $Locus::H_COUNTS{$class};
}

# Keep here the locus_tags AND the FIRST object associated to each locus_tag
# This is used by GetId() to return a uniq identifier, which is the locus tag when possible,
# the NarcisseId when not possible
#our %H_LOCUS_TAGS;

=head2 procedure _Init

 Title        : _Init
 Function     : Constructor, initialize the object, do not use directly, use New method instead
 Usage        : my $o_locus = New Locus::XXX('start'=>10,'end'=>100,'strand'=>-1,'locus_tag'=>'GNI','product'=>'enzyme');
                my $o_locus = New Locus::XXX('locus_tag'=>'GNI',$o_other_locus); 
                my $o_locus = New Locus::XXX('locus_tag'=>'GNI',$o_annot);

 Prerequisite : start < end, strand =~ /1|\-1|\+|\-/
 Returns      : none
 Args         : %h_param A lot of (key,value pairs), some are mandatory, some are optional:
                   Mandatory: start      position of the start (pb) of the object on the contig
                   Mandatory: end        position of the end (bp) of the object on the contig
				   Mandatory: strand     identify the strand on which is located the object; values=1/-1|+/-
				   Optional : sub_type   sub_type of the object (type of gene, type of rna, type of marker, ...)
				   Optional : -fragment_info => [ 'E' ] or [ 'S' ] or ['E', 'S' ]
				   Optional : any other pair of key/value
				   
				WARNING ! %h_param is MODIFIED BY THIS FUNCTION !!!
				$o_other_locus: Optional - An already initialized Locus object
				OR
				$o_annot: Optional - The AnnotatedSeq object inside which we live
 NOTES        : 1/ The attributes of the new Locus are initialized as follows:
                       -Init mandatory attributes from $o_other_locus if possible, else from %h_param
                       -Init other attributes from %h_param
                2/ The attributes initialized from %h_param belong to the hash $self->{_d_rh_core} (for some Reserved attributes) or to the hash $self->{_d_rh_other}
                   _d_rh_core  attributes should be retrieved by specialized getters (see GetLocusTag etc.) BUT can be retrieved by Get/GetAsString
                   _d_rh_other attributes should be ONLY retrieved by the Get and GetAsString functions
                3/ If $O_ASSEMBLY is defined, it is used to correct the start, end, strand attributes 
                   UNLESS THEY ARE INITIALIZED BY $o_locus as in this case the correction has already be done
                   If you specify $o_annot as a last parameter, the assembly will NOT BE PERFORMED
                   This is just a trick necessary in some special cases to disable the assembly (rather dirty in fact), see Locus::Mrna::__CreateCDS
 Globals      : $Locus::O_ASSEMBLY, @Locus::A_RESERVED_ATTRIBUTES %Locus::H_CORRESPONDANCE

=cut

# The reserved attributes are taken from the many file types supported (genbank, gff3)
our @A_IGNORED_ATTRIBUTES  = qw( translation phase );
our %H_RESERVED_ATTRIBUTES = ( ID=>1, gene=>1, Name=>1, locus_tag=>1, note=>1, Note=>1, score=>1, product=>1, db_xref=>1, Dbxref=>1, Alias=>1, gene_synonym =>1,protein_id=>1, transcript_id=>1 );
our @A_RESERVED_ATTRIBUTES = keys %H_RESERVED_ATTRIBUTES;
our %H_CORRESPONDANCE      = ( Name => 'gene', Note => 'note', db_xref => 'Dbxref', Alias => 'gene_synonym' );
                          
sub _Init
{
	my $self    = shift;
	my $o_locus = (ref($_[-1]) =~ /^Locus|AnnotatedSeq/) ? pop(@_) : undef;
	my %h_param = @_;
	
	$self -> { _d_rh_core }  = {};
	$self -> { _d_rh_other } = {};
	
	# Mandatory parameters
	my ( $start, $end, $strand );
	
	# fragment_info, should be set as early as possible
	if ( defined $h_param { -fragment_info } )
	{
		my $ra_fi = $h_param { -fragment_info };
		foreach my $fi (@$ra_fi)
		{
			$self -> SetFragment($fi);
		}
		delete $h_param { -fragment_info };
	}
	
	# Initialized from the other locus if any
	if ( defined $o_locus && $o_locus -> isa ( 'Locus' ) )
	{
		( $start, $end ) = $o_locus -> GetEnds();
		$strand          = $o_locus -> GetStrand();
	}
	
	# Initialized from the parameters
	else
	{
		# check to see if all required fields are here
		( $start, $end, $strand ) = @h_param { qw( start end strand ) };
		foreach my $required ( qw(start end strand) )
		{
			if ( !defined($required) )
			{
				throw LipmError::ParameterException( $required, 'missing');
			}
		}

		# Correct data for the assembly	if $O_ASSEMBLY is specified but NOT $o_locus specified
		# if we are here and $o_locus is defined, $o_locus is an AnnotatedSeq objest. This is the dirty trick
		if ( !defined($o_locus) && defined($Locus::O_ASSEMBLY) )
		{
			$start = $Locus::O_ASSEMBLY -> Relative2Absolute( $start );
			$end   = $Locus::O_ASSEMBLY -> Relative2Absolute( $end );
			if ( $strand == 0 )
			{
				throw Error::Simple( "ERROR - Strand not correctly initialized during assembly !" );
			}
			if ( $Locus::O_ASSEMBLY->GetStrand() eq '-' )
			{
				$strand = -$strand;
				my $tmp = $start;
				$start  = $end;
				$end    = $tmp;
				##$self->__SwapFragmentInfo();
			}
		}
		
		# Delete them from %h_param
		delete $h_param { start  };
		delete $h_param { end    };
		delete $h_param { strand };
		
		# EC_number validation, if wanted
		if ( defined $Locus::O_ENZYMVAL )
		{
			my $key="";
			$key='EC_number' if ( exists $h_param{'EC_number'} );
			$key='eC_number' if ( exists $h_param{'eC_number'} );
			if ( $key ne "" )
			{
				my $ec_number = $h_param{$key};
				my $ra_t_ec_number;
				if ( ref($ec_number) eq 'ARRAY' )
				{
					#$ec_number = join(',',@$ec_number);
					$ra_t_ec_number = $Locus::O_ENZYMVAL->Check(@$ec_number);
				}
				else
				{
					$ra_t_ec_number = $Locus::O_ENZYMVAL->Check($ec_number);
				}
				$h_param{$key}=$ra_t_ec_number;
			}
		}
	}

	# Setting the sub_type if necessary, using SetSubType for the sanity checks
	if ( exists($h_param{sub_type}) )
	{
		my $sub_type = $h_param {sub_type};
		delete $h_param { sub_type };
		$self -> SetSubType ( $sub_type );
	}
	else
	{
		$self -> SetDefaultSubType();
	}

	# Sanity checks
	if ( $start > $end )
	{
		my $sub_type = $self -> GetSubType();
		my $locus_tag = $self -> GetLocusTag();
		throw LipmError::ParameterException( 'start and end','value',-text => "$sub_type $locus_tag: start should be < end (start=$start end=$end)" );
	}
	if ( $strand != 1 && $strand != -1 && $strand != 0 )
	{
		my $sub_type = $self -> GetSubType();
		my $locus_tag = $self -> GetLocusTag();
		throw LipmError::ParameterException( 'strand','value',-text => "$sub_type $locus_tag: $start..$end: strand should be one of -1,1 (not $strand )" );
	}
	
	# Keep required data in _d_rh_core hash
	$self -> { _d_rh_core } -> { start } = $start;
	$self -> { _d_rh_core } -> { end }   = $end;
	$self -> { _d_rh_core } -> { strand } = $strand;

	# the 'ignored' attributes (very few, they are computed, not read)
	foreach my $ignored ( @Locus::A_IGNORED_ATTRIBUTES )
	{
		delete $h_param{$ignored};
	}
	
	# The 'reserved' attributes
	foreach my $reserved ( @Locus::A_RESERVED_ATTRIBUTES )
	{
		if ( defined $h_param{$reserved} )
		{
			# replace Note by note etc.
			# NB We DO NOT consider pathologic cases where we have Note="...";note="...", the second will replace the first !
			my $key = (exists $Locus::H_CORRESPONDANCE{$reserved}) ? $Locus::H_CORRESPONDANCE{$reserved} : $reserved;
			$self -> { _d_rh_core } -> { $key } = $h_param{$reserved};
			delete $h_param{$reserved};
        }
	}

	my $rh_other = $self -> { _d_rh_other };
	@$rh_other{ keys(%h_param) } = values(%h_param);
	
	#my @a_keys = keys %h_param;
	#if ( @a_keys =! 0 )
	#{
	#	$self -> { _d_rh_other } = \%h_param;
	#}

	# The parent
    $$self{ __o_parent } = undef;
}

=head2 procedure RelCoords

 Title        : RelCoords
 Usage        : $o_locus->RelCoords();
 Prerequisites: The locus must have been completely created with all its children (NOT CHECKED)
                The locus coordinates must be Relative to the current contig (NOT CHECKED)

                The InitAssembly function must have been called, and the current contig must be correctly selected
 Procedure    : Using Assembly::Absolute2Relative, change the start/end positions of this locus AND all its children
                relative to the current contig
 
 Args         : $inside bool If &TRUE (the default) the procedure is recursively called for all children
                             If &FALSE only this object is modified
 Globals      : $Locus::O_ASSEMBLY

=cut
sub RelCoords
{
	my $self = shift;
	my ($inside) = @_;
	$inside = &TRUE unless defined $inside;

	if ( !defined $Locus::O_ASSEMBLY )
	{
		throw Error::Simple( "ERROR - Please call Locus::InitAssembly before calling AbsCoords !" );
	}

	my ($start,$end) = $self -> GetEnds();
	my $strand = $self -> GetStrand();
	$start = $Locus::O_ASSEMBLY -> Absolute2Relative( $start );
	$end   = $Locus::O_ASSEMBLY -> Absolute2Relative( $end );

	if ( $Locus::O_ASSEMBLY->GetStrand() eq '-' )
	{
		$strand = -$strand;
		my $tmp = $start;
		$start  = $end;
		$end    = $tmp;
		##$self->__SwapFragmentInfo();
	}

	#print STDERR $self->GetpkExtended()."......";
	# We do not have any Setter for those data
	$self -> { _d_rh_core } -> { start } = $start;
	$self -> { _d_rh_core } -> { end }   = $end;
	$self -> { _d_rh_core } -> { strand } = $strand;
	#print STDERR $self->GetpkExtended()."\n";
	
	# Call recursively the function for all children
	if ( $inside )
	{
		my $ra_children = $self->GetAllChildren();
		for my $o_child ( @$ra_children )
		{
			$o_child -> RelCoords(&FALSE);
		}
	}
}

=head2 procedure AbsCoord

 Title        : AbsCoords
 Usage        : $o_locus->AbsCoords();
 Prerequisites: The locus must have been completely created with all its children (NOT CHECKED)
                The locus coordinates must be Absolute (ie relative to the chromosome) (NOT CHECKED)
                The InitAssembly function must have been called, and the current contig must be correctly selected
 Procedure    : Using Assembly::Relative2Absolute, change the start/end positions of this locus AND all its children
                relative to the current contig
 
 Args         : $inside bool If &TRUE (the default) the procedure is recursively called for all children
                             If &FALSE only this object is modified
 Globals      : $Locus::O_ASSEMBLY

=cut
sub AbsCoords
{
	my $self = shift;
	my ($inside) = @_;
	$inside = &TRUE unless defined $inside;

	if ( !defined $Locus::O_ASSEMBLY )
	{
		throw Error::Simple( "ERROR - Please call Locus::InitAssembly before calling AbsCoords !" );
	}

	my ($start,$end) = $self -> GetEnds();
	my $strand = $self -> GetStrand();
	$start = $Locus::O_ASSEMBLY -> Relative2Absolute( $start );
	$end   = $Locus::O_ASSEMBLY -> Relative2Absolute( $end );

	if ( $Locus::O_ASSEMBLY->GetStrand() eq '-' )
	{
		$strand = -$strand;
		my $tmp = $start;
		$start  = $end;
		$end    = $tmp;
		##$self->__SwapFragmentInfo();
	}

	#print STDERR $self->GetpkExtended()."......";
	# We do not have any Setter for those data
	$self -> { _d_rh_core } -> { start } = $start;
	$self -> { _d_rh_core } -> { end }   = $end;
	$self -> { _d_rh_core } -> { strand } = $strand;
	#print STDERR $self->GetpkExtended()."\n";
	
	# Call recursively the function for all children
	if ( $inside )
	{
		my $ra_children = $self->GetAllChildren();
		for my $o_child ( @$ra_children )
		{
			$o_child -> AbsCoords(&FALSE);
		}
	}
}

=head2 procedures and function Select,Deselect,IsSelected

 Title        : Select,Deselect,IsSelected
 Usage        : $o_locus->Deselect($inside);
                $o_locus->Select($inside);
                if ( $o_locus -> IsSelected() ... )
 Procedures   : Set or remove the 'deseelected' flag
 Function     : Return a boolean, it true the locus is selected, if false it is not
                When a Locus is created, it is selected (the __deselected attribute is not set, so that the default does
                not increase memory print)
                Deselect sets the attribute, Select removes it
                Deselect and Select calls the same method on the embedded $o_locus:
                     - When a Gene is selected/deselected, all Mrnas or Ncrnas children of this gene are selected/deselected, too
                     - When an Mrna is selected/deselected, the ExonSet/Cds/Utr objects inside it are selected/deselected
                     - When an ExonSet/Utr/Cds is selected/deselected, the corresponding ExonSet/Cdsregion/UtrRegion are selected/deselected, too
                Deselect/Select are NOT defined for CdsRegion, ExonSet, UtrRegion, because it does not make sense to select/deselect a single region
                IsSelected is defined but calls the Parent's IsSelected
                IsSelected should be called by the AnnotatedSeqFileWriter::XXX::Dump methods (any format)
 Prerequisite : All the biological obejcts should be loaded, because the following could happen:
                    - Deselect a $o_gene object
                    - Create a new mrna with the AddmRNA method
                    - The gene will be deselectd, but NOT the mrna, which could lead to badly formatted files (for instance 
                    a Parent=xxx in gff3 without any corresponding ID= field)
 
 Args         : $inside bool If &TRUE (the default) All objects inside this object are selected
                             If &FALSE Only THIS object is selected/deselected
                             WARNING this DOES NOT work for EVERY locus type:
                               this does not work for CDS, ExonSet, Utr, because it should be a nonsense selecting only the container
                               this does not work for Repeat because there nothing inside a repeat
                               this works for Gene, Mrna, Rna, MiscRegion
 Globals      : none

=cut
sub Select
{
	my $self = shift;
	throw LipmError::ObjectException($self,'abstract method');
}
sub Deselect
{
	my $self = shift;
	throw LipmError::ObjectException($self,'abstract method');
}	
sub IsSelected
{
	my $self = shift;
	return ( defined $self -> { __deselected  } ) ? &FALSE:&TRUE; 
}

=head2 procedure SetParent

 Title        : SetParent
 Usage        : $o_locus->setParent( $o_parent );
 Procedure    : Set the parent to $o_parent
 Prerequisite : The parent should not already set, (it is possible to set the parent only once - see ResetParent for an exception to this rule)
                The method could be overloaded in the derived class, to control the parent type
                BUT YOU SHOULD CALL $self->SUPER::SetParent from the overloaded function
                In the Locus class, the parent could be set to any object, the only restrictions are:
                - The parent should 'can' GetParent (thus should be a Locus object)
                - OR the parent shoud 'can' GetContigLabel, GetSpecies and GetKingdom 
                However this is not checked here
 
 Args         : $o_parent : The parent locus
                $adj_boundaries (def &TRUE): If TRUE, adjust the parent boudaries
 Globals      : none

=cut
sub SetParent
{
	my $self      = shift;
	my ($o_parent, $adj_boundaries) = @_;
	$adj_boundaries = &TRUE if ( ! defined $adj_boundaries );
	
	if ( defined $self -> { __o_parent } )
	{
		throw LipmError::ObjectException($self,'!attribute', -text => "Parent is already set - " . $self -> GetpkExtended() );
	}
	else
	{
		$self -> { __o_parent } = $o_parent;
	}
	
	if ( ! defined $self -> GetAncestor() )
	{
		throw LipmError::ObjectException($self,'!attribute', -text => "Missing link in the parents chain, may be the " . $self -> GetSubType() . " is declared BEFORE the gene - " . $self -> GetpkExtended() );
	}
	
	# Adjust the parent boudaries
	$o_parent -> AdjustBoundaries( $self ) if ( $adj_boundaries == &TRUE );
	
	# Update the sequence-region of the Ancestor
	$self -> GetAncestor() -> UpdateSeqRegion( $self -> GetEnds() );
	
	# set the parent of the regions - Could not be done before !
	my @a_regions = $self -> GetRegions();
	foreach my $o_region (@a_regions)
	{
		# GetRegions may return the locus itself, and the parentship may be already done (especially if called from ResetParent)
		if ( $o_region != $self && ! defined ($o_region -> GetParent()) )
		{
			$o_region -> SetParent ( $self );
		}
	}
}

=head2 procedure AddTo

 Abstract function - Each top-level child implements this function
 Call one the AddGene, AddNcRna, AddRepeat, AddMiscRegion in AnnotatedSeq
 
 Usage: $o_locus -> AddTo ( $o_annot );

=cut
sub AddTo { my $self = shift; throw LipmError::ObjectException($self,'abstract method'); };

=head2 procedures ResetParent,_ResetParent

 Title        : ResetParent
 Usage        : $o_locus->ResetParent( $o_parent );
 Procedure    : Clear __o_parent and call AddTo
 Prerequisite : The parent should already be set to an AnnotatedSeq object (usable ONLY for high level loci)
 
 Args         : $o_parent : The new parent, an AnnotatedSeq object
 Globals      : none
 NOTE         : Use this function with caution !
                Data are no more coherent, as a resetted locus is STILL DECLARED in its old parent !
                The CALLING FUNCTION must correct this
                
=cut

sub ResetParent
{
	my $self      = shift;
	my ($o_parent) = @_;
	
	if ( ref($o_parent) ne 'AnnotatedSeq' )
	{
		throw LipmError::ParameterException( 'o_parent', 'type', -text => 'should be an AnnotatedSeq object' );
	}
	if ( !defined $self -> { __o_parent } )
	{
		throw LipmError::ObjectException($self,'!attribute', -text => "Parent NOT YET set - " . $self -> GetpkExtended() );
	}
	my $o_old_parent = $self->{ __o_parent };
	if ( ref($o_old_parent) ne 'AnnotatedSeq' )
	{
		throw LipmError::ParameterException( '!attribute', -text => "The current parent should be an AnnotatedSeq, not a " . ref($o_old_parent) . " - " . $self->GetpkExtended() );
	}
	$self -> { __o_parent } = undef;
	$self->AddTo($o_parent);
}

=head2 procedure AdjustBoundaries

 Title    : AdjustBoundaries
 Procedure: Adjust the boundaries of this locus, (start/end) insuring the child is INSIDE the parent
            this is done recursively, until the ancestor is reached (see AnnotatedSeq::AdjustBoundaries)
            TODO ==> Throw an exception is the child's strand is not the same as my strand
                     Le pb est dans le traitement d'erreur: que faire de ces objets ?
 Usage    : $self -> AdjustBoundaries ( $o_child);
 Args     : $o_child, the child used to adjust the boundaries - This SHOULD BE a child of $self (not checked !)

=cut

sub AdjustBoundaries
{
	my $self    = shift;
	my ($o_child) = @_;
	#print STDERR "BOUDARIES SELF=" . ref($self) . " CHILD=" . ref($o_child) . "\n";
	
	if ( $self -> {_d_rh_core } -> {strand} == 0 && $o_child -> {_d_rh_core } -> {strand} == 0 )
	{
		my $pks = $self -> GetpkExtended();
		my $pkc = $o_child -> GetpkExtended();
		throw Error::Simple ( "parent = $pks - child = $pkc - both strands unknown" );
	}
	elsif ( $self -> {_d_rh_core } -> {strand} == 0 && $o_child -> {_d_rh_core } -> {strand} != 0 )
	{
		my $pks = $self -> GetpkExtended();
		my $pkc = $o_child -> GetpkExtended();
		$self -> {_d_rh_core } -> {strand} = $o_child -> {_d_rh_core } -> {strand};
		warn $self ->GetWarningHeader() . "parent = $pks - child = $pkc - forcing to the child strand\n";

	}
	elsif ( $self -> {_d_rh_core } -> {strand} != 0 && $o_child -> {_d_rh_core } -> {strand} == 0 )
	{
		my $pks = $self -> GetpkExtended();
		my $pkc = $o_child -> GetpkExtended();
		$o_child -> {_d_rh_core } -> {strand} = $self -> {_d_rh_core } -> {strand};
		warn $self ->GetWarningHeader() . "parent = $pks - child = $pkc - forcing to the parent strand\n";
	}
	elsif ( $self -> {_d_rh_core } -> {strand} != $o_child -> {_d_rh_core } -> {strand} )
	{
		my $pks = $self -> GetpkExtended();
		my $pkc = $o_child -> GetpkExtended();
		#throw Error::Simple ( "parent = $pks - child = $pkc - both strands known, but different" );
		warn $self ->GetWarningHeader() . "parent = $pks - child = $pkc - both strands different, forcing to the parent strand\n";
	}
		
	# If necessary, change the start/end of $self
	my $changed = &FALSE;
	if ( $self -> {_d_rh_core } -> { start } > $o_child -> {_d_rh_core } -> { start } )
	{
		$self -> {_d_rh_core } -> { start } = $o_child -> {_d_rh_core } -> { start };
		$changed = &TRUE;
	}
	if ( $self -> {_d_rh_core } -> { end } < $o_child -> {_d_rh_core } -> { end } )
	{
		$self -> {_d_rh_core } -> { end } = $o_child ->{_d_rh_core } -> { end };
		$changed = &TRUE;
	}
	
	# If something was changed, propagate the change through the parents chain
	$self -> GetParent() -> AdjustBoundaries ( $self) if ( $changed );
}

=head2 procedure RevComp

 Title        : RevComp
 Procedure    : "Reverse complement" the locus, ie compute the coordinates on the other strand
                NOTE - The children of this locus are NOT modified, so that you may get inconsistent results using this method !
 Usage        : my $loc_id = $o_locus->revComp($seq_length)
 Args         : optional: The sequence length, if not passed get the sequence length from the ancestor, throwing an exception if 0
                It is better passing the sequence length, because the calls to GetAncestor may be somewhat tedious

=cut
sub RevComp
{
	my $self = shift;
	my ($seq_length) = @_;
	
	if ( not defined ($seq_length) )
	{
		$seq_length = $self -> GetSequenceLength();
		if ( $seq_length==0 )
		{
			throw Error::Simple ("Locus::RevComp: the sequence could not be loaded, cannot kown its length");
		}
	}

	my ($s,$e) = $self -> GetEnds();
	$self->{_d_rh_core } -> {'start'} = $seq_length - $e + 1;
	$self->{_d_rh_core } -> {'end'}   = $seq_length - $s + 1;
	$self->{_d_rh_core } -> {'strand'}= -$self->{_d_rh_core } -> {'strand'};
}

=head2 procedure NewOrigin

 Title        : NewOrigin
 Usage        : $o_locus -> NewOrigin ( $new_origin, $seq_length);
 Procedure    : Change the origin of the coordinates. For x=$start and $end, do:
                     if ( $x >= $origin ) then $x = $x - $new_origin + 1;
                     else $x = $seq_length - $new_origin + $x + 1;
                Throw an exception if the locus is on both sides of the replication origin
 Args         : required: $new_origin
                optional: The sequence length, if not passed get the sequence length from the ancestor, throwing an exception if 0
                It is better passing the sequence length, because the calls to GetAncestor may be somewhat tedious

=cut
sub NewOrigin
{
	my $self = shift;
	my ($new_origin, $seq_length) = @_;
	
	if ( not defined ($seq_length) )
	{
		$seq_length = $self -> GetSequenceLength();
		if ( $seq_length==0 )
		{
			throw Error::Simple ("Locus::RevComp: the sequence could not be loaded, cannot kown its length");
		}
	}

	my ($s,$e) = $self -> GetEnds();
	if ( $s >= $new_origin && $e >= $new_origin )
	{
		$self->{_d_rh_core } -> {'start'} = $s - $new_origin + 1;
		$self->{_d_rh_core } -> {'end'}   = $e - $new_origin + 1;
	}
	elsif ( $s < $new_origin &&  $e < $new_origin )
	{
		$self->{_d_rh_core } -> {'start'} = $seq_length - $new_origin + $s + 1;
		$self->{_d_rh_core } -> {'end'}   = $seq_length - $new_origin + $e + 1;
	}
	else
	{
		throw Error::Simple ("Locus::NewOrigin: the locus is on bothsides of $new_origin - " . $self->GetpkExtended() . " (NewOrigin=$new_origin)");
	}
}

=head2 function GetLocusTag

 Title        : GetLocusTag
 Function     : Return the locus tag
 Usage        : my $loc_id = $o_locus->GetLocusTag();
 Returns      : The locus tag
 Args         : none

=cut
sub GetLocusTag
{
	my $self = shift;
	return $self -> Get ( 'locus_tag' );
}

=head2 function SetLocusTag

 Title     : SetLocusTag
 Procedure : Set the locus tag, the current locus_tag being moved to old_locus_tag
 Args      : The new locus tag
 
=cut
sub SetLocusTag
{
	my $self = shift;
	my ($new_locus_tag) = @_;
	
	my $old_locus_tag = $self->GetLocusTag();
	return if ( $old_locus_tag eq $new_locus_tag );
	if ( $old_locus_tag ne '' )
	{
		$self -> AddValue ( 'old_locus_tag', $old_locus_tag);
		$self -> Delete   ( 'locus_tag' );
	}
	$self -> Set ( 'locus_tag', $new_locus_tag );
}

=head2 function GetParentLocusTag

 Title        : GetParentLocusTag
 Function     : Walk through the parent chain, returning the first non-blank locus tag
 Usage        : my $p_loc_id = $o_locus->GetParentLocusTag();
 Returns      : The Parent locus tag, or ""
 Args         : none

=cut
sub GetParentLocusTag
{
	my $self = shift;
	
	my $o_locus = $self;
	my $p_locus_tag = "";
	#print ref($self).' ';
	while ((my $o_parent=$o_locus->GetParent())->isa("Locus"))
	{
		$p_locus_tag = $o_parent->GetLocusTag();
		#print $p_locus_tag . " # ";
		last if ($p_locus_tag ne "");
		$o_locus=$o_parent;
	} 
	#print "\n";
	return $p_locus_tag;
}

=head2 function GetId

 Title        : GetId
 Function     : Return the ID if defined, else the NarcisseId (data come from a gff3 file)
                Else, return the NarcisseId
 Usage        : my $id = $o_locus->GetId();
 Returns      : the locus tag or the id
 Args         : none

=cut
sub GetId
{
	my $self = shift;
	my $id   = ($Locus::GENERATE_ID==&FALSE) ? $self -> Get ('ID' ) : $self -> GetNarcisseId();
	return ( $id ne "" ) ? $id : $self->GetNarcisseId();
}

=head2 function GetParentId

 Title     : GetParentId
 Function  : Returns Id of the parent, or "" if already at top of the chain
 Usage     : my $p_id = $o_locus->GetParentId();
 Returns   : the parent's id, or ""
 Args      : none
 NOTE      : GetParentId is overloaded in some derived classes (ie CdsRegion, ExonSet, UtrRegion)    

=cut
sub GetParentId	
{
	my $self = shift;
	my $o_parent = $self -> GetParent();
	return $o_parent->can("GetId")?$o_parent->GetId():'';
}

=head2 function GetGeneNameOrLocusTag
 Title     : GetGeneNameOrLocusTag
 Function  : Return the 'gene' attribute if any, the 'locus tag' attribute if not, and maybe GetId as a last resort (see arg)
 Arg      : flg_id = optional, defaults to  &FALSE, '' may be returned
            if &TRUE, return GetId as a last resort, so that '' will never returned
            TODO - SUPPRIMER CET ARGUMENT, IL N'EST ACTUELLEMENT PAS UTILISE
 Return   : gene name of locus tag

=cut
sub GetGeneNameOrLocusTag
{
	my $self = shift; 
	my ($flg_id) = @_;
	$flg_id = &FALSE unless ( defined $flg_id );
	
	return (($flg_id == &TRUE) ? 
		$self -> Get ( 'gene' ) || $self -> Get ( 'locus_tag' ) || $self -> GetId() :
		$self -> Get ( 'gene' ) || $self -> Get ( 'locus_tag' ));
}

=head2 function GetGeneNameSimple
 Title     : GetGeneNameSimple
 Function  : Return the 'gene' member
  Return   : The gene name, or '' if no gene name found

=cut
sub GetGeneNameSimple
{
	my $self = shift;
	return $self -> Get ( 'gene' );
}

=head2 function GetGeneName
 Title    : GetGetGeneName
 Function : Call the parent's GetgeneName if possible, or GetGeneNameOrLocusTag if at top of chain
 Usage    : my $gene_name = $o_locus -> GetGeneName();
 Return   : The gene name, the locus_tag, or ''
 
=cut
sub GetGeneName
{
	my $self = shift;
	my $o_parent = $self -> GetParent();
	if ( defined($o_parent) && $o_parent -> can('GetGeneName') )
	{
		return $o_parent -> GetGeneName( @_ );
	}
	else
	{
		return $self->GetGeneNameOrLocusTag();
	}
}

=head2 function GetProduct,GetNote,GetDbxref,GetAlias

 Title        : GetProduct/GetNote/GetDbXref
 Function     : Return the product, note, dx_xref if found
                The tag may be provided by GFF3 (Note, Dbxref) or ny genbank/embl ( note, db_xref)
 Usage        : my $gene_name = $o_mcds->GetProduct();
 args         : If &TRUE, GetAsString
 Returns      : The product

=cut
sub GetProduct
{
	my $self = shift;
	my ($as_string) = @_;
	return ($as_string) ? $self -> GetAsString ( 'product' ) : $self -> Get ( 'product' );
}
sub GetNote
{
	my $self = shift;
	my ($as_string) = @_;
	return ($as_string) ? $self -> GetAsString ( 'note' ) : $self -> Get ( 'note' );
}
sub GetDbXref
{
	my $self = shift;
	my ($as_string) = @_;
	return ($as_string) ? $self -> GetAsString ( 'Dbxref' ) : $self -> Get ( 'Dbxref' );
}

# GetDbXref4gff3 = return dbxref terms, excluding GO or SO terms
sub GetDbXref4gff3
{
	my $self = shift;
	my ($as_string) = @_;
	
	my $ra_xref= $self->GetAsArrayConsolidated('Dbxref');
	my @a_xref1;
	foreach my $x (@$ra_xref)
	{
		push (@a_xref1,$x) unless ($x=~ /^[SG]O:/);
	}
	my @a_xref2 = sort @a_xref1;
	return ($as_string) ? join ',',@a_xref2 : \@a_xref2;
}

# getOntologyTerms = return dbxref terms belonging to GO or SO, and ontology_term
sub GetOntologyTerms
{
	my $self = shift;
	my ($as_string) = @_;

	
	my $ra_xref= $self->GetAsArrayConsolidated('Dbxref','ontology_term','Ontology_term');
	my @a_xref1;
	foreach my $x (@$ra_xref)
	{
		my @a_tmp = split ' ',$x; # sometimes SO:0000147 SO:0005845
		foreach my $y (@a_tmp)
		{
			push (@a_xref1,$y) if ($y=~ /^[SG]O:/);
		}
	}
	my @a_xref2 = sort @a_xref1;
	if ( $as_string == &TRUE )
	{
		my $rvl = join ',',@a_xref2;
		return $rvl;
	}
	else
	{
		return \@a_xref2;
	}
}

sub GetAlias
{
	my $self = shift;
	my $as_string = @_;
	return ($as_string) ? $self -> GetAsString ( 'gene_synonym' ) : $self -> Get ( 'gene_synonym' );
}

=head2 function GetEnds

 Title        : GetEnds,GetStart,GetEnd
 Function     : Return the locus ends (start,end)
 Usage        : my ($start,$end) = $o_locus->GetEnds();
 Returns      : The ends of the locus
 Args         : none
 Globals      : none

=cut
sub GetEnds
{
	my ($self) = @_;
	return ($self->{_d_rh_core } -> {'start'},$self->{_d_rh_core } -> {'end'});
}
sub GetStart
{
	my ($self) = @_;
	return $self->{_d_rh_core } -> {'start'};
}
sub GetEnd
{
	my ($self) = @_;
	return $self->{_d_rh_core } -> {'end'};
}

=head2 function setStart,SetEnd

 Title        : SetStart,SetEnd
 Procedure    : Change the start or end position of the locus, taking care of its coherency
 Usage        : $o_locus->SetStart($new_position);
 Args         : The new start or end position
 Globals      : none

=cut
sub SetStart
{
	my $self = shift;
	my ($start) = @_;
	my $end = $self->GetEnd();
	
	if ( $start > $end )
	{
		throw Error::Simple ("Locus::SetStart: start ($start) > end ($end) ! - " . $self->GetpkExtended() );
	}
	else
	{
		$self -> { _d_rh_core } -> { start } = $start;
		my $o_parent = $self -> GetParent();
		$o_parent -> AdjustBoundaries($self);
	}
}
sub SetEnd
{
	my $self = shift;
	my ($end) = @_;
	my $start = $self->GetStart();
	
	if ( $start > $end )
	{
		throw Error::Simple ("Locus::SetStart: start ($start) > end ($end) ! - " . $self->GetpkExtended() );
	}
	else
	{
		$self -> { _d_rh_core } -> { end } = $end;
		my $o_parent = $self -> GetParent();
			$o_parent -> AdjustBoundaries($self);
	}
}

=head2 function GetStrand

 Title        : GetStrand
 Function     : Returns -1/+1
 Usage        : my $strand = $o_locus->GetStrand( &FALSE );
 Returns      : The strand (-1/+1 or -/+)
 Args         : If &FALSE (default) return -1/+1/0
                If &TRUE return -/+/. (useful for GFF3)
 NOTE - GetStrand(&TRUE) may write a NON VALIDATED gff3 file (if '.' is returned)

=cut

sub GetStrand
{
	my $self   = shift;
	my ($gff3) = @_;
	$gff3 = &FALSE unless defined $gff3;
	
	my $strand = $self -> {_d_rh_core } -> {strand};
	
	if ($gff3 == &TRUE)
	{
		$strand = '-' if $strand eq '-1';
		$strand = '+' if $strand eq  '1';
		$strand = '.' if $strand eq  '0';
		return $strand;
	}
	else
	{
		$strand = -1 if $strand eq '-';
		$strand = +1 if $strand eq '+';
		$strand = 0  if $strand eq '.';
		return $strand+0;
	}
}
=head2 function Getpk

 Title        : Getpk
 Function     : Returns a key, unique for a given position (pk = Postion Key)
 Usage        : my $key = $o_locus->Getpk();
 Returns      : A key: "strandstart-end" (ex: '+1000-1100' or '-1000-1100')
 Args         : None

=cut
sub Getpk
{
	my $self = shift;
	return $self -> GetStrand(&TRUE) . $self -> GetStart() . '..' . $self -> GetEnd();
};
=head2 function GetpkExtended

 Title        : GetpkExtended
 Function     : Returns the pk together with gene and GetLocusTag() 
 Usage        : my $key = $o_locus->GetpkExtended();
 Returns      : A key: "strandstart-end-gene-locus" (ex: 'mRNA:+1000-1100-IGHV-LOC000012')
 Args         : None

=cut
sub GetpkExtended
{
	my $self = shift;
	my $type = ref($self);
	$type =~ s/Locus:://;
	my $n  = $self -> GetGeneNameSimple ();
	my $lt = $self -> GetLocusTag();
        if ( $n ne '' || $lt ne '' )
	{
		return $type . ':' . $self -> Getpk() . " ($n $lt)";
	}
	else
	{
		return $type . ':' . $self -> Getpk();
	}
};

=head2 function ComparePosition

 Title        : ComparePosition
 Function     : Compare the position of $self with the Locus passed by parameter
                Return 1, -1, 0
                $self       -----*...........*---------   )
                $o_locus    -------------------*....*--   ) Return -1 (no overlap)

                $self       ---------------------*--*--   )
                $o_locus    ---*.........*-------------   ) Return 1 (no overlap)
                
                $self       --------*.........*--------   )
                $o_locus    ---*.........*-------------   ) Return 0 (overlap)
 Note         : "No overlap" means No overlap at all when $Locus::ALLOWED_OVERLAP==0 (default behaviour)
 
 Usage        : if ( $o_locus->ComparePosition($o_other_locus) != 0 ) ...
 Returns      : 1, -1, 0
 Global       : $Locus::ALLOWED_OVERLAP
 See also     : __CmpRegions - __CompRegions throws an exception when there is an overlap
                InitAllowedOverlap
 Args         : The Locus compared

=cut
sub ComparePosition
{
	my $self = shift;
	my ($o_locus) = @_;

	my ($s,$e)   = $self    -> GetEnds();
	my ($sl,$el) = $o_locus -> GetEnds();
	
	return -1 if ($e<$sl+$Locus::ALLOWED_OVERLAP);
	return 1  if ($s>$el-$Locus::ALLOWED_OVERLAP);
	return 0;
}
=head2 function IsRegionOverlap

 Title        : IsRegionOverlap
 Function     : Return &TRUE if there is some overlap between one of the self's regions and one of the $o_locus's regions
                (in the following, ... shows a region, --- shows interregion sequences)
 
                $self       -*...*-----*.....*----*...*-   )
                $o_locus    ------*..*---*..*-----------   ) Return &TRUE

                $self       -*...*-----*.....*----*...*-   )
                $o_locus    ------*..*--------*..*------   ) Return &FALSE

 Prerequisite : The regions in both Locus must be sorted
 Usage        : if ( $o_locus->IsOverlapRegion($o_locus) ) ...
 Returns      : &TRUE, &FALSE
 Args         : The Locus compared

=cut
sub IsRegionOverlap
{
	my $self = shift;
	my ($o_locus) = @_;
	
	my @a_self_regions  = $self    -> GetRegions();
	my @a_locus_regions = $o_locus -> GetRegions();
	
	# May be we have this situation:
	#   $self    --*..*---*...*------------------
	#   $o_locus ------------------*..*---*...*-- 
	return &FALSE if ($a_self_regions[-1]->ComparePosition($a_locus_regions[0])==-1);

	# May be we have this situation:
	#   $o_locus --*..*---*...*------------------
	#   $self    ------------------*..*---*...*-- 
	return &FALSE if ($a_self_regions[0]->ComparePosition($a_locus_regions[-1])==1);
	
	# Both sets of regions are mixed
	my $i=0;
	foreach my $o_l_region (@a_locus_regions)
	{
		for ( my $j=$i; $j<@a_self_regions; $j++ )
		{
			my $c = $a_self_regions[$j]->ComparePosition($o_l_region);
			return &TRUE if ($c==0);
			last if ($c==1);
		}
	}
	return &FALSE;
}

=head2 GetOverlapNb

 Title        : GetOverlapNb
 Function     : Return the number of bases overlapping between $self and the locus passed by parameter
 Usage        : my $overlap = $o_locus -> GetOvelapNb ( $o_other_locus );
 Returns      : The number of overlapping bases, 0 if none or if different strands
 Args         : The Locus compared

=cut
sub GetOverlapNb
{
	my $self = shift;
	my ($o_locus) = @_;
	
	return 0 if ($o_locus -> GetStrand() != $self -> GetStrand());
	return 0 if ($o_locus -> GetEnd() < $self -> GetStart());
	return 0 if ($self    -> GetEnd() < $o_locus -> GetStart());
	return (&Min($self->GetEnd(),$o_locus->GetEnd()) - &Max($self->GetStart(),$o_locus->GetStart()) + 1);
}

=head2 function GetNarcisseId

 Title        : GetNarcisseId
 Function     : Return the narcisse Id, computed by some child class
 Usage        : my $id = $o_locus -> GetNarcisseId()
 Returns      : the Narcisse Id
 Args         : none
 Globals      : none

=cut

sub GetNarcisseId
{
	my ($self) = @_;
	return $self -> { __narcisse_id };
}

=head2 function GetRegions

 Title        : GetRegions
 Function     : Returns the regions associated to a composite locus like a Cds or an ExonSet or an Utr
                Return a string like "1..10,20..30" or an array of Locus
                If there is NO region, consider that there is only one region inside the locus: from start to end
                and return the locus itself. 
                WARNING: GetRegions will ALWAYS return something, but may be not always the same locus type !!!
 Usage        : my $region_string = $o_locus->GetRegions();
                my @a_regions = $o_locus->GetRegions();
 Returns      : The regions, string or array formatted, '' or self if no array in the locus
 Args         : none

=cut

sub GetRegions
{
	my ($self) = @_;
	
	if ( defined $self -> {__ra_regions} )
	{
		my $ra_regions = $self -> {__ra_regions};
		if ( wantarray ) 
		{
			return @$ra_regions;
		}
		else
		{
			my @a_region_ends;
			foreach my $o_region ( @$ra_regions )
			{
				my ( $s,$e) = $o_region -> GetEnds();
				push ( @a_region_ends, "$s..$e" );
			}
			return join(',',@a_region_ends);
		}
	}
	else
	{
		if ( wantarray )
		{
			return ( $self );
            return ();
		}
		else
		{
			my ($s,$e) = $self -> GetEnds();
			return "$s..$e";
            return '';
		}
	}
}
=head2 function GetRegion0

 Title        : GetRegion0,GetregionLast
 Function     : Returns the FIRST region, of undef
 Returns      : The region nb 0, or undef if no region
 Args         : none

=cut
sub GetRegion0
{
	my $self = shift;
	if ( defined $self -> {__ra_regions} )
	{
		my $ra_regions = $self -> {__ra_regions};
		return $ra_regions -> [0];
	}
	else
	{
		return undef;
	}
}
sub GetRegionLast
{
	my $self = shift;
	if ( defined $self -> {__ra_regions} )
	{
		my $ra_regions = $self -> {__ra_regions};
		return $ra_regions -> [-1];
	}
	else
	{
		return undef;
	}
}

=head2 function GetNbRegions

 Title        : GetNbRegions
 Function     : Returns the number of regions inside this complete locus, or 0 if it is not a composite locus
 Usage        : my $nb = $o_locus->GetNbRegions();
 Returns      : The number of regions
 Args         : none

=cut
sub GetNbRegions
{
    my $self = shift;
    if ( defined ( $self -> {__ra_regions} ) )
    {
        return @{$self -> {__ra_regions}}+0;
    }
    else
    {
        return 0;
    }
}

=head2 function GetLength

 Title    : GetLength
 Function : Returns the length of the locus
 Returns  : the locus length (end-start+1)
 
=cut
sub GetLength
{
	my $self = shift;
	my ($s,$e) = $self -> GetEnds();
	return abs($e-$s+1);
}

=head2 function GetRegionsLength

 Title    : GetRegionsLength
 Function : Returns the length of a composite locus
 Usage    : my $length = $locus->GetRegionsLength()
 Returns  : the total length (adding the length of each region)
 Note     : Is there is no region, GetRegion returns $self, thus GetRegionsLength retunrs the length of the locus itself
 Args     : none

=cut

sub GetRegionsLength
{
	my $self = shift;

	my @a_regions   = $self -> GetRegions();
	my $len=0;
	foreach my $o_region (@a_regions)
	{
		$len += $o_region -> GetLength();
	}
	return $len;
}

=head2 function GetDefaultSubType

 Title        : GetDefaultSubType
 Usage        : my $def_type = $o_locus->GetDefaultSubType
 Function     : return the default sub_type
 Returns      : The default sub_type
 Args         : none

=cut
sub GetDefaultSubType
{
	my $self = shift;
	throw LipmError::ObjectException($self,'abstract method');
}

=head2 function GetSubType

 Title        : GetSubType
 Function     : Returns the sub_type of locus 
 Usage        : my $sub_type = $o_locus->GetSubType();
 Returns      : The sub_type ( ex: for Ncrna, return miRNA, ncRNA etc.)
 Args         : none

=cut

sub GetSubType
{
	my ($self) = @_;
	return $self -> { __sub_type };
}

=head2 procedure SetSubType

 Title        : SetSubType
 Usage        : $o_ncrna->SetSubType('miRNA');
 Function     : Sets the sub_type of locus. The sub_type is NOT checked
 Args         : The sub_type to set

=cut

sub SetSubType
{
	my $self    = shift;
	my ($sub_type)  = @_;
	$self -> { __sub_type } = $sub_type;
}

=head2 procedure SetDefaultSubType

 Title        : SetDefaultSubType
 Usage        : $o_ncrna->SetDefaultSubType();
 Function     : Sets the default sub_type of locus
 Args         : None

=cut

sub SetDefaultSubType
{
	my $self = shift;
	my $sub_type = $self->GetDefaultSubType();
	$self -> { __sub_type } = $sub_type;
}

=head2 procedure AddExon

 Title        : AddExon
 Usage        : $o_locus->AddExon( $o_exon );
 Procedure    : Add an Exon to the ExonSet object inside the Locus (Rna or Gene), may be creating the ExonSet
 Args         : $o_exon: The region to add to the ExonSet
                NOTE - $o_exon MAY be modified by this procedure
 Globals      : none

=cut
sub AddExon
{
	my $self = shift;
	my ($o_exon) = @_;
	
	if ( ref($o_exon) ne 'Locus::Exon')
	{
		throw LipmError::ParameterException( 'o_exon', 'type', -text => 'should be a Locus::Exon object' );
	}

	if ( ! defined $self -> { __o_exonset } )
	{
		my $o_exonset       = New Locus::ExonSet ( $o_exon );
		$o_exonset -> SetParent ( $self );
		$self -> { __o_exonset } = $o_exonset;
	}

	my $o_exonset = $self -> { __o_exonset };
	$o_exonset -> AddRegion ( $o_exon );
}

=head2 procedure AddRegion

 Title        : AddRegion
 Procedure    : Add a new region to a composite locus like a Cds/ExonSet/Utr
                See also AddMiscRegion for adding a region to a misc_region
 Usage        : $o_locus->AddRegion($o_locus);
 Returns      : none
 Args         : $o_locus: the region (should be a Exon/UtrRegion/CdsRegion) to add to the locus (should be a ExonSet/Utr/Cds)
 Globals      : none

=cut

sub AddRegion
{
    my $self = shift;
    my ($o_locus) = @_;
	
	my ($start,$end) = $o_locus -> GetEnds();
	
	if ( ! defined($$self{__ra_regions}) )
	{
		$$self{__ra_regions} = [];
	}
    my $ra_regions = $$self{__ra_regions};

	my $max   = 0;
	$max = $ra_regions->[-1]->GetEnd() if ( @$ra_regions > 0 );
	if ( $start <= $max )
	{
		# MANU 7/10/2009 Old behaviour: if not well ordered, we giveup
		#                New behaviour: if not well ordered, we reorder
		# Carp::confess("AddRegion: locations must be added in increasing order\nDebug: prev=$$self{'__regions'}, curr=$start,$end\n");
		$self->__AddSorted($o_locus);
	}
	else
	{
		push (@$ra_regions, $o_locus);
	}

	# Tell the new region its rank in the array of regions
	$o_locus -> __SetRank ( @$ra_regions+0);
	
	# Set the parent of the new region, if the parent of this locus is already set.
	if ( defined($self -> GetParent() ) )
	{
		$o_locus -> SetParent( $self );
	}
}

=head2 procedure AddUnsortedRegion

 Title        : AddUnsortedRegion
 Function     : Add a new MiscRegion, WITHOUT modifying the start/end of the parent (see MiscRegion:SetParent)
                and without checking the subregions: they can overlap
 Usage        : $o_region = $o_gene->AddMiscRegion($o_ncrna)
 Returns      : The added region
 Args         : $o_region: A Locus::MiscRegion object to add
 Globals      : none

=cut

sub AddUnsortedRegion
{
	my ($self,$o_region) = @_;
	throw LipmError::ParameterException( 'o_region', 'type', -text => 'should by a Locus::MiscRegion object' ) unless ref $o_region eq 'Locus::MiscRegion';
	
	if ( ! defined($$self{__ra_regions}) )
	{
		$$self{__ra_regions} = [];
	}
	my $ra_regions = $self -> { __ra_regions };
	push @$ra_regions, $o_region;

	if ( defined($self -> GetParent() ) )
	{
		$o_region -> SetParent ( $self );
	}

	return $o_region;
}

=head2 procedure __SetRank

 Title        : __SetRank
 Procedure    : Set the rank (used only for a region, the rank is its position in the regions array) of a locus, only if not already set
                The rank is used in the NarcisseId computation (see CdsRegion::SetParent, Exon::SetParent, UtrRegion::SetParent)
 Usage        : $o_locus->SetRank($rank);
 Returns      : none
 Args         : $rank: The rank to set
 Globals      : none

=cut
sub __SetRank
{
	my $self   = shift;
	my ($rank) = @_;		
	if ( defined ($self -> { __rank }) )
	{
		my $locus_tag = $self -> GetLocusTag();	
		my $sub_type  = $self -> GetSubType();
		my ($s, $e )  = $self -> GetEnds();
		throw LipmError::ObjectException($self,'!attribute', -text => "$sub_type $locus_tag $s..$e: Rank is already set ");
	}
	else
	{
		$self -> { __rank } = $rank;
	}
}

=head2 procedure __GetRank

 Title    : __GetRank
 Function : Accessor for the __rank, throw an exception is not already set (means that there is something wrong in the program logic)

=cut
sub __GetRank
{
	my $self = shift;
	if ( ! defined ($self -> { __rank }) )
	{
		my $locus_tag = $self -> GetLocusTag();	
		my $sub_type  = $self -> GetSubType();
		my ($s, $e )  = $self -> GetEnds();
		throw Error::Simple("The rank is not yet set for this ".ref($self)." ($locus_tag $sub_type [$s..$e]");
	}
	else
	{
		return $self -> { __rank };
	}
}
		
=head2 procedure __AddSorted

 Title     : __SortAndUpdate
 Procedure : Same as AddRegion, but inserts the new region so that $self->{'__ra_regions'} is kept sorted
 Usage     : Called by AddRegion if necessary
 Args      : $o_locus : The locus to insert
 Globals   : none
 NOTE      : Was called __SortAndUpdate

=cut
sub __AddSorted
{
    my $self      = shift;
    my ($o_locus) = @_;

    my $ra_regions = $self -> { __ra_regions };
    push @$ra_regions,$o_locus;
	my @a_sorted_regions    = sort __CmpRegions @$ra_regions;
    $self->{ __ra_regions } = \@a_sorted_regions;
}
# See also ComparePosition
sub __CmpRegions
{
	my ($s_a,$e_a) = $a -> GetEnds();
	my ($s_b,$e_b) = $b -> GetEnds();
	
	return -1 if ($e_a<$s_b+$Locus::ALLOWED_OVERLAP);
	return 1 if ($e_b<$s_a+$Locus::ALLOWED_OVERLAP);
	
	# NOTE - If 2 regions are identical, this IS an overlap - So , we NEVER return 0
	# return 0 if ($s_a==$s_b and $e_a==$e_b);

	throw Error::Simple ( "Locus::AddRegion: overlap between those regions: ".ref($a)." start=".$a->GetStart()." and ".ref($b)." start=".$b->GetStart() );
}

=head2 Function GetSequence

 Title        : GetSequence
 Precondition : The __genomic_sequence field of the locus ancestor must be initialized
 Function     : Returns the sequence of a locus, initializing the __length private field
 Usage        : my $seq = $o_locus->GetSequence();
 Returns      : $seq (a copy of the subsequence corresponding to the locus)
 Args         : none
 Globals      : none
=cut

sub GetSequence
{
	my $self = shift;
	
	my $seq   = '';
	my $len   = 0;
	my $ancestor = $self -> GetAncestor() ;
	return "" if (!defined $ancestor);
	
	my $r_seq = $ancestor -> GetGenomicSequence();
	if ( $$r_seq ne '' )
	{
		my @a_regions = $self -> GetRegions();
		my $strand    = $self -> GetStrand();
	
		my $seqlen = length($$r_seq);
	
		foreach my $o_region (@a_regions)
		{
			my ($start,$end) = $o_region -> GetEnds();
			
			# Correct with the contig offset, because we don't have the whole sequence in memory
			#if ( defined ($Locus::O_ASSEMBLY) )
			#{
			#	my $offset = $Locus::O_ASSEMBLY -> GetStart () - 1;
			#	$start -= $offset;
			#	$end   -= $offset;
			#}
			
			my $len = $end-$start+1;
			if ( $seqlen < $end )
			{
				my $msg = "ASSEMBLY ERROR for the " . $self -> GetSubType() . " " . $self -> GetLocusTag();
				my $strand = $o_region -> GetStrand();
				$msg .= " coordinates out of range: " . $self->GetpkExtended() . "sequence length=$seqlen";
				throw Error::Simple($msg);
			}
			$seq .= substr ($$r_seq,$start-1,$len);
		}
		if ( $strand < 0 )
		{
			$seq = &UtilRevComp($seq);
		}
		$len = length($seq);
	}
	$self -> { __length } = $len;
	return $seq;
}

=head2 Function GetSequenceWithEnvironment

 Title        : GetSequenceWithEnvironment
 Precondition : The __genomic_sequence field of the locus ancestor must be initialized
 Function     : Returns the sequence of a locus, with several nucleotides BEFORE the sequence and AFTER the sequence
 Usage        : my $seq = $o_locus->GetSequence();
 Returns      : $seq (a copy of the subsequence corresponding to the locus)
 Args         : none
 Globals      : none
=cut
sub GetSequenceWithEnvironment
{
	my $self = shift;
	my ($len_env) = @_;
	my $seq_len = $self->GetSequenceLength();
	return '' if ( $seq_len==0);
	
	my $r_seq = $self -> GetAncestor() -> GetGenomicSequence();
	my ($start, $end) = $self -> GetEnds();
	my $strand = $self -> GetStrand();

	my $s_pre = &Locus::UtilGetSeqPre ($r_seq,$start,$len_env);
	my $s_post= &Locus::UtilGetSeqPost($r_seq,$end,$len_env);
	
	if ( $strand > 0 )
	{
		return $s_pre . $self -> GetSequence() . $s_post;
	}
	else
	{
		return &Locus::UtilRevComp($s_post) . $self -> GetSequence() . &Locus::UtilRevComp($s_pre);
	}
}

=head2 Function GetSequenceLength

 Title        : GetSequenceLength
 Function     : Returns the sequence length, calling GetSequence if necessary
 Usage        : my $seq = $o_locus->GetSequenceLength($r_seq);
 Returns      : The sequence length
 Globals      : none
=cut

sub GetSequenceLength
{
	my $self = shift;
	if (!defined $self -> { __length })
	{
		$self->GetSequence();
	}
	return $self -> { __length };
}

=head2 function GetParent

 Title    : GetParent
 Function : Return the parent locus
 Usage    : my $o_parent = $o_locus -> GetParent();
 Args     : none
 Return   : an object, generally Locus, but may be some other class (at top of string)
 Globals  : none

=cut
sub GetParent
{
	my $self = shift;
	return $self -> { __o_parent };
}

=head2 function GetAllChildren

 Title    : GetAllChildren
 Function : Return the children loci, ie the loci embedded by this locus
            This is a recursive function
 Usage    : my $ra_children = $o_locus -> GetAllChildren();
 Args     : none
 Return   : A ref to an array containing the children of the locus, as well as the children of the children, etc.
 Globals  : none

=cut
sub GetAllChildren
{
	my $self = shift;
	my @a_children;
	
	my $o_exon_set = $self->GetExonSet();
	if ( defined $o_exon_set )
	{
		push @a_children, $o_exon_set, @{$o_exon_set->GetAllChildren()};
	}
	
	my @a_regions = $self -> GetRegions();
	if ( $a_regions[0] ne $self )
	{
		foreach my $o_region (@a_regions)
		{
			push @a_children, $o_region, @{$o_region->GetAllChildren()};
		}
	}
	return \@a_children;
}

=head2 procedure GetAncestor

 Title    : GetAncestor
 Function : Return the ancestor locus, ie the farest parent (should be an AnnotatedSeq object)
 Usage    : my $o_ancestor = $o_locus -> GetAncestor();
 Args     : none
 Return   : an AnnotatedSeq object
 Globals  : none

=cut
sub GetAncestor
{
	my $self = shift;
	return undef unless defined $self -> GetParent();
	return (defined ($self -> GetParent() -> can ("GetAncestor")) ? $self -> GetParent() -> GetAncestor() : $self -> GetParent());
}

=head2 Functions GetKingdom, GetSpecies,GetContigLabel,GetContigInfo

 Title    : GetKingdom, GetSpecies,GetContigLabel
 Function : Call GetKingdom, Getspecies, GetContigLabel from the locus ancestor (see GetAncestor)
 Usage    : $ctg = $o_locus -> GetContigLabel();
 Args     : none
 Globals  : none

=cut
sub GetKingdom
{
	my $self = shift;
	return $self -> GetAncestor() -> GetKingdom();
}
sub GetSpecies
{
	my $self = shift;
	return $self -> GetAncestor() -> GetSpecies();
}
sub GetContigLabel
{
	my $self = shift;
	return $self -> GetAncestor() -> GetContigLabel();
}
sub GetChromosomeLabel
{
	my $self = shift;
	if ( ! defined ( $self->GetAncestor() ))
	{
		print $self -> GetpkExtended()."\n";
		use Carp;
		Carp::confess();
	}
	return $self -> GetAncestor() -> GetChromosomeLabel();
}

# See Locus::CdsRegion
# This is mainly useful for gff3

sub GetScore
{
	my $self = shift;
	my $rvl  = $self -> Get ( 'score' );
	return $rvl eq '' ? '.' : $rvl; 
}

sub GetPhase
{
	return '.';
}

=head2 function GetBioPerlFeature

 Title    : GetBioPerlFeature
 Function : Return a bioperl representation of the locus
            This function calls the protected function _GetbioPerlFeature, 
            which may be overloaded in the derived classes
 Usage    : my $f = $o_locus -> GetBioPerlFeature();
 Args     : none
 Return   : A Bio::SeqFeature::Generic object
 TODO - MAINTENANT DANS AnnotatedSeqFileWriter::UsingBioPerl !!! (__GetQualifiers)
=cut
sub GetBioPerlFeature
{
	my $self = shift;
	return $self -> _GetBioPerlFeature ( @_ );
}

=head2 function _GetBioPerlFeature

 Title    : _GetBioPerlFeature
 Function : Create and return new Bio::SeqFeature::Generic object, add several tag values to this object
 Usage    : my $f = $self -> GetBioPerlFeature();
 Args     : $rh_ignored (optional) The list of ignored features
 Return   : A Bio::SeqFeature::Generic object
 TODO - MAINTENANT DANS AnnotatedSeqFileWriter::UsingBioPerl !!!

=cut
sub _GetBioPerlFeature
{
	my $self         = shift;
	my $sequence_loaded = ($self -> GetAncestor() -> GetLength()!=0);
	
	eval "use Bio::SeqFeature::Generic";
	if ( $@ ne "" )
	{
		throw Error::Simple ( $@ );
	}
	
	my %h_params = ();
	my $rh_core = $self -> { _d_rh_core };
	@h_params { qw( -start -end -strand ) } = @$rh_core{ qw ( start end strand ) };
	
	my $f = new Bio::SeqFeature::Generic ( %h_params );

	my $gene = $self -> GetGeneName();
	$self -> _AddTagValues( $f,'gene',$gene)      if ( $gene ne '' );

	my $locus_tag = $self -> GetLocusTag(&TRUE);
	$self -> _AddTagValues( $f,'locus_tag',$locus_tag)      if ( $locus_tag ne '' );

	my $note = $self -> GetNote();
	$self -> _AddTagValues( $f,'note',$note)      if ( $note ne '' );

	my $product = $self -> GetProduct();
	$self -> _AddTagValues( $f,'product',$product)      if ( $product ne '' );

	my $dbxref = $self -> GetDbXref();
	$self -> _AddTagValues( $f,'db_xref',$dbxref) if ( $dbxref ne "" );
	
	my $alias = $self -> GetAlias();
	$self -> _AddTagValues( $f,'gene_synonym',$alias) if ( $alias ne "" );

	my $ra_tags = $self -> GetOtherKeys();
	foreach my $tag ( @$ra_tags )
	{
		$self -> _AddTagValues( $f,$tag,$self -> Get ( $tag ));
	}
	return $f;
}

=head2 procedure _AddTagValues

 Title    : _AddTagValues
 Procedure: Call add_tag_values, adding one or several tag values to a feature
 Usage    : my $l = $self -> _GetBioPerlRegions();
 Args     : none
 Return   : A Bio::SeqFeature::Split object

=cut
sub _AddTagValues
{
	my $self = shift;
	my ($f, $tag, $value) = @_;
	if ( ref($value) eq 'ARRAY' )
	{
		foreach my $v ( @$value )
		{
			$f -> add_tag_value ( $tag, $v );
		}
	}
	else
	{
		$f -> add_tag_value ( $tag, $value );
	}
}
		
=head2 function _GetBioPerlRegions

 Title    : _GetBioPerlRegions
 Function : When a locus is made of several regions, this function may be used to create
            a new Bio::Location::Split location object, fill this object with some Bio::Location::Simple objects,
            and return the location created.
 Usage    : my $l = $self -> _GetBioPerlRegions();
 Args     : none
 Return   : A Bio::SeqFeature::Split object

=cut

sub _GetBioPerlRegions
{
	my $self = shift;
	eval "use Bio::Location::Simple;use Bio::Location::Split;";
	if ( $@ ne "" )
	{
		throw Error::Simple ( $@ );
	}
	my $l = new Bio::Location::Split;
	my @a_regions = $self -> GetRegions();
	if ( @a_regions != 0)
	{
		if ( $a_regions[0] -> GetStrand() == -1 )
		{
			@a_regions = reverse @a_regions;
		}
		foreach my $o_region (@a_regions)
		{
			next unless defined($o_region);
			$l -> add_sub_Location ( Bio::Location::Simple->new(-start  => $o_region->GetStart(),
																-end    => $o_region->GetEnd(),
																-strand => $o_region->GetStrand())
									);
		}
		$l -> splittype ( $self->_GetSplitType() );
	}
	return $l;
}

=head2 function GetExonSet

 Title        : GetExonSet
 Usage        : my $o_exonset = $o_mrna->GetExonSet();
 Function     : Accessor to the ExonSet object
 Args         : none
 Globals      : none

=cut
sub GetExonSet{
	my $self = shift;
	if ( defined $self -> { __o_exonset } )
	{
		return $self -> { __o_exonset };
	}
	else
	{
		return undef;
	}
}

=head2 function Get

 Title    : Get
 Usage    : my $ra_value = $o_locus->Get('tag');
 Function : Returns the value of a given tag (= attribute)
            'tag' is set only when initializing the object (see _Init for the details)
 Returns  : The tag value if found, or ''
            May be a string if there is only 1 value, or a $ra_ if there are several values
 Args     : 'tag' The tag name

=cut

sub Get
{
	my $self        = shift;
	my ($attribute) = @_;

	if ( defined( $self -> { _d_rh_core } -> { $attribute } ) )
	{
		return $self -> { _d_rh_core } -> { $attribute };
	}
	elsif ( defined ( $self -> { _d_rh_other } -> { $attribute } ) )
	{
		return $self -> { _d_rh_other } -> { $attribute };
	}
	else
	{
		return '';
	}
}

=head2 function _Get0

 Title    : _Get0
 Usage    : my $ra_value = $o_locus->Get('tag');
 Function : Calls Get, if no value (return '') call Get on the FIRST REGION 
            This is useful for composite loci like Cds, Exonset, etc
 Returns  : The tag value if found, or ''
 Args     : 'tag' The tag name

=cut
sub _Get0
{
	my $self        = shift;
	my ($attribute) = @_;
	my $value = $self -> Get ( $attribute );
	if ( $value eq '' )
	{
		my $r0 = $self -> GetRegion0();
		$value = $r0 -> Get ( $attribute ) if ( defined $r0 );
	}
	return $value;
}
	
=head2 procedure Delete

 Title    : Delete
 Usage    : $o_locus->Delete('tag');
 Function : Delete the given tag
 Args     : 'tag' The tag name

=cut

sub Delete
{
	my $self        = shift;
	my ($attribute) = @_;

	if ( defined( $self -> { _d_rh_core } -> { $attribute } ) )
	{
		delete $self -> { _d_rh_core } -> { $attribute };
	}
	elsif ( defined ( $self -> { _d_rh_other } -> { $attribute } ) )
	{
		delete $self -> { _d_rh_other } -> { $attribute };
	}
}

=head2 procedure DeleteExcept

 Title    : DeleteExcept
 Usage    : $o_locus->DeleteExcept('locus_tag');
 Function : Delete all the 'Other' attributes, except those specified (locus_tag in the example above)
 Args     : none

=cut
sub DeleteExcept
{
	my $self = shift;
	my (@a_attributes) = @_;
	my %h_other;
	foreach my $k (@a_attributes)
	{
		if ( exists $self->{_d_rh_other}->{$k} )
		{
			$h_other{$k}=$self->{_d_rh_other}->{$k};
		}
	}
	$self -> { _d_rh_other } = \%h_other;
}

#=head2 procedure DeleteOther

 #Title    : DeleteOther
 #Usage    : $o_locus->DeleteOther();
 #Function : Delete all the 'Other' attributes
 #Args     : none

#=cut
#sub DeleteOther
#{
	#my $self = shift;
	#$self -> { _d_rh_other } = {};
#}

=head2 procedure Set

 Title    : Set
 Usage    : $o_locus->Set('attribute','value');
 Function : Set the given attribute, should NOT be already set
 Args     : 'attribute' The attribute name
            'value' The value to set

=cut

sub Set
{
	my $self        = shift;
	my ($attribute,$value) = @_;
	my $v = $self -> Get ( $attribute );
	if ( $v ne '' ) 
	{
		throw Error::Simple ("Locus::$attribute is already set, please call Delete before Set - " . $self->GetpkExtended() );
	}
	if (exists $H_RESERVED_ATTRIBUTES{$attribute})
	{
		$self -> { _d_rh_core } -> { $attribute } = $value;
	}
	else
	{
		$self -> { _d_rh_other } -> { $attribute } = $value;
	}
}
=head2 procedure AddValue

 Title    : AddValue
 Usage    : $o_locus->AddValue('attribute','value');
 Function : If the attribute is not set, call Set
            Else, add a value to a multivalued attribute
 Args     : 'attribute' The attribute name (multivalued)
            'value' The value to set

=cut
sub AddValue
{
	my $self = shift;
	my ($attribute,$value) = @_;
	my $v = $self -> Get ( $attribute );
	if ( $v eq '' )
	{
		$self->Set($attribute, $value);
	}
	else
	{
		if ( ref($v) eq 'ARRAY' )
		{
			push(@$v,$value);
		}
		else
		{
			my $ra_v = [$v,$value];
			$self->Delete($attribute);
			$self->Set($attribute,$ra_v);
		}
	}
}

	
=head2 function GetAsString

 Title    : GetAsString
 Usage    : my $value = $o_locus->Get('tag');
 Function : Returns the value of a given tag (= attribute) as a string, even if there are several values
 Returns  : The tag value if found, or ''
 Args     : 'tag' The tag name

=cut

sub GetAsString
{
	my $self = shift;
	my ($attribute) = @_;
	my $value = $self -> Get ( $attribute );
	if ( ref($value) eq 'ARRAY' )
	{
		return join ',', @$value;
	}
	else
	{
		return $value;
	}
}

=head2 function GetAsStringConsolidated

 Title    : GetAsStringConsolidated
 Usage    : my $value = $o_locus->GetAsStringConsolidated('tag', $o_locus1, $o_locus2, ...);
 Function : Call GetAsString for the different loci specified, and consolidate values
 Returns  : The tag value if found, or ''
 Args     : 'tag' The tag name
            $o_locus1, ... One or several loci to consolidate with $self
            Each $o_locus may be undef, and the array may be empty

=cut
sub GetAsStringConsolidated_OBSOLETE
{
	my $self = shift;
	my ($tag, @a_o_loci) = @_;
	
	# Same as GetAsString if there is nothing to consolidate
	return $self->GetAsString($tag) if ( @a_o_loci==0 );
	
	# Get values
	my $val_self = $self -> GetAsString($tag);
	my @a_val;
	for my $o_locus ( @a_o_loci )
	{
		push @a_val,$o_locus->GetAsString($tag) if (defined $o_locus && $o_locus->IsSelected());
	}
	
	# Consolidate values, unless equal
	my @a_cval;
	push @a_cval,$val_self if ( $val_self ne '' );
	for my $val (@a_val) 
	{
		push @a_cval,$val if ( $val ne '' && $val ne $val_self );
	}
	return join ' ', @a_cval;
}

sub GetAsStringConsolidated
{
	my $self = shift;
	my $ra_val = $self->GetAsArrayConsolidated(@_);
	return join ' ', @$ra_val;
}

=head2 function GetAsArrayConsolidated

 Title    : GetAsArrayConsolidated
 Usage    : my $ra_value = $o_locus->GetAsArrayConsolidated('tag1', 'tag2',..., $o_locus1, $o_locus2, ...);
 Function : Call Get for the different loci specified, and consolidate values in an array
 Returns  : A ra, may be the array is empty
 Args     : 'tag','tag2',... The tag names
            $o_locus1, ... One or several loci to consolidate with $self
            Each $o_locus may be undef, and the array may be empty

=cut
sub GetAsArrayConsolidated
{
	my $self = shift;
	my @a_tags;
	my @a_o_loci;
	foreach my $arg (@_)
	{
		if ( ref($arg) eq '' )
		{
			push @a_tags,$arg;
		}
		else
		{
			push @a_o_loci,$arg;
		}
	}
	unshift @a_o_loci, $self;

	# Get values
	my @a_val;
	for my $o_locus ( @a_o_loci )
	{
		for my $t ( @a_tags )
		{
			push @a_val,$o_locus->Get($t) if (defined $o_locus && $o_locus->IsSelected());
		}
	}
	
	# We have some scalars together with some $ra_
	my @a_cval;
	for my $val ( @a_val)
	{
		if ( ref($val) eq 'ARRAY' )
		{
			push @a_cval, @$val;
		}
		elsif ( $val ne "" )
		{
			push @a_cval, $val;
		}
	}
	
	# pour garantir l'unicite
	my %h_uniq;
	map { $h_uniq{$_}++ } @a_cval;
	my @a_uniq = keys %h_uniq;
	return \@a_uniq;
}
sub GetAsArrayConsolidated_OBSOLETE
{
	my $self = shift;
	my ($tag, @a_o_loci) = @_;
	
	unshift @a_o_loci, $self;

	# Get values
	my @a_val;
	for my $o_locus ( @a_o_loci )
	{
		push @a_val,$o_locus->Get($tag) if (defined $o_locus && $o_locus->IsSelected());
	}
	
	# We have some scalars together with some $ra_
	my @a_cval;
	for my $val ( @a_val)
	{
		if ( ref($val) eq 'ARRAY' )
		{
			push @a_cval, @$val;
		}
		elsif ( $val ne "" )
		{
			push @a_cval, $val;
		}
	}
	
	# pour garantir l'unicite
	my %h_uniq;
	map { $h_uniq{$_}++ } @a_cval;
	my @a_uniq = keys %h_uniq;
	return \@a_uniq;
}

=head2 function GetOtherKeys

 Title    : GetOtherKeys
 Usage    : my $ra_keys = $o_locus->GetOtherKeys();
 Function : Returns the list of "other" (ie non-standard) attributes
 Returns  : A ref to an array
 Args     : $rh_ignored (optional): The keys of this hash will NEVER be returned 

=cut
sub GetOtherKeys
{
	my $self = shift;
	my ($rh_ignored) = @_;
	
	my @a_keys = keys %{$self -> { _d_rh_other }};
	if ( defined $rh_ignored )
	{
		my @a_returned_keys;
		foreach my $k ( @a_keys )
		{
			push @a_returned_keys, $k unless ( exists $rh_ignored -> { $k } );
		}			
		return \@a_returned_keys;
	}
	else
	{
		return \@a_keys;
	}
}

=head2 procedure GetConsolidatedOtherKeys

 Title     : GetConsolidatedOtherKeys
 Usage     : my $ra_other_keys = $o_locus->GetConsolidatedOtherKeys($rh_ignored,$o_locus1,$o_locus2,...)
 Function  : Consolidate the 'other' keys of this CDS and the o_loci specified
 Args      : $rh_ignored, see GetOtherKeys
             $o_locus1,$o_locus2 The loci to consolidate
 Return    : The keys in a ra
 Globals   : none

=cut
sub GetConsolidatedOtherKeys
{
	my $self = shift;
	my ($rh_ignored,@a_o_loci) = @_;
	
	my @a_other_keys = ();
	push @a_other_keys, @{$self->GetOtherKeys($rh_ignored)};
	
	for my $o_locus ( @a_o_loci)
	{
		push @a_other_keys, @{$o_locus->GetOtherKeys($rh_ignored)};
	}
	
	# pour garantir l'unicite
	my %h_uniq;
	map { $h_uniq{$_}++ } @a_other_keys;
	my @a_uniq = keys %h_uniq;
	return \@a_uniq;
}

=head2 function GetConsistency

 Title        : GetConsistency
 Function     : Return the consistency flag
 Usage        : my $consis = $o_locus->GetConsistency();
 Returns      : 'C', 'I', 'FS','FE' or 'FSFE'
 Args         : none
 Globals      : none

=cut

sub GetConsistency
{
	my ($self) = @_;
	return (exists $self -> { __consistency })?$self -> { __consistency } : 'C';
}

=head2 procedure SetFragment,SetInconsistent

 Title        : SetFragment,SetInconsistent
 Procedure    : Set the consistency flag to 'FS' (fragment, did not find start codon), 'FE' (fragment, did not find stop codon) or 'I' (inconsistent)
                when marked as 'I' it is not possible to return to the 'FS' or 'FE' state
                when marked as 'F' or 'I', it is not possible to return to the 'C' state
                The consistency status may be FS, FE, FSFE, C or I.
                Mark the locus as a pseudogene, unless already done
 Usage        : $o_locus->SetFragment('S') - fragment at start
				$o_locus->SetFragment('E') - fragment at end
				$o_locus->SetFragment('F') - fragment somewhere
                $o_locus->SetInconsistent()
 Args         : $type = 'F','S','E'
 Globals      : none

=cut

sub SetFragment
{
	my $self = shift;
	my ( $type) = @_;
	$self->{__consistency}="" unless exists $self->{__consistency};
	return if ( $self -> { __consistency } eq 'I' );
	return if ( $type !~ /^[FSE]$/ );
	$self -> { __consistency } = "" if ( $self -> { __consistency } eq 'C' );
	if ( $type eq 'F' )
	{
		$self -> { __consistency } = "FSFE";
	}
	elsif ( $self -> { __consistency } !~ /F$type/ )
	{
		$self -> { __consistency } .= "F$type";
	}
	$self->__SetPseudo();

	return;
}

sub SetInconsistent
{
	my $self = shift;
	$self -> { __consistency } = 'I';
	$self->__SetPseudo();
	delete $self -> { __seq_prot };
	return;
}

sub __SetPseudo
{
	my $self = shift;
	if ( $self -> Get("pseudogene") eq "" )
	{
		$self -> Set( "pseudogene","unknown");
	}
}

#=head2 procedure __SwapFragmentInfo

 #Called when the locus changes its strand during assembly or disassembly
 #start/end is changed too, so we must exchange fragment information:
	#'FE' <=> 'FS'
	#'FEFS' or 'I' unchanged

#=cut
#sub __SwapFragmentInfo
#{
	#my $self = shift;
	#if (exists ($self -> { __consistency }) && length ($self -> { __consistency }) == 2)
	#{
		#$self -> { __consistency } = ($self -> { __consistency } eq 'FE')?'FS':'FE';
	#}
#}

=head2 SyncConsistency
 Title :  SyncConsistency
 Procedure : Synchronize the consistency between the First region, the Last region and the object
             This is useful for Cds or Mrna
 Usage     : $o_cds->SyncConsistency()
 
=cut
sub SyncConsistency
{
	my $self = shift;
	
	my $self_consistency = $self -> GetConsistency();
	
	my @a_regions  = $self -> GetRegions();
	if ( $self->GetStrand() > 0 )
	{
		# Region (start or end) => Locus
		$self->SetFragment('S') if ($a_regions[0]->GetConsistency() =~ /FS/);
		$self->SetFragment('E') if ($a_regions[-1]->GetConsistency() =~ /FE/);
		
		# Locus => Region (start or end)
		$a_regions[0]->SetFragment('S')  if ($self_consistency =~ /FS/);
		$a_regions[-1]->SetFragment('E') if ($self_consistency =~ /FE/);
	}
	else
	{
		# Region (start or end) => Locus
		$self->SetFragment('S') if ($a_regions[-1]->GetConsistency() =~ /FS/);
		$self->SetFragment('E') if ($a_regions[0]->GetConsistency() =~ /FE/);
		
		# Locus => Region (start or end)
		$a_regions[-1]->SetFragment('S')  if ($self_consistency =~ /FS/);
		$a_regions[0]->SetFragment('E') if ($self_consistency =~ /FE/);
	}
	if ( $self_consistency eq 'I' )
	{
		map { $_->SetInconsistent() } @a_regions;
	}
}

# ============================================================================================
#
# Some utility functions - Those functions are NOT methods ! They are JUST ordinary functions !
#
# ============================================================================================

=head2 UtilGetSeqPre
 
 Title :  UtilGetSeqPre
 Usage :  my $pre_seq = UtilGetSeqPre($r_seq,$pos,$len);
 Procedure : Given a position on a reference sequence, return the $len nucleotides BEFORE $pos
             If $pos is near the start of the sequence, pad with ' ' characters, so that the length of the returned sequence will be ALWAYS $len
 Example   : $r_seq -> GAGACAGTCAGT
                       ^ ^      ^  
                       1 3      10
             UtilGetSeqPre($r_seq,10,5) -> CAGTC
             UtilGetSeqPre($r_seq,3,5   -> ...GA
 Args : $r_seq A reference towards the sequence (may be very long)
        $pos   The position to consider
        $len   The length to retrieve
 
=cut
sub UtilGetSeqPre
{
	my ($r_seq,$pos,$len) = @_;
	my $len1 = ($pos>$len)?$len:$pos-1;
	if ( $len == $len1 )
	{
		return substr $$r_seq,$pos-$len1-1,$len1;
	}
	else
	{
		return ' 'x($len-$len1) . substr ($$r_seq,0,$len1);
	}
}

=head2 UtilGetSeqPost
 
 Title :  UtilGetSeqPost
 Usage :  my $post_seq = UtilGetSeqPost($r_seq,$pos,$len);
 Procedure : Given a position on a reference sequence, return the $len nucleotides AFTER $pos
             If $pos is near the end of the sequence, pad with ' ' characters, so that the length of the returned sequence will be ALWAYS $len
 Example   : $r_seq -> GAGACAGTCAGT
                       ^ ^      ^  
                       1 3      10
             UtilGetSeqPost($r_seq,3,5)  -> ACAGT
             UtilGetSeqPost($r_seq,10,5) -> GT...
 Args : $r_seq A reference towards the sequence (may be very long)
        $pos   The position to consider
        $len   The length to retrieve
 
=cut
sub UtilGetSeqPost
{
	my ($r_seq,$pos,$len) = @_;
	my $seq_len = length($$r_seq);
	my $len1 = ($pos+$len<=$seq_len)?$len:$seq_len-$pos;
	if ( $len == $len1 )
	{
		return substr $$r_seq,$pos,$len1;
	}
	else
	{
		return substr ($$r_seq,$pos,$len1) . ' 'x($len-$len1);
	}
}

sub UtilRevComp
{
	my ($seq) = @_;
	$seq = reverse($seq);
	$seq =~ tr/ATGCatgc/TACGtacg/;
	return $seq;
}

=head2 procedure SafeDumper

 Title        : SafeDumper
 Usage        : $o_locus->SafeDumper();
 Function     : Return a printable representation of the object
                The field __o_parent, if set, is modified, the Data::Dumper is called
                It is necessary modifying this field to avoid huge and stupid dumps, because Dumper follows the __o_dumper reference
                So, this method MAY BE USED from the perl -d debugger:
                DO NOT TYPE    p Dumper($o_gene)
                PREFER         p $o_gene->SafeDumper()
 
 Args         : none
 Globals      : none

=cut
sub SafeDumper
{
	my $self = shift;
	my $rvl;
	if ( defined $self->{__o_parent} )
	{
		my $o_parent = $self->{__o_parent};
		$self->{__o_parent} = ref $o_parent;
		$rvl = Dumper($self);
		$self->{__o_parent} = $o_parent;
	}
	else
	{
		$rvl = Dumper($self);
	}
	return $rvl;
}


=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut


1;
