#
# $Id: FileCompressedAppend.pm 1081 2013-03-05 13:47:04Z manu $
#

package FileCompressedAppend;

use strict;
use warnings;

# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=head1 NAME

FileCompressedAppend - To conveniently append data to a compressed file

=head1 SYNOPSIS
 my $o_compressed = New FileCompressedAppend('filename.gz');	# The file is automatically decompressed now
 my $file_name    = $o_compressed -> GetFileName();             # Return filename (NOT filename.gz)
 my $fh_comp      = $o_compressed -> GetFileHandler();			# Return the file, opened in append mode
 $fh_comp -> close();                                           # Just close the file, don't compress it
 $o_compressed = '';                                            # Call destructor, compressiong the file back
 
=head1 DESCRIPTION

 This object helps you appending data to a compressed file. 
    - The function General::GetCompressPath is called, returning the path of the compression program
    - If the file exists it is uncompressed by the constructor
    - The file is uncompressed so you may append data as needed
    - The destructor calls the compression program again , compressing the file back
 Please note that if the file is NOT compressed (compression path is ''), this object is not really useful, but it works
 
=cut

use General;
use Error qw(:try);
use base qw( LipmObject );
use Carp;

BEGIN
{
    our $VERSION = do {my @r = (q$Rev: 1081 $ =~ /\d+/g); $r[0]};
}


sub _Init
{
	my $self = shift;
	
	my ($file_name, $mode) = @_;
	$mode = 'a' unless defined($mode);
	unlink ($file_name) unless ($mode eq 'a');

	my $compress_path = &GetCompressPath($file_name);
	if ( $compress_path ne '' )
	{
		if ( -e $file_name)
		{
			system "$compress_path -d $file_name";
			my $sts = $? >> 8;
			if ($? == -1) 
			{
				croak "failed to execute: $!\n";
			}
			elsif ($? & 127) 
			{
				croak sprintf ("$compress_path -d died with signal %d, %s coredump\n",($? & 127),  ($? & 128) ? 'with' : 'without');
			}
			elsif ( $sts != 0 )
			{
				croak sprintf ("$compress_path -d exited with value %d\n", 0 );
			}
		}
		$file_name =~ s/\.\w+$//;
	}	
	$self -> _Set ( '__compress_path', $compress_path );
	$self -> _Set ( '__file_name', $file_name );
	return $self;
}

sub GetFileName
{
	my $self = shift;
	return $self -> _Get ( '__file_name' );
}

sub GetFileHandler
{
	my $self = shift;
	if ( $self -> _Get ( '__file_handler' ) eq '' )
	{
		my $file_name = $self -> _Get ( '__file_name' );
		my $fh_handler= new IO::File ( $file_name, 'a' ) || croak ("ERROR - Cannot open $file_name for appendign data" );
		$self -> _Set ( '__file_handler', $fh_handler );
	}
	return $self -> _Get ( '__file_handler' );
};

sub DESTROY
{
	my $self = shift;
	my $compress_path = $self -> _Get ( '__compress_path' );
	my $file_name     = $self -> _Get ( '__file_name' );
	if ( -e $file_name && $compress_path ne '' )
	{
		system "$compress_path -f $file_name";
		my $sts = $? >> 8;
		if ($? == -1) 
		{
			croak "failed to execute: $!\n";
		}
		elsif ($? & 127) 
		{
			croak sprintf ("$compress_path -d died with signal %d, %s coredump\n",($? & 127),  ($? & 128) ? 'with' : 'without');
		}
		elsif ( $sts != 0 )
		{
			croak sprintf ("$compress_path -d exited with value %d\n", 0 );
		}
	}
}
	
=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;
