#
# $Id: GFF3.pm 945 2012-06-27 07:51:23Z manu $
#

package Splitter::GFF3;

use strict;
use warnings;

use base qw( Splitter );

# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=head1 NAME

 Splitter::Fasta - Split a fasta file

=head1 SYNOPSIS
 See Splitter

=head1 DESCRIPTION
 See the description in the Splitter abstract class
 
=cut

use General;
use Error qw( :try );
use Error::Simple;
use LipmError::ParameterException;

BEGIN
{
    our $VERSION = do {my @r = (q$Rev: 945 $ =~ /\d+/g); $r[0]};
}

=head function GetFileHeader
 return the gff3 standard header

=cut
sub GetFileHeader
{
	return "##gff-version 3\n";
}

=head2 procedure _Index

 Title        : _Index
 Usage        : $self -> _Index()
 Prerequisite : The line describing the different contigs must be ordered, ie all lines corresponding to some
                contig must be together in the file:
                
                BAD:
                contig1 1   10  bla-bla
                contig2	1   10  bla-bla
                contig1 20  30  bla-bla
                contig2 20  30  bla-bla
                
                GOOD:
                contig1 1   10  bla-bla
                contig1 20  30  bla-bla
                contig2	1   10  bla-bla
                contig2 20  30  bla-bla
 Function     : Initializes the _h_contigs hash (key = contig name, val = ref to an array of offsets offset in the file)
                The contig name is the FIRST WORD of the line (column 0)
                Initializes also the _rh_contig_size hash with the number of lines for each contig
 Args         : None
 Globals      : none

=cut

sub _Index
{
	my $self = shift;
	
	my $fh_contigs     = $self -> _Get ( '_fh_contigs' );
	my $rh_contigs     = $self -> _Get ( '_rh_contigs' );
	my $rh_contig_size = $self -> _Get ( '_rh_contig_size' );
	my $id     = "";
	my $offset = 0;
	my $line = "";
	$fh_contigs -> seek(0,0);
	my ($last_contig, $current_contig) = ("","");
	while( $offset=$fh_contigs->tell(), $line = <$fh_contigs> )
	{
		# Skip comments or empty lines, 
		next if ( $line =~ /^#/);
		next if ( $line =~ /^$/);
		
		($current_contig) = split /\t/, $line, 2;
		$rh_contig_size -> { $current_contig } += 1;
		if ( $current_contig ne $last_contig )
		{
			$last_contig = $current_contig;
			if ( ! exists $rh_contigs -> { $last_contig } )
			{
				my $ra_offsets = [];
				$rh_contigs -> { $last_contig } = $ra_offsets;
			}
			push ( @{$rh_contigs -> { $current_contig }}, $offset );
		}
	}
}

=head2 function _GetContigs

 Title    : _GetContigs
 Usage    : return $self->_GetContigs( \@a_contigs );
 Prerequisite : The array should NOT be empty 
 Function : For each id in the array, return the corresponding contig
            If some id is not found, throw an exception
            NOTE - If you want to retrieve some contigs, this is much more performant than using the Regex version,
            as we do not have to scan all the indexed keys !
 Returns  : $r_contigs, a ref to string with the extracted contigs, formatted as in the original file
 Args     : $ra_contigs, ... A list of contigs to retrieve

=cut

sub _GetContigs
{
	my $self     = shift;
	my ( $ra_contigs ) = @_;

	my $fh_contigs = $self -> _Get ( '_fh_contigs' );
	my $rh_contigs = $self -> _Get ( '_rh_contigs' );
	
	# Avoiding memory allocation in the loop for performance reasons
	my $contig;
	my ($ra_offsets,$offset,$line,$id,$current_contig_id);
	foreach $id (@$ra_contigs)
	{
		#locate the contig in the file, may be at several offsets
		$ra_offsets = $rh_contigs -> { $id };
		foreach $offset ( @$ra_offsets )
		{
			$fh_contigs -> seek( $offset, 0 );
			while( $line = <$fh_contigs> )
			{
				next if ( $line =~ /^#/ || $line =~ /^$/ );
				($current_contig_id) = split ( /\t/, $line, 2 );
				last	if ($current_contig_id ne $id);
				$contig .= $line;
			}
		}
	}
	return \$contig;
}		

=head2 procedure _CopyContigs

 Title    : _CopyContigs
 Usage    : $self->_CopyContigs( $dst_filename, $regexp );
 Function : Copy each contig whose id is retrieved with $regexp
            The method GetContigs is NOT used, to have a lower memory print.
 Args     : $dst_filename The destination filename (should be writeable)
            $regexp, a regular expression, see GetContigNames

=cut
sub _CopyContigs
{
	my $self    = shift;
	my ( $dst_filename, $ra_contigs ) = @_;
	my $is_file = ref($dst_filename) eq '';

	my $fh_contigs = $self -> _Get ( '_fh_contigs' );
	my $rh_contigs = $self -> _Get ( '_rh_contigs' );
	my $fh_dst     = ($is_file) ? &GetStreamOut( $dst_filename ) : $dst_filename;
	print $fh_dst $self -> GetFileHeader() if ($is_file);

	# Avoiding memory allocation in the loop for performance reasons
	my ($ra_offsets,$offset,$line,$id,$current_contig_id);
	foreach $id (@$ra_contigs)
	{
		#locate the contig in the file
		$ra_offsets = $rh_contigs -> { $id };
		foreach $offset ( @$ra_offsets )
		{
			$fh_contigs -> seek( $offset, 0 );
			while( $line = <$fh_contigs> )
			{
				next if ( $line =~ /^#/ || $line =~ /^$/ );
				($current_contig_id) = split ( /\t/, $line, 2 );
				last	if ($current_contig_id ne $id);
				print $fh_dst $line;
			}
		}
	}
}		

=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;
