#
# $Id: Gene.pm 1336 2013-12-04 13:35:33Z manu $
#

package Locus::Gene;

# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


=head1 NAME

Locus::Gene - describes a protein coding gene

=head1 SYNOPSIS

 my $o_gene = New Locus::Gene('start'=>integer,'end'=>integer,'strand'=>'-1|1|+|-','locus_tag'=>string,parent=>$o_locus);

=head1 DESCRIPTION

 A Gene extends the Locus class, mainly including an array to store the mRNAs (Locus::Mrna) and storing ncRNA (Locus::Ncrna)

=cut

=head1 NAME

Locus::Gene - describes a protein coding gene

=head1 SYNOPSIS

=head1 DESCRIPTION

=cut

use strict;
use warnings;
use Carp;

use Error qw(:try);
use Data::Dumper;
use General;

use Locus;
use base qw( Locus );

sub GetDefaultSubType { return 'gene' };

BEGIN
{
    our $VERSION = do {my @r = (q$Rev: 1336 $ =~ /\d+/g); $r[0]};
}
	
=head2 procedure _Init

 Title        : _Init
 Function     : Constructor, initialize the object, do not use directly, use New method instead
 Usage        : my $o_gene = New Gene('start'=>10,'end'=>100,'strand'=>-1,'locus_tag'=>'GNI','type'=>'gene','parent' => $o_annot,'product'=>'enzyme');
 Prerequisite : start < end, strand =~ /1|\-1|\+|\-/
 Returns      : none
 Args         : Mandatory: start      position of the start (pb) of the object on the contig
                Mandatory: end        position of the end (bp) of the object on the contig
                Mandatory: strand     identify the strand on which is located the object; values=1/-1|+/-
                Mandatory: locus_tag  accession on the molecule (expected to by unique)
                Optional: any other pair of key/value
 Globals      : none

=cut

sub _Init
{
	my $self = shift;
	$self->SUPER::_Init(@_);

	# An array of Locus::Mrna objects
	$$self{'__a_mrnas'} = [];
	
	# An array of Locus::Rna objects
	$$self{'__a_ncrnas'} = [];

	# An array of Locus::MiscRegion objects
	$$self{'__a_regions'} = [];	
	## EXPERIMENTAL - An 'invisible' RNA (will NOT BE PRINTED, only for containing the exons if necessary)
	##$$self{'__o_inv_rna'} = undef;
	
	# A generic type
	$$self{'__type'} = 'gene';
	
}

=head2 procedures Select,Deselect

 See Locus.pm for the generic documentation
 we select/deselect all the Locus::Mrna objects, or the Locus::Ncrna object

=cut
sub Select
{
	my $self = shift;
	my ( $inside ) = @_;
	$inside = &TRUE unless (defined $inside);
	if ( $inside)
	{
		my $ra_mrnas  = $self -> {'__a_mrnas'};
		my $ra_ncrnas = $self -> {'__a_ncrnas'};
		my $ra_regions= $self -> {'__a_regions'};
		foreach my $o_rna ( @$ra_mrnas,@$ra_ncrnas,@$ra_regions )
		{
			$o_rna -> Select ();
		}
	}
	delete $self -> { __deselected };
}
sub Deselect
{
	my $self   = shift;
	my ( $inside ) = @_;
	$inside = &TRUE unless (defined $inside);
	if ($inside)
	{
		my $ra_mrnas = $self -> {'__a_mrnas'};
		my $ra_ncrnas = $self -> {'__a_ncrnas'};
		my $ra_regions= $self -> {'__a_regions'};
		foreach my $o_rna ( @$ra_mrnas,@$ra_ncrnas,@$ra_regions )
		{
			$o_rna -> Deselect ();
		}
	}
	$self -> { __deselected } = &TRUE;
}

=head2 procedure SetParent

 Title        : SetParent
 Usage        : $o_gene->SetParent( $o_annot );
 Procedure    : Set the parent
 Prerequisite : The parent should not already set
 
 Args         : $o_annot : The parent, an AnnotatedSeqfile.pm object (also a contig !)
 Globals      : none

=cut
sub SetParent
{
	my $self = shift;
	my ($o_annot) = @_;
	
	if ( ref($o_annot) ne 'AnnotatedSeq' )
	{
		throw LipmError::ParameterException( 'o_parent', 'type', -text => 'should be an AnnotatedSeq object' );
	}
	$self->SUPER::SetParent( $o_annot );
	
	my $nc_id = $o_annot -> GetNarcisseId();
	$nc_id .= 'g' . $self -> GetCount();
	$self -> { __narcisse_id } = $nc_id;
}

=head2 procedure AddTo

=cut
sub AddTo { 
	my $self = shift;
	my ($o_annot) = @_;
	$o_annot->AddGene($self);
}

=head2 function GetAllChildren

 Title    : GetAllChildren
 Function : Return the children loci, ie the loci embedded by this locus
            This is a recursive function
 Usage    : my $ra_children = $o_locus -> GetAllChildren();
 Args     : none
 Return   : A ref to an array containing the children of the locus, as well as the children of the children, etc.
 Globals  : none

=cut
sub GetAllChildren
{
	my $self = shift;
	my @a_children = @{$self->SUPER::GetAllChildren()};
	
	if ( $self -> IsProteinCoding() )
	{
		my $ra_mrnas = $self -> GetMrnas();
		foreach my $o_mrna ( @$ra_mrnas )
		{
			push @a_children, $o_mrna, @{$o_mrna->GetAllChildren()};
		}
	}
	
	my $ra_ncrnas = $self -> GetNcrnas();
	if ( defined $ra_ncrnas )
	{
		foreach my $o_ncrna ( @$ra_ncrnas )
		{
			push @a_children, $o_ncrna, @{$o_ncrna->GetAllChildren()};
		}
	}
	return \@a_children;
}

=head2 GetNbMrnas

 Title    : GetNbMrnas
 Function : Accessors to the number of mRNAS in this gene
 Usage    : my $nb = $o_gene -> GetNbMrnas();
 Return   : The nb of mrnas
 
=cut
sub GetNbMrnas
{
	my $self = shift;
	my $ra_mrna = $self -> { '__a_mrnas' };
	return scalar @$ra_mrna;
}

=head2 GetMrnas,GetNcrnas,GetMiscRegions

 Title    : GetMrnas
 Function : Accessor to the array of mRNA
 Usage    : my $ra_mrnas = $o_gene -> GetMrnas();
 Arg      : None
 Return   : A ref to an array of mRNAs objects
 
=cut
sub GetMrnas
{
	my $self = shift;
	if ($self -> GetSubType() eq 'ncRNA_gene' )
	{
		throw Error::Simple ("GetMrna is called for a ncRNA_gene - " . $self->GetpkExtended() );
	}
	return $self -> { __a_mrnas };
}
sub GetNcrnas
{
	my $self = shift;
	return $self -> { __a_ncrnas };
}
sub GetMiscRegions
{
	my $self = shift;
	return $self -> { __a_regions };
}

=head2 GetMrna, GetNcrna

 Title    : GetMrna
 Function : Accessors to a single mRNA or ncRNA, throw an exception if GetMrna is called for a non coding gene or if GetNcrna is called for a coding gene
 Usage    : my $o_mrna = $o_gene -> GetMrna($index);
            my $o_ncrna= $o_gene -> GetNcrna();
 Arg      : $index, the index of the mRNA to retrieve
 Return   : Locus::Mrna/Locus::Ncrna found at index $index or undef if none
 
=cut
sub GetMrna
{
	my $self = shift;
	my ($index) = @_;

	if ($self -> GetSubType() eq 'ncRNA_gene' )
	{
		throw Error::Simple ("ERROR - GetMrna is called for a ncRNA_gene - " . $self->GetpkExtended());
	}
	
	my $ra_mrnas = $self -> { '__a_mrnas' };
	return $ra_mrnas -> [ $index ];
}

sub GetNcrna
{
	my $self = shift;
	my ($index) = @_;
	if (!defined ($index) ) { 
		throw Error::Simple ("GetNcrna is called without parameter");
	}
	
	my $ra_ncrna = $self -> { '__a_ncrnas' };
	return $ra_ncrna->[$index];
}

=head2 procedure AddmRNA

 Title        : AddmRNA
 Function     : Add a new mRNA, maybe modifying the start/end of the gene and the type
                NOTE - the type is changed to 'protein_coding_gene'.
                       Thus a protein coding gene is a gene with AT LEAST 1 mRNA
 
 Usage        : my $o_mrna = $o_gene->AddmRNA($o_mrna)
 Returns      : The added mRNA
 Args         : $o_mrna: A Locus::Mrna object to add
 Globals      : none

=cut

sub AddmRNA
{
	my ($self,$o_mrna) = @_;
	throw LipmError::ParameterException( 'o_mrna', 'type', -text => 'should by a Locus::Mrna object' ) unless ref $o_mrna eq 'Locus::Mrna';
	
	my $type = $self -> GetSubType();
	$self -> SetSubType ( 'protein_coding_gene' );
	
	my $ra_mrna = $self -> { __a_mrnas };
	push @$ra_mrna, $o_mrna;
	my ($mrna_start,$mrna_end) = $o_mrna -> GetEnds();
	
	$o_mrna -> SetParent ( $self );

	return $o_mrna;
}

=head2 procedure AddNcrna

 Title        : AddNcrna
 Function     : Add a new Ncrna, maybe modifying the start/end of the gene and the type
                NOTE - If the current type is 'gene', the type is changed to ncRNA_gene
                       If the current type is 'protein_coding_gene', the type is unchanged
                       Thus a protein coding gene is a gene with AT LEAST 1 mRNA
 Usage        : $o_ncrna = $o_gene->AddNcrna($o_ncrna)
 Returns      : The added RNA
 Args         : $o_ncrna: A Locus::Ncrna object to add
 Globals      : none

=cut

sub AddNcrna
{
	my ($self,$o_ncrna) = @_;
	throw LipmError::ParameterException( 'o_ncrna', 'type', -text => 'should by a Locus::Ncrna object' ) unless ref $o_ncrna eq 'Locus::Ncrna';
	
	my $type = $self -> GetSubType();
	if ( $type eq 'gene' )
	{
		$self -> SetSubType ( 'ncRNA_gene' );
	}

	my $ra_ncrna = $self -> { __a_ncrnas };
	push @$ra_ncrna, $o_ncrna;
	$o_ncrna -> SetParent ( $self );

	return $o_ncrna;
}

=head2 function IsProteinCoding

 Title        : IsProteinCoding
 Function     : Return true if the gene type is "protein_coding_gene"
 Usage        : my $isproteincoding = $o_gene->IsProteinCoding();
 Prerequisite : none
 Returns      : Return true/false

=cut

sub IsProteinCoding
{
	my($self) = @_;
	return ( $self -> GetSubType() eq 'protein_coding_gene' );
}

=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;
