#
# $Id: Fasta.pm 946 2012-06-27 11:09:58Z manu $
#

package Splitter::Fasta;

use strict;
use warnings;

use base qw( Splitter );

# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=head1 NAME

 Splitter::Fasta - Split a fasta file

=head1 SYNOPSIS
 See Splitter

=head1 DESCRIPTION
 See the description in the Splitter abstract class
 Splitter::Fasta looks like, but cannot be exchanged with, FastaReader
 
=cut

use General;
use Error qw(:try);
use LipmError::ParameterException;

BEGIN
{
    our $VERSION = do {my @r = (q$Rev: 946 $ =~ /\d+/g); $r[0]};
}

=head2 procedure _Index

 Title        : _Index
 Usage        : $self -> _Index()
 Function     : Initializes the _h_contigs hash (key = contig name, val = offset in the file)
                The contig name is the FIRST WORD of the header
                Initializes also the _rh_contig_size hash with the number of bytes for each sequence 
                (each byte is counted: for the header, the bases, the \n)
               
 Args         : None
 Globals      : none

=cut

sub _Index
{
	my $self = shift;
	
	my $fh_contigs = $self -> _Get ( '_fh_contigs' );
	my $rh_contigs = $self -> _Get ( '_rh_contigs' );
	my $rh_contig_size = $self -> _Get ( '_rh_contig_size' );
	
	my $id     = "";
	my $offset = 0;
	my $s_ctg  = 0;
	my $p_id   = '';
	my $line   = "";
	$fh_contigs -> seek(0,0);
	while( $line = <$fh_contigs> )
	{
		# Skip all lines except headers
		next if( $line !~ /^>/ );

		# The line is a header
		#     1/ If it is formatted like this (from genbank):
		#        >gi|89058234|ref|NT_113952.1| Homo sapiens..., the key is NT_113952.1
		#     2/ Else the key is the first word of the header
		# The value is the position of the start of line
		#
		my $line_length = length($line);
		chomp $line;
		$id = $line;
		
		# It is a file downloaded from genbank
		if ( $id =~ /^>gi\|/ )
		{
			my @a_fields = split ( '\|', $id );
			$id = $a_fields[3];
		}
		else
		{
			$id =~ s/^>([\w.\-]+)\W*.*/$1/;
		}
		
		$offset = $fh_contigs->tell() - $line_length;

		# Each contig should be uniq in the file
		if( defined( $rh_contigs -> { $id } ) )
		{
			my $txt = "you have more than one sequence with the id -$id-\n";
			throw LipmError( -text => $txt );
		}

		$rh_contigs     -> { $id } = $offset;
		if ( $p_id ne '' )
		{
			$rh_contig_size -> { $p_id } = $offset - $s_ctg + 1;
			$s_ctg = $offset;
		}
		$p_id = $id;
	}
	$rh_contig_size -> { $p_id } = $fh_contigs->tell() - $s_ctg + 1;	
}

=head2 function _GetContigs

 Title    : GetContigs
 Usage    : return $o_splitter->_GetContigs( \@a_contigs );
 Prerequisite : The array should NOT be empty 
 Function : For each id in the array, return the corresponding contig
            If some id is not found, throw an exception
            NOTE - If you want to retrieve some contigs, this is much more performant than using the Regex version,
            as we do not have to scan all the indexed keys !
 Returns  : $r_contigs, a ref to string with the extracted contigs, formatted as in the original file
 Args     : $ra_contigs, ... A list of contigs to retrieve

=cut
sub _GetContigs
{
	my $self           = shift;
	my ( $ra_contigs ) = @_;

	my $fh_contigs = $self -> _Get ( '_fh_contigs' );
	my $rh_contigs = $self -> _Get ( '_rh_contigs' );
	my $seq = "";
	# Avoiding memory allocation in the loop for performance reasons
	my ($offset,$line,$id);
	foreach $id (@$ra_contigs)
	{
		#locate the seq in the file
		$offset = $rh_contigs -> { $id };
		$line = "";
		$fh_contigs -> seek( $offset, 0 );
		
		# the header...
		$seq .= <$fh_contigs>;
		
		# ...then the sequence
		while( $line = <$fh_contigs> )
		{
			last	if( $line =~ /^>/ );
			$seq .= $line;
		}
	}
	return \$seq;
}	

=head2 procedure _CopyContigs

 Title    : _CopyContigs
 Usage    : $o_splitter->_CopyContigs( $dst_filename, \@a_contigs );
 Function : Copy each contig to dst_filename (should be writable)
            For the meaning of the second parameter, see the corresponding Get functions...
            ...But please note that the method GetContigs is NOT used, to get a lower memory print.
 Args     : $dst_filename The destination filename (should be writeable)
            $ra_contigs A list of contigs to copy

=cut
sub _CopyContigs
{
	my $self = shift;
	my ( $dst_filename, $ra_contigs ) = @_;

	my $fh_contigs = $self -> _Get ( '_fh_contigs' );
	my $rh_contigs = $self -> _Get ( '_rh_contigs' );
	my $fh_dst     = (ref($dst_filename) eq '') ? &GetStreamOut( $dst_filename ) : $dst_filename;
	my $id     = "";
	my $offset = 0;
	my $line   = "";
	foreach $id (@$ra_contigs)
	{
		#locate the seq in the file
		$offset = $rh_contigs -> { $id };
		$fh_contigs -> seek( $offset, 0 );
		
		# the header...
		my $h = <$fh_contigs>;
		print $fh_dst $h;
		
		# ...then the sequence
		while( $line = <$fh_contigs> )
		{
			last	if( $line =~ /^>/ );
			print $fh_dst $line;
		}
	}
}

=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;
